/* Written by James Henderson and Ivan Titov */

#include "cluster.h"

#define CHECK_PTR(x) if ((x) == NULL) { printf("Can't allocate memory block\n"); exit(-1);}
#define CL_ALLOC_SIZE 10

#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>




/** no attempt to make it work fast (done once on initalization) 
 *  @param parent the parent node to which to attach (if NULL - assuming creating root)
 *  @return created node
 * */
struct cluster *cl_create_child(struct cluster *parent) {
    
    struct cluster *child;

    child = (struct cluster*) malloc(sizeof(struct cluster));
    CHECK_PTR(child);
    child->cluster_id = -1;
    child->child_qty = 0;
    child->parent = parent;
    child->childs = NULL;
    child->layer_num = 0;
    
    
    if (parent != NULL) {    
        parent->childs = (struct cluster**) realloc(parent->childs, 
                (parent->child_qty + 1) * sizeof(struct cluster*));
        CHECK_PTR(parent->childs);
        parent->childs[parent->child_qty] = child;
        parent->child_qty++;
        child->layer_num = parent->layer_num + 1;
    }

    return child;
}

/** internal freeing function */
void _cl_free(struct cluster* c) {
    int idx;
    for (idx = 0; idx < c->child_qty; idx++) {
        _cl_free(c->childs[idx]);
    }
    free(c->childs);
    free(c);
}

/** freeing of all the clustering structure */
void cl_free(struct clustering *clustering) {    
    _cl_free(clustering->root);   
    free(clustering->clusters_arr);
    free(clustering);
}

/** method for creating clustering structure from cluster root */
struct clustering *cl_create_clustering(struct cluster *root) {

    struct clustering *wrapper = (struct clustering*) malloc(sizeof(struct clustering));
    CHECK_PTR(wrapper); 
    
    wrapper->root = root;
    wrapper->clusters_arr = NULL;
    wrapper->cluster_qty = -1;


    wrapper->cluster_qty = _cl_enumerate(wrapper->root,  &(wrapper->clusters_arr));
    return wrapper;
}   

/** 
 * enumerates the clusters 
 * @param cluster  cluster subclusters of which should be enumerated
 * @param pcluster_arr ptr to cluster array
 * @param cluster_arr_size cluster_arr_size
 * @return new cluster array size
 */
/*
int _cl_enumerate(struct cluster  *start_cluster, struct cluster ***pcluster_arr, int cluster_arr_size) {
    int child_idx;
    int new_size;

    // if leaf cluster 
    if (start_cluster->child_qty == 0) {
        return cluster_arr_size;
    }
    
    new_size = start_cluster->child_qty + cluster_arr_size;

    
    *pcluster_arr = (struct cluster**) realloc(*pcluster_arr, new_size * sizeof(struct cluster *));
    CHECK_PTR(pcluster_arr);

    
    //fill refs to childs 
    for (child_idx = 0; child_idx < start_cluster->child_qty; child_idx++) {
        (*pcluster_arr)[cluster_arr_size + child_idx] = start_cluster->childs[child_idx];
        start_cluster->childs[child_idx]->cluster_id = cluster_arr_size + child_idx;
    }


    //fill childs
    for (child_idx = 0; child_idx < start_cluster->child_qty; child_idx++) {
        new_size = _cl_enumerate(start_cluster->childs[child_idx], pcluster_arr, new_size);
    }

    return new_size;
    
}
*/
int _cl_enumerate(struct cluster *root, struct cluster *** pcluster_arr) {
    
    struct _cl_list *head, *tail;
    struct cluster* curr; 
    
    
    head = tail = NULL;  
    
    _cl_add_to_tail(&head, &tail, root);
    
    int cluster_id = -1;
    int child_id;

    //we do not take root into account!
   
    while (head != NULL && tail != NULL) {
        curr = _cl_pop_from_head(&head, &tail);         
        if (cluster_id != -1) {
            (*pcluster_arr) = (struct cluster**) realloc((*pcluster_arr), (cluster_id + 1) * sizeof(struct cluster*));
            (*pcluster_arr)[cluster_id] = curr;
            curr->cluster_id = cluster_id;         
        }
        cluster_id++;

        for (child_id = 0; child_id < curr->child_qty; child_id++) {
            _cl_add_to_tail(&head, &tail, curr->childs[child_id]);
        }
    }

    return cluster_id;
}




char *_cl_make_message(const char *fmt, ...) {

    int n, size;
    char *msg;
    va_list ap;

    size = 1;
    
    msg = malloc(size);
    CHECK_PTR(msg);

    while (1) {

        va_start(ap, fmt);
        n = vsnprintf (msg, size, fmt, ap);
        va_end(ap);
        
        if (n > -1 && n < size) {
            return msg;
        }
        if (n > -1)  {
            size = n + 1; 
        } else  { 
            size *= 2;  
        }
        msg = realloc (msg, size);
        CHECK_PTR(msg);
    }
}

/** debug cluster output */
void cl_print_cluster(struct cluster* c) {
    _cl_print_cluster(c, "");
}


void _cl_print_cluster(struct cluster* c, char* prefix) {

    char* new_prefix;
    int child_idx;

    if (c->child_qty == 0) {
        printf("%s\n", prefix);    
        return;
    }  
    for (child_idx = 0; child_idx < c->child_qty; child_idx++) {
        _cl_print_cluster(c->childs[child_idx], _cl_make_message("%s %d", prefix, child_idx));
        
    }
}

/** debug clustering output */
void cl_print_clustering(struct clustering* cluster_str) {
    int cluster_idx;
    
    cl_print_cluster(cluster_str->root);

    for (cluster_idx = 0; cluster_idx < cluster_str->cluster_qty; cluster_idx++) {
        printf("%d ", cluster_str->clusters_arr[cluster_idx]->cluster_id);
    }
    
    printf("\n");
}

/**
 * Read clusters from buffer (buffer contains the quantity of childs for clustering nodes in
 * width-first order
 * @param buffer buffer to read from
 * @param separs used separaters
 * @return the resulting clustering
 */
struct clustering  *cl_read_clustering(char* buffer, char* separs) {
    
    struct _cl_list *head, *tail;
    char *child_qty_str;
    struct cluster *root, *curr, *child;
    struct clustering *cluster_str;
    int child_qty, child_id;
    
    head = tail = NULL;
    root = cl_create_child(NULL);
    _cl_add_to_tail(&head, &tail, root);
    while((child_qty_str = strtok(buffer, separs)) != NULL) {
        buffer = NULL;
        child_qty = atoi(child_qty_str);

        if (child_qty < 0) {
            return NULL;           
        }
        
        curr = _cl_pop_from_head(&head, &tail);
        
        for (child_id = 0; child_id < child_qty; child_id++) {
           
           child = cl_create_child(curr);           
           _cl_add_to_tail(&head, &tail, child);
           
        }
    }    
  
    /** zeros are assumed at the end */ 
    while (head != NULL && tail != NULL) {
        curr = _cl_pop_from_head(&head, &tail);
    }

    cluster_str = cl_create_clustering(root);
    return cluster_str;
}


/**  
 * returns the specification which is the quantity of childs for the node in 
 * the traverse order: parent - first child  -... - last child 
 */ 
char *cl_get_specification(struct clustering *cluster_str) {

    struct _cl_list *head, *tail;
    struct cluster *curr;
    char *spec, *swap;
    int spec_size, child_id;
        
    spec_size  = CL_ALLOC_SIZE;
    
    spec = (char*) malloc(spec_size * sizeof(char));
    CHECK_PTR(spec);
    spec[0] = 0;
    
    //hack - each has at least one output (nonbranching only at the root level)
    if (cluster_str->cluster_qty == 0) {
        strcpy(spec, "1 0");
        return spec;
    }
    
    head = tail = NULL;
    
    _cl_add_to_tail(&head, &tail, cluster_str->root);
     
            
    while (head != NULL && tail != NULL) {
        curr = _cl_pop_from_head(&head, &tail); 
        
        //skip one branching structures
        if (curr->child_qty <= 1) {
            if (strlen(spec) + strlen("0 ") >= spec_size) {
                spec = (char*) realloc(spec,  (strlen(spec) + strlen("0 ") + 1) * sizeof(char) * 2);
                CHECK_PTR(spec);
                spec_size = (strlen(spec) + strlen("0 ") + 1) * 2;
            }
            strcat(spec, "0 ");
            continue;
        }
        
        //spec += curr->child_qty + " "
        swap = (char*) malloc(5);
        CHECK_PTR(swap);
        sprintf(swap, "%d ", curr->child_qty); 
        
        if (strlen(spec) +  strlen(swap) >= spec_size) {
            spec = (char*) realloc(spec,  (strlen(spec) + strlen(swap) + 1) * sizeof(char) * 2);
            CHECK_PTR(spec);
            spec_size = (strlen(spec) + strlen(swap) + 1) * 2;
        }
        strcat(spec, swap);
        free(swap); 
        
    
        for (child_id  = 0; child_id < curr->child_qty; child_id++) {
            _cl_add_to_tail(&head, &tail, curr->childs[child_id]);
        }
    } // while queue isn't empty

    return spec; 
}



/** 
 * adds elem to the tail of the list
 * @param head head of the list
 * @param tail tail of the list
 * @param elem the elem of the node to add
 */
void _cl_add_to_tail(struct _cl_list **head, struct _cl_list **tail, struct cluster *elem) {

    struct _cl_list *new_node; 
   
    assert((*head == NULL && *tail == NULL) || (*head != NULL && *tail != NULL));

    new_node = (struct _cl_list*) malloc(sizeof(struct _cl_list));
    CHECK_PTR(new_node);
    new_node->elem = elem;
    new_node->next = NULL;
    new_node->prev = NULL;
    
    
    if (*head == NULL) {
        /** no list yet */
        *tail = *head = new_node;
    } else {
        (*tail)->next = new_node;
        new_node->prev = (*tail);
        *tail = new_node;
    }
}

/**
 * gets and removes element from the list head
 * @param head the list head
 * @param tail the list tail
 * @return cluster
 */
struct cluster* _cl_pop_from_head(struct _cl_list **head, struct _cl_list **tail) {
    struct cluster *c;
    
    if (*head == NULL || *tail == NULL) {
        return NULL;
    }
   
    c = (*head)->elem;
    
    //last elem in list
    if (*head == *tail) {
        free(*head);
        *head = NULL;
        *tail = NULL;
    } else {
        *head = (*head)->next;
        free((*head)->prev);
        (*head)->prev = NULL;
    }

    return c;
}


void _cl_print_list(struct _cl_list *head, struct _cl_list *tail) {
    
    while (head != NULL) {
        head = head->next;
    }
    
}

/**
 * Creates flat structure (root and childs)
 * @param child_qty root childs qty
 * @return resulting clustering
 */
struct clustering *cl_create_one_level(int child_qty) {
    struct cluster *root;
    int child_id;
        
    root =  cl_create_child(NULL);
    
    for (child_id = 0; child_id < child_qty; child_id++) {
        cl_create_child(root);
        
    }
    
    return cl_create_clustering(root);
} 

/**
 * Determines the branching factor of the clustering
 * @param cluster_str the cluster structure to analyze
 * @return the max branching factor
 */
int cl_get_max_branch_fact(struct clustering *cluster_str) {
    int cluster_id;
    int max_fact = 0;

    for (cluster_id = 0; cluster_id < cluster_str->cluster_qty; cluster_id++) {
        if (cluster_str->clusters_arr[cluster_id]->child_qty > max_fact) {
            max_fact = cluster_str->clusters_arr[cluster_id]->child_qty;
        }
    }

    return max_fact;
}

/**
 * frees the list (without freeing the elements)
 * @param head head  of the list
 */
void _cl_free_list(struct _cl_list *head) {
    
    struct _cl_list *swap;

    while(head != NULL) {
        swap = head;
        head = head->next;
        free(swap);
    }
}

/**
 * Creates new clustering based on given, extending the leafs which are marked in is_clusterid_prolonged with 1 adding 
 * subclusters (using given branch factor). 
 * @param is_clusterid_extended maps cluster_id to whether it has to be extended or not
 * @param branch_factor the branch factor for the clusters (nodes)  to be extended
 * @return the clustering 
 */
struct clustering* cl_extend_clustering(struct clustering  *init_clustering, int is_clusterid_extended[], int branch_factor) {

    struct clustering *new_clustering  = (struct clustering*) malloc(sizeof(struct clustering));
    CHECK_PTR(new_clustering);
    
    new_clustering->cluster_qty = 0;
    new_clustering->root = NULL;
    new_clustering->clusters_arr = NULL;
    
    if (init_clustering->root != NULL) {

        new_clustering->root = cl_create_child(NULL);
        _cl_extend_subtree(new_clustering->root, init_clustering->root, is_clusterid_extended, branch_factor);        
        new_clustering->cluster_qty = _cl_enumerate(new_clustering->root,  &(new_clustering->clusters_arr));
 
    }
    return new_clustering;
}


/**
 * Extends the cluster in a given subtree. 
 * @param new_cluster new cluster which subtree should be added (and extention of needed clusters should be performed in it)
 * @param old_cluster the cluster form which to copy
 * @param is_clusterid_extended maps cluster_id to whether it has to be extended or not
 * @param branch_factor the branch factor for the clusters (nodes)  to be extended 
 */
void _cl_extend_subtree(struct cluster *new_cluster, struct cluster *old_cluster, int is_clusterid_extended[], int branch_factor) {
    int child_idx;
    if (old_cluster->child_qty == 0) {
        //if it has to be prolonged
        if (is_clusterid_extended[old_cluster->cluster_id]) {
            for (child_idx = 0; child_idx < branch_factor; child_idx++) {
                cl_create_child(new_cluster);
            }
        } else {
            //do nothing
        }
    } else {        
        //creating the list of childs 
        for (child_idx = 0; child_idx < old_cluster->child_qty; child_idx++) {
            cl_create_child(new_cluster);    
            _cl_extend_subtree(new_cluster->childs[child_idx], old_cluster->childs[child_idx], is_clusterid_extended, branch_factor);
        }
    }
}


/**
 * Removes the node from the clustering.
 * It is assumed that this node is leaf (if not error returned)
 * clusters_arr is not adjusted (NULL is placed), so it should be used carefully after (quick hack!)
 * @param cluster_str the clustering to deal with
 * @param cluster_id the cluster_id of the leaf to remove
 * @return negative if error, else if OK
 */
 
int cl_remove_leaf_wo_list_change(struct clustering *cluster_str, int cluster_id) {
    
    struct cluster *leaf, *parent;
    int leaf_child_id, child_id;
    
    if (cluster_str->cluster_qty <= cluster_id) {
        return -1;    
    } 
    
    leaf = cluster_str->clusters_arr[cluster_id];
    
    if (leaf->child_qty > 0) {
        return -1;
    }   
    
    parent = leaf->parent;
    leaf_child_id = -1;

    for (child_id = 0; child_id < parent->child_qty; child_id++) {
        if (parent->childs[child_id]->cluster_id == cluster_id) {
            leaf_child_id = child_id;
        }
    }
    
    if (leaf_child_id == -1) {
        return -1;
    }
    
    _cl_free(leaf);
    
    memmove(parent->childs + leaf_child_id, parent->childs + leaf_child_id + 1, 
            sizeof(struct cluster*) * (parent->child_qty - leaf_child_id - 1));
    parent->childs = (struct cluster**) realloc(parent->childs, (parent->child_qty - 1) * sizeof(struct cluster*));
    parent->child_qty--;
    
    cluster_str->clusters_arr[cluster_id] = NULL;
    return 1;
    
}

void  _cl_test_cluster()  {
    
    struct cluster *root, *child1;
    struct clustering *cluster_str;
    
    char* text  = malloc(1000);   
            
    printf("Creating root\n");
    root = cl_create_child(NULL);
    printf("Appending child\n");
    child1 = cl_create_child(root);
    printf("Appending child\n");
    cl_create_child(root);
    printf("Appending child\n");
    cl_create_child(child1);
    printf("Appending child\n");
    cl_create_child(child1);
    printf("Creating clustering\n");
    cluster_str = cl_create_clustering(root);
    printf("Printing clustering\n");
    cl_print_clustering(cluster_str);
    
    printf("max branch factor: %d\n", cl_get_max_branch_fact(cluster_str)); 
    
    printf("Freeing\n");
    cl_free(cluster_str);
    
    strcpy(text, "3 3 3 2 0 0 0 2 0 0 0 0 0 0");
            
    printf("Reading: %s\n", text);
    cluster_str = cl_read_clustering(text, " ");
    printf("Printing clustering\n");
    cl_print_clustering(cluster_str);

    printf("max branch factor: %d\n", cl_get_max_branch_fact(cluster_str));
    
    printf("Freeing\n");
    cl_free(cluster_str);
    
}


/*
int main(int argc, char** argv) {
    _cl_test_cluster();
}

*/

