/* copyright James Henderson 2005 */

/* Code for converting treebank2 trees to a version of a left corner
   derivation called "predictive LR", extended with tree transformations for
   Chomsky adjunction ("modification"), nonbranching nodes ("conflation"), and
   heads.  The resulting derivation is output in the format needed to train
   SSN parsers.
*/
/* This code is much more complicated than it needs to be, because it was
   written for a completely different purpose and converted to its current
   purpose.  By now I have forgotten how it works, but my best guess is the
   following.  Allowable modifications and conflations are specified in some
   arrays initialized at the top of the code, along with the head table.
   After reading in the tree, each nonterminal node of the tree is annotated
   with the way in which it is introduced into the tree, called "triggering",
   (modification, conflation, or default to projection).  Nonterminals are
   also added to two arrays which record which nodes are introduced between
   the prediction of each two preterminals.  The node projected from the
   preterminal in position p is held in lexical_triggerings[p].  All other
   nodes projected between position p and p+1 are held in
   nonlexical_triggerings[p+1].  When printing the output format, a stack is
   constructed to record the state of the parser at that point.  For each
   position in the sentence, after printing the previous position's
   preterminal prediction, printing starts with the projection in
   lexical_triggerings.  Then printing alternates between printing the
   attachments/modifications for the nodes on the stack down to the triggering
   child of the next projection in nonlexical_triggerings, and printing the
   next projection in nonlexical_triggerings.  When all projections in
   nonlexical_triggerings have been printed, it prints the the
   attachments/modifications for the nodes on the stack which end after this
   position.
*/

/* tree transform control */
#define CONFLATIONS   1 /* perform allowed conclations */
#define MODOFOCATIONS 1 /* perform allowed modifications */

/* output format control */
#define DISPOUTPUT 1 /* display the output format */
#define IDOUTPUT   1 /* use id numbers instead of tag/label texts */
#define FILEWORDS  1 /* write words to <datafile>.words when computing 
			outputs. */
#define HEADOUTPUT 1 /* mark head projections and attachments in output. */
#define LONGOUTPUT 0 /* use a more human-readable format for output */
#define DISPRULES  0 /* display the rule format, augmented as below */
#define DISPINFO   0 /* expanded version of rule format */
#define DISPPAR    0 /* show the parent of the site with the rule format */
#define DISPALL    0 /* even display nontriggering nonlexical projections */
#define ONLYLABELS 0 /* only display the nonterminal labels, w/ conflations */
#define ONLYTAGS   0 /* only display the preterminal labels */
#define DISPCONT   0 /* display right context */ 
#define DISPINP    0 /* display input */
#define DISPTREE   0 /* display target tree */
#define TREEPOS    0 /* display position numbers when displaying target tree */

#define DEBUG    0
#define DEBUG_ALLOC    0
#define DEBUG_INP 0
#define DEBUG_STRUCMOD 0
#define DEBUG_TRIG 0
#define DEBUG_READOUT 0
void bugbreak() {}

#include <stdlib.h>
/* #include "/usr/include/stdlib.h" ?*/
#include <stdio.h>
#include <math.h>
#include <string.h>

#define MAXLABEL 100  /* max length of label string */
#define MAXDATALINE 1000  /* max length of one line of treebank */
#define MAXSENTLENGTH 1000
#define MAXWORDLENGTH 100 

#define UNTRIGGER "***"
#define NONTRIGGER "_"


/** THE HEAD RULE TABLE (note: NP and NX are special cases) **/

char *head_projections[25] = {
  "ADJP", "ADVP", "CONJP", "FRAG", "INTJ", "LST", "NAC", "PP", "PRN", "PRT",
  "QP", "RRC", "S", "SBAR", "SBARQ", "SINV", "SQ", "UCP", "VP", "WHADJP",
  "WHADVP", "WHNP", "WHPP", "X",
  NULL};
int head_directions[25] = {
  0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0};
char *head_categories[25][21] = {
  {"NNS", "QP", "NN", "$", "ADVP", "JJ", "VBN", "VBG", "ADJP", "JJR", "NP",
   "NX", "JJS", "DT", "FW", "RBR", "RBS", "SBAR", "RB", NULL}, 
  {"RB", "RBR", "RBS", "FW", "ADVP", "TO", "CD", "JJR", "JJ", "IN", "NP",
   "NX", "JJS", "NN", NULL}, 
  {"CC", "RB", "IN", NULL}, 
  {NULL},
  {NULL},
  {"LS", ":", NULL}, 
  {"NN", "NNS", "NNP", "NNPS", "NP", "NX", "NAC", "EX", "$", "CD", "QP", "PRP",
   "VBG", "JJ", "JJS", "JJR", "ADJP", "FW", NULL}, 
  {"IN", "TO", "VBG", "VBN", "RP", "FW", NULL}, 
  {NULL},
  {"RP", NULL}, 
  {"$", "IN", "NNS", "NN", "JJ", "RB", "DT", "CD", "NCD", "QP", "JJR",
   "JJS", NULL}, 
  {"VP", "NP", "NX", "ADVP", "ADJP", "PP", NULL}, 
  {"TO", "IN", "VP", "S", "SBAR", "ADJP", "UCP", "NP", "NX", NULL}, 
  {"WHNP", "WHPP", "WHADVP", "WHADJP", "IN", "DT", "S", "SQ", "SINV",
   "SBAR", "FRAG", NULL}, 
  {"SQ", "S", "SINV", "SBARQ", "FRAG", NULL}, 
  {"VBZ", "VBD", "VBP", "VB", "MD", "VP", "S", "SINV", "ADJP", "NP", "NX",
   NULL},  
  {"VBZ", "VBD", "VBP", "VB", "MD", "VP", "SQ", NULL}, 
  {NULL},
  {"TO", "VBD", "VBN", "MD", "VBZ", "VB", "VBG", "VBP", "VP", "ADJP", "NN",
   "NNS", "NP", "NX", NULL}, 
  {"CC", "WRB", "JJ", "ADJP", NULL}, 
  {"CC", "WRB", NULL}, 
  {"WDT", "WP", "WP$", "WHADJP", "WHPP", "WHNP", NULL}, 
  {"IN", "TO", "FW", NULL}, 
  {NULL}};


/** THE "GRAMMAR" **/
/* (when adding entries, also check array sizes) */

/* these lists define allowable modification (Chomsky) adjunctions.  Each
   element in modification_projections can adjoin at any of the sites listed
   in the corresponding element of modification_sites, if the root has the
   same label as the site. */ 
char *modification_projections[20] = {
  "PP",                           
  "SBAR",                    
  "ADJP",      
  "RRC",       
  "IN",
  "CC",                             
  ",",                         
  ":",                                     
  "PRN",           
  "ADVP",          
  "X",         
  NULL};
char *modification_sites[20][15] = {
  {"NP","ADJP","PP","ADVP","WHNP",NULL}, 
  {"NP","ADJP","ADVP",NULL},
  {"NP",NULL}, 
  {"NP",NULL}, 
  {"S",NULL}, 
  {"NP","VP","S","ADJP","SBAR","PP","UCP","NX","INTJ",NULL}, 
  {"NP","VP","S","ADJP","SBAR","PP","UCP","NX","INTJ",NULL}, 
  {"S","NP","FRAG","PP","VP","SBAR",NULL}, 
  {"NP","S",NULL}, 
  {"NP","S",NULL}, 
  {"NP",NULL}};

/* these lists define allowable conflations.  Any nonbranching levels
   between nonterminals are conflations if the child is in
   conflation_projections and the parent is in conflation_sites. */
char *conflation_projections[10] = {
  "S",           "VP",            "NP",       
  /* "QP",        */
  NULL};
char *conflation_sites[10][5] = {
  {"SBAR",NULL}, {"S","SQ",NULL}, {"S",NULL}
  /* {"NP",NULL}, */
  };



/** TAG AND LABEL CODES **/

int tag_id(char *str)
{
  if (!strcmp(str, "#"))
    return(0);
  else if (!strcmp(str, "$"))
    return(1);
  else if (!strcmp(str, "''"))
    return(2);
  else if (!strcmp(str, ","))
    return(3);
  else if (!strcmp(str, "-LRB"))
    return(4);
  else if (!strcmp(str, "-RRB"))
    return(5);
  else if (!strcmp(str, "."))
    return(6);
  else if (!strcmp(str, ":"))
    return(7);
  else if (!strcmp(str, "^^^"))
    return(8);
  else if (!strcmp(str, "``"))
    return(9);
  else if (!strcmp(str, "CC"))
    return(10);
  else if (!strcmp(str, "CD"))
    return(11);
  else if (!strcmp(str, "DT"))
    return(12);
  else if (!strcmp(str, "EX"))
    return(13);
  else if (!strcmp(str, "FW"))
    return(14);
  else if (!strcmp(str, "IN"))
    return(15);
  else if (!strcmp(str, "JJ"))
    return(16);
  else if (!strcmp(str, "JJR"))
    return(17);
  else if (!strcmp(str, "JJS"))
    return(18);
  else if (!strcmp(str, "LS"))
    return(19);
  else if (!strcmp(str, "MD"))
    return(20);
  else if (!strcmp(str, "NN"))
    return(21);
  else if (!strcmp(str, "NNP"))
    return(22);
  else if (!strcmp(str, "NNPS"))
    return(23);
  else if (!strcmp(str, "NNS"))
    return(24);
  else if (!strcmp(str, "PDT"))
    return(25);
  else if (!strcmp(str, "POS"))
    return(26);
  else if (!strcmp(str, "PRP"))
    return(27);
  else if (!strcmp(str, "PRP$"))
    return(28);
  else if (!strcmp(str, "RB"))
    return(29);
  else if (!strcmp(str, "RBR"))
    return(30);
  else if (!strcmp(str, "RBS"))
    return(31);
  else if (!strcmp(str, "RP"))
    return(32);
  else if (!strcmp(str, "SYM"))
    return(33);
  else if (!strcmp(str, "TO"))
    return(34);
  else if (!strcmp(str, "UH"))
    return(35);
  else if (!strcmp(str, "VB"))
    return(36);
  else if (!strcmp(str, "VBD"))
    return(37);
  else if (!strcmp(str, "VBG"))
    return(38);
  else if (!strcmp(str, "VBN"))
    return(39);
  else if (!strcmp(str, "VBP"))
    return(40);
  else if (!strcmp(str, "VBZ"))
    return(41);
  else if (!strcmp(str, "WDT"))
    return(42);
  else if (!strcmp(str, "WP"))
    return(43);
  else if (!strcmp(str, "WP$"))
    return(44);
  else if (!strcmp(str, "WRB"))
    return(45);
  else
    {
      fprintf(stderr, "Error: unrecognized tag %s\n", str);
      return(99);
    }
}

int label_id(char *str)
{
  if (!strcmp(str, ""))
    return(0);
  else if (!strcmp(str, "ADJP"))
    return(1);
  else if (!strcmp(str, "ADVP"))
    return(2);
  else if (!strcmp(str, "CONJP"))
    return(3);
  else if (!strcmp(str, "FRAG"))
    return(4);
  else if (!strcmp(str, "INTJ"))
    return(5);
  else if (!strcmp(str, "LST"))
    return(6);
  else if (!strcmp(str, "NAC"))
    return(7);
  else if (!strcmp(str, "NP"))
    return(8);
  else if (!strcmp(str, "NX"))
    return(9);
  else if (!strcmp(str, "PP"))
    return(10);
  else if (!strcmp(str, "PRN"))
    return(11);
  else if (!strcmp(str, "PRT"))
    return(12);
  else if (!strcmp(str, "QP"))
    return(13);
  else if (!strcmp(str, "RRC"))
    return(14);
  else if (!strcmp(str, "S"))
    return(15);
  else if (!strcmp(str, "S-NP"))
    return(16);
  else if (!strcmp(str, "S-VP"))
    return(17);
  else if (!strcmp(str, "SBAR"))
    return(18);
  else if (!strcmp(str, "SBAR-S"))
    return(19);
  else if (!strcmp(str, "SBAR-S-VP"))
    return(20);
  else if (!strcmp(str, "SBARQ"))
    return(21);
  else if (!strcmp(str, "SINV"))
    return(22);
  else if (!strcmp(str, "SQ"))
    return(23);
  else if (!strcmp(str, "SQ-VP"))
    return(24);
  else if (!strcmp(str, "UCP"))
    return(25);
  else if (!strcmp(str, "VP"))
    return(26);
  else if (!strcmp(str, "WHADJP"))
    return(27);
  else if (!strcmp(str, "WHADVP"))
    return(28);
  else if (!strcmp(str, "WHNP"))
    return(29);
  else if (!strcmp(str, "WHPP"))
    return(30);
  else if (!strcmp(str, "X"))
    return(31);
  else if (!strcmp(str, "SBAR-S-NP"))
    return(32);
  else
    {
      fprintf(stderr, "Error: unrecognized label %s\n", str);
      return(99);
    }
}


/** DATATYPE DEFINITIONS **/

struct tree
{
  struct tree *parent; /* parent.  Includes conflation projections. */
  struct tree_list *children;  /* in reverse order.  NULL if preterm.
				  Includes conflation projections. */
  char label[MAXLABEL + 1];
  int start; /* first word */
  int end;   /* last word + 1 */
  struct tree *trigger; /* pointer to the trigger preterm, NULL if not set */
  struct tree_list *trigger_child; /* pointer to trigger child in children */
  struct tree_list *adjunction_child; /* pointer to the child in children
					 of which this is an adjunction.
					 Skips to lowest of conflation
					 projections. */ 
  struct tree_list *modification_child; /* pointer to the child in children
					   of which this is an modification.
					   Parent removed from tree. */
  struct tree *attachment; /* site to which this is an attachment, if
			      checked in trigger test */
  int conflation; /* flag if this was a conflation projection */
  int toplexical; /* flag if this was a toplexical projection */
  int modifier; /* flag if attached to parent with modification */
  int head; /* flag if head of parent */
  int site_id; /* ID in output format */
};

struct tree_list
{
  struct tree *elem;  
  struct tree_list *next;
  struct tree_list *prev;
};

struct triggering
{
  struct tree *preterm;  
  struct tree_list *projection;
};


/** GLOBAL VARIABLES **/

int sentence_number;
char sentence_words[MAXSENTLENGTH][MAXWORDLENGTH + 1];
struct triggering *lexical_triggerings[MAXSENTLENGTH];
struct triggering *nonlexical_triggerings[MAXSENTLENGTH];
struct tree_list *sentence_untriggered;
int next_site_id;  /* for allocating site id's */

/** PROCEDURE DECLARATIONS **/

int main(int , char *[]);
void compute_triggers(FILE *, FILE *);
void initialize_triggers();
int remove_empty(struct tree *);
void relabel_conflations(struct tree *);
struct tree *remove_modification(struct tree *);
void set_heads(struct tree *);

struct tree *read_sentence(FILE *);
struct tree_list *start_const(int , struct tree_list *);
struct tree_list *start_root(int , struct tree_list *);
char *read_nonterm(char *, struct tree_list *);
struct tree_list *end_const(int , struct tree_list *);
char *read_terminal(char *, int );
char *skip_space(char *);

void set_triggers(struct tree * );
int nonempty(struct tree *);
struct tree *find_modification_trigger(struct tree_list *, struct tree *);
struct tree *find_conflation_trigger(struct tree_list *, struct tree *);
struct tree *get_trigger(struct tree *, struct tree *, int );
struct tree *get_nonlex_trigger(struct tree *);
struct tree *skip_modifications(struct tree *); 
struct tree_list *skip_conflations(struct tree_list *);
struct tree *nonconflation_parent(struct tree *); 

void display_triggers(int );
void display_trig(struct triggering *);
void display_site(struct tree *, struct tree *);
void display_backward(struct tree_list *);
void display_labels(int );
void display_labs(struct triggering *);
void display_tags(int );

void display_outputs(struct tree *);
struct tree_list *display_position_outputs(int , struct tree_list *);
struct tree_list *adjoin_site(struct tree_list *, struct tree *);
struct tree_list *display_end_outputs(int , struct tree_list *, struct tree *);

void transform_brackets(FILE *);
void simplify_brackets(FILE *);
void compute_brackets(FILE *);
void compute_inputs(FILE *);
struct tree *read_outputs(FILE *, int );
void expand_conflations(struct tree *);

void display_tree(struct tree *, int);
int display_subtree(struct tree *, char *, int , int);

struct tree *create_tree();
void free_tree(struct tree *);
void free_all_tree(struct tree *);
struct tree_list *create_tree_list(struct tree *, struct tree_list *);
void free_tree_list(struct tree_list *);
void free_all_tree_list(struct tree_list *);
void pluckfree_tree_list(struct tree_list *);
struct tree_list *popfree_tree_list(struct tree_list *);
struct triggering *create_triggering();
void clear_triggering(struct triggering *);
void init_triggering(struct triggering *);
void free_triggering(struct triggering *);
void free_projection(struct triggering *);

int find_str(char *, char **);

void usage_error(char *);
void error(char *);



/** CODE **/

/* Parse the command line and do the right thing. */
int main(int argc, char *argv[])
{
  int nextarg;
  char data_filename[FILENAME_MAX];
  char output_filename[FILENAME_MAX];
  char words_filename[FILENAME_MAX];
  FILE *datafp;
  FILE *wordsfp;

  int errno = 0;
  nextarg = 1;

  if (argc < 2)  
    usage_error(argv[0]);

  if (strcmp(argv[nextarg], "-h") == 0)
    usage_error(argv[0]);

  if (strcmp(argv[nextarg], "-d2o") == 0 ||
      *(argv[nextarg]) != '-')
    {
      if (*(argv[nextarg]) == '-')
	nextarg++;
      strcpy(data_filename, argv[nextarg]);
      nextarg++;
      if ((datafp=fopen(data_filename, "r")) == NULL)
	error("Error: can't open data file.");
      strcpy(words_filename, data_filename);
      strcat(words_filename, ".words");
      if ((wordsfp=fopen(words_filename, "w")) == NULL)
	error("Error: can't open words file to write.");
      compute_triggers(datafp, wordsfp);
      fclose(wordsfp);
      fclose(datafp);
    }

  else if (strcmp(argv[nextarg], "-d2b") == 0)
    {
      nextarg++;
      strcpy(data_filename, argv[nextarg]);
      nextarg++;
      if ((datafp=fopen(data_filename, "r")) == NULL)
	error("Error: can't open data file.");
      transform_brackets(datafp);
      fclose(datafp);
    }

  else if (strcmp(argv[nextarg], "-d2g") == 0)
    {
      nextarg++;
      strcpy(data_filename, argv[nextarg]);
      nextarg++;
      if ((datafp=fopen(data_filename, "r")) == NULL)
	error("Error: can't open data file.");
      simplify_brackets(datafp);
      fclose(datafp);
    }

  else if (strcmp(argv[nextarg], "-o2b") == 0)
    {
      nextarg++;
      strcpy(data_filename, argv[nextarg]);
      nextarg++;
      if ((datafp=fopen(data_filename, "r")) == NULL)
	error("Error: can't open data file.");
      compute_brackets(datafp);
      fclose(datafp);
    }

  else if (strcmp(argv[nextarg], "-o2d") == 0)
    {
      nextarg++;
      strcpy(data_filename, argv[nextarg]);
      nextarg++;
      if ((datafp=fopen(data_filename, "r")) == NULL)
	error("Error: can't open data file.");
      compute_inputs(datafp);
      fclose(datafp);
    }


  return(errno);
}


void compute_triggers(FILE *data, FILE *words)
{
  struct tree *target;
  int i;
  
#if DEBUG_INP
  printf("\n");
#endif

  initialize_triggers();
  sentence_number = 1;
  while ((target = read_sentence(data)) != NULL)
    {
      remove_empty(target);
      sentence_untriggered = NULL; 
      for (i = 0; i <= target->end; i++)
	{
	  clear_triggering(lexical_triggerings[i]);
	  init_triggering(nonlexical_triggerings[i]);
	}
      set_triggers(target);
#if DEBUG_STRUCMOD
      display_tree(target, 0);
#endif
      relabel_conflations(target);
      set_heads(target);
#if DISPTREE
      display_tree(target, 0);
#endif
#if ONLYLABELS
      display_labels(target->end);
#else
#if ONLYTAGS
      display_tags(target->end);
#else
#if DISPRULES
      display_triggers(target->end);
#endif
#if DISPOUTPUT
      display_outputs(target);
#endif
#endif
#endif
#if FILEWORDS
      for (i = 0; i < target->end; i++)
	{
	  fprintf(words, "%s ", sentence_words[i]);
	}
      fprintf(words, "%s\n", sentence_words[target->end]);
#endif

      free_tree_list(sentence_untriggered);
      for (i = 0; i <= target->end; i++)
	{
	  free_projection(lexical_triggerings[i]);
	  free_projection(nonlexical_triggerings[i]);
	}
      free_all_tree(target);
#if DEBUG
      printf("Done sentence %d.\n", sentence_number); 
#endif
      sentence_number++;
      /* printf("\n"); */
    }

}


void transform_brackets(FILE *data)
{
  struct tree *target;
  int i;
  
#if DEBUG_INP
  printf("\n");
#endif

  initialize_triggers();
  sentence_number = 1;
  while ((target = read_sentence(data)) != NULL)
    {
      remove_empty(target);
      sentence_untriggered = NULL; 
      for (i = 0; i <= target->end; i++)
	{
	  clear_triggering(lexical_triggerings[i]);
	  init_triggering(nonlexical_triggerings[i]);
	}
      set_triggers(target);
      relabel_conflations(target);
      set_heads(target);

      display_tree(target, 0);

      free_tree_list(sentence_untriggered);
      for (i = 0; i <= target->end; i++)
	{
	  free_projection(lexical_triggerings[i]);
	  free_projection(nonlexical_triggerings[i]);
	}
      free_all_tree(target);
#if DEBUG
      printf("Done sentence %d.\n", sentence_number); 
#endif
      sentence_number++;
      /* printf("\n"); */
    }

}


void simplify_brackets(FILE *data)
{
  struct tree *target;
  int i;
  
#if DEBUG_INP
  printf("\n");
#endif

  initialize_triggers();
  sentence_number = 1;
  while ((target = read_sentence(data)) != NULL)
    {
      remove_empty(target);
      display_tree(target, 1);
      free_all_tree(target);
#if DEBUG
      printf("Done sentence %d.\n", sentence_number); 
#endif
      sentence_number++;
      /* printf("\n"); */
    }

}


void compute_brackets(FILE *data)
{
  struct tree *result;
  int i;
  
#if DEBUG_READOUT
  printf("\n");
#endif

  sentence_number = 1;
  while ((result = read_outputs(data, 0)) != NULL)
    {
      display_tree(result, 0);

      free_all_tree(result);
#if DEBUG
      printf("Done sentence %d.\n", sentence_number); 
#endif
      sentence_number++;
      /* printf("\n"); */
    }
}


void compute_inputs(FILE *data)
{
  struct tree *target;
  struct tree_list *tl;
  int i;
  
#if DEBUG_INP
  printf("\n");
#endif

  initialize_triggers();
  sentence_number = 1;
  while ((target = read_outputs(data, 1)) != NULL) /* read into tree format */
    {
      if (target->children != NULL)
	{
	  if (target->children->next != NULL)
	    {
	      for (tl = target->children; tl->next->next != NULL; 
		   tl = tl->next)
		{}
	      if (strcmp(tl->next->elem->label, "^^^") == 0)
		{   /* remove initializing terminal */
		  free_tree(tl->next->elem);
		  free(tl->next);
#if DEBUG_ALLOC
		  printf("\n%d tree list freed\n", tl->next);
#endif
		  tl->next = NULL;
		  target->start++;
		}
	    }
	  else if (strcmp(target->children->elem->label, "^^^") == 0)
	    {   /* remove initializing terminal */
	      free_tree(target->children->elem);
	      free(target->children);
#if DEBUG_ALLOC
	      printf("\n%d tree list freed\n", target->children);
#endif
	      target->children = NULL;
	      target->start++;
	    }
	}
      display_tree(target, 0);
      free_all_tree(target);
#if DEBUG
      printf("Done sentence %d.\n", sentence_number); 
#endif
      sentence_number++;
      /* printf("\n"); */
    }
}


void initialize_triggers()
{
  int i;

  for (i = 0; i <= MAXSENTLENGTH; i++)
    {
      lexical_triggerings[i] = create_triggering();
      nonlexical_triggerings[i] = create_triggering();
    }
}


/* modify root to remove empty branches.  Return 1 if root is non-empty, 0
   if empty.  frees empty subtrees, except root. */
int remove_empty(struct tree *root)
{
  int res;
  struct tree_list *tl, *ntl;

  res = 0;
  if (root->children == NULL) /* preterm case */
    {
      if (nonempty(root) != 0)
	res = 1;
    }
  else
    {
      for (tl = root->children; tl != NULL; tl = ntl)
	{
	  ntl = tl->next;
	  if (remove_empty(tl->elem) == 0)  /* recursive call */
	    {
	      if (root->children == tl)
		root->children = ntl;
	      free_tree(tl->elem);
	      pluckfree_tree_list(tl);
	    }
	  else
	    res = 1;
	}
    }

  return(res);
}


/* relabel root's children according to conflations.  Changes root's
   adjunction_child to skip conflations.  set_triggers must have already
   run.  Modifications already removed. */ 
void relabel_conflations(struct tree *root)
{
  struct tree_list *tl;
  struct tree *child, *confl;

  if (root->children != NULL) {
    for (tl = root->children; tl != NULL; tl = tl->next) {
      child = tl->elem;
      relabel_conflations(child);

      if (child->conflation != 0) {
	if (child->children == NULL)
	  printf("\nWarning: conflation with preterm %d:%2d.\n",
		 sentence_number, child->start);
	else {
	  if (child->children->next != NULL)
	    printf("Warning: conflation with multiple children %d:%2d.\n",
		   sentence_number, child->start);
	  confl = child->children->elem;
#if DEBUG_STRUCMOD
	  printf("relabelling conflation (%s (%s (%s -> ", 
		 root->label, child->label, confl->label);
#endif
	  strcat(child->label, "-");
	  strcat(child->label, confl->label);
	  /* was used to change tree *
	     tl->elem = confl;
	     confl->parent = child->parent;
	     */
#if DEBUG_STRUCMOD
	  printf("(%s (%s   (%d:%2d-%2d).\n", 
		 root->label, child->label, sentence_number, child->start,
		 child->end); 
#endif

	  strcpy(confl->label, child->label);
	  confl->modifier = child->modifier; /* propagate modifier flag */
	  /* propogate effects to lower conflations */
	  while (confl->conflation != 0) {
	    if (confl->children == NULL) {
	      printf("\nWarning: conflation with preterm %d:%2d.\n",
		     sentence_number, child->start);
	      confl = NULL;
	    }
	    else {
	      if (confl->children->next != NULL)
		printf("Warning: conflation with multiple children %d:%2d.\n",
		       sentence_number, child->start);
	      confl = confl->children->elem;
	      strcpy(confl->label, child->label);
	      confl->modifier = child->modifier; /* propagate modifier flag */
	    }
	  }
	}
      }
    }
    if (root->adjunction_child != NULL &&
	root->adjunction_child->elem->conflation != 0)
      {
#if DEBUG_STRUCMOD
	printf("  skipping conflations for adjunction.\n");
#endif
	root->adjunction_child = skip_conflations(root->adjunction_child);
      }
  }
}


/* if root is a modification then remove it from the tree and attach its
   other children to its modification child.  trigger child is
   marked as a modifier. 
   root is not removed from trigger lists and not freed.  set_triggers must
   have already run on children.  Conflations have not yet been
   "relabeled". */
struct tree *remove_modification(struct tree *root)
{
  struct tree_list *tl;
  struct tree_list *leftch;
  struct tree *mod;
  struct tree *res;
  struct tree_list *projtl;

  res = root;

  if (root->modification_child != NULL) 
    {
      leftch = root->children;
      while (leftch->next != NULL)
	leftch = leftch->next;
      if (root->parent == NULL ||
	  leftch->elem != root->modification_child->elem ||
	  leftch->prev == NULL ||
	  leftch->prev->elem != root->trigger_child->elem ||
	  leftch->elem->children == NULL)
	{
	  printf("\nWarning: *** modification in incorrect configuration %d:%2d\n", 
		 sentence_number, root->start);
	}
      else 
	{
	  mod = root->modification_child->elem;
	  if (mod->conflation != 0)
	    {
	      printf("Warning: obselete part of code");
	    }
	  else
	    {
#if DEBUG_STRUCMOD
	      printf("removing modification (%s (%s (%s) (%s -> ", 
		     root->parent->label, root->label, mod->label,
		     root->trigger_child->elem->label);
#endif
	      /* mark trigger child as modifier */
	      root->trigger_child->elem->modifier = 1;
	      /* append children */
	      for (tl = root->trigger_child; tl != NULL; tl = tl->prev)
		tl->elem->parent = mod;
	      root->trigger_child->next = mod->children;
	      mod->children->prev = root->trigger_child;
	      mod->children = root->children;
	      mod->end = root->end;
	      /* remove root from tree */
	      tl = root->parent->children;
	      while (tl->elem != root)
		tl = tl->next;
	      tl->elem = mod;
	      mod->parent = root->parent;
	      res = mod;
#if DEBUG_STRUCMOD
	      printf("(%s (%s +(%s  (%d:%2d-%2d).\n", 
		     mod->parent->label, mod->label,
		     root->trigger_child->elem->label, 
		     sentence_number, root->start, root->end);
#endif
	    }
	}
    }

  return(res);
}


/* mark all heads in tree. */
void set_heads(struct tree *root)
{
  struct tree_list *ch;
  struct tree *headcht;
  int rootind;
  int i, j;
  char *cat;
  char root_label[MAXLABEL + 1];
  char child_label[MAXLABEL + 1];

  if (root->children == NULL)    /* preterms have no head */
    return;

  /* children of conflations are only heads if their conflation projections
     are */
  if (root->conflation != 0) 
    {
      root->children->elem->head = root->head;
    }
  else {
    /* find the lowest conflation label of root */
    j = 0;
    for (i = 0; root->label[i] != '\0'; i++)
      {
	root_label[j] = root->label[i];
	if (root->label[i] == '-')
	  j = 0;
	else
	  j++;
      }
    root_label[j] = '\0';

    headcht = NULL;
    /* matrix root has first child as head */
    if (strcmp(root_label, "") == 0) 
      {
	headcht = root->children->elem;
      }
    /* non-NP case handled with table */
    else if (strcmp(root_label, "NP") != 0 &&
	     strcmp(root_label, "NX") != 0) 
      {
	rootind = find_str(root_label, head_projections);
	if (head_projections[rootind] == NULL)
	  {
	    printf("label = %s from %s", root_label, root->label);
	    error("Error: root label not found in head_projections[].");
	  }
	for (i = 0; head_categories[rootind][i] != NULL; i++)
	  {
	    cat = head_categories[rootind][i];
	    for (ch = root->children; ch != NULL; ch = ch->next)
	      {
		/* find the first conflation label of child */
		strcpy(child_label, ch->elem->label);
		strtok(child_label, "-");
		if (strcmp(child_label, cat) == 0)
		  {
		    if (ch->elem->modifier == 0) 
		      {
			headcht = ch->elem;
			if (head_directions[rootind] == 0) 
			  break;  /* then stop at rightmost */
		      }
		  }
	      }
	    if (headcht != NULL)
	      break;
	  }
	/* default case */
	if (headcht == NULL) {
	  for (ch = root->children; ch != NULL; ch = ch->next)
	    {
	      if (ch->elem->modifier == 0) 
		{
		  headcht = ch->elem;
		  if (head_directions[rootind] == 0) 
		    break;  /* then stop at rightmost */
		}
	    }	
	}		
	if (headcht == NULL) {
	  headcht = root->children->elem;
	}
      }
    /* NP case (not exactly the same as Collins' rules) */
    else 
      {
	for (ch = root->children; ch != NULL; ch = ch->elem->children)
	  {
	    if (ch->elem->children == NULL &&  /* rightmost preterminal */
		strcmp(ch->elem->label, "POS") == 0 &&
		root->children->elem->modifier == 0)
	      headcht = root->children->elem;  /* can't conflate with
						  preterms */ 
	  }
	if (headcht == NULL) {
	  for (ch = root->children; ch != NULL; ch = ch->next)
	    {
	      /* find the first conflation label of child */
	      strcpy(child_label, ch->elem->label);
	      strtok(child_label, "-");
	      if (strcmp(child_label, "NN") == 0 ||
		  strcmp(child_label, "NNP") == 0 ||
		  strcmp(child_label, "NNPS") == 0 ||
		  strcmp(child_label, "NNS") == 0 ||
		  strcmp(child_label, "NX") == 0 ||
		  strcmp(child_label, "POS") == 0 ||
		  strcmp(child_label, "JJR") == 0)
		{
		  if (ch->elem->modifier == 0) 
		    {
		      headcht = ch->elem;
		      break;
		    }
		}
	    }
	}
	if (headcht == NULL) {
	  for (ch = root->children; ch != NULL; ch = ch->next)
	    {
	      /* find the first conflation label of child */
	      strcpy(child_label, ch->elem->label);
	      strtok(child_label, "-");
	      if (strcmp(child_label, "NP") == 0)
		{
		  if (ch->elem->modifier == 0) 
		    {
		      headcht = ch->elem;
		    }
		}
	    }	
	}		
	if (headcht == NULL) {
	  for (ch = root->children; ch != NULL; ch = ch->next)
	    {
	      /* find the first conflation label of child */
	      strcpy(child_label, ch->elem->label);
	      strtok(child_label, "-");
	      if (strcmp(child_label, "$") == 0 ||
		  strcmp(child_label, "ADJP") == 0 ||
		  strcmp(child_label, "PRN") == 0)
		{
		  if (ch->elem->modifier == 0) 
		    {
		      headcht = ch->elem;
		      break;
		    }
		}
	    }	
	}		
	if (headcht == NULL) {
	  for (ch = root->children; ch != NULL; ch = ch->next)
	    {
	      /* find the first conflation label of child */
	      strcpy(child_label, ch->elem->label);
	      strtok(child_label, "-");
	      if (strcmp(child_label, "CD") == 0)
		{
		  if (ch->elem->modifier == 0) 
		    {
		      headcht = ch->elem;
		      break;
		    }
		}
	    }	
	}		
	if (headcht == NULL) {
	  for (ch = root->children; ch != NULL; ch = ch->next)
	    {
	      /* find the first conflation label of child */
	      strcpy(child_label, ch->elem->label);
	      strtok(child_label, "-");
	      if (strcmp(child_label, "JJ") == 0 ||
		  strcmp(child_label, "JJS") == 0 ||
		  strcmp(child_label, "RB") == 0 ||
		  strcmp(child_label, "QP") == 0)
		{
		  if (ch->elem->modifier == 0) 
		    {
		      headcht = ch->elem;
		      break;
		    }
		}
	    }	
	}		
	/* default case */
	if (headcht == NULL) { 
	  for (ch = root->children; ch != NULL; ch = ch->next)
	    {
	      if (ch->elem->modifier == 0) 
		{
		  headcht = ch->elem;
		  break;
		}
	    }	
	}		
	if (headcht == NULL) {
	  headcht = root->children->elem;
	}
      }
    if (headcht == NULL)
      error("can't find head.");
    /* set head flag */
    headcht->head = 1;
  }

  /* recursive calls */
  for (ch = root->children; ch != NULL; ch = ch->next)
    set_heads(ch->elem);
}



/* read in a sentence from data file pointer */
struct tree *read_sentence(FILE *data)
{
  struct tree_list *stack;
  struct tree *last;
  char buffer[MAXDATALINE];
  char *inpp;
  int position;
  int line; 

  stack = NULL;
  last = NULL;
  position = 0;
  line = -1;
  while (fgets(buffer, MAXDATALINE, data) != NULL)
    {
      line++;
#if DISPINP
      printf(" %5d:%3d.%1d %s", sentence_number, position, line, buffer);
#endif
      inpp = buffer;
      while (inpp[0] != '\n')
	{
	  if (inpp >= buffer+MAXDATALINE)
	    error("Error: no newline found or line too long.");

	  inpp = skip_space(inpp);
	  if (inpp[0] == '\n')
	    break;  /* go get new line */
	  else if (inpp[0] == '(')
	    {
	      if (stack == NULL)
		{
		  stack = start_root(position, stack);
		  position++;
		}
	      else
		stack = start_const(position, stack);
	      inpp++;
	      inpp = read_nonterm(inpp, stack);
	    }
	  else if (inpp[0] == ')')
	    {
	      last = stack->elem;
	      stack = end_const(position, stack);
	      inpp++;
	    }
	  else
	    {
	      inpp = read_terminal(inpp, position);
	      if (nonempty(stack->elem) != 0) /* check if this preterm empty */
		{
		  position++;
		  line = -1;
		}
	    }
	}
      if (stack == NULL && last != NULL)
	break;
    }
  return(last);
}
	  

/* add a new tree to stack with start position, and add it to the children
   of the previous tree in stack, if one exists */
struct tree_list *start_const(int position, struct tree_list *stack)
{
  struct tree_list *res;
  struct tree *newtree;
  struct tree_list *newchildren;

  newtree = create_tree();
  newtree->start = position;
  if (stack != NULL)
    { /* add tree to parent's children (in reverese order) */
      newtree->parent = stack->elem;
      newchildren = create_tree_list(newtree, stack->elem->children);
      stack->elem->children = newchildren;
    }

  res = create_tree_list(newtree, stack);
  return(res);
}

/* create a start preterm and add a root tree to stack with start position,
   and add root to the children of the previous tree in stack, if one exists */
struct tree_list *start_root(int position, struct tree_list *stack)
{
  struct tree_list *res;
  struct tree *newtree;
  struct tree_list *newchildren;
  struct tree *preterm;

  newtree = create_tree();
  newtree->start = position;

  preterm = create_tree();
  preterm->start = position;
  strcpy(preterm->label, "^^^");
  strcpy(sentence_words[position], "^^^");
  preterm->end = position + 1;

  preterm->parent = newtree;
  newchildren = create_tree_list(preterm, NULL);
  newtree->children = newchildren;

  if (stack != NULL)
    { /* add tree to parent's children (in reverese order) */
      newtree->parent = stack->elem;
      newchildren = create_tree_list(newtree, stack->elem->children);
      stack->elem->children = newchildren;
    }

  res = create_tree_list(newtree, stack);
  return(res);
}

/* set label of top tree in stack and return new input pointer */
char *read_nonterm(char *inputp, struct tree_list *stack)
{
  char *res;
  char *cpp;

  res = inputp;
  while (res[0] != '-' && res[0] != '=' && res[0] != '|' && 
	 res[0] != ')' && res[0] != '(' && 
	 res[0] != ' ' && res[0] != '\t' && res[0] != '\n')
    {
#if DEBUG_INP
      printf("n:%c ", res[0]);
#endif
      res++;
    }
  if (res == inputp && res[0] == '-' && res[1] != 'N') 
    {                                         /* to handle -LRB and -RRB */ 
#if DEBUG_INP
      printf("n:%c ", res[0]);
#endif
      res++;
      while (res[0] != '-' && res[0] != '=' && res[0] != '|' && 
	     res[0] != ')' && res[0] != '(' && 
	     res[0] != ' ' && res[0] != '\t' && res[0] != '\n')
	{
#if DEBUG_INP
	  printf("n:%c ", res[0]);
#endif
	  res++;
	}
    }
  
  if ((res - inputp) >= MAXWORDLENGTH)
    error("Error: Nonterminal/Preterminal label too long.");

  for (cpp = inputp; cpp < res; cpp++)
    stack->elem->label[(int) (cpp - inputp)] = cpp[0];
  stack->elem->label[(int) (res - inputp)] = '\0';

  /* skip extra features */
  while (res[0] != ')' && res[0] != '(' && 
	 res[0] != ' ' && res[0] != '\t' && res[0] != '\n')
    {
#if DEBUG_INP
      printf("f:%c ", res[0]);
#endif
      res++;
    }

  return(res);
}

/* pop top tree from stack, set its end position. */
struct tree_list *end_const(int position, struct tree_list *stack)
{
  struct tree_list *res;

  if (stack == NULL)
    error("Error: Bad data format, too many )'s.");

  stack->elem->end = position;
  res = popfree_tree_list(stack);

  return(res);
}

/* set position's value in sentence_words and return new input pointer */
char *read_terminal(char *inputp, int position)
{
  char *res;
  char *cpp;

  res = inputp;
  while (res[0] != ')' && res[0] != '(' && res[0] != ' ' && res[0] != '\t' 
	 && res[0] != '\n')
    {
#if DEBUG_INP
      printf("t:%c ", res[0]);
#endif
      res++;
    }
  
  if ((res - inputp) >= MAXWORDLENGTH)
    error("Error: Terminal word too long.");

  for (cpp = inputp; cpp < res; cpp++)
    sentence_words[position][(int) (cpp - inputp)] = cpp[0];
  sentence_words[position][(int) (res - inputp)] = '\0';

  return(res);
}


/* return new input pointer to next nonspace character */
char *skip_space(char *inputp)
{
  char *res;

  res = inputp;
  while (res[0] == ' ' || res[0] == '\t')
    {
#if DEBUG_INP
      printf("s:%c ", res[0]);
#endif
      res++;
    }

  return(res);
}



/* set values for trigger events in root in lexical_triggerings and
   nonlexical_triggerings.  removes modifications from tree, but adds them
   to projections.  conflations are left in tree but not added to
   projections. */   
void set_triggers(struct tree *root)
{
  struct tree_list *ch;
  struct tree_list *leftch;
  struct tree_list *secondch;
  struct tree *cht;
  struct tree *trig;
  struct tree_list *trigch;
  struct tree_list *adch;
  struct tree_list *mdch;
  struct tree *atdom;
  struct tree *nontrig;
  int cnflag;
  int topflag;
  char *nttrigger;
  struct tree *adtrig, *mdtrig, *attrig, *extrig, *cntrig, *toptrig, *tadtrig;

  leftch = root->children;
  if (leftch != NULL)
    while (leftch->next != NULL)
      leftch = leftch->next;   /* set leftch to leftmost child */

  if (root->children == NULL)    /* preterm root */
    {
      lexical_triggerings[root->start]->preterm = root;
      root->trigger = root;
      nontrig = nonlexical_triggerings[root->end]->preterm;
      strcpy(nontrig->label, NONTRIGGER);
      strcat(nontrig->label, root->label);
      nontrig->parent = root->parent;
      nontrig->start = root->end;
      nontrig->end = root->end;
      nontrig->trigger = nontrig;
      nontrig->toplexical = 1;
    } 
  else {                                /* nonterm root */

    secondch = leftch->prev;
    trig = NULL;
    trigch = NULL;
    adch = NULL;
    mdch = NULL;
    atdom = NULL;
    cnflag = 0;
    topflag = 0;

    /* set child triggers and check each child for a trigger */

    set_triggers(leftch->elem);  /* recursive call, left child */

    if (leftch->elem->children != NULL) /* nonterm left child */
      {
	if (secondch == NULL)
	  {
#if CONFLATIONS
	    /* conflation */
	    cntrig = find_conflation_trigger(leftch, root);
	    if (cntrig != NULL)
	      {
		trig = cntrig;
		trigch = leftch;
		adch = NULL;
		mdch = NULL;
		atdom = NULL;
		cnflag = 1;
		cht = leftch->elem;
		topflag = cht->toplexical; /* inherit lexical blocking */
	      }
#endif
	  }
	else
	  {
	    set_triggers(secondch->elem);  /* recursive call, second child */
	    
#if MODOFOCATIONS
	    /* modification */ 
	    mdtrig = find_modification_trigger(secondch, root);
	    if (mdtrig != NULL)
	      {
		trig = mdtrig;
		trigch = secondch;
		adch = NULL;
		mdch = leftch;  /* assumes modifications always adjacent */
		atdom = NULL;
		cnflag = 0;
		topflag = 0;
	      }
#endif
	  }
      }
    /* defaults */
    if (trig == NULL)
      {
	if (leftch->elem->children != NULL)
	  {
	    /* default to nonlexical trigger if nonterm child */
	    trig = get_nonlex_trigger(leftch->elem);
	    trigch = leftch;
	    adch = NULL;
	    mdch = NULL;
	    atdom = NULL;
	    cnflag = 0;
	      topflag = 0;
	  }
	else
	  {
	    /* default to lexical trigger if preterm child */
	    trig = leftch->elem;
	    trigch = leftch;
	    adch = NULL;
	    mdch = NULL;
	    atdom = NULL;
	    cnflag = 0;
	      topflag = 0;
	  }
      }
    
    /* set triggers of other children, left to right */
    if (secondch != NULL)
      {
	if (secondch->elem->trigger == NULL) /* maybe done above */
	  set_triggers(secondch->elem);  /* recursive call, second child */
	for (ch = secondch->prev; ch != NULL; ch = ch->prev) 
	  set_triggers(ch->elem);  /* recursive call, other children */ 
      }

    /* if a trigger was found, set it */
    if (trig != NULL)
      {
	root->trigger = trig;
	root->trigger_child = trigch;
	root->adjunction_child = adch;
	root->modification_child = mdch;
	root->attachment = atdom;
	root->conflation = cnflag;
	root->toplexical = topflag;
	if (root->conflation == 0) 
	  {      /* conflations aren't new projection levels */
	    if (root->trigger->toplexical != 0)
	      nonlexical_triggerings[root->trigger->end]->projection = 
		create_tree_list(root, 
	          nonlexical_triggerings[root->trigger->end]->projection);
	    else
	      lexical_triggerings[root->trigger->start]->projection = 
		create_tree_list(root, 
	          lexical_triggerings[root->trigger->start]->projection);
	  }
	/* if a modification, then remove from tree */
	root = remove_modification(root); 
      }
    else
      {
	printf("****Warning: Current version assumes everything triggered by something.  %4d:%2d-%2d\n", sentence_number, root->start, root->end);
	sentence_untriggered = create_tree_list(root, sentence_untriggered);
      }

  }

#if DEBUG_TRIG
  printf("set_trigger: top=%d ", root->toplexical);
  if (root->trigger->toplexical)
    display_trig(nonlexical_triggerings[root->trigger->end]);
  else
    display_trig(lexical_triggerings[root->trigger->start]);
#endif
}

/* test if tr has any non-empty terminals under it. */
int nonempty(struct tree *tr)
{
  int res;
  struct tree_list *ch;

  res = 0;
  if (tr->children == NULL)  /* preterm case */
    {
      if ((tr->label)[0] != '\0')
	res = 1;
    }
  else  /* nonterm case */
    for (ch = tr->children; ch != NULL; ch = ch->next)
      {
	if (nonempty(ch->elem) != 0) 
	  {
	    res = 1;
	    break;
	  }
      }

  return(res);
}


/* find a modification trigger for root in nonterminal children->elem, or
   return NULL.  assumes set_triggers has already run on children->elem. */
struct tree *find_modification_trigger(struct tree_list *children, 
				       struct tree *root)
{
  struct tree *proj;
  struct tree *site;
  struct tree *res;
  int proj_ind;
  int modifica_ind;

  res = NULL;

  if (children != NULL && 
      children->next != NULL)  /* modifications need a site */
    {
      proj = children->elem;
      site = children->next->elem;
      /* modifications must have the same label on the root and site, and
	 the site cannot be a conflation */
      if (strcmp(root->label, site->label) == 0 &&
	  site->conflation == 0)
	{
	  proj_ind = find_str(proj->label, modification_projections);
	  if (modification_projections[proj_ind] != NULL && 
	      proj->trigger != NULL)
	    {
	      modifica_ind = find_str(site->label, 
				      modification_sites[proj_ind]);
	      if (modification_sites[proj_ind][modifica_ind] != NULL)
		{
		  res = get_trigger(proj, site, 0); /* not blocked by toplex */ 
		}
	    }
	}
    }

  return(res);
}


/* find a conflation trigger for root in nonterminal children->elem, or
   return NULL.  assumes set_triggers has already run on children->elem. */
struct tree *find_conflation_trigger(struct tree_list *children, 
				     struct tree *root)
{
  struct tree *proj;
  struct tree *res;
  int proj_ind;
  int confl_ind;

  res = NULL;

  /* children->elem must be the only child and a nonterminal */
  if (children != NULL &&
      children->next == NULL &&
      children->prev == NULL &&
      children->elem->children != NULL) 
    {
      proj = children->elem;
      proj_ind = find_str(proj->label, conflation_projections);
      if (conflation_projections[proj_ind] != NULL)
	{
	  confl_ind = 
	    find_str(root->label, conflation_sites[proj_ind]);
	  if (conflation_sites[proj_ind][confl_ind] != NULL)
	    {
	      res = get_trigger(proj, NULL, 0); /* not blocked by toplexical
						 */  
	    }
	}

    }

  return(res);
}



/* return the lexical or nonlexical trigger for projecting from proj, given
   adjunsite if this is an adjunction.  blockflag != 0 if projection blocked
   by toplex feature. */
struct tree *get_trigger(struct tree *proj, struct tree *adjunsite, 
			 int blockflag)
{
  struct tree *res;
  struct tree *projroot;

#if DEBUG_TRIG
  printf("get_trigger: %s *:%d  %4d:%3d\n", 
	 proj->label, (adjunsite == NULL), sentence_number, proj->start);
#endif

  projroot = skip_modifications(proj); /* necessary anymore? */

  res = projroot->trigger;

  return(res);
}

/* return the nonlexical trigger for the last preterm in proj. */
struct tree *get_nonlex_trigger(struct tree *proj)
{
  struct tree *res;
  struct tree *projroot;

#if DEBUG_TRIG
  printf("get_nonlex_trigger: %s *:%d  %4d:%3d\n", 
	 proj->label, (proj->adjunction_child == NULL), sentence_number, proj->start);
#endif

  /* projroot = skip_modifications(proj); */

  res = nonlexical_triggerings[proj->end]->preterm;

  return(res);
}

/* skip over modification adjunctions and return projection root */
struct tree *skip_modifications(struct tree *t)
{
  struct tree *res;

  for (res = t; res->modification_child != NULL; 
       res = res->modification_child->elem)
    {}

  return(res);
}

/* skip over conflation projections and return the lowest nonconflation
   ancester of t (or NULL). */
struct tree *nonconflation_parent(struct tree *t)
{
  struct tree *res;

  res = t->parent;
  while (res != NULL && res->conflation != 0)
    res = res->parent;

  return(res);
}

/* find the children list of the first descendent of tl (or tl itself) with
   non-conflation children.. */
struct tree_list *skip_conflations(struct tree_list *tl)
{
  struct tree_list *res;

  res = tl;
  while (res != NULL && res->elem->conflation != 0)
    res = res->elem->children; 

  if (res == NULL)
    {
      printf("\nWarning: ***bottomless conflation. (%d)\n", sentence_number);
    }

  return(res);
}



/* send each triggering event to stdout.  */
void display_triggers(int length)
{
  int i;
  struct tree_list *tl;
  struct tree *t;

  for (i = 1; i < length; i++)  /* skip "^^^" in position 0 */
    {
#if DEBUG_TRIG
      printf("triggers: %3d:%2d\n", sentence_number, i);
#endif
      display_trig(lexical_triggerings[i]);
      display_trig(nonlexical_triggerings[i + 1]);
    }

  for (tl = sentence_untriggered; tl != NULL; tl = tl->next)
    {
      t = tl->elem;
      if (nonempty(t) != 0)
	{
#if DEBUG_TRIG
	  printf("untrigger: %3d:%2d\n", sentence_number, i);
#endif
	  if (t->parent != NULL)
	    {
	      printf("%-3s %-6s  ^%-6s [", 
		     UNTRIGGER, t->label, t->parent->label);
	      display_backward(t->children);
	      printf("] %3d:%2d-%d\n", sentence_number, t->start, t->end);
	    }
	  /*
	    else
	    printf("%-3s root  ^\n", UNTRIGGER);
	  */
	}
    }
}

/* display the trigger projection tgi */
void display_trig(struct triggering *tgi)
{
  struct tree_list *tl;
  struct tree *prj;
  struct tree *maxprj;
  struct tree *t;
  struct tree_list *backprj;
  int numadjun;
  int numlev;
  int levadjun;
  int numconfl;
  int i;
  
  numadjun = 0;
  levadjun = 0;
  numlev = 0;
  numconfl = 0;
  if (tgi->preterm != NULL && nonempty(tgi->preterm) != 0
#if ! DISPALL
      && (tgi->preterm->toplexical == 0 || tgi->projection != NULL)
#endif
      )

    {
      printf("%-6s ", tgi->preterm->label); /* trigger */
      if (tgi->projection != NULL)
	{
	  maxprj = tgi->projection->elem;
	  backprj = tgi->projection;
	  while (backprj->next != NULL)
	    backprj = backprj->next;
	}
      else
	{
	  maxprj = tgi->preterm;
	  backprj = NULL;
	}
      for (tl = backprj; tl != NULL; tl = tl->prev)
	{       /* context */
	  prj = tl->elem;
	  if (prj->adjunction_child != NULL)
	    {
	      numadjun++;
	      numlev++;
	      if (numadjun <= 1)
		levadjun = numlev;
	      else
		printf(" ****(%d)", prj->adjunction_child->elem->start);
#if DISPINFO
	      printf("*%-6s", prj->adjunction_child->elem->label);
#else
	      display_site(prj->adjunction_child->elem, tgi->preterm);
#endif
	    }
	  else if (prj->modification_child != NULL)
	    {  /* modifications printed as parent */
	      if (prj != maxprj)
		{
		  printf("+ ");
		}
	    }
	  else if (prj->attachment != NULL)
	    {
#if DISPINFO
	      printf("|%-6s", " ");
#endif
	    }
	  else if (prj->conflation != 0)
	    {}
	  else if (tgi->preterm->toplexical != 0)
	    { /* nonlexical extensions are like adjunctions, except chaining 
		 is allowed */
	      numadjun++;
	      numlev++;
	      if (numadjun <= 1)
		levadjun = numlev;
#if DISPINFO
	      printf("*%-6s", prj->trigger_child->elem->label);
#else
	      display_site(prj->trigger_child->elem, tgi->preterm);
#endif
	    }
#if DISPCONT
	  else if (prj->trigger_child->prev != NULL)
	    {
	      numlev++;
#if DISPINFO
	      printf(">%-6s", prj->trigger_child->prev->elem->label);
#endif
	    }
#endif
	  else
	    {
	      numlev++;
#if DISPINFO
	      printf("<%-6s", " ");
#endif
	    }
	}
#if DISPINFO
      if (maxprj->modification_child != NULL)
	{
	  printf(" ^+%-6s", maxprj->modification_child->elem->label);
	}
      else if (maxprj->parent != NULL &&
	       maxprj->parent->trigger != NULL &&
	       maxprj->parent->trigger->start < tgi->preterm->start)
	{                                 /* potential attachment */
	  printf(" ^<%-6s", maxprj->parent->label);
	}
#if DISPCONT
      else if (maxprj->parent != NULL)
	{
	  printf(" ^>%-6s", maxprj->parent->label);
	}
#endif
      else
	printf(" ^-%-6s", " ");
      printf(" ->  %-6s", tgi->preterm->label); /* trigger is base of
						   projection */ 
#else
      if (numadjun > 0)
	{}  /* site already printed */
      else if (maxprj->modification_child != NULL)
	{
	  display_site(skip_modifications(maxprj), tgi->preterm);
	}
      else
	{
	  display_site(nonconflation_parent(maxprj), tgi->preterm);
	}
      printf(" ->  ");
      for (i = 1; i <= numlev; i++)
	{
	  if (numadjun > 0 && levadjun == i)
	    {
	      printf("*");
	    }
	  else
	    {
	      printf("|");
	    }
	}
      for (; i <= 3; i++)
	printf(" ");
      if (maxprj->modification_child != NULL)
	{
	  printf("+");
	}
      else
	{
	  printf("^");
	}
      printf(" %d: ", numlev);
#endif
      for (tl = backprj; tl != NULL; tl = tl->prev)
	{       /* projection constituents */
	  if (tl->elem->conflation != 0)
	    {
	      printf("\\%-6s", tl->elem->label);
	      numconfl++;
	    }
	  else if (tl->elem->modification_child == NULL)
	    printf(" %-6s", tl->elem->label);
	}
#if DISPINFO
      for (tl = backprj; tl != NULL; tl = tl->prev)
	{              /* constituent extra left children */
	  if (tl->elem->conflation == 0 && 
	      tl->elem->modification_child == NULL)
	    {
	      printf(" ["); 
	      if (tl->elem->adjunction_child != NULL)
		display_backward(tl->elem->adjunction_child->next);
	      /* else if (tl->elem->modification_child != NULL)
		 display_backward(tl->elem->modification_child->next);
	      */
	      else if (tl->elem->trigger_child != NULL)
		display_backward(tl->elem->trigger_child->next);
	      else
		printf(" ***");
	      printf("]");
	    }
	}
#else
      for (numlev += numconfl; numlev < 3; numlev++)
	printf(" %-6s", " ");
#endif
#if DISPINP
      printf(" %4d:%2d", sentence_number, tgi->preterm->start);
#endif
      printf("\n");
    }
}


/* given an attachment, adjunction or modification to location, display
   labels of the node which were the site and site's parent at
   position-lexical in the derivation.  */
void display_site(struct tree *location, struct tree *trigger)
{
  struct tree *site;
  struct tree *par;

  site = location;
  while (site != NULL && site->trigger != NULL && 
	 site->trigger->start >= trigger->start)
    {
      if (site->trigger->start == trigger->start && 
	  trigger->toplexical == 0 &&
	  site->trigger->toplexical != 0)
	break;  /* nonlexical triggers are before lexical */
      site = site->parent;
    }
  if (site != NULL && site->trigger != NULL)
    {                      
      printf("%-6s", site->label);  /* site label */
    }
  else
    {
      printf("%-6s", "--");
    }

#if DISPPAR
  par = site->parent;
  while (par != NULL && 
	 (par->trigger == NULL || par->trigger->start >= trigger->start))
    {
      if (par->trigger->start == trigger->start && 
	  trigger->toplexical == 0 &&
	  par->trigger->toplexical != 0)
	break;  /* nonlexical triggers are before lexical */
      par = par->parent;
    }
  if (par != NULL && par->trigger != NULL)
    {                      
      printf("^%-6s", par->label);  /* parent label */
    }
  else
    {
      printf("^%-6s", "--");
    }
#endif
}


/* display tl's labels in backward order. */
void display_backward(struct tree_list *tl)
{
  if (tl != NULL)
    {
      display_backward(tl->next);
      printf(" %-6s", tl->elem->label);
    }
}


/* send each nonterminal label to stdout.  */
void display_labels(int length)
{
  int i;
  struct tree_list *tl;
  struct tree *t;

  for (i = 1; i < length; i++)  /* skip "^^^" in position 0 */
    {
#if DEBUG_TRIG
      printf("labels: %3d:%2d\n", sentence_number, i);
#endif
      display_labs(lexical_triggerings[i]);
      display_labs(nonlexical_triggerings[i + 1]);
    }

  for (tl = sentence_untriggered; tl != NULL; tl = tl->next)
    {
      t = tl->elem;
      if (nonempty(t))
	{
#if DEBUG_TRIG
	  printf("untrigger: %3d:%2d\n", sentence_number, i);
#endif
	  if (t->parent != NULL)
	    {
	      printf("%-6s\n", t->label);
	    }
	  /*
	    else
	    printf("%-3s root  ^\n", UNTRIGGER);
	  */
	}
    }
}

/* display the labels for the trigger projection tgi */
void display_labs(struct triggering *tgi)
{
  struct tree_list *tl;
  struct tree_list *backprj;
  int numconfl;


  if (tgi->preterm != NULL && nonempty(tgi->preterm) != 0 &&
      tgi->projection != NULL)
    {
      backprj = tgi->projection;
      while (backprj->next != NULL)
	backprj = backprj->next;

      /* projection constituents */
      tl = backprj;
      if (tl->elem->conflation != 0)
	{
	  printf(" \\%-6s", tl->elem->label);
	}
      else if (tl->elem->modification_child == NULL)
	{
	  printf("%-6s", tl->elem->label);
	}
      /*
      else if (tl->prev != NULL)
	{ * modification with something above? *
	  printf(" ****(%d) %-6s", tl->elem->start, tl->elem->label);
	}
      */
      else
	{
	  return;  /* don't print \n below */
	}
      numconfl = 0;
      for (tl = backprj->prev; tl != NULL; tl = tl->prev)
	{       
	  if (tl->elem->conflation != 0)
	    {
	      printf(" \\%-6s", tl->elem->label);
	      numconfl++;
	    }
	  else if (tl->elem->modification_child == NULL)
	    {
	      for (; numconfl < 2; numconfl++)
		printf("  %-6s", " ");
#if DISPINP
	      printf("  %d %4d:%2d", tgi->preterm->toplexical,
		     sentence_number, tgi->preterm->start);
#endif
	      printf("\n%-6s", tl->elem->label);
	      numconfl = 0;
	    }
	  /*
	  else if (tl->prev != NULL)
	    { * modification with something above? *
	      for (; numconfl < 2; numconfl++)
		printf("  %-6s", " ");
#if DISPINP
	      printf("  %d %4d:%2d", tgi->preterm->toplexical,
		     sentence_number, tgi->preterm->start);
#endif
	      printf("\n error?: above proj = %s(%d)", 
		     tl->prev->elem->label, tl->prev->elem->start);
	      printf("\n ****(%d) %-6s", tl->elem->start, tl->elem->label);
	      numconfl = 0;
	    }
	  */
	}
      for (; numconfl < 2; numconfl++)
	printf("  %-6s", " ");
#if DISPINP
      printf("  %d %4d:%2d", tgi->preterm->toplexical,
	     sentence_number, tgi->preterm->start);
#endif
      printf("\n");
    }
}


/* send each preterminal labels to stdout.  */
void display_tags(int length)
{
  int i;
  struct tree_list *tl;
  struct tree *t;

  for (i = 0; i < length; i++)  /* include "^^^" in position 0 */
    {
#if DEBUG_TRIG
      printf("tags: %3d:%2d\n", sentence_number, i);
#endif
      printf("%-6s\n", lexical_triggerings[i]->preterm->label);
    }
}





/* display all the network outputs for root */
void display_outputs(struct tree *root)
{
  int position;
  struct tree_list *newstack;

  root->site_id = 0;
  next_site_id = 1;
#if LONGOUTPUT
  printf("%3d,%3d :  | %3d = %s\n", 0, -1, root->site_id, root->label);
#endif
  newstack = create_tree_list(root, NULL);

  for (position = 1; position < root->end; position++)
    newstack = display_position_outputs(position, newstack);
#if LONGOUTPUT
  printf("\n.");
#endif
  printf("\n");

}

/* display all the network outputs that occur at position given the right
   frontier stack, plus prediction step for position-1.  returns the right
   frontier for use at position+1 */  
struct tree_list *display_position_outputs(int position,
 					   struct tree_list *stack)
{
  struct tree_list *newstack;
  struct tree *nextsite;
  struct tree *preterm;
  struct tree_list *proj, *backproj;
  struct tree_list *nexttop;
  struct tree *nexttoppar;
  int head;

  if (stack == NULL)
    {
      printf("Warning: ***no parent context %d:%-2d\n", 
	     sentence_number, position);
      return(NULL);
    }
  newstack = stack;
  preterm = lexical_triggerings[position]->preterm;

  proj = lexical_triggerings[position]->projection;
  if (proj != NULL && 
      proj->elem->modification_child != NULL)
    {
      proj = proj->next; /* skip modification */
    }

  /* predict word at position */
  preterm->site_id = next_site_id;
  next_site_id++;
#if LONGOUTPUT
  printf("\n%3d,%3d :  ", position - 1, newstack->elem->site_id);
#endif
  printf("> ");
#if LONGOUTPUT
  printf("%3d = ", preterm->site_id);
#endif
#if IDOUTPUT
  printf("%d %s\n", tag_id(preterm->label), sentence_words[position]);
#else
  printf("%s %s\n", preterm->label, sentence_words[position]);
#endif
  /* lookahead step */
#if LONGOUTPUT
  printf("%3d,%3d>:  ", position, newstack->elem->site_id);
#endif
  newstack = create_tree_list(preterm, newstack);

  /* treat extensions as first nonlexical projection */
  if (proj != NULL)
    {
      backproj = proj;
      while (backproj->next != NULL)
	backproj = backproj->next;
      for (; backproj != proj; backproj = backproj->prev)
	{ /* shouldn't actually be used (?) */
printf("maybe should be used\n");
	  nextsite = backproj->elem;
	  newstack = 
	    display_end_outputs(position, newstack, 
				(skip_conflations(nextsite->trigger_child))->elem);
#if LONGOUTPUT
	  printf("\n%3d,%3d :  ", position, newstack->elem->site_id);
#endif
	  head = newstack->elem->head;
	  newstack = adjoin_site(newstack, nextsite);
	  printf("|");
#if HEADOUTPUT
	  if (head)
	    printf("h");
#endif
	  printf(" ");
#if LONGOUTPUT
	  printf("%3d = ", nextsite->site_id);
#endif
#if IDOUTPUT
	  printf("%d\t", label_id(nextsite->label));
#else
	  printf("%s\t", nextsite->label);
#endif
	}
      nextsite = proj->elem;
      /* end finished sites, if needed  */
      newstack = 
	display_end_outputs(position, newstack, 
			    (skip_conflations(nextsite->trigger_child))->elem);
      /* extension */
#if LONGOUTPUT
      printf("\n%3d,%3d :  ", position, newstack->elem->site_id);
#endif
      head = newstack->elem->head;
      newstack = adjoin_site(newstack, nextsite);
      printf("|");
#if HEADOUTPUT
      if (head)
	printf("h");
#endif
      printf(" ");
#if LONGOUTPUT
      printf("%3d = ", nextsite->site_id);
#endif
#if IDOUTPUT
      printf("%d\t", label_id(nextsite->label));
#else
      printf("%s\t", nextsite->label);
#endif
    }
  
  /* nonlexical projections */
  proj = nonlexical_triggerings[position + 1]->projection;
  if (proj != NULL)
    {
      backproj = proj;
      while (backproj->next != NULL)
	backproj = backproj->next;
      for (; backproj != NULL; backproj = backproj->prev)
	{
	  if (backproj->elem->modification_child == NULL)
	    {
	      nextsite = backproj->elem;
	      /* end finished sites */
	      newstack = 
		display_end_outputs(position, newstack, 
				    (skip_conflations(nextsite->trigger_child))->elem);
	      /* nonlexical projection */
#if LONGOUTPUT
	      printf("\n%3d,%3d :  ", position, newstack->elem->site_id);
#endif
	      head = newstack->elem->head;
	      newstack = adjoin_site(newstack, nextsite);
	      printf("|");
#if HEADOUTPUT
	      if (head)
		printf("h");
#endif
	      printf(" ");
#if LONGOUTPUT
	      printf("%3d = ", nextsite->site_id);
#endif
#if IDOUTPUT
	      printf("%d\t", label_id(nextsite->label));
#else
	      printf("%s\t", nextsite->label);
#endif
	    }
	}
    }

  /* end finished sites */
  nexttop = newstack;
  while (nexttop != NULL && nexttop->elem->end <= position + 1)
    {
      nexttoppar = nonconflation_parent(nexttop->elem);
      if (nexttoppar == NULL ||
	  ((nexttoppar->adjunction_child == NULL ||
	    nexttoppar->adjunction_child->elem != nexttop->elem) &&
	   (nexttoppar->trigger->toplexical == 0 ||
	    (skip_conflations(nexttoppar->trigger_child))->elem != 
	     nexttop->elem))) /* added skip_... without understanding */
	nexttop = nexttop->next;
      else 
	break;
    }
  if (nexttop != NULL)
    newstack = display_end_outputs(position, newstack, nexttop->elem);
  else
    newstack = display_end_outputs(position, newstack, NULL);

  return(newstack);
}


/* pop stack once (element freed), add site to stack, set its site_id,
   return new stack. */
struct tree_list *adjoin_site(struct tree_list *stack, struct tree *site)
{
  struct tree_list *newstack;
  
  if (site->adjunction_child != NULL && 
      site->adjunction_child->elem != stack->elem)
    {
      printf("\nWarning: *** bad adjunction %d:%-2d\n", 
	     sentence_number, stack->elem->end);
      printf("top:%d, adj:%d\n", stack->elem->site_id, 
	     site->adjunction_child->elem->site_id);
      bugbreak();
    }
  if (site->adjunction_child == NULL && 
      (site->trigger_child == NULL ||
       (skip_conflations(site->trigger_child))->elem != stack->elem))
    printf("\nWarning: *** bad extension %d:%-2d\n", 
	   sentence_number, stack->elem->end);
  newstack = create_tree_list(site, popfree_tree_list(stack));
  site->site_id = next_site_id;
  next_site_id++;
 
  return(newstack);
}

/* pop stack until site found (elements freed), display their
   attachment/modifier outputs.  if site is NULL or not in stack, then do
   whole stack. */ 
struct tree_list *display_end_outputs(int position, struct tree_list *stack,
 				      struct tree *site)
{
  struct tree_list *newstack;
  struct tree *nextsite;
  
  newstack = stack;
  while (newstack != NULL && newstack->elem != site)
    { /* attach/modify (end) */
      nextsite = newstack->elem;
#if LONGOUTPUT
      printf("\n%3d,%3d :  ", position, nextsite->site_id);
#endif
      if (nextsite->parent == NULL)
	printf(".\n");
      else if (nextsite->modifier != 0)
	printf("+\t");
#if HEADOUTPUT
      else if (nextsite->head != 0)
	printf("^h\t");
#endif
      else
	printf("^\t");
      newstack = popfree_tree_list(newstack);
    }
  if (newstack == NULL && site != NULL)
    {
      printf("\nWarning: site %d not found in stack %d by display_end_outputs (%d:%d)\n",
	     site, stack, sentence_number, position);
      bugbreak();
    }
  return(newstack);
}



/* read the short output format for one sentence into a tree.  If no more
   sentences then return NULL.  If format_flag != 0 then reintroduce
   chomsky adjunction for modifications and conflated nodes.  Currently
   doesn't set trigger relationships */ 
struct tree *read_outputs(FILE *data, int format_flag)
{
  struct tree *res;
  char buffer[MAXDATALINE];
  char *inpp;
  int position;
  struct tree_list *stack;
  struct tree_list *tl;
  struct tree *new_preterm;
  struct tree *new_nonterm;
  struct tree *top;
  struct tree *subtop;
  int i, j;
  int head;

  position = 0;
  next_site_id = 1;

  stack = start_root(position, NULL);
  res = stack->elem;
  position++;

  while (stack != NULL && fgets(buffer, MAXDATALINE, data) != NULL)   {

    inpp = buffer;

    while (inpp[0] != '\0') {

#if DEBUG_READOUT
      printf("S:");
      for (tl = stack; tl != NULL; tl = tl->next)
	printf(" %d,", tl->elem->site_id);
      printf("\n");
      printf("I %d,%d: %s", position, stack->elem->site_id, buffer);
#endif

      if (inpp[0] == '>')   /* prediction */
	{
	  inpp += 2;
	  new_preterm = create_tree();
	  new_preterm->start = position;
	  new_preterm->site_id = next_site_id;
	  next_site_id++;
	  for (i = 0; inpp[i] != ' ' && inpp[i] != '\t' && inpp[i] != '\n' 
		 && i < MAXLABEL; i++)
	    new_preterm->label[i] = inpp[i];
	  new_preterm->label[i] = '\0';
	  inpp += i + 1;
	  for (i = 0; inpp[i] != ' ' && inpp[i] != '\t' && inpp[i] != '\n'
		 && i < MAXWORDLENGTH;  
	       i++)
	    sentence_words[position][i] = inpp[i];
	  sentence_words[position][i] = '\0';
	  inpp += i + 1;

	  stack = create_tree_list(new_preterm, stack); /* push preterm */
	  position++;
	}
      else if (inpp[0] == '|')   /* projection */
	{
	  if (inpp[1] == 'h') {
	    head = 1;
	    inpp++;
	  }
	  else
	    head = 0;
	  inpp += 2;
	  top = stack->elem;
	  /* create projected nonterm */
	  new_nonterm = create_tree();
	  new_nonterm->start = top->start;
	  new_nonterm->site_id = next_site_id;
	  next_site_id++;
	  for (i = 0; inpp[i] != ' ' && inpp[i] != '\t' && inpp[i] != '\n'
		 && i < MAXLABEL; i++) 
	    new_nonterm->label[i] = inpp[i];
	  new_nonterm->label[i] = '\0';
	  inpp += i + 1;
	  new_nonterm->children = create_tree_list(top, 
						   new_nonterm->children);
	  /* set new projection site features */
	  top->parent = new_nonterm;
	  top->end = position;
	  top->head = head;
	  /* update stack */
	  stack = popfree_tree_list(stack);  /* pop projection site */
	  stack = create_tree_list(new_nonterm, stack); /* push projected */
	}
      else if (inpp[0] == '^')   /* attachment */
	{
	  if (inpp[1] == 'h') {
	    head = 1;
	    inpp++;
	  }
	  else
	    head = 0;
	  inpp += 2;
	  top = stack->elem;
	  subtop = stack->next->elem;
	  subtop->children = 
	    create_tree_list(top, subtop->children);
	  top->parent = subtop;
	  top->end = position;
	  top->head = head;
	  stack = popfree_tree_list(stack);  /* pop attached nonterm */
	}
       else if (inpp[0] == '+')   /* modification */
	{
	  inpp += 2;
	  top = stack->elem;
	  subtop = stack->next->elem;
	  if (format_flag != 0)
	    {
	      /* create Chomsky adjunction nonterm */
	      new_nonterm = create_tree();
	      new_nonterm->start = subtop->start;
	      new_nonterm->site_id = subtop->site_id;
	      /* set and reset labels */
	      for (i = 0; subtop->label[i] != '\0' && i < MAXLABEL; i++)
		new_nonterm->label[i] = subtop->label[i];
	      new_nonterm->label[i] = '\0';
	      while (subtop->label[i] != '-' && i > 0)
		i--;
	      if (i > 0) /* relabel site with last conflated label */
		{
		  for (j = 0; 
		       subtop->label[i + j + 1] != '\0' && j < MAXLABEL; 
		       j++)
		    subtop->label[j] = subtop->label[i + j + 1];
		  subtop->label[j] = '\0';
		}
	      /* do Chomsky adjunction */
	      new_nonterm->children = 
		create_tree_list(subtop, new_nonterm->children);
	      subtop->parent = new_nonterm;
	      subtop->end = top->start;
	      new_nonterm->children = 
		create_tree_list(top, new_nonterm->children);
	      top->parent = new_nonterm;
	      top->end = position;
	      /* update stack */
	      stack = popfree_tree_list(stack);  /* pop modifier */
	      stack = popfree_tree_list(stack);  /* pop modification site */
                                       	         /* push modification */
	      stack = create_tree_list(new_nonterm, stack); 
	    }
	  else
	    {
	      subtop->children = 
		create_tree_list(top, subtop->children);
	      top->parent = subtop;
	      top->end = position;
	      top->modifier = 1;
	      stack = popfree_tree_list(stack);  /* pop modifier */
	    }
	}
       else if (inpp[0] == '.')   /* end of sentence */
	{
	  inpp += 2;
	  stack->elem->end = position;
	  if (stack->elem != res)
	    {
	      printf("Warning: ***stack not empty at end of sentence %d:%-2d\n", 
		     sentence_number, position);
	    }
	  while (stack != NULL)
	    stack = popfree_tree_list(stack);  /* empty stack */
	  /* empty stack finishes loop */
	}
      else if (inpp[0] == '\n')   /* blank line */
	{
	  inpp++;
	}
      else
	{
	  printf("Warning: ***bad symbol at position %d in line: %s  %d:%-2d\n", 
		 (int) inpp - (int) &buffer, buffer, sentence_number, 
		 position);
	}
    }
  }

  if (next_site_id == 1)
    return(NULL);  /* no sentence read */

  if (format_flag)  /* expand conflations */
    {
      expand_conflations(res);
    }

  return(res);
}


/* replace nodes labeled with a conflation by multiple nodes, one per conflated label. */
void expand_conflations(struct tree *root)
{
  struct tree *new_nonterm;
  struct tree_list *tl;
  int i, j;

  for (i = 0; root->label[i] != '\0' && i < MAXLABEL; i++)
    if (root->label[i] == '-' && i != 0)
      {
	root->label[i] = '\0';
	new_nonterm = create_tree();
	new_nonterm->start = root->start;
	new_nonterm->end = root->end;
	new_nonterm->site_id = root->site_id;
	for (j = 0; root->label[j + i + 1] != '\0' && j < MAXLABEL; j++)
	  new_nonterm->label[j] = root->label[j + i + 1];
	new_nonterm->label[j] = '\0';
	new_nonterm->children = root->children;
	for (tl = new_nonterm->children; tl != NULL; tl = tl->next)
	  {
	    tl->elem->parent = new_nonterm;
	  }
	root->children = create_tree_list(new_nonterm, NULL);
	new_nonterm->parent = root;
	break;
      }

  for (tl = root->children; tl != NULL; tl = tl->next)
    {
      expand_conflations(tl->elem);
    }
}



/* display all the branches in root */
void display_tree(struct tree *root, int goldflag)
{
  char *indent;
  char blanks[71] = 
    "                                                                      ";
  indent = blanks + 70;

  display_subtree(root, indent, root->start, goldflag);
  printf("\n\n");

}

/* returns end position of subtree */
int display_subtree(struct tree *root, char *indent, int position, int goldflag)
{
  struct tree_list *ch;
  char *newindent;
  int newposition;

  if (goldflag != 0 && !strcmp(root->label, "^^^"))
    return(position + 1);

  if (strlen(indent) >= 70)
    indent += strlen(indent);

  if (root->conflation == 0) {

    if (goldflag == 0) {
      printf("\n%s", indent);
      if (root->modifier != 0)
	printf("+");
      if (root->head != 0)
	printf("h");
    }
    else
      printf(" ");
    if (goldflag != 0 && !strcmp(root->label, ""))
      printf("(TOP");
    else
      printf("(%s", root->label);

    if (root->children == NULL)
      {
	if (root->start != position)
	  printf(" ****missing terminal :%2d\n%s  ", position, indent);
	printf(" %s", sentence_words[root->start]);
#if TREEPOS
	printf(" :%2d)", root->start);
#endif
	printf(")");
	if (goldflag != 0)
	  printf("\n%s", indent);
	newposition = root->start + 1;
      }
    else
      {
	newindent = indent - 2;
	ch = root->children;
	while (ch->next != NULL)
	  ch = ch->next;
	newposition = position;
	for (; ch != NULL; ch = ch->prev)
	  {
	    newposition =
	      display_subtree(ch->elem, newindent, newposition, goldflag);
	  }

	printf(" )");
      }
  }
  else {
    newposition =
      display_subtree(root->children->elem, indent, position, goldflag);
  }

  return(newposition);
}


/* return a pointer to a new tree. */
struct tree *create_tree()
{
  struct tree *res;

  res = (struct tree *) malloc(sizeof(struct tree));
#if DEBUG_ALLOC
  printf("\n%d tree created\n", res);
#endif

  res->parent = NULL;
  res->children = NULL;
  res->label[0] = '\0';
  res->trigger = NULL; /* flag for trigger not set */
  res->trigger_child = NULL;
  res->adjunction_child = NULL;
  res->modification_child = NULL;
  res->attachment = NULL;
  res->conflation = 0;
  res->toplexical = 0;
  res->modifier = 0;
  res->head = 0;
  res->site_id = -1;

  return(res);
}

/* free the contents of tree t, including children list but not children */
void free_tree(struct tree *t)
{
  if (t != NULL)
    {
      free_tree_list(t->children);
      free(t);
#if DEBUG_ALLOC
      printf("\n%d tree freed\n", t);
#endif
    }
}

/* free the contents of tree t, including children */
void free_all_tree(struct tree *t)
{
  if (t != NULL)
    {
      free_all_tree_list(t->children);
#if DEBUG_READOUT
      printf("Freeing node %d - %d\n", t->start, t->end);
#endif
      free(t);  /* children list conses already freed */
#if DEBUG_ALLOC
      printf("\n%d tree freed\n", t);
#endif
    }
}

/* return a pointer to a tree_list cons includeing elem and rest. */
struct tree_list *create_tree_list(struct tree *elem, struct tree_list *rest)
{
  struct tree_list *res;

  res = (struct tree_list *) malloc (sizeof(struct tree_list));
#if DEBUG_ALLOC
  printf("\n%d tree list created\n", res);
#endif

  res->elem = elem;
  res->next = rest;
  res->prev = NULL;
  if (rest != NULL)
    rest->prev = res;

  return(res);
}

/* free the space used by all the cons's in tl, but not its trees. */
void free_tree_list(struct tree_list *tl)
{
  struct tree_list *head;
  struct tree_list *rest;

  for (head = tl; head != NULL; head = rest)
    {
      rest = head->next;
      free(head);
#if DEBUG_ALLOC
      printf("\n%d tree list freed\n", head);
#endif
    }
}

/* free the space used by all the cons's in tl, plus its trees and subtrees. */
void free_all_tree_list(struct tree_list *tl)
{
  struct tree_list *head;
  struct tree_list *rest;

  for (head = tl; head != NULL; head = rest)
    {
      rest = head->next;
      free_all_tree(head->elem);
      free(head);
#if DEBUG_ALLOC
      printf("\n%d tree list freed\n", head);
#endif
    }
}

/* free the space used by the cons tl, but not its tree.  update rest of
   list to remove this cons */
void pluckfree_tree_list(struct tree_list *tl)
{
  if (tl->prev != NULL)
    tl->prev->next = tl->next;
  if (tl->next != NULL)
    tl->next->prev = tl->prev;
  free(tl);
#if DEBUG_ALLOC
  printf("\n%d tree list freed\n", tl);
#endif
}

/* free the space used by the first cons in tl and return a pointer to
   the next cons.  Does not free the element in the cons. */
struct tree_list *popfree_tree_list(struct tree_list *tl)
{
  struct tree_list *res;

  if (tl == NULL)
    error("Error: can't pop an empty tlist.");

  res = tl->next;
  if (res != NULL)
    res->prev = NULL;
  free(tl);
#if DEBUG_ALLOC
  printf("\n%d tree list freed\n", tl);
#endif

  return(res);
}


/* return a pointer to a new trigger. */
struct triggering *create_triggering()
{
  struct triggering *res;

  res = (struct triggering *) malloc(sizeof(struct triggering));
#if DEBUG_ALLOC
  printf("\n%d triggering created\n", res);
#endif

  res->preterm = NULL;
  res->projection = NULL;

  return(res);
}

/* set all pointers to NULL. */
void clear_triggering(struct triggering *tg)
{
  tg->preterm = NULL;
  tg->projection = NULL;
}


/* free space used and create new preterm. */
void init_triggering(struct triggering *tg)
{
  free_tree(tg->preterm);
  tg->preterm = create_tree();

  free_tree_list(tg->projection);
  tg->projection = NULL;
}

/* free the contents of trigger t, but not its children */
void free_triggering(struct triggering *tg)
{
  if (tg != NULL)
    {
      free(tg);
#if DEBUG_ALLOC
      printf("\n%d triggering freed\n", tg);
#endif
    }
}

/* free any modification nodes in tg->projection, and free the list
   tg->projection.  must be done before the tree is freed.  */
void free_projection(struct triggering *tg)
{
  struct tree_list *tl, *tln;

  tl = tg->projection;
  while (tl != NULL)
    {
      tln = tl->next;
      if (tl->elem->modification_child != NULL)  
	{   /* modifications need to be freed, but not their children lists,
	       except for the modification child cons */
	  free(tl->elem->modification_child);
#if DEBUG_ALLOC
	  printf("\n%d tree list freed\n", tl->elem->modification_child);
#endif
	  free(tl->elem);
#if DEBUG_ALLOC
	  printf("\n%d tree freed\n", tl->elem);
#endif
	}
      free(tl);
#if DEBUG_ALLOC
      printf("\n%d tree list freed\n", tl);
#endif
      tl = tln;
    }
  tg->projection = NULL;
}


/* returns an index for the first instance of s in str_array, or an index
   for NULL.  str_array must end in NULL. */
int find_str(char *s, char **str_array)
{
  int res;

  for (res = 0; str_array[res] != NULL; res++)
    if (strcmp(s, str_array[res]) == 0)
      break;

  return(res);
}


void usage_error(char *name)
{
  printf("Usage:\n  %s -h | datafile | -d datafile  | -d2b datafile | -d2g datafile | -o2d outputfile | -o2b outputfile\n  d: bracketed strings, b: custom bracket format, o: network output format, g: gold-like tree format\n", 
	 name);  
  exit(0);
}

void error(char *mes)
{
  printf("\n%s\n", mes); 
  exit(0);
}

