/* Written by James Henderson and Ivan Titov */
#ifndef _SSN_COMMON_TYPES_H_
#define _SSN_COMMON_TYPES_H_

#include "ssn_common_consts.h"

#define CHECK_PTR(x) if ((x) == NULL) { printf("Can't allocate\n"); exit(-1);}


/** Data Structure Types **/

/* Define the layer state types */
struct input {
    int input1;                 /* tag of input */
    int input2[MAXNUMBINS];     /* tag-word of input input */
    int output2[MAXCLUSTERLEVELS];      /* word of output (word_id, realoutput = wordoutputs + word_id,  
                                           memoutput = word_id  - memoffset + wordoutputs)  */
    int cl_levels_out;
};
struct hidrho {
    double act[MAXNUMRHO];      /* output activation */
    double err[MAXNUMRHO];      /* net error input */
};
struct hidtau {
    double act[MAXNUMTAU];      /* output activation */
    double err[MAXNUMTAU];      /* net error input */
};
struct output {
    double act[MAXNUMOUTMEM];   /* output activation (times grandnorm before
                                   normalization).  First NUMSTRUC are for
                                   structure */
    double del[MAXNUMOUTMEM];   /* error delta term */
};
struct target {
    double act[MAXNUMTARMEM];   /* target activation, other than word
                                   predicitons. */
};

/* new unified phase state type */
struct phase_state {
    int phase;                  /* phase number */
    int period;                 /* current period */
    int outind;                 /* index of output in output_hist (and combin_hist) */
    struct phase_state *previous;       /* state in previous period */
    double hidrho_act[MAXNUMRHO];       /* hidrho activations */
    double hidrho_err[MAXNUMRHO];       /* hidrho net error inputs */
};
/* list of hidden states. */
struct phase_state_list {
    struct phase_state *elem;
    struct phase_state *next;
    struct phase_state *prev;
};


/* Define types for computation phase control structures */
struct stack {
    int *top;                   /* pointer to top of stack in array, or NULL */
    int *subtop;                /* pointer to element below top in stack in array, or NULL */
    int *content;               /* the array for the stack, 0 is bottom */
};
struct phase_spec {
    int birth;                  /* period created */
    int death;                  /* period after which computation stops */
    struct stack states;        /* periods when phase has or needs outputs */
    int *computed;              /* pointer to last period in states.content computed */
    int *nextcomp;              /* pointer to next period in states.content to be computed */
    int trigger_op;             /* the operation used to create. PREDOUT is for preterms
                                   and ROOTFLAG is for the matrix root. */
    int trigger_parent;         /* phase of lowest node above during computation */
    int trigger_child;          /* phase which caused adjunction or projection, or -1 */
    int trigger_label;          /* chosen nonterminal OR preterminal label */
    int head_child;             /* phase of head child, or -1. */
    int dying_op;               /* the last operation, PROJOUT or HDPROUT or ATTOUT or
                                   HDATOUT or MODOUT. */
    int proj_depth;             /* how deep are descendents projections */
};


/* Define the types for searching for the best parse */
/* An option is a node in a possible derivation.  They are partitioned
     between the queue and the current best derivations (and the garbage). */
struct option {
    int period;                 /* period of state */
    int phase;                  /* phase of state or -1 if not allocated */

    double lprob;               /* log probability of preceeding derivation */
    double ranking_lprob;       /* log prob plus an estimate of probability
                                   of the subsequent derivation */
    int preterm;                /* preterm/nonterm */
    int previous_output;        /* preceding operation's output unit */
    struct option *previous_option;     /* preceding option in derivation */
    struct option_stack *next_options;  /* following options in derivation, 
                                           if set */
    int trigger_label;
    int trigger_output;
    /* stack after preceding derivation (option's phase is top) */
    int stack_subtop;           /* subtop phase or -1 if not yet allocated 
                                   or -2 if subtop is in stack_rest */
    int stack_rest_rel;         /* number of elements - 1 in the last
                                   period's phase stack which are in the
                                   rest of the stack */
    int prev_pred_ind;          /* index in previous (or own, if a
                                   prediction) best_options of the
                                   prediction option which this option was
                                   derived from */
    int proj_depth;             /* how many projections in a row */
    struct option *context_option;      /* the option for the last calculated
                                           period of phase, or NULL. */
    struct option *trigparent_option;   /* the option for the last calculated
                                           period of the triggering parent's
                                           phase, or NULL. */
    struct option *trigchild_option;    /* the option for the last calculated
                                           period of the triggering child's
                                           phase, or NULL. */
    struct option *headchild_option;    /* the option for the last calculated
                                           period of the head child's
                                           phase, or NULL. */
    int trigger_period;         /* period when triggered */
    int left_edge;              /* period of leftmost dominated word */
    /* instructions for calculations */
    int new_phase;              /* flag for allocation of new phase. */
    int new_state;              /* flag for allocation of state,
                                   (combination,) and outputs. */
    int new_subtop;             /* flag for allocation of subtop_state. */
    int in_best;                /* flag to prevent freeing because still
                                   being pursued. */
    int in_target;              /* 1 if in target, 0 otherwise. */
    int num_descendents;        /* number of descendents pursued. */
    int is_pred_opt;                /** whether predicted option */
    /* layer activations */
    struct hidrho *state;       /* hidden activations of phase's state */
    struct hidrho *subtop_state;        /* hidden activations of subtop's state */
    struct output *outputs;     /* output layer activations of state */
    int firstout;               /* index of first relevant output (in memory space) */
    int lastout;                /* index of last relevant output (in memory space) */
    int memoffset;              /* memory offset for word outputs */

    /* for backpropagation */
    double backward_prob;       /* probability of going from this
                                   decision to any complete parse. */
                                    /** for SIGMOIDDISC this should be
					called global_slack. **/
#if PLUGLEAKS
    double backward_knownout;   /* sum of the outputs which contribute
                                   known values to backward_prob. */
#if ! LOCALESTIMATE
    double pred_loss;
    double pred_loss_weight;
#endif
#endif

    struct cluster *dec_cluster;        /* decision cluster (decision to be made between its childs) */



};


/* list of options.  Used for queue, etc. */
struct option_stack {
    struct option *elem;
    struct option_stack *next;
    struct option_stack *prev;
};
/* list of list of options.  Used for displaying derivations. */
struct option_stack_stack {
    struct option_stack *elem;
    struct option_stack_stack *next;
    struct option_stack_stack *prev;
};

/* node of tree structure.  For representing target tree. */
struct tree {
    struct tree *parent;        /* parent.  Includes conflation projections. */
    struct tree_list *children; /* in reverse order.  NULL if preterm.
                                   Includes conflation projections. */
    char label[MAXLABELLENGTH + 1];
    int start;                  /* first word */
    int end;                    /* last word + 1 */
    struct tree *trigger;       /* pointer to the trigger preterm, NULL if not set */
    struct tree_list *trigger_child;    /* pointer to trigger child in children */
    struct tree_list *adjunction_child; /* pointer to the child in children
                                           of which this is an adjunction.
                                           Skips to lowest of conflation
                                           projections. */
    struct tree_list *modification_child;       /* pointer to the child in children
                                                   of which this is an modification.
                                                   Parent removed from tree. */
    struct tree_list *head_child;       /* pointer to the child in children which is 
                                           this constituent's head. */
    struct tree *attachment;    /* site to which this is an attachment, if
                                   checked in trigger test */
    int conflation;             /* flag if this was a conflation projection */
    int toplexical;             /* flag if this was a toplexical projection */
    int modifier;               /* flag if attached to parent with modification */
    int head;                   /* flag if head of parent */
    int site_id;                /* ID in output format */

};
/* list of trees. */
struct tree_list {
    struct tree *elem;
    struct tree_list *next;
    struct tree_list *prev;
};
#endif

