/* Written by James Henderson and Ivan Titov */
#include "cluster.h"
#include "list.h"

#include "ssn_common_consts.h"
#include "ssn_common_types.h"
#include "ssn_common_api.h"
#include "ssn_common_globals.h"



/* Declaration of the methods for internal use (@ivan: I suggest using leading underscore to mark them)  */
/* 
 * read in the filenames (stems) (filename [datafilename [validationfile]]) 
 * from the command line, starting at argument position offset. 
 * and set all the needed files on the basis of these stems
 */
void _process_names(int argc, char *argv[], int offset);
/* main routine: read the argument line and branch accordingly */
int main(int argc, char *argv[])
{
    int offset;
    int ival;
    double dval;
    FILE *fp;
    double score;
    int tpbflag, tcbflag;

    if (argc < 2) {
        usage_help(argv[0]);
    }

    /* for implied defaults below */
    tpbflag = 0;
    tcbflag = 0;

    /* defaults */
    autoreduce = 2; /* -av -nb */
    saveoldwgt = 0;
    schedreduce = 0;
    autoreduce_reg = 1;
    joint_error_flag = 1; /* -jerr */
    numtopparses = 0;
    numtopderivs = 0;
    readcandidates = 0;
    readtraincandidates = 0;
    stdin_word_input = 0;
    stdout_bracketed_output = 0;
    max_cl_levels = 255;
    bracketed_output = 0;
    idtext_input = 0;

    /* allow options to be in any order, until filename options reached */
    for (offset = 1; offset < argc; ) {
    
      if (strcmp(argv[offset], "-h") == 0) {
        usage_help(argv[0]);
      }

      else if (strcmp(argv[offset], "-seed") == 0) {
        offset++;
        seed = atoi(argv[offset]);
        offset++;
      }

      else if (strcmp(argv[offset], "-sched") == 0) {
        schedreduce = 1;
        offset += 1;
        dval = 1.0 - atof(argv[offset]);
        if (dval < 0.0 || dval > 1.0) {
	  sprintf(warning_text,
		  "scheduled reduction rate %f out of range %d-%d, using %d.",
		  1.0 - dval, 0.0, 1.0, 1.0 - sched_red_rate);
	  warning(warning_text);
        } else
	  sched_red_rate = 1.0 - dval;
        offset += 1;
      }

      else if (strcmp(argv[offset], "-autorate") == 0) {
        offset += 1;
	dval = atof(argv[offset]);
	if (dval < 0.0 || dval > 1.0) {
	  sprintf(warning_text,
		  "learning rate reduction rate %f out of range %d-%d, using %d.",
		  dval, 0.0, 1.0, 1.0 - eta_red_rate);
	  warning(warning_text);
	} else
	  eta_red_rate = 1.0 - dval;
	offset += 1;
      }

      else if (strcmp(argv[offset], "-autoreg") == 0) {
        autoreduce_reg = 1;
	dval = atof(argv[offset + 1]);
	if (dval < 0.0 || dval > 1.0) {
	  sprintf(warning_text,
		  "regularization reduction rate %f out of range %d-%d, using %d.",
		  dval, 0.0, 1.0, 1.0 - reg_red_step);
	  warning(warning_text);
	} else
	  reg_red_step = 1.0 - dval;
        offset += 2;
      }


      else if (strcmp(argv[offset], "-condtrain") == 0) {
        joint_error_flag = 0;
        offset++;
      } 

      else if (strcmp(argv[offset], "-predbeam") == 0) {
        ival = atoi(argv[offset + 1]);
        if (ival <= 0 || ival > MAXPREDBEAM) {
	  sprintf(warning_text,
		  "prediction beam %d out of range %d-%d, using %d.",
		  ival, 1, MAXPREDBEAM, predbeam);
	  warning(warning_text);
        } else
	  predbeam = ival;
        offset += 2;
      }

      else if (strcmp(argv[offset], "-trainbeam") == 0) {
        ival = atoi(argv[offset + 1]);
        if (ival < 0 || ival > MAXPREDBEAM) {
	  sprintf(warning_text,
		  "training prediction beam %d out of range %d-%d, using %d.",
		  ival, 0, MAXPREDBEAM, training_predbeam);
	  warning(warning_text);
        } else {
	  training_predbeam = ival;
	  tpbflag = 1;
	}
        offset += 2;
      }

      else if (strcmp(argv[offset], "-candbeam") == 0) {
        ival = atoi(argv[offset + 1]);
        if (ival <= 0 || ival > MAXPREDBEAM) {
	  sprintf(warning_text,
		  "candidate prediction beam %d out of range %d-%d, using %d.",
		  ival, 1, MAXPREDBEAM, candidate_predbeam);
	  warning(warning_text);
        } else
	  candidate_predbeam = ival;
        offset += 2;
      }

      else if (strcmp(argv[offset], "-traincandbeam") == 0) {
        ival = atoi(argv[offset + 1]);
        if (ival < 0 || ival > MAXPREDBEAM) {
	  sprintf(warning_text,
		  "training candidate prediction beam %d out of range %d-%d, using %d.",
		  ival, 0, MAXPREDBEAM, training_candidate_predbeam);
	  warning(warning_text);
        } else {
	  training_candidate_predbeam = ival;
	  tcbflag = 1;
	}
        offset += 2;
      }

      else if (strcmp(argv[offset], "-lookaheadbeam") == 0) {
        ival = atoi(argv[offset + 1]);
        if (ival <= 0 || ival > MAXPREDBEAM) {
	  sprintf(warning_text,
		  "lookahead prediction beam %d out of range %d-%d, using %d.",
		  ival, 1, MAXPREDBEAM, lookahead_predbeam);
	  warning(warning_text);
        } else
	  lookahead_predbeam = ival;
        ival = atoi(argv[offset + 2]);
        if (ival <= 0) {
	  sprintf(warning_text,
		  "lookahead %d out of range %d-, using %d.",
		  ival, 1, lookahead);
	  warning(warning_text);
        } else
	  lookahead = ival;
        offset += 3;
      }

      else if (strcmp(argv[offset], "-predfactor") == 0) {
        expected_max_pred_factor = atof(argv[offset + 1]);
        offset += 2;
      }

      else if (strcmp(argv[offset], "-branchfactor") == 0) {
        ival = atoi(argv[offset + 1]);
        if (ival <= 0 || ival > MAXPHASEBEAM) {
	  sprintf(warning_text,
		  "branching factor %d out of range %d-%d, using %d.",
		  ival, 1, MAXPHASEBEAM, phasebeam);
	  warning(warning_text);
        } else
	  phasebeam = ival;
        offset += 2;
      }

      else if (strcmp(argv[offset], "-finprob") == 0) {
        finishing_prob = atof(argv[offset + 1]);
        offset += 2;
      }

      else if (strcmp(argv[offset], "-projbias") == 0) {
        projection_bias = atof(argv[offset + 1]);
        offset += 2;
      }

      else if (strcmp(argv[offset], "-top") == 0) {
        offset++;
        numtopparses = atof(argv[offset]);
        offset++;
      } 

      else if (strcmp(argv[offset], "-deriv") == 0) {
        offset++;
        numtopderivs = atof(argv[offset]);
        offset++;
      } 

      else if (strcmp(argv[offset], "-cand") == 0) {
        readcandidates = 1;
        offset++;
        strcpy(candfile, argv[offset]);
	if (strcmp(candfile, "-") == 0) {
	  stdin_candidate_input = 1;
	}
	else {
	  stdin_candidate_input = 0;
	}	    
        strcat(candfile, ".deriv");
        offset++;
      } 

      else if (strcmp(argv[offset], "-traincand") == 0) {
        readtraincandidates = 1;
        offset++;
        strcpy(traincandfile, argv[offset]);
        strcat(traincandfile, ".deriv");
        offset++;
      } 
      else if (strcmp(argv[offset], "-stdin") == 0) {
        stdin_word_input = 1;
        offset++;
      } 
      else if (strcmp(argv[offset], "-stdout") == 0) {
        stdout_bracketed_output = 1;
        offset++;
      } 
      else {
	break;  /* no more unordered options */
      }
    }

    /* implied defaults */
    testing_predbeam = predbeam;
    if (joint_error_flag != 0) 
        training_predbeam = 0;
    else if (tpbflag == 0)
        training_predbeam = predbeam;
    testing_candidate_predbeam = candidate_predbeam;
    if (tcbflag == 0)
        training_candidate_predbeam = candidate_predbeam;

    if (offset >= argc) {
      error("Error: insufficient options.");
    }
    else if (strcmp(argv[offset], "-eval") == 0) {
      if (stdin_word_input != 0) {
	warning("'-stdin' overridden by '-eval'; taking word input from target.");
	stdin_word_input = 0;
      }
      if (stdout_bracketed_output != 0) {
	warning("'-stdout' overridden by '-eval'.");
	stdout_bracketed_output = 0;
      }
      _process_names(argc, argv, offset + 1);
      evaluation();
    }

    else if (strcmp(argv[offset], "-parsedata") == 0) {
      if (stdin_word_input != 0) {
	warning("'-stdin' overridden by '-parsedata'; taking word input from target.");
	stdin_word_input = 0;
      }
      bracketed_output = 1;
      _process_names(argc, argv, offset + 1);
      evaluation();
    }

    else if (strcmp(argv[offset], "-parse") == 0) {
      offset++;
      idtext_input = 1;
      bracketed_output = 1;
      _process_names(argc, argv, offset);
      evaluation();
    }
    
    else if (strcmp(argv[offset], "-train") == 0) {
      int learn_cycles;

      learn_cycles = atoi(argv[offset + 1]);
      if (learn_cycles == 0)
	error
	  ("Error: Zero training cycles no longer supported.  Use -e.");
      if (learn_cycles < 0)
	error("Error: Training cycles must be > 0.");

      _process_names(argc, argv, offset + 2);
      learning(learn_cycles);
    }

    else if (strcmp(argv[offset], "-valtrain") == 0) {
      int test_freq, learn_cycles;

      test_freq = atoi(argv[offset + 1]);
      if (test_freq <= 0)
	error("Error: Testing frequencey must be > 0.");
      learn_cycles = atoi(argv[offset + 2]);
      if (learn_cycles < 0)
	error("Error: Training cycles must be >= 0.");

      _process_names(argc, argv, offset + 3);
      validation_learning(test_freq, learn_cycles);
    }

    else {
      printf("Error: invalid option: %s\n", argv[offset]);
      usage_error(argv[0]);
    }

    return 0;
}

