/* Left Corner Simple Synchrony Network Parser */
/*           James Henderson, August 2001 - November 2005  */
/*           Ivan Titov, January 2004 - November 2005  */
/* Generative and Discriminative models of parsing with SSNs */
/*   for the Penn Treebank corpus. */
/*  */
/* Written by James Henderson and Ivan Titov */
/*  */
/* This work was supported mostly by the Swiss National Science Foundation, */
/* under Division 2 grants  and  , */
/* and to some extent by the EC under the FP6 project 507802. */


#include "cluster.h"
#include "list.h"

#include "ssn_common_consts.h"
#include "ssn_common_types.h"
#include "ssn_common_api.h"
#include "ssn_common_globals.h"



/* Declaration of the methods for internal use (@ivan: I suggest using leading underscore to mark them)  */
/* 
 * read in the filenames (stems) (filename [datafilename [validationfile]]) 
 * from the command line, starting at argument position offset. 
 * and set all the needed files on the basis of these stems
 */
void _process_names(int argc, char *argv[], int offset);

/* Global data structures */
char filename[FILENAMESIZE];
char inffile[FILENAMESIZE];
char weightfile[FILENAMESIZE];
char newweightfile[FILENAMESIZE];
char datafile[FILENAMESIZE];
char outfile[FILENAMESIZE];
char brafile[FILENAMESIZE];
char gldfile[FILENAMESIZE];
char resfile[FILENAMESIZE];
char topfile[FILENAMESIZE];
char candfile[FILENAMESIZE];
char traincandfile[FILENAMESIZE];
char valdatafile[FILENAMESIZE];
char valweightfile[FILENAMESIZE];
char retestfile[FILENAMESIZE];
char testingfile[FILENAMESIZE];
char neededfile[FILENAMESIZE];
char valneededfile[FILENAMESIZE];
char testneededfile[FILENAMESIZE];
#if FILEPROG
char progfile[FILENAMESIZE];
FILE *progfp;
#endif



#if FILEWARN|FILEDISP
char warnfile[FILENAMESIZE];
FILE *warnfp;
#endif
char warning_text[1000];

/* Define the computation phase control structures */
struct phase_spec *phase_hist;  /* phase_hist[p] is the spec of p.  When
                                   testing, can only use for birth features?? */
struct stack *deriv_order;      /* derivation_order[t] is the phases computed at
                                   t in derivation order.  site of
                                   post-prediction computation is first and may
                                   occur again later in ordering, as with
                                   adjunction newterm if exists. */
int *max_phase;                 /* max_phase[t] is the highest phase index used up to t */
int next_phase;                 /* next phase index to be used as a new phase */
struct stack *phase_stacks[MAXPREDBEAM];        /* stack of phases on right
                                                   frontier for each previous
                                                   prediction option REMOVE */
struct stack *phase_stack_scratch[MAXPREDBEAM]; /* for copying REMOVE */
struct stack *target_phase_stack;       /* stack of phases used by target REMOVE */
char **sentence_words;          /* text of words in sentence */
char **sentence_tagwordids;     /* text of tag and word id spec in sentence */
int next_site_id;               /* like next_phase but used in read_target */
struct stack *free_phases;      /* list of free phases used in testing */
int next_outind;                /* next output index to be used */

/* Define the structures for searching for the best parse during testing */
struct option_stack *option_queue;      /* list of options without computed phase
                                           states, sorted by probability */
struct option *winning_option;  /* the final best option */
double worst_best_option_lprob; /* the lowest log probability of the
                                   best_options */
double expected_max_pred_lprob = 0.0; /* assume the final option before
				   starting the next word will never be more
				   probable than this */
struct option_stack *garbage_option_stack = NULL;       /* unallocated conss */
struct option *garbage_option = NULL;   /* unallocated options */
struct option *garbage_option_state = NULL;     /* unallocated options with
                                                   state,output arrays  */
int garbage_option_state_cnt = 0;       /* count of states allocated */
#if CHECKANTIGARBAGE
struct option_stack *anti_garbage_options = NULL;       /* allocated options */
#endif

struct option *candidate_derivations = NULL;    /* structure which holds the
                                                   candidate derivations.  */
#if LOSTTAIL
double lost_probability;        /* used to estmate prob of unpersued options */
#endif
#if ! LOCALESTIMATE
double *ave_finish_probs;       /* ave_finish_probs[t] is the
                                   probability-weighted average of the 
                                   multiplication of the
                                   prediction probabilities for words t
                                   through the end. */
double *total_finish_probs;     /* total_finish_probs[t] is the
                                   probability-weighted total of the
                                   multiplication of the
                                   prediction probabilities for word t
                                   through the end. */
double *total_finish_prob_weights;      /* total_finish_prob_weights[t] is the total
                                           of the probability weights used in
                                           total_finish_probs[t].  */
#endif
/* structures which hold the target derivation.  list conses 
   allocated on need and never freed.  could do same with options.  */
struct option_stack *target_derivation; /* list of target options in
                                           derivation order. */
struct option_stack *target_derivation_next;    /* next free cons in list */
struct option ***best_options;  /* best_options[t][b] is the option after
                                   period t with the bth highest ranking
                                   probability, or NULL */
struct option_stack *training_options;  /* list of all the options which are
                                           included in target_derivation or
                                           best_options, listed in order of
                                           creation. */
struct option_stack *training_options_next;     /* next free cons in list */
struct option_stack *scratch_options;   /* for reordering other lists. */
struct option_stack *scratch_options_next;      /* next free cons in list */

/* Define the arrays for layer states */
struct input *input_hist;       /* input_hist[t] is the input at t */
struct hidrho **hidrho_hist;    /* (hidrho_hist[t])[p] is state at */
/* period t and phase p */
struct hidrho hidrho_initstate; /* used for (hidrho_hist[0])[0] */
struct hidrho *hidrho_constant_net;     /* hidrho_constant_net[p] memoizes net
                                           input and delta (not actually error)
                                           used for p at every period. */
struct hidrho hidrho_nullstate; /* zero context for new phases (nonzero not
                                   needed with INTRORHOINP=1) */
struct output *output_hist;     /* output_hist[p] is state at current period
                                   and phase (or index) p */
struct target **target_hist;    /* (target_hist[t])[p] is target output at */
/* period t and phase p */
/* could mimic output_hist to save space */


double default_error = 0.0;     /* use for -log(0.0) (error for unnormalized
                                   values) */


/* For keeping a record of structures */
struct tree *target_tree = NULL;
struct tree *output_tree = NULL;


/* global weight state variables */
int weight_age = 0;             /* the number training epochs which the current
                                   weights have gone through. (in .wgt header) */
double weight_pretrnerr = -1.0; /* training error before/during
                                   weightfile. (in .wgt header) */


/* Define each set of weights */
/* input links (rho input = rightmost word of left-corner child, 
   tau input = most recent word) */
double link_r1_r[MAXNUMRHO][MAXNUMINP1];        /* rho input1 to rho hidden weight */
double link_r2_r[MAXNUMRHO][MAXNUMINP2];        /* rho input2 to rho hidden weight */
double link_t1_r[MAXNUMRHO][MAXNUMINP1];        /* tau input1 to rho hidden weight */
double link_t2_r[MAXNUMRHO][MAXNUMINP2];        /* tau input2 to rho hidden weight */
#if HEADWORDS
double link_h1_r[MAXNUMRHO][MAXNUMINP1];        /* head input1 to rho hidden weight */
double link_h2_r[MAXNUMRHO][MAXNUMINP2];        /* head input2 to rho hidden weight */
#endif
/* context links */
#if ! LABELCONTEXTINP
double link_r_r[MAXNUMRHO][MAXNUMRHO];  /* rho context to rho hidden weight */
#endif
double link_bir_r[MAXNUMRHO];   /* newphase input to rho hidden weight */
double link_tro_r[MAXNUMRHO][NUMTRIG];  /* triggering operation to rho hidden
                                           weight */
#if LABELTRIGPARENTINP
double link_trp_r[MAXNUMRHO][MAXNUMLAB];        /* triggering parent to rho
                                                   hidden weight */
#else
double link_trp_r[MAXNUMRHO][MAXNUMRHO];        /* triggering parent to rho
                                                   hidden weight */
#endif
#if LABELTRIGCHILDINP
double link_trc_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];   /* triggering child to rho
                                                           hidden weight */
#else
double link_trc_r[MAXNUMRHO][MAXNUMRHO];        /* triggering child to rho
                                                   hidden weight */
#endif
double link_lab_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];   /* chosen non/pre-term
                                                           label to rho hidden
                                                           weight */
double link_pred_r[MAXNUMRHO];  /* prev prediction to rho hidden weight */
#if CHILDCONTEXT
#if LABELCHILDINP
double link_chd_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];   /* previous child to rho
                                                           hidden weight */
#else
double link_chd_r[MAXNUMRHO][MAXNUMRHO];        /* previous child to rho
                                                   hidden weight */
#endif
#endif
#if HEADHIDDEN
#if LABELHEADCHILDINP
double link_hdc_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];   /* head child to rho
                                                           hidden weight */
#else
double link_hdc_r[MAXNUMRHO][MAXNUMRHO];        /* head child to rho
                                                   hidden weight */
#endif
#endif
#if CHOPCONTEXT
double link_cho_r[MAXNUMRHO][NUMCHOP];  /* previous child operation to rho
                                           hidden weight */
#endif
#if HEADOUTS
double link_nhd_r[MAXNUMRHO];   /* not headed to rho hidden weight */
#endif
/* output (and combination) links */
double link_r_o[MAXNUMOUT][MAXNUMRHO];  /* rho hidden to output weight */

/* define bias weights for each layer */
double bias_r[MAXNUMRHO];       /* rho hidden bias */
double bias_o[MAXNUMOUT];       /* output bias */


/* Define weight deltas (out[i]*del[j], no eta)       */
double linkdel_r1_r[MAXNUMRHO][MAXNUMINP1];     /* rho input1 to rho hidden */
double linkdel_r2_r[MAXNUMRHO][MAXNUMINP2];     /* rho input2 to rho hidden */
double linkdel_t1_r[MAXNUMRHO][MAXNUMINP1];     /* tau input1 to rho hidden */
double linkdel_t2_r[MAXNUMRHO][MAXNUMINP2];     /* tau input2 to rho hidden */
#if HEADWORDS
double linkdel_h1_r[MAXNUMRHO][MAXNUMINP1];     /* head input1 to rho hidden */
double linkdel_h2_r[MAXNUMRHO][MAXNUMINP2];     /* head input2 to rho hidden */
#endif
#if ! LABELCONTEXTINP
double linkdel_r_r[MAXNUMRHO][MAXNUMRHO];       /* rho context to rho hidden */
#endif
double linkdel_bir_r[MAXNUMRHO];        /* newphase input to rho hidden */
double linkdel_tro_r[MAXNUMRHO][NUMTRIG];       /* triggering operation to rho
                                                   hidden weight */
#if LABELTRIGPARENTINP
double linkdel_trp_r[MAXNUMRHO][MAXNUMLAB];     /* triggering parent to rho
                                                   hidden weight */
#else
double linkdel_trp_r[MAXNUMRHO][MAXNUMRHO];     /* triggering parent to rho
                                                   hidden weight */
#endif
#if LABELTRIGCHILDINP
double linkdel_trc_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];        /* triggering child to
                                                                   rho hidden weight */
#else
double linkdel_trc_r[MAXNUMRHO][MAXNUMRHO];     /* triggering child to rho
                                                   hidden weight */
#endif
double linkdel_lab_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];        /* chosen non/pre-term
                                                                   label to rho hidden
                                                                   weight */
double linkdel_pred_r[MAXNUMRHO];       /* prev prediction to rho hidden */
#if CHILDCONTEXT
#if LABELCHILDINP
double linkdel_chd_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];        /* previous child to rho
                                                                   hidden weight */
#else
double linkdel_chd_r[MAXNUMRHO][MAXNUMRHO];     /* previous child to rho
                                                   hidden weight */
#endif
#endif
#if HEADHIDDEN
#if LABELHEADCHILDINP
double linkdel_hdc_r[MAXNUMRHO][MAXNUMLAB + MAXNUMINP1];        /* head child to
                                                                   rho hidden weight */
#else
double linkdel_hdc_r[MAXNUMRHO][MAXNUMRHO];     /* head child to rho
                                                   hidden weight */
#endif
#endif
#if CHOPCONTEXT
double linkdel_cho_r[MAXNUMRHO][NUMCHOP];       /* previous child operation to rho
                                                   hidden weight */
#endif
#if HEADOUTS
double linkdel_nhd_r[MAXNUMRHO];        /* not headed to rho hidden weight */
#endif
double linkdel_r_o[MAXNUMOUT][MAXNUMRHO];       /* rho hidden to output */

/* define bias deltas */
double biasdel_r[MAXNUMRHO];    /* rho hidden */
double biasdel_o[MAXNUMOUT];    /* output */


int max_cl_levels = 255;
/* Default layer sizes */
int numinp1 = 46;
int numinp2 = 0;
#define numbins  1              /* should help compiling optimization */
int numinp3 = 0;
int numrho = 10;
int numout = 629;
int numoutmem = 399;
int numtarmem = 111;
int numlabels = 33;
int numout1 = 46;
int numpredwords = 512;
int labeloutputs = NUMSTRUC;    /* never reset */
#if HEADOUTS
int hdlabeloutputs = NUMSTRUC + 33;
int tagoutputs = NUMSTRUC + 33 + 33;
int wordoutputs = NUMSTRUC + 33 + 33 + 46;
#else
int tagoutputs = NUMSTRUC + 33;
int wordoutputs = NUMSTRUC + 33 + 46;
#endif



/** clustering of outputs for each out1 */
struct clustering **out_clusters_arr;


/* default input format specifications */
int inpwrdpos = 1;
int outwrdpos = 1;


/* default link numbers */
int numlink_tro = NUMTRIG;      /* never reset */
#if LABELTRIGPARENTINP
int numlink_trp = MAXNUMLAB;
#else
int numlink_trp = MAXNUMRHO;
#endif
#if LABELTRIGCHILDINP
int numlink_trc = MAXNUMLAB + MAXNUMINP1;
#else
int numlink_trc = MAXNUMRHO;
#endif
int numlink_lab = MAXNUMLAB + MAXNUMINP1;
#if CHILDCONTEXT
#if LABELCHILDINP
int numlink_chd = MAXNUMLAB + MAXNUMINP1;
#else
int numlink_chd = MAXNUMRHO;
#endif
#endif
#if HEADHIDDEN
#if LABELHEADCHILDINP
int numlink_hdc = MAXNUMLAB + MAXNUMINP1;
#else
int numlink_hdc = MAXNUMRHO;
#endif
#endif
#if CHOPCONTEXT
int numlink_cho = NUMCHOP;      /* never reset */
#endif

/* arrays of first word ids for each tag, and last+1 for tag+1 */
int inp1_first_word[MAXNUMINP1 + 1];
int out1_first_word[MAXNUMOUT1 + 1];
int out1_bin_first_word[MAXNUMOUT1][MAXNUMBINS + 1];


/** Parameters and default values **/

/* learning parameters */
int seed = 1;                   /* random weight seed seed */
double eta = 0.0002;            /* learning rate */
double mom = 0.9;               /* momentum */
double reg = 0.08;              /* multiplicative weight decay term /eta */
double decay = 0.99998;         /* 1-reg*eta */
int structypecount = 0;         /* number of structure outputs in training set */
double relreglabels = 1.0;      /* regularization for label outpus relative to
                                   structure outputs */
double relregtags = 1.0;        /* regularization for tag outpus relative to
                                   structure outputs */
double relregwords[MAXNUMOUT1]; /* regularization for word outpus relative 
                                   to structure outputs */

/* computation parameters */
int readcandidates = 0;
int readtraincandidates = 0;

/* learning parameter adjustment variables.  */
/* if autoreduce!=0 then eta = eta_inf*eta_red.  */
int autoreduce = 2;             /* flag for automatically reducing learning rate */
int saveoldwgt = 0;             /* flag for backtracking to old weight on error increase */
double eta_red = 1.0;           /* percentage of eta_inf remaining after reduction  (in
                                   .wgt header) */
double eta_inf = 0.0002;          /* eta taken from .inf file */
double eta_red_rate = 0.5;  /* muiltiplied by eta and added to compute
                                   eta_red during reduction */
int schedreduce = 0;            /* flag for scheduled reduction of eta */
double sched_red_rate = 0.5;        /* if used, every cycle the eta
                                           are multiplied by this rate */
/* if autoreduce_reg!=0 then reg = reg_inf-reg_red.  */
int autoreduce_reg = 1;         /* flag for automatically reducing reg */
double reg_red = 0.0;           /* subtracted from reg_inf for reduction  (in .wgt
                                   header) */
double reg_red_step = 0.5;      /* muiltiplied by reg and added to compute
                                   reg_red during reduction */
double reg_inf = 0.08;          /* reg taken from .inf file */
double finishing_prob = 1.0;    /* the probability to be used as the estimate
                                   for finishing constituents on the stack. */
double projection_bias = 0.0;   /* the amount to add to the net input of the
                                   PROJOUT output when testing */

/* testing parameters */
int predbeam = 50;
int testing_predbeam = 50;
int training_predbeam = 50;
int candidate_predbeam = 50;
int testing_candidate_predbeam = 50;
int training_candidate_predbeam = 50;
int lookahead_predbeam = 1;
int lookahead = 0;              /* 0 is flag for no lookahead pruning */
int phasebeam = 5;
int joint_error_flag = 1;
double expected_max_pred_factor = 5.0; 
    /* assume predictions won't be more than this times average in 
       best_options */

/* output format parameters */
int bracketed_output = 0;
int idtext_input = 0;
int numtopparses = 0;
int numtopderivs = 0;
int stdin_word_input = 0;
int stdin_candidate_input = 0;
int stdout_bracketed_output = 0;

/* global network state variables */
int sequence_space;             /* the sequence length before which space allocation
                                   must next occur. */
int state_space;                /* the number of phase states before which space allocation
                                   must next occur (not used?) */
int output_space;               /* the number of phases/indices before which space
                                   allocation must next occur */
int *period_phase_space;        /* period_phase_space[t] is the highest phase
                                   index allocated for period t */
int phase_space;                /* the number of phases before which space allocation must
                                   next occur for things which do not refer to period.
                                   phase_space >= max_t(period_phase_space[t]) */
int sentence_length;            /* the length of the last sentence to finish being
                                   input plus 1. */


/* Records of performance */
double sentence_prob;           /* estimated probability of current sentence */
double sentence_error;          /* the global error for current sentence */
double total_err;               /* the total error */
double total_local_err;         /* local computation of total_err */
double total_local_err_correction;      /* total_err = total_local_err + this */
int num_items;                  /* number of outputs   */
double total_targets;           /* sum of all the target values */
double total_struc_err, total_label_err, total_tag_err, total_word_err;
double total_struc_targets, total_label_targets;
double total_tag_targets, total_word_targets;
int num_sent;
int total_sent_correct;
int sent_unparsed = 0;
int total_const_correct, total_const_target, total_const_output;
int total100_const_correct, total100_const_target, total100_const_output;
int total40_const_correct, total40_const_target, total40_const_output;
int total_const_struct_correct;
int total_pch_correct, total_pch_target, total_pch_output;
#if RECHECKTESTING
int retest_correct, retest_target, retest_output;
#endif



/*********************** Start of Code ***********************/

/* print the usage information */
void usage_message(char *name)
{
    printf("Usage:\n");
    printf
        ("%s [-h] [-seed seed] [-sched rate] [-autorate step] [-autoreg step]\n",
         name);
    printf
        ("        [-condtrain] [-predbeam beam] [-trainbeam beam] [-candbeam beam]\n");
    printf
        ("        [-traincandbeam beam] [-lookaheadbeam beam lookahead]\n");
    printf
        ("        [-predfactor factor] [-branchfactor factor] [-finprob prob]\n");
    printf
        ("        [-projbias bias] [-top n] [-deriv n]\n");
    printf
        ("        [-cand candidatefile] [-traincand candidatefile] [-stdin] [-stdout]\n");
    printf
        ("        -eval | -parsedata | -parse | -train epochs | -valtrain frequency epochs\n");
    printf
        ("        filename [datafilename [validationfilename]]\n\n");
}

/* print the usage information and exit */
void usage_error(char *name)
{
    usage_message(name);
    error("");
}

/* print the long version usage information and exit */
void usage_help(char *name)
{
    printf("SSN Parser (copyright James Henderson and Ivan Titov)\n\n");
    usage_message(name);
    printf("  -h                   - display this message and exit.\n");
    printf
        ("  -seed seed           - used to initialize weight seeding (default %d).\n", 
	 seed);
    printf
        ("  -sched rate          - reduce learning rate on a schedule where\n");
    printf
        ("                         reduction=(1-rate)^cycle (default none).\n");
    printf
        ("  -autorate step       - automatically reduce learning rate by step proportion\n");
    printf
        ("                         of current rate when either training error or\n");
    printf
        ("                         validation error increases (default %e).\n",
	 1.0 - eta_red_rate);
    printf
        ("  -autoreg step        - automatically reduce weight decay regularization by\n");
    printf
        ("                         step proportion when validation error increases (default %e).\n",
	 1.0 - reg_red_step);
    printf
        ("  -condtrain           - use the conditional probability error rather than the\n");
    printf
        ("                         joint probability error for training.  Requires.\n");
    printf
        ("  -predbeam beam       - the width of the beam to be applied after each\n");
    printf
        ("                         prediction when searching for the best parse\n");
    printf
        ("                         (default %d).\n",
	 predbeam);
    printf
        ("  -trainbeam beam      - the width of the beam to be applied after each\n");
    printf
        ("                         prediction when training (default %d).\n",
	 predbeam);
    printf
        ("  -candbeam beam       - the width of the beam to be applied after each\n");
    printf
        ("                         prediction when reading the testing candidates\n");
    printf
        ("                         (default %d).\n",
	 predbeam);
    printf
        ("  -traincandbeam beam  - the width of the beam to be applied after each\n");
    printf
        ("                         prediction when reading the training candidates\n");
    printf
        ("                         (default %d).\n",
	 predbeam);
    printf
        ("  -lookaheadbeam beam lookahead  - like -predbeam except the pruning for a\n");
    printf
        ("                         given prediction is done based on the sum of the\n");
    printf
        ("                         derivations which descend from that prediction\n");
    printf
        ("                         lookahead predictions later (defaults to none).\n");
    printf
        ("  -predfactor factor   - the factor multiplied by the current average of\n");
    printf
        ("                         prediction probabilities in the beam to estimate\n");
    printf
        ("                         the maximum possible future prediction probability,\n");
    printf
        ("                         for pruning (default %e).\n",
	 expected_max_pred_factor);
    printf
        ("  -branchfactor width  - the width of the branching allowed from each\n");
    printf
        ("                         decision when searching for the best parse\n");
    printf
        ("                         (default %d).\n",
	 phasebeam);
    printf
        ("  -finprob prob        - the probability to be used as the estimate for\n");
    printf
        ("                         finishing constituents on the stack (default %e).\n",
	 finishing_prob);
    printf
        ("  -projbias bias       - the amount to add to the net input of the project\n");
    printf
        ("                         output when testing (default %e).\n",
	 projection_bias);
    printf
        ("  -top n               - create a .top file and save the n best parses\n");
    printf
        ("                         (default %d).\n",
	 numtopparses);
    printf
        ("  -deriv n             - print n best derivations to stdout (default %d).\n",
	 numtopderivs);
    printf
        ("  -cand candidatefile  - read the candidate derivations for testing from\n");
    printf
        ("                         <candidatefile>.deriv, or from stdin if\n");
    printf
        ("                         <candidatefile> = '-' (defaults to none).\n");
    printf
        ("  -traincand candidatefile  - read the candidate derivations for training from\n");
    printf
        ("                         <candidatefile>.deriv (defaults to none).\n");
    printf
        ("  -stdin               - with '-parse', read tagged word id input from stdin\n");
    printf
        ("                         (from candidate derivations if '-cand -').\n");
    printf
        ("  -stdout              - with '-parse' or '-parsedata', send parses to stdout.\n");
    printf("\n");
    printf
        ("  -eval                - append testing recall and precision to .b file.\n");
    printf
        ("                         Requires data file in training format.\n");
    printf
        ("  -parsedata           - create files with target (filename.gld) and output\n");
    printf
        ("                         (filename.res) trees.  Requires data file in training\n");
    printf
        ("                         format.\n");
    printf
        ("  -parse               - take idtext as input and create file\n");
    printf
        ("                         (filename.datafilename.res) with output trees.\n");
    
    
    printf
        ("  -train epochs        - train for epochs, append error to .out file.\n");
    printf
        ("  -valtrain frequency epochs  - same as -train except does -eval every\n");
    printf
        ("                         frequency, and saves best .wgt in .best.wgt.\n");
    printf("\n");
    printf
        ("  filename             - stem for .inf, .wgt, .out, .prog, .warn, and .b files.\n");
    printf
        ("  datafilename         - stem for .dat training file (defaults to filename).\n");
    printf
        ("  validationfilename   - stem for .dat testing file (defaults to datafilename).\n");
    error("");
}
/* 
 * read in the filenames (stems) (filename [datafilename [validationfile]]) 
 * from the command line, starting at argument position offset. 
 * and set all the needed files on the basis of these stems
 */
void _process_names(int argc, char *argv[], int offset) {
    char *file_name_stem, *data_name, *val_name;

    if (argc >= offset + 1) {
        file_name_stem = (char*) calloc(strlen(argv[offset]) + 1, sizeof(char));
        strcpy(file_name_stem, argv[offset]);
    } else {
        printf("Error: not enough options\n");
        usage_error(argv[0]);
    } 
    
    if (argc >= offset + 2) {
        data_name = (char*) calloc(strlen(argv[offset + 1]) + strlen(".dat") +  1, sizeof(char));  
        strcpy(data_name, argv[offset + 1]);
    } else {
        data_name = (char*) calloc(strlen(file_name_stem) + strlen(".dat") +  1, sizeof(char));  
        strcpy(data_name, file_name_stem);
    }
    strcat(data_name, ".dat");

    if (argc >= offset + 3) {
        val_name = (char*) calloc(strlen(argv[offset + 2]) + strlen(".dat") + 1, sizeof(char));
        strcpy(val_name, argv[offset + 2]);
        strcat(val_name, ".dat");

    } else {
        val_name = (char*) calloc(strlen(data_name) + strlen(".dat") +  1, sizeof(char));
        strcpy(val_name, data_name);
    }
    
    set_all_file_names(file_name_stem, data_name, val_name);
    
    free(file_name_stem);
    free(data_name);
    free(val_name);
    
}

/** set all the file names on the basis of given names: first is stem, other two includes suffix */
void set_all_file_names(char *file_name_stem, char *data_name, char *val_name) {

    strcpy(filename, file_name_stem);
    strcpy(weightfile, filename);
    strcat(weightfile, ".wgt");
    strcpy(newweightfile, filename);
    strcat(newweightfile, ".new.wgt");
    strcpy(valweightfile, filename);
    strcat(valweightfile, ".best.wgt");
    strcpy(inffile, filename);
    strcat(inffile, ".inf");
    strcpy(outfile, filename);
    strcat(outfile, ".out");
    strcpy(brafile, filename);
    strcat(brafile, ".b");
    strcpy(gldfile, filename);
    strcat(gldfile, ".gld");
    strcpy(resfile, filename);
    strcat(resfile, ".res");
    strcpy(topfile, filename);
    strcat(topfile, ".top");
#if FILEPROG
    strcpy(progfile, filename);
    strcat(progfile, ".prog");
#endif
#if FILEWARN|FILEDISP
    strcpy(warnfile, filename);
    strcat(warnfile, ".warn");
#endif
#if RECHECKTESTING
    strcpy(retestfile, filename);
    strcat(retestfile, ".check.dat");
#endif



    
    strcpy(datafile, data_name);
    strcpy(valdatafile, val_name);
}



/* run evaluation */
void evaluation()
{
    FILE *fp;
    int wpe;

    /*
       printf("initialising\n");
     */
    initialize();
    /*
       printf("reading weights\n");
     */
    read_weights(weightfile);
    wpe = weight_pretrnerr;
    if ((fp = fopen(newweightfile, "r")) != NULL) {
        fclose(fp);
        read_weights(newweightfile);
        weight_pretrnerr = wpe;
    }
    /*    display_net(); */

    /* don't test on datafile *
       strcpy(testingfile, datafile);  
       testing_epoch();
       if (idtext_input == 0)
         write_testing(0);
       if (strcmp(datafile,valdatafile) != 0)
       {
     */

    if (bracketed_output != 0) {        
        /* delete old gld and res files */
        if (idtext_input == 0) {
	    if ((fp = fopen(gldfile, "w")) != NULL)
	        fclose(fp);
	}
	if (stdout_bracketed_output == 0) {
	    if ((fp = fopen(resfile, "w")) != NULL)
	        fclose(fp);
	}
        if (numtopparses > 0) {
            if ((fp = fopen(topfile, "w")) != NULL)
                fclose(fp);
        }
    }

    strcpy(testingfile, valdatafile);   /* test on valdatafile */
    testing_epoch();
    if (idtext_input == 0)
        write_testing(0);
}

/* run learning for cycles epochs. */
void learning(int cycles)
{
    int i, c, step;
    int nextincrease;           /* number of cycles before eta gets increased */
    double wgttrnerr;           /* training error after weightfile */
    int wgtcycle;               /* value of c when weightfile created */
    int oldage;                 /* used in cycle cancellation */
    double oldred;              /* used in cycle cancellation */
    FILE *ofp;
    FILE *fp;
    double wpe;
    int wa;

    initialize();

    wgttrnerr = -1.0;

    read_weights(weightfile);

    wa = weight_age;
    wpe = weight_pretrnerr;
    if ((fp = fopen(newweightfile, "r")) != NULL) {
        fclose(fp);
        read_weights(newweightfile);
        wgtcycle = 0;
        if (wa != weight_age)
            wgttrnerr = weight_pretrnerr;
        weight_pretrnerr = wpe;
    } else if (autoreduce != 0 && saveoldwgt != 0) {
        save_weights(weightfile, 0.0);
        copy_weights(weightfile, newweightfile);
    }
    /* display_net(); */

    /* write training parameters to outfile as a record. */
    if ((ofp = fopen(outfile, "a")) != NULL) {
        fprintf(ofp,
                "\nStarting on %s.  eta=%.2e, mom=%.3f, reg=%.1e, beam=%d.\n",
                datafile, eta, mom, reg, training_predbeam);
        fclose(ofp);
    }


    if (autoreduce)
        step = 1;
    else if (cycles > 2000)
        step = 100;
    else if (cycles > 10)
        step = cycles / 10;
    else
        step = 1;

    for (c = 1; c <= cycles; c++) {
        if ((weight_age + 1) % step == 0 || c == cycles) {
            if (step > 1 && autoreduce != 0)    /* otherwise already saved */
                save_weights(newweightfile, 0.0);
            read_inf_changes();
            training_epoch();
            if (schedreduce != 0 && eta_red * sched_red_rate >= MAXRED) {
                eta_red *= eta_red_rate;
                eta = eta_inf * eta_red;
                decay = 1.0 - reg * eta;
            }
            if (autoreduce != 0 && wgttrnerr >= 0 &&
                total_err >
                wgttrnerr + REDCRIT * (weight_pretrnerr - wgttrnerr) &&
                weight_pretrnerr >= 0 && eta_red * eta_red_rate >= MAXRED)
            {
                if (saveoldwgt != 0) {
                    oldage = weight_age;
                    oldred = eta_red;
                    read_weights(weightfile);
                    copy_weights(weightfile, newweightfile);
                    if (eta_red > oldred)       /* weightfile's might be unreduced */
                        eta_red = oldred;
                }
                eta_red *= eta_red_rate;
                eta = eta_inf * eta_red;
                decay = 1.0 - reg * eta;
                if ((ofp = fopen(outfile, "a")) != NULL) {
                    if (saveoldwgt != 0) {
                        fprintf(ofp,
                                "Cancelled %d-%d cycles (Error: %.2f), Eta = %.2e.\n",
                                weight_age, oldage, total_err, eta);
                    } else {
                        fprintf(ofp,
                                "Reduced learning rate at cycle %d (Error: %.2f), Eta = %.2e.\n",
                                weight_age, total_err, eta);
                    }
                    fclose(ofp);
                } else {
                    if (saveoldwgt != 0) {
                        printf
                            ("Cancelled %d-%d cycles (Error: %.2f), Eta = %.2e.\n",
                             weight_age, oldage, total_err, eta);
                    } else {
                        printf
                            ("Reduced learning rate at cycle %d (Error: %.2f), Eta = %.2e.\n",
                             weight_age, total_err, eta);
                    }
                }
                if (saveoldwgt != 0) {
                    c = wgtcycle;       /* return to last test training epoch */
                    wgttrnerr = weight_pretrnerr;
                    continue;   /* skip to next loop
                                 */
                }
            } else {
                if (autoreduce == 0)
                    save_weights(weightfile, 0.0);
                else if (total_err <= wgttrnerr || wgttrnerr < 0.0) {
                    weight_pretrnerr = wgttrnerr;
                    wgttrnerr = total_err;
                    wgtcycle = c - 1;
                    if (saveoldwgt != 0) {
                        copy_weights(newweightfile, weightfile);
                    }
                    save_weights(newweightfile, 0.0);
                }
                if ((ofp = fopen(outfile, "a")) != NULL) {
                    fprintf(ofp,
                            "Cycle%4d: Err= %.0f (Pstruc=%.4f Plab=%.4f Ptag=%.4f Pword=%.4f).\n",
                            weight_age, total_err,
                            exp(-total_struc_err / total_struc_targets),
                            exp(-total_label_err / total_label_targets),
                            exp(-total_tag_err / total_tag_targets),
                            exp(-total_word_err / total_word_targets));
                    fclose(ofp);
                } else
                    printf
                        ("Cycle%4d: Err= %.0f (Pstruc=%.4f Plab=%.4f Ptag=%.4f Pword=%.4f).\n",
                         weight_age, total_err,
                         exp(-total_struc_err / total_struc_targets),
                         exp(-total_label_err / total_label_targets),
                         exp(-total_tag_err / total_tag_targets),
                         exp(-total_word_err / total_word_targets));
            }
            if (total_err == 0)
                break;
        } else {
            read_inf_changes();
            training_epoch();
            if (schedreduce != 0 && eta_red * sched_red_rate >= MAXRED) {
                eta_red *= eta_red_rate;
                eta = eta_inf * eta_red;
                decay = 1.0 - reg * eta;
            }
        }
    }                           //for c from 1 to cycles
    if ((ofp = fopen(outfile, "a")) != NULL) {
        fprintf(ofp,
                "%3d: Total error: %f=%f+%f, ave prob: %f\n  (%.0f/%d targets ones).\n",
                weight_age, total_err, total_local_err,
                total_local_err_correction,
                exp(-1.0 * total_err / total_targets), total_targets,
                num_items);
        fprintf(ofp,
                "  P(structure)=%.4f^%.0f, P(label)=%.4f^%.0f,\n  P(tag)=%.4f^%.0f, P(word)=%.4f^%.0f\n\n",
                exp(-total_struc_err / total_struc_targets),
                total_struc_targets,
                exp(-total_label_err / total_label_targets),
                total_label_targets,
                exp(-total_tag_err / total_tag_targets), total_tag_targets,
                exp(-total_word_err / total_word_targets),
                total_word_targets);
        fclose(ofp);
    } else {
        printf
            ("%3d: Total error: %f=%f+%f, ave prob: %f\n  (%.0f/%d targets ones).\n",
             weight_age, total_err, total_local_err,
             total_local_err_correction,
             exp(-1.0 * total_err / total_targets), total_targets,
             num_items);
        printf
            ("  P(structure)=%.4f^%.0f, P(label)=%.4f^%.0f,\n  P(tag)=%.4f^%.0f, P(word)=%.4f^%.0f\n\n",
             exp(-total_struc_err / total_struc_targets),
             total_struc_targets,
             exp(-total_label_err / total_label_targets),
             total_label_targets, exp(-total_tag_err / total_tag_targets),
             total_tag_targets, exp(-total_word_err / total_word_targets),
             total_word_targets);
    }


}
//!EM_CLUSTERING

/* training with stopping on cross validation */
void validation_learning(int test_freq, int cycles) /**/
{
    int i, c, step;
    int nextincrease;           /* number of cycles before eta gets increased */
    double score;
    double lastscore;
    int numsub;
    double wgttrnerr;           /* training error after weightfile */
    int wgtcycle;               /* value of c when weightfile created */
    int oldage;                 /* used during cycle cancellation */
    double oldred;              /* used in cycle cancellation */
    FILE *bfp, *wfp;
    FILE *ofp;
    FILE *fp;
    double wpe;
    int wa;
    double rr;
    double max_score = 0;
    

    initialize();
    strcpy(testingfile, valdatafile);   /* all testing on valdatafile */

    numsub = 0;
    wgttrnerr = -1.0;

    /* if valweightfile exists then read max_score from header, or calculate */
    if ((wfp = fopen(valweightfile, "r")) != NULL) {
        fclose(wfp);
        max_score = read_weights(valweightfile);
        if (max_score <= 0.0) {
            testing_epoch();
#if ERRVALIDATE
            max_score = exp(-1.0 * total_err / total_targets);
#else
            max_score = ((double) (total_const_correct * 200))
                / ((double) (total_const_target + total_const_output));
#endif
            write_testing(0);
        }
        lastscore = max_score;
    } else {
        max_score = 0.0;
        lastscore = 0.0;
    }
    read_weights(weightfile);
    wa = weight_age;
    wpe = weight_pretrnerr;
    if ((fp = fopen(newweightfile, "r")) != NULL) {
        fclose(fp);
        read_weights(newweightfile);
        wgtcycle = 0;
        if (wa != weight_age)
            wgttrnerr = weight_pretrnerr;
        weight_pretrnerr = wpe;
    } else if (autoreduce != 0 && saveoldwgt != 0) {
        save_weights(weightfile, 0.0);
        copy_weights(weightfile, newweightfile);
    }
    /* display_net(); */

    /* write training parameters to outfile as a record. */
    if ((ofp = fopen(outfile, "a")) != NULL) {
        fprintf(ofp,
                "\nStarting on %s.  eta=%.2e, mom=%.3f, reg=%.1e, beam=%d.\n",
                datafile, eta, mom, reg, training_predbeam);
        fclose(ofp);
    }

    if (test_freq % 5 == 0)     /* test_freq must be divisable by step */
        step = test_freq / 5;
    else
        step = 1;

    for (c = 1; c <= cycles; c++) {

        /* calculate training error every step cycles  (normally always true)*/
        if ((weight_age + 1) % step == 0 || c == cycles) {
            if (step > 1 && autoreduce != 0)    /* otherwise already saved */
                save_weights(newweightfile, 0.0);
            read_inf_changes();
            training_epoch();
            if (schedreduce != 0 && eta_red * sched_red_rate >= MAXRED) {
                eta_red *= eta_red_rate;
                eta = eta_inf * eta_red;
                decay = 1.0 - reg * eta;
            }
            if (autoreduce != 0 && wgttrnerr >= 0 &&
                total_err >
                wgttrnerr + REDCRIT * (weight_pretrnerr - wgttrnerr) &&
                weight_pretrnerr >= 0 && eta_red * eta_red_rate >= MAXRED)
            {
                if (saveoldwgt != 0) {
                    oldage = weight_age;
                    oldred = eta_red;
                    read_weights(weightfile);
                    copy_weights(weightfile, newweightfile);
                    if (eta_red > oldred)       /* weightfile's eta_red maybe less */
                        eta_red = oldred;
                }
                eta_red *= eta_red_rate;
                eta = eta_inf * eta_red;
                decay = 1.0 - reg * eta;
                if ((ofp = fopen(outfile, "a")) != NULL) {
                    if (saveoldwgt != 0) {
                        fprintf(ofp,
                                "Cancelled %d-%d cycles (Error: %.2f), Eta = %.2e.\n",
                                weight_age, oldage, total_err, eta);
                    } else {
                        fprintf(ofp,
                                "Reduced learning rate at cycle %d (Error: %.2f), Eta = %.2e.\n",
                                weight_age, total_err, eta);
                    }
                    fclose(ofp);
                } else {
                    if (saveoldwgt != 0) {
                        printf
                            ("Cancelled %d-%d cycles (Error: %.2f), Eta = %.2e.\n",
                             weight_age, oldage, total_err, eta);
                    } else {
                        printf
                            ("Reduced learning rate at cycle %d (Error: %.2f), Eta = %.2e.\n",
                             weight_age, total_err, eta);
                    }
                }
                if (saveoldwgt != 0) {
                    c = wgtcycle;       /* return to last test training epoch */
                    wgttrnerr = weight_pretrnerr;
                    continue;   /* skip to next loop */
                }
            } else {
                if (autoreduce == 0)
                    save_weights(weightfile, 0.0);
                else if (total_err <= wgttrnerr || wgttrnerr < 0.0) {
                    weight_pretrnerr = wgttrnerr;
                    wgttrnerr = total_err;
                    wgtcycle = c - 1;
                    if (saveoldwgt != 0) {
                        copy_weights(newweightfile, weightfile);
                    }
                    save_weights(newweightfile, 0.0);
                }
                if ((ofp = fopen(outfile, "a")) != NULL) {
                    fprintf(ofp,
                            "Cycle%4d: Err= %.0f (Pstruc=%.4f Plab=%.4f Ptag=%.4f Pword=%.4f).\n",
                            weight_age, total_err,
                            exp(-total_struc_err / total_struc_targets),
                            exp(-total_label_err / total_label_targets),
                            exp(-total_tag_err / total_tag_targets),
                            exp(-total_word_err / total_word_targets));
                    fclose(ofp);
                } else
                    printf
                        ("Cycle%4d: Err= %.0f (Pstruc=%.4f Plab=%.4f Ptag=%.4f Pword=%.4f).\n",
                         weight_age, total_err,
                         exp(-total_struc_err / total_struc_targets),
                         exp(-total_label_err / total_label_targets),
                         exp(-total_tag_err / total_tag_targets),
                         exp(-total_word_err / total_word_targets));
            }
            if (total_err == 0)
                break;
        } else {
            read_inf_changes();
            training_epoch();
            if (schedreduce != 0 && eta_red * sched_red_rate >= MAXRED) {
                eta_red *= eta_red_rate;
                eta = eta_inf * eta_red;
                decay = 1.0 - reg * eta;
            }
        }

        /* calculate validation error every test_freq cycles */
        if (weight_age % test_freq == 0 || c == cycles) {
            if ((ofp = fopen(outfile, "a")) != NULL) {
                fprintf(ofp,
                        "%3d: Total error: %f=%f+%f, ave prob: %f\n  (%.0f/%d targets ones).\n",
                        weight_age, total_err, total_local_err,
                        total_local_err_correction,
                        exp(-1.0 * total_err / total_targets),
                        total_targets, num_items);
                fprintf(ofp,
                        "  P(structure)=%.4f^%.0f, P(label)=%.4f^%.0f,\n  P(tag)=%.4f^%.0f, P(word)=%.4f^%.0f\n\n",
                        exp(-total_struc_err / total_struc_targets),
                        total_struc_targets,
                        exp(-total_label_err / total_label_targets),
                        total_label_targets,
                        exp(-total_tag_err / total_tag_targets),
                        total_tag_targets,
                        exp(-total_word_err / total_word_targets),
                        total_word_targets);
                fclose(ofp);
            } else {
                printf
                    ("%3d: Total error: %f=%f+%f, ave prob: %f\n  (%.0f/%d targets ones).\n",
                     weight_age, total_err, total_local_err,
                     total_local_err_correction,
                     exp(-1.0 * total_err / total_targets), total_targets,
                     num_items);
                printf
                    ("  P(structure)=%.4f^%.0f, P(label)=%.4f^%.0f,\n  P(tag)=%.4f^%.0f, P(word)=%.4f^%.0f\n\n",
                     exp(-total_struc_err / total_struc_targets),
                     total_struc_targets,
                     exp(-total_label_err / total_label_targets),
                     total_label_targets,
                     exp(-total_tag_err / total_tag_targets),
                     total_tag_targets,
                     exp(-total_word_err / total_word_targets),
                     total_word_targets);
            }

            /* test on valdatafile and if best save in valweightfile */
            testing_epoch();
#if ERRVALIDATE
            score = exp(-1.0 * total_err / total_targets);
#else
            score = ((double) (total_const_correct * 200))
                / ((double) (total_const_target + total_const_output));
#endif
            if (score > max_score) {
                max_score = score;
                numsub = 0;
                save_weights(valweightfile, score);
                write_testing(1);
            } else {
                write_testing(0);
                /* if (prewgttrnerr >= 0) * don't count if auto-reducing */
                numsub++;
            }

            /* if validation error increases */
            if (lastscore > score) {
                /* termination test, given validation error has increased */
                if (numsub >= STOPAFTER) {
                    if ((bfp = fopen(brafile, "a")) == NULL)
                        error
                            ("Error: Cannot open bracket file to write data.");
                    fprintf(bfp,
                            "%3d: Ending with best const F1: %9.6f%%\n\n",
                            weight_age, max_score);
                    fclose(bfp);

                    break;      /* exit cycle loop */
                }
                /* for -av option */
                if (autoreduce == 2 && wgttrnerr >= 0 &&
                    weight_pretrnerr >= 0
                    && eta_red * eta_red_rate >= MAXRED) {
                    eta_red *= eta_red_rate;
                    eta = eta_inf * eta_red;
                    decay = 1.0 - reg * eta;
                    if ((ofp = fopen(outfile, "a")) != NULL) {
                        fprintf(ofp,
                                "Reduced learning rate at cycle %d, Eta = %.2e.\n",
                                weight_age, eta);
                        fclose(ofp);
                    } else
                        printf
                            ("Reduced learning rate at cycle %d, Eta = %.2e.\n",
                             weight_age, eta);
                }
#if REGULARIZE
                /* auto reduction of regularization */
                if (autoreduce_reg != 0 && reg_inf > reg_red) {
                    rr = reg_red + (reg * reg_red_step);
                    if (reg_inf > rr)
                        reg_red = rr;
                    else
                        reg_red = reg_inf;
                    reg = reg_inf - reg_red;
                    decay = 1.0 - reg * eta;
                    if ((ofp = fopen(outfile, "a")) != NULL) {
                        fprintf(ofp,
                                "Reduced weight decay at cycle %d, Reg = %.2e.\n",
                                weight_age, reg);
                        fclose(ofp);
                    } else
                        printf
                            ("Reduced weight decay at cycle %d, Reg = %.2e.\n",
                             weight_age, reg);
                }
#endif
            }

            lastscore = score;
        }

    }
}
//!EM_CLUSTERING




/* Trains the network on each sentence in the data set, updating after */
/* each sentence.   A summary of the error */
/* is computed. */
void training_epoch()
{
    FILE *dfp, *fptrace;
    FILE *nfp;
    FILE *cfp;
    int i, p;
    struct option_stack *ol;

    predbeam = training_predbeam;
    candidate_predbeam = training_candidate_predbeam;

#if DISPITER
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp, "Starting training cycle %d\n", weight_age + 1);
        fclose(warnfp);
    } else
#endif
        printf("Starting training cycle %d\n", weight_age + 1);
#endif

    if ((dfp = fopen(datafile, "r")) == NULL)
        error("Error: Cannot open (data)filename.dat to read data.");
    if (readtraincandidates != 0) {
        if ((cfp = fopen(traincandfile, "r")) == NULL)
            error
                ("Error: Cannot open traincandidatefile.deriv to read data.");
    } else
        cfp = NULL;

    total_err = 0.0;
    total_local_err = 0.0;
    total_local_err_correction = 0.0;
    num_items = 0;
    total_targets = 0.0;
    total_struc_err = 0.0;
    total_label_err = 0.0;
    total_tag_err = 0.0;
    total_word_err = 0.0;
    total_struc_targets = 0.0;
    total_label_targets = 0.0;
    total_tag_targets = 0.0;
    total_word_targets = 0.0;

    num_sent = 0;

    while (opt_comp_sentence_act(dfp, cfp) >= 0)        /* test for end of file */
    {
        opt_comp_sentence_del(sentence_length);

        update_weights(sentence_length);

        num_sent++;
    }

    fclose(dfp);
    if (readtraincandidates != 0) {
        if (cfp != NULL)
	    fclose(cfp);
    }

    weight_age++;               /* increment current weight's number of training epochs */

    predbeam = testing_predbeam;
    candidate_predbeam = testing_candidate_predbeam;
}


/* load the next sentence from datafile_p into the input and target
     derivation, compute sentence outputs.  Also computes derivation
     alternatives if joint_error_flag==0, and sets target ancesters (?***).
     Returns sentence length (including `^^^'), or -1 if end of file.  Assumes
     doing training. */
int opt_comp_sentence_act(FILE * datafile_p, FILE * candfile_p)
{
    char buffer[MAXDATALINE];
    int i, t, p, b;
    int length;
    int flag;
    struct option *opt;
    struct option_stack *ol;

#if DISPSEN
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
#if DISPSEN==2
        fprintf(warnfp, "Sen%d, ", num_sent + 1);
#else
        fprintf(warnfp, "Sentence %d\n", num_sent + 1);
#endif
        fclose(warnfp);
    } else
#endif
    {
#if DISPSEN==2
        printf("Sen%d, ", num_sent + 1);
        fflush(stdout);
#else
        printf("Sentence %d\n", num_sent + 1);
#endif
    }
#endif
#if FILEPROG
    if ((progfp = fopen(progfile, "w")) != NULL) {
        fprintf(progfp, "Computing sentence %d, during cycle %d.\n",
                num_sent + 1, weight_age + 1);
        fprintf(progfp, "    prior error per sentence: %f,\n",
                total_err / ((double) num_sent));
        fprintf(progfp,
                "    -ln(P(.)): tot=%.5f,str=%.5f,lab=%.5f,tag=%.5f,wrd=%.5f.\n",
                total_err / total_targets,
                total_struc_err / total_struc_targets,
                total_label_err / total_label_targets,
                total_tag_err / total_tag_targets,
                total_word_err / total_word_targets);
        fclose(progfp);
    }
#endif

#if CHECKANTIGARBAGE
    if (anti_garbage_options != NULL) {
        warning("not all options freed");
        for (ol = anti_garbage_options; ol != NULL; ol = ol->next)
            display_option(ol->elem, 0);
    }
#endif

    /* load all periods for sentence */
    init_phases_for_options();  /* initialize with matrix root */
    length = read_target_derivation(datafile_p);        /* sets target_derivation */
    if (length < 0) {           /* flag for end of file */
        /* free the unused initializing target option */
        if (target_derivation_next != target_derivation)
            free_option_branch(target_derivation->elem);
        target_derivation_next = target_derivation;
        return (-1);
    }
    for (t = 0; t <= length; t++)
        best_options[t][0] = NULL;

    comp_option_list_act(target_derivation, target_derivation_next);


    if (joint_error_flag == 0) {
        if (readtraincandidates != 0) {
            if (read_candidate_derivations(candfile_p, 1, 0) == -1) {
                warning
                    ("Not enough candidate derivations in traincandidatefile.");
            }
        } else
            candidate_derivations = NULL;
        comp_sentence_alternatives(target_derivation,
                                   target_derivation_next, length);
    }

    return (length);
}



/* Compute the error and delta terms for current sentence, given derivations
     in target_derivation and best_options. Frees all options */
void opt_comp_sentence_del(int length)
{
    int i, t, p, b;
    int bin;
    int *pred_word;
    struct option *opt, *prevopt;
    struct option_stack *ol, *ol2;
    struct option *last_target_opt;
    double norm;
    double *herr;
    double bp;
    double target_logp;
    double slack, total_slack;  /* derivative of slack vector norm */
    struct option **last_best_options;  /* best_options[length - 1] */
    int child_id;
    int memoffset, mem_out, layer_idx;



    last_best_options = best_options[length - 1];

    if (target_derivation == target_derivation_next)
        warning("this shouldn't happen");

     /** for SIGMOIDDISC=1 should remove sub-threshold derivations here,
      or never generate them at all in comp_sentence_alternatives(). **/

    /* put all the options in a single list, in an order compatible with all
       the derivation orderings */
    training_options_next = training_options;
    for (ol = target_derivation; ol != target_derivation_next;
         ol = ol->next) {
        training_options_next->elem = ol->elem;
        training_options_next =
            incr_option_stack_next(training_options_next);
    }
    /*{{{ */
    if (joint_error_flag == 0) {                
        /* use scratch_options as scratch, and in_best as flag */
        for (b = 0; b < predbeam && last_best_options[b] != NULL; b++) {
            for (opt = last_best_options[b];
                 opt != NULL && opt->in_target == 0;
                 opt = opt->previous_option) {
                opt->in_best = 1;
            }
        }
        for (b = 0; b < predbeam && last_best_options[b] != NULL; b++) {
            scratch_options_next = scratch_options;
            for (opt = last_best_options[b]; opt != NULL;
                 opt = opt->previous_option) {
                if (opt->in_best != 0 && opt->in_target == 0) {
                    scratch_options_next->elem = opt;
                    scratch_options_next =
                        incr_option_stack_next(scratch_options_next);
                    opt->in_best = 0;
                } else
                    break;
            }
            /* new options are now in reverse order */
            for (ol = scratch_options_next->prev; ol != NULL;
                 ol = ol->prev) {
             /** just for error detection **/
/*                for (ol2 = training_options_next->prev; ol2 != NULL;
                     ol2 = ol2->prev) {
                    if (ol->elem == ol2->elem)
                        warning("option in training list twice");
                } */
                training_options_next->elem = ol->elem;
                training_options_next =
                    incr_option_stack_next(training_options_next);
            }
        }
    }



    /*}}} */
    /* eo joint_error_flag == 0 */
    /* */
    /* compute output deltas */
    last_target_opt = target_derivation_next->prev->elem;
    if (joint_error_flag == 0) {        /*{{{ */
#if PLUGLEAKS
#if ! LOCALESTIMATE
        last_target_opt->pred_loss = 1.0;
        last_target_opt->pred_loss_weight = 1.0;
        for (b = 0; b < predbeam && last_best_options[b] != NULL; b++) {
            last_best_options[b]->pred_loss = 1.0;
            last_best_options[b]->pred_loss_weight = 1.0;
        }
        for (ol = training_options_next->prev; ol != NULL; ol = ol->prev) {
            if (ol->elem->is_pred_opt) {
                total_finish_probs[ol->elem->period] +=
                    ol->elem->pred_loss * exp(ol->elem->lprob);
                total_finish_prob_weights[ol->elem->period] +=
                    exp(ol->elem->lprob);
            } else if (ol->elem->pred_loss_weight != 0)
                ol->elem->pred_loss /= ol->elem->pred_loss_weight;
            else
                warning("zero pred_loss_weight shouldn't happen.");
            if (ol->elem->previous_option != NULL) {
                ol->elem->previous_option->pred_loss +=
                    ol->elem->pred_loss *
                    ol->elem->previous_option->outputs->act[ol->elem->
                                                            previous_output];
                if (!ol->elem->previous_option->is_pred_opt
                    && ol->elem->previous_option->firstout <=
                    ol->elem->previous_option->lastout) {
                    ol->elem->previous_option->pred_loss_weight +=
                        ol->elem->previous_option->outputs->act[ol->elem->
                                                                previous_output];
                } else {
                    ol->elem->previous_option->pred_loss_weight = 1.0;
                }
            }
        }
        ave_finish_probs[length] = 1.0;
        for (t = length - 1; t >= 0; t--) {
            ave_finish_probs[t] =
                total_finish_probs[t] / total_finish_prob_weights[t];
        }
#endif//!LOCALESTIMATE
        last_target_opt->backward_prob = exp(last_target_opt->lprob);
        last_target_opt->backward_knownout = 1.0;
        for (b = 0; b < predbeam && last_best_options[b] != NULL; b++) {
            last_best_options[b]->backward_prob =
                exp(last_best_options[b]->lprob);
            last_best_options[b]->backward_knownout = 1.0;
        }
#else
        /* normalize final target and last_best_options probs to get backward_probs */
        norm =
            tail_estimate(exp(target_derivation_next->prev->elem->lprob),
                          last_best_options, predbeam);
        norm += exp(last_target_opt->lprob);
        for (b = 0; b < predbeam && last_best_options[b] != NULL; b++) {
            norm += exp(last_best_options[b]->lprob);
        }
        sentence_prob = norm;   /* includes tail estimate */
        last_target_opt->backward_prob =
            exp(last_target_opt->lprob) / norm;
        /* last_target_opt->backward_prob -= 1.0; */
        for (b = 0; b < predbeam && last_best_options[b] != NULL; b++) {
            last_best_options[b]->backward_prob =
                exp(last_best_options[b]->lprob) / norm;
        }
#endif
        for (ol = training_options_next->prev; ol != NULL; ol = ol->prev) {
#if PLUGLEAKS
          /** Here was bug, removed in dglc3f (condition below isn't needed) */
/*            if (ol->elem->previous_output <= predtag_to_output(0) ||
                ol->elem->previous_output >=
                predtag_to_output(numout1 - 1)) {  */
#if LOCALESTIMATE
                if (ol->elem->backward_knownout != 0)
                    ol->elem->backward_prob /= ol->elem->backward_knownout;
                else
                    warning("zero backward_knownout shouldn't happen.");
#else
                /* default sum */
                ol->elem->backward_prob +=
                    exp(ol->elem->lprob) *
                    ave_finish_probs[ol->elem->period] * (1.0 -
                                                          ol->elem->
                                                          backward_knownout);
#endif
           /* }  */
#endif
            if (ol->elem->previous_option != NULL) {
                ol->elem->previous_option->backward_prob +=
                    ol->elem->backward_prob;
#if PLUGLEAKS
                if (!ol->elem->previous_option->is_pred_opt &&
                    ol->elem->previous_option->firstout <=
                    ol->elem->previous_option->lastout)
                    ol->elem->previous_option->backward_knownout +=
                        ol->elem->previous_option->outputs->act[ol->elem->
                                                                previous_output];
                else
                    ol->elem->previous_option->backward_knownout = 1.0;
#endif
            }
        }
#if PLUGLEAKS
        /* normalize all backward_probs using total of all derivations */
        norm = training_options->elem->backward_prob;
        sentence_prob = norm;   /* includes estimates of pruned */
        /* printf("total sentence probability = %e\n", norm); */
        for (ol = training_options_next->prev; ol != NULL; ol = ol->prev) {
            ol->elem->backward_prob /= norm;
        }
#endif
    }
    /* eo jerr == 0 */
    /* */
    /*}}} */

    for (ol = training_options_next->prev; ol != NULL; ol = ol->prev) {
        comp_default_output_deltas(ol->elem);
    }
    /* update output deltas for non-default cases */
    for (ol = training_options_next->prev; ol != NULL; ol = ol->prev) {
         opt = ol->elem;
         prevopt = opt->previous_option;
         if (prevopt != NULL && prevopt->firstout <= prevopt->lastout) {
             if (joint_error_flag != 0
#if ONLYPREDCONDDEL
                 || !prevopt->is_pred_opt
#endif
                 ) {
                 prevopt->outputs->del[opt->previous_output] +=
                     (double) opt->in_target;
             } else {
#if ! PLUGLEAKS
                prevopt->outputs->del[opt->previous_output] +=
                    (double) opt->in_target - opt->backward_prob;
#else
                if (prevopt->is_pred_opt) {
                    prevopt->outputs->del[opt->previous_output] +=
                        (double) opt->in_target - opt->backward_prob;
                } else {
#if LOCALESTIMATE
                    prevopt->outputs->del[opt->previous_output] +=
                        (double) opt->in_target - opt->backward_prob
                        + prevopt->outputs->act[opt->previous_output]
                        * prevopt->backward_prob;
#else
                    prevopt->outputs->del[opt->previous_output] +=
                        (double) opt->in_target - opt->backward_prob
                        + prevopt->outputs->act[opt->previous_output]
                        * exp(prevopt->lprob) *
                        ave_finish_probs[opt->period];
#endif
                }

#endif
            }
        }

        if (opt->new_state != 0 && opt->next_options != NULL) {
            herr = opt->state->err;
            for (i = 0; i < numrho; i++)
                herr[i] = 0.0;
        }

    }
    /********************************************************
    for (ol = training_options_next->prev; ol != NULL; ol = ol->prev) {
      printf("Deltas/outputs for %d at %d, after %d, target? %d, slack %f:\n  ", 
	     ol->elem->phase, ol->elem->period, ol->elem->previous_output,
	     ol->elem->in_target, ol->elem->backward_prob);
      for (i = ol->elem->firstout; i <= ol->elem->lastout; i++)
	printf("%f, ", ol->elem->outputs->del[i]);
      printf("\n  ");
      for (i = ol->elem->firstout; i <= ol->elem->lastout; i++)
	printf("%f, ", ol->elem->outputs->act[i]);
      printf("\n");
      if (ol->elem->previous_option != NULL) {
	printf("  deltas for previous (%d at %d, after %d, target? %d, slack %f):\n    ", 
	       ol->elem->previous_option->phase, ol->elem->previous_option->period, ol->elem->previous_option->previous_output,
	       ol->elem->previous_option->in_target, ol->elem->previous_option->backward_prob);
	for (i = ol->elem->previous_option->firstout; i <= ol->elem->previous_option->lastout; i++)
	  printf("%f, ", ol->elem->previous_option->outputs->del[i]);
	printf("\n    ");
	for (i = ol->elem->previous_option->firstout; i <= ol->elem->previous_option->lastout; i++)
	  printf("%f, ", ol->elem->previous_option->outputs->act[i]);
	printf("\n");


      }
    }
    ********************************************************/

    /* compute total output error */
    if (joint_error_flag == 0) {
        sentence_error = -(last_target_opt->lprob - log(sentence_prob));
        total_local_err_correction += log(sentence_prob);
    } else {
        sentence_error = -last_target_opt->lprob;
    }
    total_err += sentence_error;

    /* compute hidden errors and link deltas */
/*}}} */



    /* hidden rho errors initialized above */
    for (ol = training_options_next->prev; ol != NULL; ol = ol->prev) {
#if ERRSUMMARY
        opt_comp_err_summary(ol->elem);

#else
        total_err = 1.0;
#endif
        opt_comp_hidden_del(ol->elem);


#if CHECKACTDEL
        opt_check_output_act_del(ol->elem, 1);
#endif
    }


/**********************************

#if DEBUGCOMP || DEBUGTRAIN
    display_states(max_phase[length]);
    printf("Computed phases for each period,\n");
    for (t = 0; t <= length; t++)
      {
        printf("  %2d: ", t);
        display_stack(&deriv_order[t]);
      }
    printf("Max phase for each period,\n");
    for (t = 0; t <= length; t++)
      printf("  period %2d: %3d\n", t, max_phase[t]);
#endif

****************************/

    /* free all the options used */
    if (training_options_next != training_options)
        free_option_branch(training_options->elem);
    training_options_next = training_options;
}


/* find best alternatives to derivation and set best_options.  */
void comp_sentence_alternatives(struct option_stack *derivation,
                                struct option_stack *derivation_end,
                                int length)
{
    int i, t, p, b;
    struct option_stack *queue;
    struct option_stack *deriv;
    struct option_stack *ol, *ol2, *ol3;
    struct option *opt;
    double alternative_threshold;

    /* no alternatives needed is predbeam == 0 */
    if (predbeam == 0)
        return;

    if (derivation == derivation_end)
        error("empty target derivation");

    alternative_threshold = log(0.0);

    /** bug resolved in dglc3e: worst_best_threshold wasn't properly initialized => wrong result generated_options 
     *  for invokations before search_options (options are removed)
     * */
    worst_best_option_lprob = alternative_threshold;
    expected_max_pred_lprob = 0.0; 
    
    /* start after first prediction */
    for (deriv = derivation;
         deriv->elem->period <= 0 && deriv != derivation_end;
         deriv = deriv->next) {
    }
    /* no incorrect alternatives to first prediction */
    for (b = 0; b < predbeam; b++)
        best_options[0][b] = NULL;

    /* find the best alternative options for the sentence */
    for (t = 1; t < length; t++) {
/*}}} */

        queue = NULL;
        /* deriv initialized above (no alternative to last option) */
        for (; deriv != derivation_end->prev && deriv->elem->period <= t;
             deriv = deriv->next) {
            if (!deriv->elem->is_pred_opt) {
                ol = generated_options(deriv->elem, length);
                /* remove target and excess options from list */
                if (ol != NULL) {
                    if (ol->elem->previous_output ==
                        deriv->next->elem->previous_output) {
                        if (ol->elem != deriv->next->elem)
                            free_option_descendents(ol->elem);
                        ol = popfree_option_stack(ol);
                    } else {
                        i = 1;
                        for (ol2 = ol->next; ol2 != NULL; ol2 = ol3) {
                            i++;
                            ol3 = ol2->next;
                            if (ol2->elem->previous_output ==
                                deriv->next->elem->previous_output) {
                                if (ol2->elem != deriv->next->elem)
                                    free_option_descendents(ol2->elem);
                                pluckfree_option_stack(ol2);
                            } else if (i > phasebeam) {
                                free_option_descendents(ol2->elem);
                                pluckfree_option_stack(ol2);
                            }
                        }
                    }
                }               /* eo ol != NULL */
                if (ol != NULL)
                    queue = mergesort_options(queue, ol);
            }                   /* eo != is_pred_opt */
        }
        /* sets best_options[t] */
        search_options(t, queue, alternative_threshold, length);

        if (predbeam != 0 && best_options[t][0] == NULL &&
            candidate_derivations == NULL)
            warning("no predicting option found.");
        if (predbeam != 0) {
/*}}} */
            /* check for zero prob options, except top */
            for (b = predbeam - 1; b >= 1; b--)
                if (best_options[t][b] != NULL) {
                    if (isinf(best_options[t][b]->lprob)) {
                        warning
                            ("zero probability derivation in best. Resetting.");
                        if (b != 0) {
                            best_options[t][b]->lprob =
                                best_options[t][b - 1]->lprob - log(2.0);
                            best_options[t][b]->ranking_lprob =
                                best_options[t][b - 1]->ranking_lprob -
                                log(2.0);
                        } else {
                            best_options[t][b]->lprob = log(0.1);
                            best_options[t][b]->ranking_lprob = log(0.1);
                        }
                    }
                }

            /* set previous prediction option indexes for next period */
            for (b = 0; b < predbeam && best_options[t][b] != NULL; b++) {
                best_options[t][b]->prev_pred_ind = b;
            }
        }
    }
/*}}} */

}


/* set opt's output deltas to their value given nothing known about future
     derivation.  */
void comp_default_output_deltas(struct option *opt)
{
    int out;
    double bp;
    int *pred_word;

    /* no output computed for end cases */
    if (opt->firstout > opt->lastout)
        return;

    if (joint_error_flag != 0) {
        if (opt->in_target != 0) {
            for (out = opt->firstout; out <= opt->lastout; out++) {
                opt->outputs->del[out] = -opt->outputs->act[out];
            }
        } else {
            for (out = opt->firstout; out <= opt->lastout; out++) {
                opt->outputs->del[out] = 0.0;
            }
        }
    }

    
    else if (opt->is_pred_opt) {
#if CHECKOVERFLOW
        if (!in_range(opt->firstout, opt->lastout, 0, MAXNUMOUTMEM))
            error("CHECKOVERFLOW test failed.");
#endif
        bp = opt->backward_prob - (double) opt->in_target;
        for (out = opt->firstout; out <= opt->lastout; out++) {
            opt->outputs->del[out] = opt->outputs->act[out] * bp;
        }
    } else {
#if ONLYPREDCONDDEL || ( PLUGLEAKS && LOCALESTIMATE )
        if (opt->in_target != 0) {
            for (out = opt->firstout; out <= opt->lastout; out++) {
                opt->outputs->del[out] = -opt->outputs->act[out];
            }
        } else {
            for (out = opt->firstout; out <= opt->lastout; out++) {
                opt->outputs->del[out] = 0.0;
            }
        }
#else
#if PLUGLEAKS && ! LOCALESTIMATE
        bp = opt->backward_prob - (double) opt->in_target -
            exp(opt->lprob) * ave_finish_probs[opt->period];
#else
        bp = opt->backward_prob - (double) opt->in_target;
#endif
        for (out = opt->firstout; out <= opt->lastout; out++) {
            opt->outputs->del[out] = opt->outputs->act[out] * bp;
        }
#endif
    }
//not SIGMOIDDISC
}


/* set opt's output deltas to zero.  */
void zero_output_deltas(struct option *opt)
{
    int out;

    for (out = opt->firstout; out <= opt->lastout; out++) {
        opt->outputs->del[out] = 0.0;
    }
}

/* estimate of the probability mass for the options less probable than
     best derivations top, minus target derivation prob tarp, given beam width
     bw.  Estimated based on results from lcch_hwd80k_best2.crossval100.top. */
double tail_estimate(double tarp, struct option **top, int bw)
{
    double res;
    double length;
    double factor;

#define ASSUMETARTOP 1          /* assume top[0]->prob > tarp >= top[bw - 1]->prob */

    /* length = (double) sentence_length; */
    length = 24.34;
    /* factor = exp(0.277*length - 4.0) * 10.0; */
    /* factor = exp(TAIL_DEFAULT_EXP * (24.34 - length) / 2.0 ); */
    factor = 1.0;

#if LOSTTAIL
    if (bw == 0)
        res = 1.0;              /* makes error approximately equivalent to joint error */
    else
        res = lost_probability;
#elif ASSUMETARTOP
    res = factor *
        /* top[0]->prob * */
        exp(TAIL_AVE_BEST_EXP * length) *
        exp((TAIL_LENGTH_EXP + TAIL_BEAM_EXP * log(bw + 2)) * length);
#else
    if (bw == 0)
        res = factor * exp(TAIL_DEFAULT_EXP * length)
            - tarp;
    else if (tarp >= exp(top[bw - 1]->lprob)) {
        if (tarp >= exp(top[0]->lprob))
            res = factor *
                /* tarp * */
                exp(TAIL_AVE_BEST_EXP * length) *
                exp((TAIL_LENGTH_EXP + TAIL_BEAM_EXP * log(bw + 2)) *
                    length);
        else
            res = factor *
                /* top[0]->prob * */
                exp(TAIL_AVE_BEST_EXP * length) *
                exp((TAIL_LENGTH_EXP + TAIL_BEAM_EXP * log(bw + 2)) *
                    length);
    } else
        res = factor *
            /* top[0]->prob *  */
            exp(TAIL_AVE_BEST_EXP * length) *
            exp((TAIL_LENGTH_EXP + TAIL_BEAM_EXP * log(bw + 1)) * length)
            - tarp;
#endif

    /* printf("tail = %e\n", res); */

    if (res < 0.0)
        return (0.0);
    else
        return (res);
}



/* with output error from opt, update link delta terms to the output (and
     combination) layers, and pass error to hidden layer.  If last error for
     hidden layer,  update link delta terms to hidden units, and pass error to
     previous hidden layers.  */
void opt_comp_hidden_del(struct option *opt)
{
    int i, j, jm;
    int t, p;
    int bin;
    int out1;
    int *out2;
    int s, e;
    int memoffset;
    /* used to avoid recalculating components of weight changes */
    double t_p_del;
    double t_del;
    double debug_pdel;
    double debug_ndel;
    double *output_act;
    double *target_act;
    double *output_del;
    double *hidrho_act;
    double *hidrho_err;
    double *linksto;
    double *linkdelsto;

    int prevp;
    int prevop;
    int *pp;
    int contextQ;
    double *hidrho_current_act;
    double *hidrho_current_err;
#if ! LABELCONTEXTINP
    double *hidrho_context_act;
    double *hidrho_context_err;
#endif
    struct phase_spec *hist_p;
    int inp1_p, *inp2_p, inp3_p;
    int inp1_t, *inp2_t, inp3_t;
    int inp1_h, *inp2_h, inp3_h;
    int hdwdp;
    double cldel;
#if ! LABELTRIGPARENTINP
    struct hidrho *parent_state;
#endif
#if ! LABELTRIGCHILDINP
    struct hidrho *trigchild_state;
#endif
#if ! LABELCHILDINP
#if CHILDCONTEXT
    struct hidrho *child_state;
#endif
#endif
#if HEADHIDDEN && ! LABELHEADCHILDINP
    struct hidrho *headchild_state;
#endif
    double debug_err;
    struct option *headopt;

    /* final option has no output */
    if (opt->firstout > opt->lastout)
        return;

    t = opt->period;
    p = opt->phase;
    out1 = input_hist[t + 1].input1;    /* next inp1 */
    memoffset = opt->memoffset;
    out2 = input_hist[t + 1].output2;

    output_act = opt->outputs->act;
    output_del = opt->outputs->del;
    hidrho_act = opt->state->act;
    hidrho_err = opt->state->err;



    s = opt->firstout;
    e = opt->lastout;
    memoffset = opt->memoffset;

    j = s + memoffset - 1;      /* j is for links and true output index */
#if CHECKOVERFLOW
    if (!in_range(s, e, 0, MAXNUMOUTMEM) || j >= MAXNUMOUT - 1)
        error("CHECKOVERFLOW test failed.");
#endif
    for (jm = s; jm <= e; jm++) {       /* jm is for outputs in memory */
        j++;
        if (output_del[jm] < -1.0 || output_del[jm] > 1.0) {
            sprintf(warning_text,
                    "Output delta out of range for %d: %f (%d:%2d,%2d)\n",
                    j, output_del[jm], num_sent + 1, t, p);
            warning(warning_text);
        }

        /* update hidden-output links and update hidden errors */
        t_p_del = output_del[jm];
        if (t_p_del != 0.0) {
            biasdel_o[j] += t_p_del;    /* update output bias delta */
            {
                linksto = link_r_o[j];
                linkdelsto = linkdel_r_o[j];
                for (i = 0; i < numrho; i++) {
                    hidrho_err[i] += t_p_del * linksto[i];
                    linkdelsto[i] += t_p_del * hidrho_act[i];
                }
            }                   /* eo  if j != predword_to_output(inp1_first_word[out1 + 1] - 1)) */
        }
    }                           /* for jm */


    /* case when hidden error is complete */
    if (opt->new_state != 0) {

        if (opt->previous_option == NULL) {
            for (i = 0; i < numrho; i++) {
                hidrho_initstate.err[i] += hidrho_err[i];
                if (isnan(hidrho_initstate.err[i]))
                    warning("NaN nan n'nan nah");
            }
        } else {

#if TAULOOKAHEAD
            inp1_t = input_hist[t + 1].input1;
#else
            inp1_t = input_hist[t].input1;
#endif
#if TAULOOKAHEAD
            inp2_t = input_hist[t + 1].input2;
#else
            inp2_t = input_hist[t].input2;
#endif
            if (opt->preterm != 0) {
                prevp = -1;
            /** @todo: should be smth different, but prevop is used only in CHOPCONTEXT */
                prevop = PREDOUT;
            } else {
                prevp = opt->previous_option->phase;
                prevop = opt->previous_option->previous_output;
            }
            hidrho_current_act = opt->state->act;
            hidrho_current_err = opt->state->err;
            if (opt->context_option != NULL) {
                contextQ = 1;
#if ! LABELCONTEXTINP
                hidrho_context_act = opt->context_option->state->act;
                hidrho_context_err = opt->context_option->state->err;
#endif
            } else {
                contextQ = 0;
            }
#if ! LABELTRIGPARENTINP
            if (opt->trigparent_option != NULL) {
                parent_state = opt->trigparent_option->state;
            } else
                parent_state = NULL;
#endif
#if ! LABELTRIGCHILDINP
            if (opt->trigchild_option != NULL) {
                trigchild_state = opt->trigchild_option->state;
            } else
                trigchild_state = NULL;
#endif
#if ! LABELCHILDINP
#if CHILDCONTEXT
            if (prevp >= 0) {
                child_state = opt->previous_option->state;
            } else
                child_state = NULL;
#endif
#endif
#if HEADHIDDEN && ! LABELHEADCHILDINP
            if (opt->headchild_option != NULL) {
                headchild_state = opt->headchild_option->state;
            } else
                headchild_state = NULL;
#endif
            inp1_p = input_hist[opt->left_edge].input1;
            inp2_p = input_hist[opt->left_edge].input2;
#if HEADWORDS
            if (opt->headchild_option != NULL) {
                /* find head preterm's phase */
                for (headopt = opt->headchild_option;
                     headopt->headchild_option != NULL;
                     headopt = headopt->headchild_option) {
                }
                hdwdp = headopt->phase;
                inp1_h = input_hist[headopt->period].input1;
                inp2_h = input_hist[headopt->period].input2;
                if (headopt->preterm == 0)
                    warning("headless constituent in training.");
            } else {
                hdwdp = -1;
            }
#endif

            for (i = 0; i < numrho; i++) {
#if TANHHIDDEN
                t_p_del =
                    hidrho_current_err[i] *
                    (1.0 -
                     (hidrho_current_act[i] * hidrho_current_act[i]));
#else
                t_p_del =
                    hidrho_current_err[i] * hidrho_current_act[i] *
                    (1.0 - hidrho_current_act[i]);
#endif

                linkdel_t1_r[i][inp1_t] += t_p_del;
                for (bin = 0; bin < numbins; bin++) {
                    if (inp2_t[bin] >= 0) {
                        linkdel_t2_r[i][inp2_t[bin]] += t_p_del;
                    } else
                        break;
                }
#if HEADWORDS
                if (hdwdp >= 0) {
                    linkdel_h1_r[i][inp1_h] += t_p_del;
                    for (bin = 0; bin < numbins; bin++) {
                        if (inp2_h[bin] >= 0) {
                            linkdel_h2_r[i][inp2_h[bin]] += t_p_del;
                        } else
                            break;
                    }
                }
#endif

                if (opt->new_phase != 0) {
                    linkdel_bir_r[i] += t_p_del;
                }
                if (opt->context_option != NULL &&
                    opt->context_option->period == t - 1) {
                    linkdel_pred_r[i] += t_p_del;
                }

#if ! LABELCONTEXTINP
                if (contextQ) {
                    linksto = link_r_r[i];
                    linkdelsto = linkdel_r_r[i];
                    for (j = 0; j < numrho; j++) {
                        hidrho_context_err[j] += linksto[j] * t_p_del;
                        linkdelsto[j] += t_p_del * hidrho_context_act[j];
                    }
                }
#endif
#if CHILDCONTEXT
#if LABELCHILDINP
                if (prevp >= 0)
                    linkdel_chd_r[i][opt->previous_option->
                                     trigger_label] += t_p_del;
#else
                if (child_state != NULL) {
                    linksto = link_chd_r[i];
                    linkdelsto = linkdel_chd_r[i];
                    for (j = 0; j < numlink_chd; j++) {
                        child_state->err[j] += linksto[j] * t_p_del;
                        linkdelsto[j] += t_p_del * child_state->act[j];
                    }
                }
#endif
#endif
#if HEADHIDDEN
#if LABELHEADCHILDINP
                if (opt->headchild_option != NULL) {
                    linkdel_hdc_r[i][opt->headchild_option->
                                     trigger_label] += t_p_del;
                }
#else
                if (headchild_state != NULL) {
                    linksto = link_hdc_r[i];
                    linkdelsto = linkdel_hdc_r[i];
                    for (j = 0; j < numlink_hdc; j++) {
                        headchild_state->err[j] += linksto[j] * t_p_del;
                        linkdelsto[j] += t_p_del * headchild_state->act[j];
                    }
                }
#endif
#endif
#if CHOPCONTEXT
                if (prevop >= NUMSTRUC - NUMCHOP && prevop < NUMSTRUC) {
                    linkdel_cho_r[i][prevop - (NUMSTRUC - NUMCHOP)] +=
                        t_p_del;
                }
#endif
#if HEADOUTS
                if (opt->headchild_option == NULL) {
                    linkdel_nhd_r[i] += t_p_del;
                }
#endif

                /* period-independent cases. */
            /** would be faster if summed and links done once **/
#if LABELTRIGPARENTINP
                if (opt->trigparent_option != NULL) {
                    linkdel_trp_r[i][opt->trigparent_option->
                                     trigger_label] += t_p_del;
                }
#else
                if (parent_state != NULL) {
                    linksto = link_trp_r[i];
                    linkdelsto = linkdel_trp_r[i];
                    for (j = 0; j < numlink_trp; j++) {
                        parent_state->err[j] += linksto[j] * t_p_del;
                        linkdelsto[j] += t_p_del * parent_state->act[j];
                    }
                }
#endif
#if LABELTRIGCHILDINP
                if (opt->trigchild_option != NULL) {
                    linkdel_trc_r[i][opt->trigchild_option->
                                     trigger_label] += t_p_del;
                }
#else
                if (trigchild_state != NULL) {
                    linksto = link_trc_r[i];
                    linkdelsto = linkdel_trc_r[i];
                    for (j = 0; j < numlink_trc; j++) {
                        trigchild_state->err[j] += linksto[j] * t_p_del;
                        linkdelsto[j] += t_p_del * trigchild_state->act[j];
                    }
                }
#endif
                biasdel_r[i] += t_p_del;
                linkdel_r1_r[i][inp1_p] += t_p_del;
                for (bin = 0; bin < numbins; bin++) {
                    if (inp2_p[bin] >= 0)
                        linkdel_r2_r[i][inp2_p[bin]]
                            += t_p_del;
                    else
                        break;
                }
                linkdel_lab_r[i][opt->trigger_label] += t_p_del;
                linkdel_tro_r[i][opt->trigger_output + 1] += t_p_del;

            }
        }
    }
//!EM_CLUSTERING && !(FISHERDISC && !FI_USE_ALL_PARAMETERS)
}




/* find first string token in buffer.  Returns NULL if none. */
char *first_token(char *buffer)
{
    char *res;

    res = strtok(buffer, " \t\n");
    /* printf("first token: %s\n", res); */
    return (res);
}



/* find next string token in buffer most recently passed to first_token.
     Error if none. */
char *next_token()
{
    char *res;

    res = strtok(NULL, " \t\n");
    /* printf("next token: %s\n", res); */
    if (res == NULL)
        error("Error: Insufficient input.");
    else {
        return (res);
    }
}



/* test whether something was computed for p at t.  If performing
     repeatedly it is more efficient to use increment/decrement_nextcomp(). */
int test_comp(int t, int p)
{
    int *sp;

    sp = phase_hist[p].states.content;
    while (*sp < t && sp != phase_hist[p].states.top)
        sp++;

    return (*sp == t);
}



/* pop an element from s. */
void pop_stack(struct stack *s)
{
    s->top = s->subtop;
    if (s->subtop != NULL) {
        if (s->subtop != s->content)
            s->subtop--;
        else
            s->subtop = NULL;
    }
}

/* push an element onto s */
void push_stack(struct stack *s, int p)
{

    s->subtop = s->top;
    if (s->top != NULL)
        s->top++;
    else
        s->top = s->content;
    *(s->top) = p;
}

/* treat top_rel as a relative pointer to some stack, and return the
     relative pointer for the pop of the stack.  Warning if underflow. */
int relative_pop_stack(int top_rel)

{
    if (top_rel < 0) {
        warning("stack underflow.");
        return (-1);
    } else {
        return (top_rel - 1);
    }
}

/* initialize s as empty */
void empty_stack(struct stack *s)
{
    s->top = NULL;
    s->subtop = NULL;
}

/* expand the size of the array for s to space.  Return flag for success. */
int realloc_stack_content(struct stack *s, int space)
{
    int rt, rs;
    int temp;

    rt = s->top - s->content;
    rs = s->subtop - s->content;
    temp = (int) s->content;
    s->content = (int *) realloc(s->content, sizeof(int[space]));
    if (s->content == NULL) {
        s->content = (int *) temp;
        sprintf(warning_text,
                "Cannot allocate the desired number of phases, = %d.",
                space);
        warning(warning_text);
        return (0);
    }

    if (s->top != NULL)
        s->top = rt + s->content;
    if (s->subtop != NULL)
        s->subtop = rs + s->content;

    return (1);
}

/* copy the content of s1 to s2 */
void copy_stack(struct stack *s1, struct stack *s2)
{
    int i;
    int *pntr;

    if (s1 == s2)
        return;

    /*
       if (s1->top != NULL)
       {
       s2->top = s2->content + (int) (s1->top - s1->content);
       for (i = 0; i <= s1->top - s1->content; i++)
       s2->content[i] = s1->content[i];
       if (s1->subtop != NULL)
       s2->subtop = s2->content + (int) (s1->subtop - s1->content);
       else 
       s2->subtop = NULL;
       }
       else
       {
       s2->top = NULL;
       s2->subtop = NULL;
       }
     */

    empty_stack(s2);
    if (s1->top != NULL) {
        for (pntr = s1->content; pntr <= s1->top; pntr++)
            push_stack(s2, *pntr);
    }

}


/* allocate and calculate the hidden state (and combination) activations for
     opt, except tau hidden units (which should be computed earlier if used).
     Also allocate the output layer, but do not calculate. */
void comp_state_act(struct option *opt)
{
    int i;


    if (opt->lastout < opt->firstout) { /* else already set (not fullproof) */
        if (opt->new_state == 0) {
            warning("shouldn't call comp_state_act with opt->new_state=0");
            return;
        }
        if (opt->outputs == NULL)
            opt->outputs = (struct output *) malloc(sizeof(struct output));
        else
            warning("how did this output get set?");
        if (opt->outputs == NULL)
            error("Error: Cannot allocate enough space for testing.\n");

        if (opt->state == NULL) {

            opt->state = (struct hidrho *) malloc(sizeof(struct hidrho));
        } else
            warning("how did this output get set?");
        if (opt->state == NULL)
            error("Error: Cannot allocate enough space for testing.\n");

        if (opt->previous_option == NULL) {
            for (i = 0; i < numrho; i++) {
                opt->state->act[i] = hidrho_initstate.act[i];
            }
        } else {
            comp_hidrho_option_act(opt);
        }


    } else
        warning("shouldn't call comp_state_act with output computed");
}




/* compute the hidden rho layer's activation for option opt.  Assumes
     opt->period >= 1.  Doesn't use nextcomp.  */
void comp_hidrho_option_act(struct option *opt)
{
    int current;
    int p;
    int prevp;
    int prevop;
    int i, j;
    int bin;
#if ! LABELCONTEXTINP
    double *hidrho_context_act;
#endif
    double *hidrho_current_act;
    double *hidrho_const_act;
    int inp1_p, *inp2_p, inp3_p;
    int inp1_t, *inp2_t, inp3_t;
    int inp1_h, *inp2_h, inp3_h;
    struct option *hdwdopt;
    double net;
    double expnet, invexpnet;
    double *linksto;
    struct option *cont;
#if ! LABELTRIGPARENTINP
    struct hidrho *parent_state;
#endif
#if ! LABELTRIGCHILDINP
    struct hidrho *trigchild_state;
#endif
#if ! LABELCHILDINP
#if CHILDCONTEXT
    struct hidrho *child_state;
#endif
#endif
#if HEADHIDDEN && ! LABELHEADCHILDINP
    struct hidrho *headchild_state;
#endif


    p = opt->phase;
    current = opt->period;
    if (opt->preterm != 0) {
        prevp = -1;
        prevop = PREDOUT;
    } else {
        prevp = opt->previous_option->phase;
        prevop = opt->previous_output;
    }

    if (current <= 0)
        error
            ("comp_hidrho_option_act() shouldn't be called with current <= 0.");


#if TAULOOKAHEAD
    inp1_t = input_hist[current + 1].input1;
#else
    inp1_t = input_hist[current].input1;
#endif
#if TAULOOKAHEAD
    inp2_t = input_hist[current + 1].input2;
#else
    inp2_t = input_hist[current].input2;
#endif

    hidrho_current_act = opt->state->act;
#if ! LABELCONTEXTINP
    if (opt->context_option == NULL) {
        hidrho_context_act = hidrho_nullstate.act;
    } else if (opt->context_option->state != NULL)
        hidrho_context_act = opt->context_option->state->act;
    else {
        warning("no state computed for context!!");
        hidrho_context_act = hidrho_nullstate.act;
    }
#endif
    hidrho_const_act = hidrho_constant_net[p].act;
#if ! LABELTRIGPARENTINP
    if (opt->trigparent_option != NULL)
        parent_state = opt->trigparent_option->state;
    else
        parent_state = NULL;
#endif
#if ! LABELTRIGCHILDINP
    if (opt->trigchild_option != NULL)
        trigchild_state = opt->trigchild_option->state;
    else
        trigchild_state = NULL;
#endif
#if ! LABELCHILDINP
#if CHILDCONTEXT
    if (prevp >= 0) {
        child_state = opt->previous_option->state;
    } else
        child_state = NULL;     /* zero child state for preterminals */
#endif
#endif
#if HEADHIDDEN && ! LABELHEADCHILDINP
    if (opt->headchild_option != NULL)
        headchild_state = opt->headchild_option->state;
    else
        headchild_state = NULL; /* zero headchild state for matrix root etc */
#endif
    inp1_p = input_hist[opt->left_edge].input1;
    inp2_p = input_hist[opt->left_edge].input2;
#if HEADWORDS
    hdwdopt = opt->headchild_option;
    if (hdwdopt != NULL) {
        while (hdwdopt->headchild_option != NULL)       /* find head preterm's option */
            hdwdopt = hdwdopt->headchild_option;
        if (hdwdopt->preterm != 0) {
            inp1_h = input_hist[hdwdopt->period].input1;
            inp2_h = input_hist[hdwdopt->period].input2;
            /*
               printf("head word of %d/%s is %d/%s.\n", 
               hist_p->trigger_label, id_label(hist_p->trigger_label), 
               hdwdopt->period, sentence_words[hdwdopt->period]);
             */
        } else
            warning("headless constituent in testing.");
    }
#endif


    /* current version has inputs from: state at previous computation,
       triggering op, triggering word, triggering parent, triggering
       child (if nonterm), label, previous structure output
       (prediction/continuation/birth), current word.  State records
       first and last word of every child. */
    /* Note that adding information about child nonterminals would require a
       different treatment for post-prediction and pre-prediction computations 
       at the prediction site */
    for (i = 0; i < numrho; i++) {
        if (opt->new_phase != 0) {
            hidrho_const_act[i] = bias_r[i]
                + link_r1_r[i][inp1_p]
                + link_tro_r[i][opt->trigger_output + 1];
            for (bin = 0; bin < numbins; bin++) {
                if (inp2_p[bin] >= 0)
                    hidrho_const_act[i] += link_r2_r[i][inp2_p[bin]];
                else
                    break;
            }
            hidrho_const_act[i] += link_lab_r[i][opt->trigger_label];
#if LABELTRIGPARENTINP
            if (opt->trigparent_option != NULL)
                hidrho_const_act[i] +=
                    link_trp_r[i][opt->trigparent_option->trigger_label];
#else
            if (parent_state != NULL) {
                linksto = link_trp_r[i];
                for (j = 0; j < numlink_trp; j++) {
                    hidrho_const_act[i] +=
                        linksto[j] * parent_state->act[j];
                }
            }
#endif
#if LABELTRIGCHILDINP
            if (opt->trigchild_option != NULL)
                hidrho_const_act[i] +=
                    link_trc_r[i][opt->trigchild_option->trigger_label];
#else
            if (trigchild_state != NULL) {
                linksto = link_trc_r[i];
                for (j = 0; j < numlink_trc; j++) {
                    hidrho_const_act[i] +=
                        linksto[j] * trigchild_state->act[j];
                }
            }
#endif
            net = link_bir_r[i];
        }
        else if (opt->context_option != NULL &&
                 opt->context_option->period == current - 1) {
            net = link_pred_r[i];
        }
        else {
            net = 0.0;
        }
        net += hidrho_const_act[i];
        net += link_t1_r[i][inp1_t];

        for (bin = 0; bin < numbins; bin++) {
            if (inp2_t[bin] >= 0)
                net += link_t2_r[i][inp2_t[bin]];
            else
                break;
        }
#if HEADWORDS
        if (hdwdopt != NULL) {
            net += link_h1_r[i][inp1_h];
            for (bin = 0; bin < numbins; bin++) {
                if (inp2_h[bin] >= 0)
                    net += link_h2_r[i][inp2_h[bin]];
                else
                    break;
            }
        }
#endif
#if ! LABELCONTEXTINP
        linksto = link_r_r[i];
        for (j = 0; j < numrho; j++)
            net += linksto[j] * hidrho_context_act[j];
#endif
#if CHILDCONTEXT
#if LABELCHILDINP
        if (prevp >= 0)
            net += link_chd_r[i][opt->previous_option->trigger_label];
#else
        if (child_state != NULL) {
            linksto = link_chd_r[i];
            for (j = 0; j < numlink_chd; j++) {
                net += linksto[j] * child_state->act[j];
            }
        }
#endif
#endif
#if HEADHIDDEN
#if LABELHEADCHILDINP
        if (opt->headchild_option != NULL)
            net += link_hdc_r[i][opt->headchild_option->trigger_label];
#else
        if (headchild_state != NULL) {
            linksto = link_hdc_r[i];
            for (j = 0; j < numlink_hdc; j++) {
                net += linksto[j] * headchild_state->act[j];
            }
        }
#endif
#endif
#if CHOPCONTEXT
        if (prevop >= NUMSTRUC - NUMCHOP && prevop < NUMSTRUC) {
            net += link_cho_r[i][prevop - (NUMSTRUC - NUMCHOP)];
        }
#endif
#if HEADOUTS
        if (opt->headchild_option == NULL) {
            net += link_nhd_r[i];
        }
#endif

#if TANHHIDDEN
        expnet = exp(2.0 * net);
        if (expnet == 0.0)
            hidrho_current_act[i] = -1.0;
        else {
            hidrho_current_act[i] = (expnet - 1.0) / (expnet + 1.0);
        }
        /* printf("act(%d,%d,%d) = %f\n", current, p, i, hidrho_current_act[i]); */
#else
        hidrho_current_act[i] = 1.0 / (1.0 + exp(-net));
#endif
        if (isnan(hidrho_current_act[i]))
            warning("NaN in hidden state.");
    }
}






/* compute the output layer's activation for current period.  Outputs are
     normalized.   Needs nextcomp set.  */
/* same as comp_output_act() but only for phase and output units between
     first_mem and last_mem, inclusive.  first_mem and last_mem are in memory
     space, not real output space.  Ignores phase_hist and assumes doing
     testing (i.e. ignores target).  */
void comp_output_act_range(struct option *opt, int first_mem, int last_mem)
{
    int inter_first_mem;
    int inter_last_mem;
    int wrdflag;
    int current;
    int phase;
    int i, j;
    int bin;
    int out1;
    int im, jm;
    double net;
    double *output_act;
    double *hidrho_act;
    double *hidtau_act;
    double *combin_act;
    double *linksto;
    int ptf, uhf, hpf;
    int layer_idx;
    int mem_out;
    int memoffset;
    int pos;

    if (opt->lastout != -1) {
        if (opt->firstout != first_mem || opt->lastout != last_mem)
            warning("incompatible requests for output computations.");
        return;
    }

    opt->firstout = first_mem;
    opt->lastout = last_mem;

    memoffset = opt->memoffset;

    phase = opt->phase;
    current = opt->period;
    out1 = input_hist[current + 1].input1;

    inter_first_mem = first_mem;
    inter_last_mem = last_mem;
#if UNHEADEDCOND && HEADOUTS
    if (opt->preterm == 0 && opt->headchild_option == NULL &&
        first_mem < NUMSTRUC)
        inter_first_mem = NUMSTRUC;
#endif
    /***
#if DEFAULTUNKNOWN
    if (inter_last_mem >= predword_to_output(0)) {
      inter_last_mem--;
      wrdflag = 1;
    }
    else
      wrdflag = 0;
#endif
    ***/


    output_act = opt->outputs->act;
    hidrho_act = opt->state->act;


    {
        for (im = inter_first_mem; im <= last_mem; im++) {
#if ! NOPREDICT
            if (im >= wordoutputs) {
                i = im + memoffset;
            } else
#endif
                i = im;
            net = bias_o[i];

            linksto = link_r_o[i];
            for (j = 0; j < numrho; j++)
                net += linksto[j] * hidrho_act[j];


            output_act[im] = exp(net);


        }                       /* eo for im  */



        if (opt->preterm != 0)
            output_act[PREDOUT] = 0.0;
#if ROOTTAGCOND
        if (opt->previous_output == PREDOUT)
            output_act[predtag_to_output(tag_id("^^^"))] = 0.0;
#endif
#if ROOTLABELCOND
        if (opt->previous_output == PROJOUT)
            output_act[projlabel_to_output(label_id(""))] = 0.0;
#endif

        normalize_output_act_range(opt, first_mem, last_mem, 0);
    }

#if CHECKACTDEL
    check_output_act_del_range(current, phase, opt->outputs, first_mem,
                               last_mem, ptf, uhf, hpf, 0);
#endif
}


/* normalize the outputs for period current.  Used for softmax output
     function.  All normalization is within an individual period,phase.
     Normalized sets are: {PREDOUT, PROJOUT, HDPROUT, ATTOUT, HDATOUT, MODOUT}
     ({PROJOUT, HDPROUT, ATTOUT, HDATOUT, MODOUT} for preterms), projected
     labels, predicted tags, each predicted word bin.  When trainingflag!=0
     normalization is not done on sets with dontcare targets.  Requires
     nextcomp to be set.  */
/* normalize the outputs for period current and phase for units between
     first_mem and last_mem, inclusive.  If computing word predictions,
     requires memoffset to be set in opt.  See normalize_output_act().  */
void normalize_output_act_range(struct option *opt, int first_mem,
                                int last_mem, int trainingflag)
{
    int current;
    int phase;
    struct output *outputs;
    int pretermflag;
    int unheadedflag;
    int hddparentflag;
    int i, j;
    int forend;
    int t, p;
    int out1;
    int memoffset;
    int bin;
    int idx_start, idx_end;
    double *output_act;

    double norm;

    current = opt->period;
    phase = opt->phase;
    outputs = opt->outputs;
    pretermflag = opt->preterm;
#if UNHEADEDCOND && HEADOUTS
    if (opt->preterm == 0 && opt->headchild_option == NULL)
        unheadedflag = 1;
    else
#endif
        unheadedflag = 0;
#if HEADEDPARENTCOND && HEADOUTS
    if (opt->trigparent_option != NULL &&
        opt->trigparent_option->headchild_option != NULL)
        hddparentflag = 1;
    else
#endif
        hddparentflag = 0;
    memoffset = opt->memoffset;

    output_act = outputs->act;

    if (in_range(PREPREDFIRST, PREPREDLAST, first_mem, last_mem)) {
        /* normalize {PREDOUT, PROJOUT, HDPROUT, ATTOUT, HDATOUT, MODOUT}
           outputs */
#if UNHEADEDCOND && HEADOUTS
        if (unheadedflag != 0) {
            norm = 1.0;
            output_act[PREDOUT] = 1.0;
            output_act[ATTOUT] = 0.0;
            output_act[MODOUT] = 0.0;
            output_act[PROJOUT] = 0.0;
            output_act[HDATOUT] = 0.0;
            output_act[HDPROUT] = 0.0;
        } else
#endif
        {
            norm = output_act[ATTOUT] + output_act[MODOUT];
            norm += output_act[PREDOUT];
            norm += output_act[PROJOUT];
#if HEADOUTS
#if HEADEDPARENTCOND && HEADOUTS
            if (hddparentflag == 0)
#endif
                norm += output_act[HDATOUT];
            norm += output_act[HDPROUT];
#endif
            if (norm > 0.0) {
                output_act[PREDOUT] /= norm;
                output_act[ATTOUT] /= norm;
                output_act[MODOUT] /= norm;
                output_act[PROJOUT] /= norm;
#if HEADOUTS
#if HEADEDPARENTCOND && HEADOUTS
                if (hddparentflag != 0)
                    output_act[HDATOUT] = 0.0;
                else
#endif
                    output_act[HDATOUT] /= norm;
                output_act[HDPROUT] /= norm;
#endif
            } else {
                sprintf(warning_text,
                        "pre-prediction normalization value = %f for word %d, sentence %d, trainingflag %d.",
                        norm, current, num_sent + 1, trainingflag);
                warning(warning_text);

            }
        }
    }


    if (in_range
        (projlabel_to_output(0), projlabel_to_output(numlabels - 1),
         first_mem, last_mem))
        /* normalize projected labels */
    {
        norm = 0.0;
        forend = projlabel_to_output(numlabels - 1);
        for (i = projlabel_to_output(0); i <= forend; i++)
            norm += output_act[i];
        if (norm > 0.0) {
            forend = projlabel_to_output(numlabels - 1);
            for (i = projlabel_to_output(0); i <= forend; i++) {
                output_act[i] /= norm;
            }
        } else {
            sprintf(warning_text,
                    "projection normalization value = %f for word %d, sentence %d, trainingflag %d.",
                    norm, current, num_sent + 1, trainingflag);
            warning(warning_text);

        }
    }
#if HEADOUTS
    if (in_range
        (hdprlabel_to_output(0), hdprlabel_to_output(numlabels - 1),
         first_mem, last_mem))
        /* normalize head projected labels */
    {
        norm = 0.0;
        forend = hdprlabel_to_output(numlabels - 1);
        for (i = hdprlabel_to_output(0); i <= forend; i++)
            norm += output_act[i];
        if (norm > 0.0) {
            forend = hdprlabel_to_output(numlabels - 1);
            for (i = hdprlabel_to_output(0); i <= forend; i++) {
                output_act[i] /= norm;
            }
        } else {
            sprintf(warning_text,
                    "head projection normalization value = %f for word %d, sentence %d, trainingflag %d.",
                    norm, current, num_sent + 1, trainingflag);
            warning(warning_text);

        }
    }
#endif


#if ! NOPREDICT
    if (in_range(predtag_to_output(0), predtag_to_output(numout1 - 1),
                 first_mem, last_mem))
        /* normalize predicted tags */
    {
        norm = 0.0;
        forend = predtag_to_output(numout1 - 1);
        for (i = predtag_to_output(0); i <= forend; i++)
            norm += output_act[i];
        if (norm > 0.0) {
            forend = predtag_to_output(numout1 - 1);
            for (i = predtag_to_output(0); i <= forend; i++) {
                output_act[i] /= norm;
            }
        } else {
            sprintf(warning_text,
                    "prediction normalization value = %f for word %d, sentence %d, trainingflag %d.",
                    norm, current, num_sent + 1, trainingflag);
            warning(warning_text);

        }
    }
#endif


#if ! NOPREDICT
    if (last_mem >= wordoutputs) {      /* otherwise tag maybe undefined */
        out1 = input_hist[current + 1].input1;

        idx_start = opt->firstout;
        idx_end = opt->lastout;


        if (in_range(idx_start, idx_end, first_mem, last_mem)) {
            /* normalize predicted words in bin */
            norm = 0.0;
            for (i = idx_start; i <= idx_end; i++) {
                norm += output_act[i];
            }
            if (norm > 0.0) {
                for (i = idx_start; i <= idx_end; i++) {
                    output_act[i] /= norm;
                }
            } else {
                sprintf(warning_text,
                        "prediction normalization value = %f for word %d, sentence %d, trainingflag %d.",
                        norm, current, num_sent + 1, trainingflag);
                warning(warning_text);
            }
        } else {
            /* do nothing */
        }
    }
#endif
    
}


/* 1 if q0-qn is within r0-rn, inclusive, 0 o.w.  Warning if 0 but ranges
     overlap. */
int in_range(int q0, int qn, int r0, int rn)
{
    if (r0 <= q0) {
        if (qn <= rn) {
            return (1);
        } else {
            if (q0 <= rn) {
                printf("%d, %d, %d, %d => ", q0, qn, r0, rn);
                warning("overlapping ranges in in_range().");
            }
            return (0);
        }
    } else if (r0 <= qn) {
        printf("%d, %d, %d, %d => ", q0, qn, r0, rn);
        warning("overlapping ranges in in_range().");
    }

    return (0);
}





/* use the deltas computed by comp_output_err(), comp_hidden_err(), */
/* and comp_prehidden_del().  reset the deltas used with momentum */
/* term.  if BATCHTRAIN then ignores length. */
void update_weights(int length)
{
    int i, j;
    int out1;
    double invlen;
    double *linksto;
    double *linkdelsto;
    double reldecay;
    double hidact;
    double tmp1;


    /* update weights and biases to output and combination units */
/* without combination layer *////*{{{*/                                       /*}}} */
    reldecay = decay;
    out1 = 0;
    for (j = 0; j < numout; j++) {
/*}}}*/
#if ! FRZBIASOUT
        bias_o[j] += eta * biasdel_o[j];
#endif
        biasdel_o[j] *= mom;
        linksto = link_r_o[j];
        linkdelsto = linkdel_r_o[j];
        for (i = 0; i < numrho; i++) {
#if ! FRZHIDOUT
            linksto[i] += eta * linkdelsto[i];
#endif
            linkdelsto[i] *= mom;
#if REGULARIZE
#if ! FRZHIDOUT
            linksto[i] *= reldecay;
#endif
#endif
        }
/*}}} */
    }
/*{{{*/                        /*}}} */
/*}}} */

    /* update input to hidden links and hidden biases. */
    for (j = 0; j < numrho; j++) {
        linksto = link_r1_r[j];
        linkdelsto = linkdel_r1_r[j];
        for (i = 0; i < numinp1; i++) {
#if ! FRZNONWRDHID
            linksto[i] += eta * linkdelsto[i];  /* update r1_r */
#endif
            linkdelsto[i] *= mom;
#if REGULARIZE
#if ! FRZNONWRDHID
            linksto[i] *= decay;
#endif
#endif
        }
        linksto = link_t1_r[j];
        linkdelsto = linkdel_t1_r[j];
        for (i = 0; i < numinp1; i++) {
#if ! FRZNONWRDHID
            linksto[i] += eta * linkdelsto[i];  /* update t1r */
#endif
            linkdelsto[i] *= mom;
#if REGULARIZE
#if ! FRZNONWRDHID
            linksto[i] *= decay;
#endif
#endif
        }
#if HEADWORDS
        linksto = link_h1_r[j];
        linkdelsto = linkdel_h1_r[j];
        for (i = 0; i < numinp1; i++) {
#if ! FRZNONWRDHID
            linksto[i] += eta * linkdelsto[i];  /* update h1r */
#endif
            linkdelsto[i] *= mom;
#if REGULARIZE
#if ! FRZNONWRDHID
            linksto[i] *= decay;
#endif
#endif
        }
#endif
        linksto = link_r2_r[j];
        linkdelsto = linkdel_r2_r[j];
        for (i = 0; i < numinp2; i++) {
#if ! FRZWRDHID
            linksto[i] += eta * linkdelsto[i];  /* update r2r */
#endif
            linkdelsto[i] *= mom;
#if REGULARIZE
#if ! FRZWRDHID
            linksto[i] *= decay;
#endif
#endif
        }
        linksto = link_t2_r[j];
        linkdelsto = linkdel_t2_r[j];
        for (i = 0; i < numinp2; i++) {
#if ! FRZWRDHID
            linksto[i] += eta * linkdelsto[i];  /* update t2r */
#endif
            linkdelsto[i] *= mom;
#if REGULARIZE
#if ! FRZWRDHID
            linksto[i] *= decay;
#endif
#endif
        }
#if HEADWORDS
        linksto = link_h2_r[j];
        linkdelsto = linkdel_h2_r[j];
        for (i = 0; i < numinp2; i++) {
#if ! FRZWRDHID
            linksto[i] += eta * linkdelsto[i];  /* update h2r */
#endif
            linkdelsto[i] *= mom;
#if REGULARIZE
#if ! FRZWRDHID
            linksto[i] *= decay;
#endif
#endif
        }
#endif
#if ! FRZBIASHID
        bias_r[j] += eta * biasdel_r[j];        /* update r */
#endif
        biasdel_r[j] *= mom;
    }

    /* update hidden to hidden links and output to hidden links. */
    for (i = 0; i < numrho; i++) {
#if ! LABELCONTEXTINP
        linksto = link_r_r[i];
        linkdelsto = linkdel_r_r[i];
        for (j = 0; j < numrho; j++) {
#if ! FRZHIDHID
            linksto[j] += eta * linkdelsto[j];  /* update rr */
#endif
            linkdelsto[j] *= mom;
#if REGULARIZE
#if ! FRZHIDHID
                    linksto[j] *= decay;
#endif
#endif
        }
#endif
#if ! FRZCONTHID
        link_bir_r[i] += eta * linkdel_bir_r[i];        /* update bir_r */
#endif
        linkdel_bir_r[i] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
        link_bir_r[i] *= decay;
#endif
#endif
        linksto = link_tro_r[i];
        linkdelsto = linkdel_tro_r[i];
        for (j = 0; j < numlink_tro; j++) {
#if ! FRZCONTHID
            linksto[j] += eta * linkdelsto[j];  /* update tro_r */
#endif
            linkdelsto[j] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
            linksto[j] *= decay;
#endif
#endif
        }
        linksto = link_trp_r[i];
        linkdelsto = linkdel_trp_r[i];
        for (j = 0; j < numlink_trp; j++) {
#if ! FRZCONTHID
            linksto[j] += eta * linkdelsto[j];  /* update trp_r */
#endif
            linkdel_trp_r[i][j] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
            link_trp_r[i][j] *= decay;
#endif
#endif
        }
        linksto = link_trc_r[i];
        linkdelsto = linkdel_trc_r[i];
        for (j = 0; j < numlink_trc; j++) {
#if ! FRZCONTHID
            linksto[j] += eta * linkdelsto[j];  /* update trc_r */
#endif
            linkdelsto[j] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
            linksto[j] *= decay;
#endif
#endif
        }
        linksto = link_lab_r[i];
        linkdelsto = linkdel_lab_r[i];
        for (j = 0; j < numlink_lab; j++) {
#if ! FRZCONTHID
            linksto[j] += eta * linkdelsto[j];  /* update lab_r */
#endif
            linkdelsto[j] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
            linksto[j] *= decay;
#endif
#endif
        }
#if ! FRZCONTHID
        link_pred_r[i] += eta * linkdel_pred_r[i];      /* update pred_r */
#endif
        linkdel_pred_r[i] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
        link_pred_r[i] *= decay;
#endif
#endif
#if CHILDCONTEXT
        linksto = link_chd_r[i];
        linkdelsto = linkdel_chd_r[i];
        for (j = 0; j < numlink_chd; j++) {
#if ! FRZCONTHID
            linksto[j] += eta * linkdelsto[j];  /* update chd_r */
#endif
            linkdelsto[j] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
            linksto[j] *= decay;
#endif
#endif
        }
#endif
#if HEADHIDDEN
        linksto = link_hdc_r[i];
        linkdelsto = linkdel_hdc_r[i];
        for (j = 0; j < numlink_hdc; j++) {
#if ! FRZCONTHID
            linksto[j] += eta * linkdelsto[j];  /* update hdc_r */
#endif
            linkdelsto[j] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
            linksto[j] *= decay;
#endif
#endif
        }
#endif
#if CHOPCONTEXT
        linksto = link_cho_r[i];
        linkdelsto = linkdel_cho_r[i];
        for (j = 0; j < numlink_cho; j++) {
#if ! FRZCONTHID
            linksto[j] += eta * linkdelsto[j];  /* update cho_r */
#endif
            linkdelsto[j] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
            linksto[j] *= decay;
#endif
#endif
        }
#endif
#if HEADOUTS
#if ! FRZCONTHID
        link_nhd_r[i] += eta * linkdel_nhd_r[i];        /* update nhd_r */
#endif
        linkdel_nhd_r[i] *= mom;
#if REGULARIZE
#if ! FRZCONTHID
        link_nhd_r[i] *= decay;
#endif
#endif
#endif
    }

    /* update initialization states.  delta = .err*eta */
    for (i = 0; i < numrho; i++) {      /* update ri */
#if ! FRZHIDHID
#if TRAININITLESS
#if TANHHIDDEN
        hidact = hidrho_initstate.act[i];
        tmp1 =
            exp(2.0 * TRAININITLESSRATE * eta * hidrho_initstate.err[i] *
                (1.0 - (hidact * hidact)))
            * ((1.0 + hidact) / (1.0 - hidact));
        hidrho_initstate.act[i] = (tmp1 - 1.0) / (tmp1 + 1.0);
#else
        hidact = hidrho_initstate.act[i];
        hidrho_initstate.act[i] =
            1.0 / (1.0 +
                   (((1.0 / hidact) - 1.0) *
                    exp(TRAININITLESSRATE * eta * hidact * (hidact - 1.0) *
                        hidrho_initstate.err[i])));
#endif
#else
#if TANHHIDDEN
        hidact = hidrho_initstate.act[i];
        tmp1 =
            exp(2.0 * eta * hidrho_initstate.err[i] *
                (1.0 - (hidact * hidact)))
            * ((1.0 + hidact) / (1.0 - hidact));
        hidrho_initstate.act[i] = (tmp1 - 1.0) / (tmp1 + 1.0);
#else
        hidact = hidrho_initstate.act[i];
        hidrho_initstate.act[i] =
            1.0 / (1.0 +
                   (((1.0 / hidact) - 1.0) *
                    exp(eta * hidact * (hidact - 1.0) *
                        hidrho_initstate.err[i])));
#endif
#endif
#endif
        hidrho_initstate.err[i] *= mom;
        if (isnan(hidrho_initstate.act[i]))
            warning("NaN nan n'nan nah");
    }

}



/* Runs the network on each sentence in testingfile and computes */
/* performance statistics, including precision/recall on constituents */
/* but not including error. */
void testing_epoch()
{
    FILE *dfp;
    FILE *nfp;
    FILE *cfp;
#if RECHECKTESTING
    FILE *rtfp;
    int rt_correct, rt_target, rt_output;
#endif
    char *str;
    int p, i;

#if DISPITER
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp, "Starting testing after cycle %d\n", weight_age);
        fclose(warnfp);
    } else
#endif
        printf("Starting testing after cycle %d\n", weight_age);
#endif

    if (idtext_input == 0 || stdin_word_input == 0) {
      if ((dfp = fopen(testingfile, "r")) == NULL)
        error("Error: Cannot open ((val)data)filename.dat to read data.");
    }
    else {
      dfp = stdin;
    }
    if (readcandidates != 0) {
        if (stdin_candidate_input == 0) {
	    if ((cfp = fopen(candfile, "r")) == NULL)
	      error("Error: Cannot open candidatefile.deriv to read data.");
	}
	else {
	    cfp = stdin;
	}
    } else
        cfp = NULL;

    total_sent_correct = 0;
    total_const_correct = total_const_output = total_const_target = 0;
    total100_const_correct = total100_const_output =
        total100_const_target = 0;
    total40_const_correct = total40_const_output = total40_const_target =
        0;
    total_const_struct_correct = 0;
    total_pch_correct = total_pch_output = total_pch_target = 0;

    num_sent = 0;
    total_err = 0.0;            /* not calculated */
    total_local_err = 0.0;      /* not calculated */
    total_local_err_correction = 0.0;
    num_items = 0;
    total_targets = 0.0;
    total_struc_err = 0.0;
    total_label_err = 0.0;
    total_tag_err = 0.0;
    total_word_err = 0.0;
    total_struc_targets = 0.0;
    total_label_targets = 0.0;
    total_tag_targets = 0.0;
    total_word_targets = 0.0;

    bias_o[PROJOUT] += projection_bias;

#if ERRVALIDATE
    while (comp_sentence(dfp) >= 0)     /* test for end of file */
#else
    while (test_sentence(dfp, cfp) >= 0)        /* test for end of file */
#endif
    {
#if ! ERRVALIDATE
        comp_structure_summary(sentence_length);
        /* comp_query(sentence_length);  * ??? */
#endif
        /*
           display_consts(sentence_length);
           display_parchild(sentence_length);
           display_histories(sentence_length);
           printf("\n");
         */

#if RECHECKTESTING
        if ((rtfp = fopen(retestfile, "r")) == NULL)
            error("Error: Cannot open filename.check.dat to read data.");
        rt_length = test_sentence(rtfp, cfp);
        test_constituents(&rt_correct, &rt_target, &rt_output, rt_length);
        if (num_sent == 0) {
            retest_correct = rt_correct;
            retest_target = rt_target;
            retest_output = rt_output;
        } else if (retest_correct != rt_correct ||
                   retest_target != rt_target ||
                   retest_output != rt_output) {
            sprintf(warning_text,
                    "Retesting on check sentence produced a different result (%dc,%dt,%do, was %dc,%dt,%do) after sentence %d.",
                    rt_correct, rt_target, rt_output, retest_correct,
                    retest_target, retest_output, num_sent + 1);
            warning(warning_text);
            printf("new check tree:");
            display_tree(output_tree);
            printf("\n");
            retest_correct = rt_correct;
            retest_target = rt_target;
            retest_output = rt_output;
        }
        fclose(rtfp);
#endif

        num_sent++;
        sentence_error = 0.0;
        
        //added because last tree wasn't freed (original it was freed before new allocation)
        if (target_tree != NULL) {
             free_all_tree(target_tree);
             target_tree = NULL;
        }
        if (output_tree != NULL) {
            free_all_tree(output_tree);
            output_tree = NULL;
        }
    }

    bias_o[PROJOUT] -= projection_bias; /* undo addition above */

    /* Display results */
/*
print_query();
*/

    if (idtext_input == 0 || stdin_word_input == 0) {
      fclose(dfp);
    }
    if (cfp != NULL)
        fclose(cfp);
}



/* load the next sentence from datafile_p into the input and target
     output arrays, search space of outputs for best parse.  Also sets target
     tree and output tree.  Returns sentence length  (including `.'), or -1 if
     end of file. */
/** should make use of perform_derivation for winning tree **/
int test_sentence(FILE * datafile_p, FILE * candfile_p)
{
    char buffer[MAXDATALINE];
    int i, t, p, b;
    int length;
    int flag;
    struct tree_list *output_tree_stack;
    struct tree *tr;
    struct option *opt;
    struct option *opt2;
    struct option *rootopt;
    struct option_stack *ol;

#if DISPSEN
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
#if DISPSEN==2
        fprintf(warnfp, "Sen%d, ", num_sent + 1);
#else
        fprintf(warnfp, "Sentence %d\n", num_sent + 1);
#endif
        fclose(warnfp);
    } else
#endif
    {
#if DISPSEN==2
        printf("Sen%d, ", num_sent + 1);
        fflush(stdout);
#else
        printf("Sentence %d\n", num_sent + 1);
#endif
    }
#endif
#if FILEPROG
    if ((progfp = fopen(progfile, "w")) != NULL) {
        fprintf(progfp, "Testing sentence %d, after cycle %d.\n",
                num_sent + 1, weight_age);
        fprintf(progfp,
                "    prior constituents correct: %d, target: %d, output: %d (F1:%5.1f%%).\n",
                total_const_correct, total_const_target,
                total_const_output, ((double) (total_const_correct * 200))
                / ((double) (total_const_target + total_const_output)));
        fclose(progfp);
    }
#endif

#if CHECKANTIGARBAGE
    if (anti_garbage_options != NULL) {
        warning("not all options freed");
        for (ol = anti_garbage_options; ol != NULL; ol = ol->next)
            display_option(ol->elem, 0);
    }
#endif

    init_phases_for_options();

    if (target_tree != NULL) 
        free_all_tree(target_tree);
    if (idtext_input == 0) {

        target_tree = read_target(datafile_p, 1);

	/* test for end of file */
	if (target_tree == NULL)
	    return (-1);            /* flag for end of file */

	length = target_tree->end;

    }
    else {
        target_tree = NULL;
	if (datafile_p != candfile_p || readcandidates == 0) { 
	    /* else done below */
	    length = read_idtext(datafile_p);
	    if (length <= 0) 
	        return (-1);            /* flag for end of file */
	}
    }

    if (readcandidates != 0) {
	if (datafile_p != candfile_p) { 
	    if (read_candidate_derivations(candfile_p, 0, 0) == -1) {
	        warning("Not enough candidate derivations in candidatefile.");
	    }
	} else {
	    length = read_candidate_derivations(candfile_p, 0, 1);
	    if (length <= 0) {
	        return(-1);            /* flag for end of file */
	    }
	}
    } else
        candidate_derivations = NULL;

    /* compute things for whole sentence */

    /* find best option sequences and update option stack */
    init_best_options(length);
    /* initialize with root nonterminal */
    best_options[0][0] = root_prediction();
    for (rootopt = best_options[0][0]; rootopt->previous_option != NULL;
         rootopt = rootopt->previous_option) {
    }
    for (b = 1; b < predbeam; b++)
        best_options[0][b] = NULL;
    output_tree_stack =
        start_root(0, best_options[0][0]->previous_option->phase, NULL);
    free_all_tree(output_tree);
    output_tree = output_tree_stack->elem;

    /* find the best option for the sentence */
    for (t = 1; t < length; t++) {

        search_options(t, NULL, log(0.0), length);      /* sets best_options[t] */

        if (best_options[t][0] == NULL)
            error("no predicting option found.");
        else {
            /* update phase_hist's states to reflect choice of best_option */
          /** no longer needed **/
            /* for (opt = best_option->previous_option; opt->period == t; 
               opt = opt->previous_option)
               {
               set_option_states(opt);
               }
             */
            /* test for zero prob options */
            for (b = 0; b < predbeam && best_options[t][b] != NULL; b++) {
                if (isinf(best_options[t][b]->lprob)) {
                    warning
                        ("zero probability derivation in best. Resetting.");
                    if (b != 0) {
                        best_options[t][b]->lprob =
                            best_options[t][b - 1]->lprob - log(2.0);
                        best_options[t][b]->ranking_lprob =
                            best_options[t][b - 1]->ranking_lprob -
                            log(2.0);
                    } else {
                        best_options[t][b]->lprob = log(0.1);
                        best_options[t][b]->ranking_lprob = log(0.1);
                    }
                }
            }

            /* set previous prediction option indexes for next period */
            for (b = 0; b < predbeam && best_options[t][b] != NULL; b++) {
                if (best_options[t][b] != NULL)
                    best_options[t][b]->prev_pred_ind = b;
            }
        }
    }

    /* set winning_option and construct its tree */
    winning_option = best_options[length - 1][0];
    if (winning_option != NULL)
        output_tree_stack =
            perform_options(output_tree_stack, 0, winning_option, 1);
    else {
        free_all_tree_list(output_tree_stack);
        output_tree_stack = NULL;
        sprintf(warning_text,
                "no parse found for sentence %d.  Should be using a flat tree.",
                num_sent + 1);
        warning(warning_text);
    }
    output_tree->end = length;
    free_tree_list(output_tree_stack);
    for (b = 0; b < numtopparses; b++) {
        opt = best_options[length - 1][b];
        if (opt != NULL) {
            tr = perform_derivation(opt);
            sprintf(buffer, "[%d #%d, P=%e]:  ", num_sent + 1, b + 1,
                    exp(opt->lprob));
            write_tree(tr, topfile, buffer);
            free_all_tree(tr);
            tr = NULL;
        }
    }
    if (numtopderivs != 0) {
        /* remove all but best numtopderivs derivations from best_options */
      /**JH serious error corrected 21/5/04, was introduced in dglc2b **/
        for (b = predbeam - 1; b >= numtopderivs; b--) {
            if (best_options[length - 1][b] != NULL) {
                free_option_branch(best_options[length - 1][b]);
                best_options[length - 1][b] = NULL;
            }
        }
        write_derivations(rootopt, stdout);
    }



    expand_conflations(output_tree);


    /* free all best_options */
    for (b = predbeam - 1; b >= 0; b--) {
        if (best_options[length - 1][b] != NULL) {
            free_option_branch(best_options[length - 1][b]);
            best_options[length - 1][b] = NULL;
        }
    }

    return (length);
}



/* search the space of possible parser actions in period and return the best
     ones by setting best_options[period].  Don't pursue any options with
     lprob below pruning_threshold.  */
void search_options(int period, struct option_stack *init_queue,
                    double pruning_threshold, int length)
{
    struct option *toadd;
    double pred_lprob;
    double pred_ranklprob;
    int pred_out_mem;
    int first_out_mem, last_out_mem;
    int pred_out1, pred_out2;
    double pred_word_lprob;
    int memoffset;
    int bin;
    struct option *inbest;
    struct option *opt;
    struct option_stack *todel;
    struct option_stack *ol, *ol2, *ol3;
    struct option **last_best_options;
    struct option **next_best_options;
    int b, bl, blr, i, b2;
    double just_predlprob;
    double next_best_options_pred_prob;
    int lookahead_mapping[MAXPREDBEAM]; /* new to prev best */
    double lookahead_probs[MAXPREDBEAM];        /* sums over probs for new best */
    int lookahead_ranking[MAXPREDBEAM]; /* reranking of prev best */
    double lookahead_ranking_probs[MAXPREDBEAM];
    int predbeam_end;           /* end of new best */
    int idx;
    struct cluster *dec_cluster;
    int child_id;
    int trig_label;


#if LOSTTAIL
    double lost_postpred;
    double total_predprob;
    int num_predprobs;
#endif

#if LOSTTAIL
    total_predprob = 0.0;
    num_predprobs = 0;
    lost_postpred = 0.0;
    if (period == 1)
        lost_probability = 0.0;
#endif
#if ! LOCALESTIMATE
    total_finish_probs[period] = 0.0;
    total_finish_prob_weights[period] = 0.0;
#endif

    last_best_options = best_options[period - 1];
    next_best_options = best_options[period];
    worst_best_option_lprob = pruning_threshold;
    expected_max_pred_lprob = 0.0; 
    for (b = 0; b < predbeam; b++) {
        next_best_options[b] = NULL;
    }

    /* initialize queue with options from option stack root */
    option_queue = init_queue;
    for (b = predbeam - 1; b >= 0; b--) {
        if (last_best_options[b] != NULL) {
            init_option_phases(last_best_options[b]);
            comp_option_states(last_best_options[b]);
        /** set_option_states(last_best_options); **/
            if (last_best_options[b]->in_target == 0) {
                ol = generated_options(last_best_options[b], length);
                /* assume probably not more than phase_beam options */
                if (ol == NULL)
                    free_option_branch(last_best_options[b]);
                else
                    option_queue = mergesort_options(option_queue, ol);
            }
        }
    }
    pred_out1 = input_hist[period + 1].input1;

    if (option_queue == NULL)
        return;                 /* sometimes when training */

    while (option_queue != NULL) {
/*}}} */
        toadd = option_queue->elem;
        option_queue = popfree_option_stack(option_queue);
        if (worth_pursuing(toadd) &&
            (toadd->previous_option == NULL ||
             toadd->previous_option->num_descendents < phasebeam ||
             toadd->previous_output == PREDOUT ||
             (period == length - 1 && (toadd->previous_output == ATTOUT ||
                                       toadd->previous_output ==
                                       HDATOUT)))) {
            /* could lead to a better option, and within phase beam limit (unless
               only way to finish) */
            if (toadd->previous_option != NULL)
                toadd->previous_option->num_descendents++;
            init_option_phases(toadd);  /* if any new phases, then will be added */
	    if (toadd->phase < 0) {
	      /* not enough space, punt all but top on queue */
	      punt_queue(option_queue);
	      init_option_phases(toadd); /* try again */
	      if (toadd->phase < 0)
		printf("Still not enough space; Aaaarrrrrrgh...\n");
	    }
/*}}} */
            //if not last option in derivation 
            //(here was  bug, fixed dglc3e) - no condition
            if (toadd->phase != 0 || toadd->period != (sentence_length - 1)) {
                comp_option_states(toadd);  /* compute network states where needed */
            }

            /* check if toadd is one of best current predicting options */
            if (toadd->previous_output == PREDOUT) {
#if NOPREDICT
                pred_out_mem = predtag_to_output(0);
            /** should completely remove this option **/
                pred_lprob = toadd->lprob;
                pred_ranklprob = toadd->ranking_lprob;
#else
                comp_output_act_range(toadd, predtag_to_output(0),
                                      predtag_to_output(numout1 - 1));

                pred_out_mem = predtag_to_output(pred_out1);
                pred_lprob =
                    toadd->lprob + log(toadd->outputs->act[pred_out_mem]);
                pred_ranklprob =
                    toadd->ranking_lprob +
                    log(toadd->outputs->act[pred_out_mem]);
#endif

                trig_label =
                    pred_out_mem - predtag_to_output(0) + numlabels;
                dec_cluster = out_clusters_arr[pred_out1]->root;

                idx = 0;
            /** creating the internal prediction nodes */
                while (idx < input_hist[period + 1].cl_levels_out
                       && idx < max_cl_levels
                       && pred_ranklprob > worst_best_option_lprob) {

                    if (candidate_derivations != NULL) {
                        if (toadd->next_options == NULL)
                            warning
                                ("should always have a word to predict");
                        else {
                            toadd = toadd->next_options->elem;
                            toadd->lprob = pred_lprob;
                            toadd->ranking_lprob = pred_ranklprob;
                            toadd->prev_pred_ind =
                                toadd->previous_option->prev_pred_ind;
                            toadd->phase = toadd->previous_option->phase;
                            toadd->stack_rest_rel =
                                toadd->previous_option->stack_rest_rel;
                        }
                    } else
                    {
                        toadd =
                            create_option(period, toadd->phase, pred_lprob,
                                          pred_ranklprob, toadd->preterm,
                                          pred_out_mem, toadd,
                                          toadd->trigger_label,
                                          toadd->trigger_output,
                                          toadd->stack_subtop,
                                          toadd->stack_rest_rel,
                                          toadd->prev_pred_ind,
                                          toadd->context_option,
                                          toadd->trigparent_option,
                                          toadd->trigchild_option,
                                          toadd->headchild_option, 0, 0, 0,
                                          1);

                    }
                    toadd->dec_cluster = dec_cluster;
                    toadd->memoffset =
                        out1_first_word[pred_out1] +
                        dec_cluster->childs[0]->cluster_id;



#if NOPREDICT
                /** should completely remove this option **/
                    pred_lprob = toadd->previous_option->lprob;
                    pred_ranklprob = toadd->previous_option->ranking_lprob;
#else
                    first_out_mem = CHILDID_TO_MEMOUT(0);
                    last_out_mem =
                        CHILDID_TO_MEMOUT(dec_cluster->child_qty - 1);

                    comp_option_states(toadd);
                    comp_output_act_range(toadd, first_out_mem,
                                          last_out_mem);

                    pred_out2 = input_hist[period + 1].output2[idx];
                    pred_lprob =
                        toadd->lprob +
                        log(toadd->outputs->
                        act[WORDID_TO_MEMOUT(pred_out2, toadd->memoffset)]);
                    pred_ranklprob =
                        toadd->ranking_lprob +
                        log(toadd->outputs->
                        act[WORDID_TO_MEMOUT(pred_out2, toadd->memoffset)]);
#endif

                    child_id =
                        WORDID_TO_CHILDID(input_hist[period + 1].
                                          output2[idx], pred_out1,
                                          dec_cluster->childs[0]->
                                          cluster_id);
                    pred_out_mem = CHILDID_TO_MEMOUT(child_id);
                    dec_cluster = dec_cluster->childs[child_id];


                    idx++;

                }


                if (pred_ranklprob <= worst_best_option_lprob) {
#if LOSTTAIL
                    lost_postpred += exp(pred_lprob);
#endif
                    free_option_branch(toadd);
                } else {

                    /* shift best options */
                    if (next_best_options[predbeam - 1] != NULL) {
		        for (opt = next_best_options[predbeam - 1]; 
			     opt != NULL; opt = opt->previous_option) {
			  if (opt->previous_output == PREDOUT) {
			    next_best_options_pred_prob -= 
			      exp(next_best_options[predbeam - 1]->lprob 
				  - opt->lprob);
			    break;
			  }
			}
#if LOSTTAIL
                        lost_postpred += exp(worst_best_option_lprob);
#endif
                        free_option_branch(next_best_options
                                           [predbeam - 1]);
                    }
		    else if (predbeam == 1 || 
			     next_best_options[predbeam - 2] != NULL) {
		      /* just filled beam */
		      next_best_options_pred_prob = 0.0;
		      for (b = 0; b < predbeam; b++) {
		        for (opt = next_best_options[b]; 
			     opt != NULL; opt = opt->previous_option) {
			  if (opt->previous_output == PREDOUT) {
			    next_best_options_pred_prob += 
			      exp(next_best_options[b]->lprob - opt->lprob);
			    break;
			  }
			}
		      }
		    }
		    for (opt = toadd; opt != NULL; 
			 opt = opt->previous_option) {
		      if (opt->previous_output == PREDOUT) {
			next_best_options_pred_prob += 
			  exp(pred_lprob - opt->lprob);
			break;
		      }
		    }
		    /* average pred prob in best times factor */
		    expected_max_pred_lprob = 
		      log(expected_max_pred_factor)
		      + log(next_best_options_pred_prob)
		      - log((double) predbeam);
                    b = predbeam - 1;
                    while (b > 0 && (next_best_options[b - 1] == NULL ||
                                     pred_ranklprob >
                                     next_best_options[b -
                                                       1]->
                                     ranking_lprob)) {

                        next_best_options[b] = next_best_options[b - 1];
                        b--;
                    }

                    /* tag prediction, push preterm */
                    if (toadd->preterm != 0) {
                        error("can no longer predict from preterms");
                    }
                    if (candidate_derivations != NULL) {
                        if (toadd->next_options == NULL)
                            warning
                                ("should always have a word to predict");
                        else {
                            next_best_options[b] =
                                toadd->next_options->elem;
                            next_best_options[b]->lprob = pred_lprob;
                            next_best_options[b]->ranking_lprob =
                                pred_ranklprob;
                            next_best_options[b]->prev_pred_ind =
                                toadd->prev_pred_ind;
                            next_best_options[b]->phase = -1;
                            next_best_options[b]->stack_rest_rel =
                                toadd->stack_rest_rel;
                        }
                    } else
                    {

                        next_best_options[b] =
                            create_option(period + 1, -1, pred_lprob,
                                          pred_ranklprob, 1, pred_out_mem,
                                          toadd, trig_label, PREDOUT,
                                          toadd->stack_subtop,
                                          toadd->stack_rest_rel,
                                          toadd->prev_pred_ind, NULL,
                                          toadd, NULL, NULL, 1, 1, 0, 0);
                    }
                    /* when using a list of best options, move above into
                       generated_options() for consistency */
                    if (next_best_options[predbeam - 1] != NULL) {
                        worst_best_option_lprob =
                            next_best_options[predbeam - 1]->ranking_lprob;
                    }
/*}}} */
                }               /* eo else pred_ranklprob <= worst_best_option_lprob */
            } /** eo if (toadd->previous_output == PREDOUT) */
            else if (period == length - 1 && toadd->phase == 0 && (toadd->previous_output == HDATOUT || toadd->previous_output == ATTOUT)) {    /* at end, best is an attachment to root */
                /* already know in best because was selected from queue */
                if (next_best_options[predbeam - 1] != NULL)
                    free_option_branch(next_best_options[predbeam - 1]);
                b = predbeam - 1;
                while (b > 0 &&
                       (next_best_options[b - 1] == NULL ||
                        toadd->ranking_lprob >
                        next_best_options[b - 1]->ranking_lprob)) {
                    next_best_options[b] = next_best_options[b - 1];
                    b--;
                }
                next_best_options[b] = toadd;
                if (next_best_options[predbeam - 1] != NULL)
                    worst_best_option_lprob =
                        next_best_options[predbeam - 1]->ranking_lprob;
            } else {
                /* add new options to queue */
                ol = generated_options(toadd, length);
                if (ol != NULL) {
                    /* don't sort options which can't be used */
                    i = 0;
                    for (ol2 = ol; ol2 != NULL; ol2 = ol3) {
                        i++;
                        ol3 = ol2->next;
                        if (i > phasebeam) {
                            free_option_branch(ol2->elem);
                            pluckfree_option_stack(ol2);
                        }
                    }
                }
                if (ol == NULL)
                    free_option_branch(toadd);
                else
                    option_queue = mergesort_options(option_queue, ol);
            }
        } else {
#if LOSTTAIL
            lost_probability += exp(toadd->lprob);
#endif

            free_option_branch(toadd);  /* if not pursued then free */
        }
    }

    if (predbeam != 0 && next_best_options[0] == NULL) {
        /* no predicting option found */
        error("no predicting option found.  Attaching to root.");
        /*
           printf("previous best:\n");
           for (b = 0; b < predbeam; b++)
           if (last_best_options[b] != NULL) {
           display_symbolic_option(last_best_options[b]);
           }
         */
/*        for (b = 0; b < predbeam; b++)
          if (last_best_options[b] != NULL) {
            next_best_options[b] = punting_option(last_best_options[b], length);
          } */
    }

#if LOSTTAIL
    /* scale lost_probability by a good prediction probability (overestimate) */
    /*
       if (predbeam != 0 && next_best_options[0] != NULL && 
       predtag_to_output(0) <= next_best_options[0]->previous_output &&
       next_best_options[0]->previous_output <= 
       predtag_to_output(numout1 - 1))
       {
       lost_probability *= 
       exp(next_best_options[0]->lprob) / 
       exp(next_best_options[0]->previous_option->lprob);
       }
     */
    if (num_predprobs != 0)
        lost_probability *= total_predprob / (double) num_predprobs;
    lost_probability += lost_postpred;
#endif

#if LOOKAHEADPRUNING
    /* prune new best which are not descended from top best at t-lookahead */
    if (lookahead != 0 && period > lookahead) {
        /* b is current ranking at t, bl is current ranking at t-lookahead, 
           blr new ranking at t-lookahead */
        for (bl = 0; bl < predbeam; bl++) {
            lookahead_probs[bl] = 0.0;
        }
        predbeam_end = predbeam;
        for (b = 0; b < predbeam; b++) {
            if (next_best_options[b] != NULL) {
                bl = next_best_options[b]->previous_option->prev_pred_ind;
                for (i = 1; i < lookahead; i++) {
                    bl = best_options[period -
                                      i][bl]->previous_option->
                        prev_pred_ind;
                }
                lookahead_mapping[b] = bl;
                lookahead_probs[bl] += exp(next_best_options[b]->lprob);
            } else {
                predbeam_end = b;
                break;
            }
        }
        for (blr = 0; blr < lookahead_predbeam; blr++) {
            lookahead_ranking[blr] = -1;
            lookahead_ranking_probs[blr] = 0.0;
            for (bl = 0; bl < predbeam; bl++) {
                if (lookahead_probs[bl] > lookahead_ranking_probs[blr]) {
                    lookahead_ranking[blr] = bl;
                    lookahead_ranking_probs[blr] = lookahead_probs[bl];
                }
            }
            if (lookahead_ranking[blr] != -1)
                lookahead_probs[lookahead_ranking[blr]] = 0.0;
        }
        for (b = predbeam_end - 1; b >= 0; b--) {
            for (blr = 0; blr < lookahead_predbeam; blr++) {
                if (lookahead_ranking[blr] == lookahead_mapping[b])
                    break;
            }
            if (blr == lookahead_predbeam) {
#if LOSTTAIL
                lost_postpred += exp(next_best_options[b]->lprob);
#endif
                free_option_branch(next_best_options[b]);
            }
        }
    }
#endif

}


/* compute the list of options generated from last.  Assumes hidden state of
     last's period,phase and last's previous' period,phase have been computed.
     Does not include tag predictions. */
struct option_stack *generated_options(struct option *last, int length)
{
    struct option_stack *res;
    struct option *new;
    struct option *newcont;
    struct option_stack *ol;
    double new_lprob;
    double new_ranklprob;
    int current;
    int site;
    int lab;
    int out;
    int newsubtop;
    int newtop;
    int newrest;
    int newpreterm;

    res = NULL;
    current = last->period;

/*{{{ */
    /* when candidates are set, use the next candidates */
    if (candidate_derivations != NULL) {
#if DEBUGCAND2/*{{{ */
        check_option_tree(candidate_derivations);
#endif/*}}} */

        if (last->previous_output == PROJOUT) {
            /* case when projection label chosen */
            if (last->in_target == 0)   /* else already computed */
                comp_output_act_range(last, projlabel_to_output(0),
                                      projlabel_to_output(numlabels - 1));
            for (ol = last->next_options; ol != NULL;) {
                new = ol->elem;
                ol = ol->next;
                if (new->in_target == 0) {
                    out = new->previous_output;
                    new->lprob =
                        last->lprob + log(last->outputs->act[out]);
                    new->ranking_lprob =
                        last->ranking_lprob + log(last->outputs->act[out]);
                    new->prev_pred_ind = last->prev_pred_ind;
                    new->phase = -1;
                    new->stack_rest_rel = last->stack_rest_rel;
                }
                if (new->ranking_lprob > worst_best_option_lprob) {
                    res = mergesort_options(res, create_option_cons(new));
                } else if (new->in_target == 0) {
                    free_option_descendents(new);
                }
            }
        }
#if HEADOUTS/*{{{ */
        else if (last->previous_output == HDPROUT) {
            /* case when head projection label chosen */
            if (last->in_target == 0)   /* else already computed */
                comp_output_act_range(last, hdprlabel_to_output(0),
                                      hdprlabel_to_output(numlabels - 1),
                                      0);
            for (ol = last->next_options; ol != NULL;) {
                new = ol->elem;
                ol = ol->next;
                if (new->in_target == 0) {
                    out = new->previous_output;
                    new->lprob =
                        last->lprob + log(last->outputs->act[out]);
                    new->ranking_lprob =
                        last->ranking_lprob + log(last->outputs->act[out]);
                    new->prev_pred_ind = last->prev_pred_ind;
                    new->phase = -1;
                    new->stack_rest_rel = last->stack_rest_rel;
                }
                if (new->ranking_lprob > worst_best_option_lprob) {
                    res = mergesort_options(res, create_option_cons(new));
                } else if (new->in_target == 0) {
                    free_option_descendents(new);
                }
            }
        }
#endif/*}}} */
        else {
            /* pre-prediction cases */
            if (last->in_target == 0)   /* else already computed */
                 comp_output_act_range(last, PREPREDFIRST, PREPREDLAST);
            for (ol = last->next_options; ol != NULL;) {
                new = ol->elem;
                ol = ol->next;
                if (new->in_target == 0) {
                    out = new->previous_output;
                    new->lprob =
                        last->lprob + log(last->outputs->act[out]);
                    new->ranking_lprob =
                        last->ranking_lprob + log(last->outputs->act[out]);
                    new->prev_pred_ind = last->prev_pred_ind;
                    if (out == PROJOUT
#if HEADOUTS/*{{{ */
                        || out == HDPROUT
#endif/*}}} */
                        || out == PREDOUT) {
                        /* for PROJOUT, delay pop and pushing nonterm until labels */
                        /* for HDPROUT, delay pop and pushing nonterm until labels */
                        /* for PREDOUT, delay everything until after tag prediction */
                        new->phase = last->phase;
                        new->stack_rest_rel = last->stack_rest_rel;
                    } else {
                        /* for ATTOUT, HDATOUT and MODOUT, pop */
                        new->phase = last->trigparent_option->phase;
                        /* new->phase = *(last->stack_rest_rel + 
                           phase_stacks[last->prev_pred_ind]->content);
                           new->stack_rest_rel = 
                           relative_pop_stack(last->stack_rest_rel);
                         */
                    }
                }
                if (new->ranking_lprob > worst_best_option_lprob) {
                    res = mergesort_options(res, create_option_cons(new));
                } else if (new->in_target == 0) {
                    free_option_descendents(new);
                }
            }
        }

    } else
/*}}} */
    {
        if (last->previous_output == PROJOUT) {
            /* case when projection label chosen */
            site = last->phase; /* site on top of stack */
            comp_output_act_range(last, projlabel_to_output(0),
                                  projlabel_to_output(numlabels - 1));
            for (lab = 0; lab < numlabels; lab++) {
                /* pop and push nonterm */
                out = projlabel_to_output(lab);
                new_lprob = last->lprob + log(last->outputs->act[out]);
                new_ranklprob =
                    last->ranking_lprob + log(last->outputs->act[out]);
                if (new_ranklprob > worst_best_option_lprob) {
                    new =
                        create_option(current, -1, new_lprob, 
                                      new_ranklprob, 0, out, last,
                                      out - projlabel_to_output(0),
                                      last->previous_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind, NULL,
                                      last->trigparent_option, last, NULL,
                                      1, 1, 0, 0);
                    res = mergesort_options(res, create_option_cons(new));
                }
            }
        }
#if HEADOUTS/*{{{ */
        else if (last->previous_output == HDPROUT) {
            /* case when head projection label chosen */
            site = last->phase; /* site on top of stack */
            comp_output_act_range(last, hdprlabel_to_output(0),
                                  hdprlabel_to_output(numlabels - 1), 0);
            for (lab = 0; lab < numlabels; lab++) {
                /* pop and push nonterm */
                out = hdprlabel_to_output(lab);
                new_lprob = last->lprob + log(last->outputs->act[out]);
                new_ranklprob =
                    last->ranking_lprob + log(last->outputs->act[out]);
                if (new_ranklprob > worst_best_option_lprob) {
                    new =
                        create_option(current, -1, new_lprob,
                                      new_ranklprob, 0, out, last,
                                      out - hdprlabel_to_output(0),
                                      last->previous_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind, NULL,
                                      last->trigparent_option, last, last,
                                      1, 1, 0, 0);
                    res = mergesort_options(res, create_option_cons(new));
                }
            }
        }
#endif/*}}} */
        else {
            /* pre-prediction cases */
            comp_output_act_range(last, PREPREDFIRST, PREPREDLAST);
            if (last->trigparent_option != NULL
#if HEADOUTS
                && (last->preterm != 0 || last->headchild_option != NULL)
#endif
                ) {
                /* can't attach, modify or project without something to pop */
                /* and can't pop if nonterm and doesn't yet have a head */

                /* for ATTOUT, HDATOUT and MODOUT, pop */
                newcont = last->trigparent_option;
                newtop = newcont->phase;
                newrest = last->stack_rest_rel;
                                             /** remove **/
                /* attach */
#if HEADOUTS
                if (current < length - 1
                    || newcont->headchild_option != NULL)
                    /* can't attach if last chance for a head */
#endif
                {
                    new_lprob =
                        last->lprob + log(last->outputs->act[ATTOUT]);
                    new_ranklprob =
                        (last->ranking_lprob - log(finishing_prob)) +
                        log(last->outputs->act[ATTOUT]);
                    if (new_ranklprob > worst_best_option_lprob) {
                        /* new_state=1 might be reset by context in same period ?? */
                        new =
                            create_option(current, newtop, new_lprob,
                                          new_ranklprob, newcont->preterm,
                                          ATTOUT, last,
                                          newcont->trigger_label,
                                          newcont->trigger_output, -2,
                                          newrest, last->prev_pred_ind,
                                          newcont,
                                          newcont->trigparent_option,
                                          newcont->trigchild_option,
                                          newcont->headchild_option, 0, 1,
                                          0, 0);
                        res =
                            mergesort_options(res,
                                              create_option_cons(new));
                    }
                }
#if HEADOUTS
                /* head attach */
                if (newcont->headchild_option == NULL) {
                    new_lprob =
                        last->lprob + log(last->outputs->act[HDATOUT]);
                    new_ranklprob =
                        (last->ranking_lprob - log(finishing_prob)) +
                        log(last->outputs->act[HDATOUT]);
                    if (new_ranklprob > worst_best_option_lprob) {
                        /* new_state=1 might be reset by context in same period ?? */
                        new =
                            create_option(current, newtop, new_lprob,
                                          new_ranklprob, newcont->preterm,
                                          HDATOUT, last,
                                          newcont->trigger_label,
                                          newcont->trigger_output, -2,
                                          newrest, last->prev_pred_ind,
                                          newcont,
                                          newcont->trigparent_option,
                                          newcont->trigchild_option, last,
                                          0, 1, 0, 0);
                        res =
                            mergesort_options(res,
                                              create_option_cons(new));
                    }
                }
#endif
                /* modify */
                if (newtop != 0
#if HEADOUTS
                    && (current < length - 1
                        || newcont->headchild_option != NULL)
#endif
                    ) {
                    /* can't modify the root */
                    /* can't modify if last chance for a head */
                    new_lprob =
                        last->lprob + log(last->outputs->act[MODOUT]);
                    new_ranklprob =
                        (last->ranking_lprob - log(finishing_prob)) +
                        log(last->outputs->act[MODOUT]);
                    if (new_ranklprob > worst_best_option_lprob) {
                        /* new_state=1 might be reset by context in same period ?? */
                        new =
                            create_option(current, newtop, new_lprob,
                                          new_ranklprob, newcont->preterm,
                                          MODOUT, last,
                                          newcont->trigger_label,
                                          newcont->trigger_output, -2,
                                          newrest, last->prev_pred_ind,
                                          newcont,
                                          newcont->trigparent_option,
                                          newcont->trigchild_option,
                                          newcont->headchild_option, 0, 1,
                                          0, 0);
                        res =
                            mergesort_options(res,
                                              create_option_cons(new));
                    }
                }
                /* for PROJOUT, delay pop and pushing nonterm until computing labels */
#if HEADOUTS
                if (current < length - 1)
                    /* can't project if last chance for a head */
#endif
                {
                    if (last->proj_depth < MAXPROJDEPTH) {
                        new_lprob =
                            last->lprob + log(last->outputs->act[PROJOUT]);
                        new_ranklprob =
                            last->ranking_lprob +
                            log(last->outputs->act[PROJOUT]);
                        new =
                            create_option(current, last->phase, new_lprob,
                                          new_ranklprob, last->preterm,
                                          PROJOUT, last,
                                          last->trigger_label,
                                          last->trigger_output,
                                          last->stack_subtop,
                                          last->stack_rest_rel,
                                          last->prev_pred_ind,
                                          last->context_option,
                                          last->trigparent_option,
                                          last->trigchild_option,
                                          last->headchild_option, 0, 0, 0,
                                          0);

                        res =
                            mergesort_options(res,
                                              create_option_cons(new));
                    }
                }
#if HEADOUTS
                /* for HDPROUT, delay pop and pushing nonterm until computing labels */
                if (last->proj_depth < MAXPROJDEPTH) {
                    new_lprob =
                        last->lprob + log(last->outputs->act[HDPROUT]);
                    new_ranklprob =
                        last->ranking_lprob +
                        log(last->outputs->act[HDPROUT]);
                    new =
                        create_option(current, last->phase, new_lprob,
                                      new_ranklprob, last->preterm,
                                      HDPROUT, last, last->trigger_label,
                                      last->trigger_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind,
                                      last->context_option,
                                      last->trigparent_option,
                                      last->trigchild_option,
                                      last->headchild_option, 0, 0, 0, 0);
                    res = mergesort_options(res, create_option_cons(new));
                }
#endif
            }
            if (current != length - 1 && last->preterm == 0) {  /* can't predict at end, or from preterms */
                /* for PREDOUT, delay everything until after tag prediction */
                new_lprob = last->lprob + log(last->outputs->act[PREDOUT]);
                new_ranklprob =
                    last->ranking_lprob + log(last->outputs->act[PREDOUT]) + 
		    log(finishing_prob);
                new =
                    create_option(current, last->phase, new_lprob,
                                  new_ranklprob, last->preterm, PREDOUT,
                                  last, last->trigger_label,
                                  last->trigger_output, last->stack_subtop,
                                  last->stack_rest_rel,
                                  last->prev_pred_ind,
                                  last->context_option,
                                  last->trigparent_option,
                                  last->trigchild_option,
                                  last->headchild_option, 0, 0, 0, 1);
                res = mergesort_options(res, create_option_cons(new));
            } else {

            }
        }
    }
    return (res);
}


/* use worst_best_option_lprob, etc to determine whether opt is worth
   continuing the search with.  */ 
int worth_pursuing(struct option *opt)
{
  double opt_ranklprob;

  opt_ranklprob = opt->ranking_lprob;

  if (opt->is_pred_opt)
    {
      if (opt_ranklprob > worst_best_option_lprob)
	return 1;
      else
	return 0;
    }
  else {
    if (opt_ranklprob > worst_best_option_lprob - expected_max_pred_lprob)
      return 1;
    else {
      return 0;
    }
  }
}


/* free all but top option on q  */ 
void punt_queue(struct option_stack *q)
{
  struct option_stack *ol;

  if (q == NULL)
    return;
  for (ol = q->next; ol != NULL; ol = popfree_option_stack(ol)) {
    free_option_branch(ol->elem);
  }
}


/* initialize the phase control structures.  phase of matrix root is 0, and
     is automatically allocated. */
void init_phases_for_options()
{
    int b;

    init_free_phases();
    if (allocate_option_phase(0, ROOTFLAG, -1, -1, label_id("")) != 0)
        warning("matrix root not 0.");
}


/* return generated prediction option for sentence root */
struct option *root_prediction()
{
    double just_predlprob;
    double pred_lprob;
    double pred_ranklprob;
    int bin;
    int pred_out_mem;
    int num_out1;
    double pred_word_lprob;
    int memoffset;
    int first_out_mem, last_out_mem;
    int pred_out1, pred_out2;
    struct option *toadd;
    struct option *new;
    struct cluster *dec_cluster;
    int idx;
    int child_id;
    int trig_label;

    double all_pred_lprob;

    all_pred_lprob = log(1.0);

    /* when candidates are set, use the root candidate prediction */
    if (candidate_derivations != NULL) {
#if DEBUGCAND2
        check_option_tree(candidate_derivations);
#endif
        toadd = candidate_derivations;
        toadd->lprob = log(1.0);
        toadd->ranking_lprob = log(1.0);
        toadd->prev_pred_ind = 0;
        toadd->phase = 0;
        toadd->stack_rest_rel = -1;
    } else
        toadd = create_option(0, 0, log(1.0), log(1.0), 0, PREDOUT, NULL,
                              label_id(""), ROOTFLAG, -2, -1, 0,
                              NULL, NULL, NULL, NULL, 1, 1, 0, 1);

    comp_state_act(toadd);

#if NOPREDICT
    pred_out_mem = predtag_to_output(0);
    /** should completely remove this option **/
    pred_lprob = toadd->lprob;
    pred_ranklprob = toadd->ranking_lprob;
#else
    pred_out1 = input_hist[1].input1;

    comp_output_act_range(toadd, predtag_to_output(0),
                          predtag_to_output(numout1 - 1));

    pred_out_mem = predtag_to_output(pred_out1);
    pred_lprob = toadd->lprob + log(toadd->outputs->act[pred_out_mem]);
    all_pred_lprob += pred_lprob;
    pred_ranklprob =
        toadd->ranking_lprob + log(toadd->outputs->act[pred_out_mem]);
#endif

    trig_label = pred_out_mem - predtag_to_output(0) + numlabels;
    dec_cluster = out_clusters_arr[pred_out1]->root;

    idx = 0;

    /** creating the internal prediction nodes */
    while (idx < input_hist[1].cl_levels_out && idx < max_cl_levels) {

        if (candidate_derivations != NULL) {
            if (toadd->next_options == NULL) {
                warning("should always have a word to predict");
            } else {
                toadd = toadd->next_options->elem;
                toadd->lprob = pred_lprob;
                toadd->ranking_lprob = pred_ranklprob;
                toadd->prev_pred_ind =
                    toadd->previous_option->prev_pred_ind;
                toadd->phase = toadd->previous_option->phase;
                toadd->stack_rest_rel =
                    toadd->previous_option->stack_rest_rel;
            }
        } else
        {
            toadd =
                create_option(0, toadd->phase, pred_lprob, pred_ranklprob,
                              toadd->preterm, pred_out_mem, toadd,
                              toadd->trigger_label,
                              toadd->trigger_output,
                              toadd->stack_subtop,
                              toadd->stack_rest_rel, toadd->prev_pred_ind,
                              toadd->context_option,
                              toadd->trigparent_option,
                              toadd->trigchild_option,
                              toadd->headchild_option, 0, 0, 0, 1);

        }
        toadd->dec_cluster = dec_cluster;
        toadd->memoffset =
            out1_first_word[pred_out1] +
            dec_cluster->childs[0]->cluster_id;



#if NOPREDICT
        /** should completely remove this option **/
        pred_lprob = toadd->previous_option->lprob;
        pred_ranklprob = toadd->previous_option->ranking_lprob;
#else
        first_out_mem = CHILDID_TO_MEMOUT(0);
        last_out_mem = CHILDID_TO_MEMOUT(dec_cluster->child_qty - 1);

        comp_option_states(toadd);
        comp_output_act_range(toadd, first_out_mem, last_out_mem);

        pred_out2 = input_hist[1].output2[idx];
        pred_lprob =
            toadd->lprob +
            log(toadd->outputs->
                act[WORDID_TO_MEMOUT(pred_out2, toadd->memoffset)]);
        pred_ranklprob =
            toadd->ranking_lprob +
            log(toadd->outputs->
                act[WORDID_TO_MEMOUT(pred_out2, toadd->memoffset)]);
        all_pred_lprob += pred_lprob;

#endif
        child_id =
            WORDID_TO_CHILDID(input_hist[1].output2[idx], pred_out1,
                              dec_cluster->childs[0]->cluster_id);
        pred_out_mem = CHILDID_TO_MEMOUT(child_id);

        dec_cluster = dec_cluster->childs[child_id];


        idx++;

    }                           /* eo while */

#if LOSTTAIL
    total_predprob += exp(all_pred_lprob);
    num_predprobs++;
#endif



#if DEBUGCAND2
    check_option_tree(candidate_derivations);
#endif
    if (candidate_derivations != NULL) {
        if (toadd->next_options == NULL)
            warning("should always have a word to predict");
        else {
            new = toadd->next_options->elem;
            new->lprob = pred_lprob;
            new->ranking_lprob = pred_ranklprob;
            new->prev_pred_ind = toadd->prev_pred_ind;
            new->phase = -1;
            new->stack_rest_rel = toadd->stack_rest_rel + 1;
        }
    } else
    {

        new =
            create_option(1, -1, pred_lprob, pred_ranklprob,
                          1, pred_out_mem, toadd,
                          trig_label,
                          PREDOUT,
                          toadd->stack_subtop, toadd->stack_rest_rel,
                          toadd->prev_pred_ind, NULL, toadd, NULL, NULL,
                          1, 1, 0, 0);

    }

    return new;
}



/* shift all options below b in best up by one. */
void pluck_best_option(int b, struct option **best)
{
    int b2;

    for (b2 = b; b2 < predbeam - 1 && best[b2 + 1] != NULL; b2++) {
        best[b2] = best[b2 + 1];
    }
    best[b2] = NULL;
}


/* allocate and initialize any phases not yet set in opt.  New phases'
     states computed by comp_option_states().  If there is not enough memory 
     for a new phase then left as -1.  */
void init_option_phases(struct option *opt)
{
    int optout;
    int hypoth;
    int par, chd;
    int prevout;

    if (opt->phase < 0) {       /* else already done, do nothing */
        if (opt->context_option == NULL) {
            if (opt->previous_option == NULL) {
                opt->phase =
                    allocate_option_phase(opt->period, ROOTFLAG, -1, -1,
                                          label_id(""));
            } else {
                prevout = opt->previous_option->previous_output;
                if (opt->previous_option->is_pred_opt) {
                    /* new preterm, top of option's stack */
                    opt->phase =
                        allocate_option_phase(opt->period, prevout,
                                              opt->trigparent_option->
                                              phase, -1,
                                              numlabels +
                                              input_hist[opt->period].
                                              input1);
                } else if (prevout == PROJOUT) {
                    opt->phase =
                        allocate_option_phase(opt->period, prevout,
                                              opt->trigparent_option->
                                              phase,
                                              opt->previous_option->phase,
                                              output_to_projlabel(opt->
                                                                  previous_output));
                }
#if HEADOUTS
                else if (prevout == HDPROUT) {
                    opt->phase =
                        allocate_option_phase(opt->period, prevout,
                                              opt->trigparent_option->
                                              phase,
                                              opt->previous_option->phase,
                                              output_to_hdprlabel(opt->
                                                                  previous_output));
                }
#endif
            }
        } else {
            opt->phase = opt->context_option->phase;
        }
    }
}


/* compute states of phases where needed.  This procedure is closely tied to 
     init_option_phases, but it is an expensive procedure and should only be
     called when it is known to be needed.  phase_hist states updated by
     set_option_states() if in best. */
void comp_option_states(struct option *opt)
{
    int optout;
    int hypoth;
    struct option *cont;

    optout = opt->previous_output;

    /** shouldn't be needed **/
    if (opt->context_option != NULL && opt->context_option->period == opt->period) {    /* then already computed */
        warning("vistigual case.");
        opt->new_state = 0;
        opt->state = opt->context_option->state;
        opt->context_option = opt->context_option->context_option;
        opt->outputs = opt->context_option->outputs;
    }

    if (opt->new_state != 0) {
        /* new state for old or new phase */
        comp_state_act(opt);
    } else if (opt->is_pred_opt) {
        //allocate own output space, but should share the state and combination with previous options 
        opt->state = opt->previous_option->state;
        if (opt->outputs == NULL) {
            opt->outputs = (struct output *) malloc(sizeof(struct output));
        }
        CHECK_PTR(opt->outputs);


    } else {
        /* state (and output) set earlier */
        opt->state = opt->previous_option->state;
        opt->outputs = opt->previous_option->outputs;
    }
}



/* allocate a new phase from free_phases.  If there is not enough memory to
     allocate a new phase then return -1. */
int allocate_option_phase(int birth, int trigop, int trigparent,
                          int trigchild, int triglabel)
{
    int res;
    int oldspace;
    int i;
    int nexttop;

    if (free_phases->top == NULL) {     /* if no free phases, add or make some */
        oldspace = phase_space;
        if (expand_phases(-1) == 0) {   /* -1 means hist_ arrays not expanded */
            warning("expand_phases returned 0.");
            return (-1);        /* if expansion failed, return -1 */
        }
        nexttop = 0;
        for (i = phase_space - 1; i >= oldspace; i--) {
            free_phases->content[nexttop] = i;
            nexttop++;
        }
        free_phases->top = &(free_phases->content[nexttop - 1]);
        free_phases->subtop = &(free_phases->content[nexttop - 2]);
    }

    res = *(free_phases->top);
    pop_stack(free_phases);

    if (res < 0) {
        sprintf(warning_text, "*** %d on free_phases list.", res);
        warning(warning_text);
    }



    return (res);
}


/* initialize free_phases with all phases in phase_space */
void init_free_phases()
{
    int i;
    int nexttop;

    nexttop = 0;
    for (i = phase_space - 1; i >= 0; i--) {
        free_phases->content[nexttop] = i;
        nexttop++;
    }
    free_phases->top = &(free_phases->content[nexttop - 1]);
    free_phases->subtop = &(free_phases->content[nexttop - 2]);
}


/* expand the memory space for free_phases and add the new phases to the
     bottom of the stack.  Returns a flag for successful memory allocation. */
int expand_free_phases(int oldsize, int newsize)
{
    int *newcontent, *oldelem;
    int nexttop;
    int p;

    newcontent = (int *) malloc(sizeof(int[newsize]));
    if (newcontent == NULL) {
        warning("Cannot allocate the desired number of phases.\n");
        return (0);
    }
    nexttop = 0;
    for (p = newsize - 1; p >= oldsize; p--) {
        newcontent[nexttop] = p;
        nexttop++;
    }
    if (free_phases->top != NULL)
        for (oldelem = free_phases->content; oldelem <= free_phases->top;
             oldelem++) {
            newcontent[nexttop] = *oldelem;
            nexttop++;
        }
#if ! NOFREEING
    free(free_phases->content);
#endif

    free_phases->content = newcontent;
    free_phases->top = &(free_phases->content[nexttop - 1]);
    free_phases->subtop = &(free_phases->content[nexttop - 2]);

    return (1);
}



/* perform all the options in the derivation ending at opt and return the
     resulting tree.  Reintroduce chomsky adjunction for modifications
     and conflated nodes, and expand conflations.  Assumes root has phase 0. */
struct tree *perform_derivation(struct option *opt)
{
    struct tree *res;
    struct tree_list *trst;

    trst = start_root(0, 0, NULL);
    res = trst->elem;
    trst = perform_options(trst, 0, opt, 1);
    res->end = opt->period + 1;
    expand_conflations(res);
    free_tree_list(trst);

    return (res);
}

/* perform newoption's operations after input_tree_period on
     input_tree_stack and return it.  If format_flag != 0 then reintroduce
     chomsky adjunction for modifications and conflated nodes.  */
struct tree_list *perform_options(struct tree_list *input_tree_stack,
                                  int input_tree_period,
                                  struct option *newoption,
                                  int format_flag)
{
    struct tree_list *tree_stack;
    int position;
    struct option_stack *derivation;
    struct option *opt;
    struct tree_list *tl;
    struct tree *new_preterm;
    struct tree *new_nonterm;
    struct tree *top;
    struct tree *subtop;
    int i, j;


    tree_stack = input_tree_stack;
    derivation = NULL;
    for (opt = newoption; opt != NULL &&
         (opt->period > input_tree_period ||
          (opt->period == input_tree_period && IS_CLUSTER_OPT(opt)));
         opt = opt->previous_option) {
        derivation = create_option_stack(opt, derivation, NULL);
        /* if (derivation->next != NULL) derivation->next->prev = derivation; */
    }

    for (; derivation != NULL;
         derivation = popfree_option_stack(derivation)) {
        opt = derivation->elem;
        position = opt->period;
        if (opt->previous_output >= predtag_to_output(0) &&
            (numout1 == 0 ||
             opt->previous_output <= predtag_to_output(numout1 - 1))) {
            new_preterm = create_tree();
            new_preterm->start = position + 1;
            new_preterm->site_id = opt->phase;
            strcpy(new_preterm->label,
                   id_tag(input_hist[position + 1].input1));
            /* push preterm */
            tree_stack = create_tree_list(new_preterm, tree_stack);
        } else if (opt->previous_output == ATTOUT) {
            top = tree_stack->elem;
            subtop = tree_stack->next->elem;
            subtop->children = create_tree_list(top, subtop->children);
            top->parent = subtop;
            top->end = position + 1;
            tree_stack = popfree_tree_list(tree_stack); /* pop attached nonterm */
        }
#if HEADOUTS
        else if (opt->previous_output == HDATOUT) {
            top = tree_stack->elem;
            subtop = tree_stack->next->elem;
            subtop->children = create_tree_list(top, subtop->children);
            subtop->head_child = subtop->children;
            top->parent = subtop;
            top->end = position + 1;
            top->head = 1;
            /* pop head attached nonterm */
            tree_stack = popfree_tree_list(tree_stack);
        }
#endif
        else if (opt->previous_output == MODOUT) {
            top = tree_stack->elem;
            subtop = tree_stack->next->elem;
            if (format_flag != 0) {
                /* create Chomsky adjunction nonterm */
                new_nonterm = create_tree();
                new_nonterm->start = subtop->start;
                new_nonterm->site_id = subtop->site_id;
                /* set and reset labels */
                for (i = 0; subtop->label[i] != '\0' && i < MAXLABELLENGTH;
                     i++)
                    new_nonterm->label[i] = subtop->label[i];
                new_nonterm->label[i] = '\0';
                while (subtop->label[i] != '-' && i > 0)
                    i--;
                if (i > 0) {    /* relabel site with last conflated label */
                    for (j = 0;
                         subtop->label[i + j + 1] != '\0'
                         && j < MAXLABELLENGTH; j++)
                        subtop->label[j] = subtop->label[i + j + 1];
                    subtop->label[j] = '\0';
                }
                /* do Chomsky adjunction */
                new_nonterm->children =
                    create_tree_list(subtop, new_nonterm->children);
                subtop->parent = new_nonterm;
                subtop->end = top->start;
                new_nonterm->children =
                    create_tree_list(top, new_nonterm->children);
                top->parent = new_nonterm;
                top->end = position + 1;
                /* update tree_stack */
                tree_stack = popfree_tree_list(tree_stack);     /* pop modifier */
                tree_stack = popfree_tree_list(tree_stack);     /* pop modification site */
                /* push modification */
                tree_stack = create_tree_list(new_nonterm, tree_stack);
            } else {
                subtop->children = create_tree_list(top, subtop->children);
                top->parent = subtop;
                top->end = position;
                top->modifier = 1;
                tree_stack = popfree_tree_list(tree_stack);     /* pop modifier */
            }
        } else if (opt->previous_output >= projlabel_to_output(0) &&
                   opt->previous_output <=
                   projlabel_to_output(numlabels - 1)) {
            top = tree_stack->elem;
            /* create projected nonterm */
            new_nonterm = create_tree();
            new_nonterm->start = top->start;
            new_nonterm->site_id = opt->phase;
            strcpy(new_nonterm->label,
                   id_label(output_to_projlabel(opt->previous_output)));
            new_nonterm->children =
                create_tree_list(top, new_nonterm->children);
            /* set new projection site features */
            top->parent = new_nonterm;
            top->end = position + 1;
            /* update tree_stack */
            tree_stack = popfree_tree_list(tree_stack); /* pop projection site */
            /* push projected */
            tree_stack = create_tree_list(new_nonterm, tree_stack);
        }
#if HEADOUTS
        else if (opt->previous_output >= hdprlabel_to_output(0) &&
                 opt->previous_output <=
                 hdprlabel_to_output(numlabels - 1)) {
            top = tree_stack->elem;
            /* create head projected nonterm */
            new_nonterm = create_tree();
            new_nonterm->start = top->start;
            new_nonterm->site_id = opt->phase;
            strcpy(new_nonterm->label,
                   id_label(output_to_hdprlabel(opt->previous_output)));
            new_nonterm->children =
                create_tree_list(top, new_nonterm->children);
            new_nonterm->head_child = new_nonterm->children;
            /* set new projection site features */
            top->parent = new_nonterm;
            top->end = position + 1;
            top->head = 1;
            /* update tree_stack */
            tree_stack = popfree_tree_list(tree_stack); /* pop projection site */
            /* push projected */
            tree_stack = create_tree_list(new_nonterm, tree_stack);
        }
#endif
        /* stack not modified *
           else if (opt->previous_output == PROJOUT || 
           opt->trigger_output == HDPROUT ||
           opt->previous_output == PREDOUT ||
           (opt->previous_output >= predtag_to_output(0) &&
           (numout1 == 0 || 
           opt->previous_output <= predtag_to_output(numout1 - 1))))
           {
           }
         */
    }


    return (tree_stack);
}



/* read one line of tagged word ids for one sentence and set input_hist[]. */ 
int read_idtext(FILE *data)
{
  char buffer[MAXDATALINE];
  int inp1, out1, inp2, out2, em_word_code;
  int bin;
  char *str, *str2, *nextstr, *out2_str;
  char *wordstr;
  struct cluster *dec_cluster;
  int position;
  int i, j;
  int pos;

  position = 0;

  strcpy(sentence_words[position], "^^^");

  if (fgets(buffer, MAXDATALINE, data) != NULL) { 

      wordstr = strtok(buffer, "_ \t\n");
      while (wordstr == NULL)  /* ignore blank lines */
	{
	  if (fgets(buffer, MAXDATALINE, data) != NULL)
	    wordstr = strtok(buffer, "_ \t\n");
	  else
	    return(0);
	}

      while (wordstr != NULL) {

	position++;
	if (position >= sequence_space - 2)     /* -2 to be safe */
	  expand_periods();
	strcpy(sentence_words[position], wordstr);

	str = strtok(NULL, " \t\n");     /* save tag and word spec */
	if (str == NULL)
	  error("Error: bad format in tag-word id input.");
	for (nextstr = str; *nextstr != '\0'; nextstr++) 
	  {}
	nextstr++;  /* save start of next word's spec */
	read_tagword_spec(str, position);  /* sets input_hist[] */
	wordstr = strtok(nextstr, "_ \t\n");  /* restart strtok on input */
      }
  }
  else
    return(0);

  position++;
  if (position >= sequence_space - 2)     /* -2 to be safe */
    expand_periods();
  set_end_input(position);

  return(position);
}


/* read one tag-word specification from str and set input_hist[] for 
   position, and return tag id (out1).  Uses strtok(). */ 
int read_tagword_spec(char *str, int position)
{
  int inp1, out1, inp2, out2, em_word_code;
  int bin;
  char *str2, *nextstr, *out2_str;
  int i, j;
  int pos;

  /* below taken mostly from read_target_derivation */

  /* mark end of spec, and make a copy */
  strcpy(sentence_tagwordids[position], strtok(str, " \t\n"));
      
  inp1 = atoi(strtok(str, ","));  /* use strtok on word spec */
  if (inp1 < 0 || inp1 >= numinp1) {
    sprintf(warning_text,
	    "input tag %d out of range [%d:%-2d]",
	    inp1, num_sent + 1, position);
    warning(warning_text);
  }
  out1 = inp1;
  inp2 = -1;
  out2 = -1;
  out2_str = NULL;
  em_word_code = -1;
  for (pos = 1; (str2 = strtok(NULL, ",")) != NULL;
       pos++) {
    if (pos == inpwrdpos) {
      inp2 = atoi(str2);
    }
    if (pos == outwrdpos) {
      out2_str =
	(char *) malloc(sizeof(char) *
			(strlen(str2) + 1));
      CHECK_PTR(out2_str);
      strcpy(out2_str, str2);
    }
  }

  /* set next period's input */
  input_hist[position].input1 = inp1;

  for (bin = 0; bin < numbins; bin++) {
    if (inp2 >= 0) {
      input_hist[position].input2[bin] =
	inp2 + inp1_first_word[inp1];
      /* @todo: below was bug found by James, fixed but, should be checked (non lexicalized input support) */
    } else if (inpwrdpos == 0) {
      input_hist[position].input2[bin] = inp2;        /* nonlexicalized */
    } else {
      sprintf(warning_text,
	      "input word spec in data doesn't include pos %d", inpwrdpos);
      error(warning_text);
    }
  }
#if ! NOPREDICT
  if (out2_str != NULL) {
    /* relative cluster ids in traverse are divided by  '- ' character */
    fill_input_clusterized(&input_hist[position], out1,
			   out2_str, "-");
  } else if (outwrdpos == 0) {
    /** @todo: nonlexicalized outputs - should be checked */
    input_hist[position].cl_levels_out = 0;

  } else {
    sprintf(warning_text,
	    "output word spec in data doesn't include pos %d", outwrdpos);
    error(warning_text);
  }
#else
  input_hist[position].cl_levels_out = 0;
#endif
  //should it be here?  
  if (out2_str != NULL) {
    free(out2_str);
  }

  return(out1);
}

/* set input_hist[] for '.' */
void set_end_input(int position)
{
  int inp1, out1, inp, inp2, out2, inpwrd, lab, em_word_code;
  int bin;

  strcpy(sentence_words[position], "^^^");
  /* dummy value for hypoth predition, or TAULOOKAHEAD=1 */
  inp1 = tag_id("^^^");
  out1 = inp1;

  input_hist[position].input1 = inp1;
  for (bin = 0; bin < numbins; bin++) {
    if (inpwrdpos >= 0) {
      input_hist[position].input2[bin] =
	inp1_first_word[inp1];
    } else {
      input_hist[position].input2[bin] = -1;  /* nonlexicalized */
    }
  }
  if (outwrdpos >= 0) {
    input_hist[position].output2[0] =
      CLUSTERID_TO_WORDID(out_clusters_arr[out1]->root->
			  childs[0]->cluster_id, out1);

  } else {
    input_hist[position].output2[0] = -1;       /* nonlexicalized */
  }
}


/* read the bracketed short output format for the top predbeam derivations
     for one sentence and set candidate_derivations.  Merge with
     target_derivations if targetflag!=0.  Set input_hist[] if inputflag!=0. 
     Previous_pred_ind always zero.  Returns sentence
     length, or -1 if no more sentences. */
int read_candidate_derivations(FILE * data, int targetflag, int inputflag)
{
    char buffer[MAXDATALINE];
    char *str;
    int out;
    int newsubtop;
    int newtop;
    int newrest;
    int newpreterm;
    struct option *newcont;
    int inp1, inp, inp2, inpwrd, lab;
    int out1;
    int bin;
    int position;
    int pretermflag;
    int i, j;
    int head;
    struct option *opt;
    struct option *opt2;
    struct option *last;
    struct option_stack *ol;
    double pred_lprob;
    double pred_ranklprob;
    int pred_out;
    int pred_inp1, *pred_word;
    double pred_word_lprob;
    int memoffset;
    int length;
    int stopflag;
    int numcand;
    struct option_stack *candidate_stack;
    struct option_stack *target_ol;
    int line;
    struct option_stack *candidate_ends;
    struct cluster *dec_cluster;
    int idx;
    int child_id;
    struct option *trig_parent;

    stopflag = 0;
    numcand = 0;
    candidate_stack = NULL;
    candidate_ends = NULL;

    /* root is phase 0, and already allocated */
    line = 0;
    position = 0;
    /**JH: setting of target_ol fixed */
    if (targetflag != 0) {
        candidate_derivations = target_derivation->elem;
        /* find PREDOUT option */
        target_ol = target_derivation;
        while (target_ol != target_derivation_next &&
               target_ol->elem->period == 0 &&
               target_ol->elem->previous_output != PREDOUT) {
            target_ol = target_ol->next;
        }
        opt = target_ol->elem;
        target_ol = target_ol->next;
    } else {
        opt =
            create_option(position, 0, 1.0, 1.0, 0, PREDOUT, NULL,
                          label_id(""), ROOTFLAG, -2, -1, MAXPREDBEAM,
                          NULL, NULL, NULL, NULL, 1, 1, 0, 1);
        candidate_derivations = opt;
    }

    if (inputflag) {
        strcpy(sentence_words[position], "^^^");
    }

    last = opt;
    while (stopflag == 0 && fgets(buffer, MAXDATALINE, data) != NULL) {
        line++;

        str = first_token(buffer);
        if (str == NULL)        /* ignore blank lines */
            continue;

        /* loop until next prediction / end of line */
        while (1) {             /* enter this loop with str set */

            if (str[0] == '{') {        /* start of sub-derivation */
                candidate_stack =
                    create_option_stack(last, candidate_stack, NULL);
            }

            else if (str[0] == ';') {   /* end and start of sub-derivation */
                if (last != NULL) {
                    candidate_ends =
                        mergesort_options(candidate_ends,
                                          create_option_cons(last));
                    numcand++;
                }
                last = candidate_stack->elem;
                break;          /* exit inner loop */
            }

            else if (str[0] == '}') {   /* end of sub-derivation */
                if (last != NULL) {
                    candidate_ends =
                        mergesort_options(candidate_ends,
                                          create_option_cons(last));
                    numcand++;
                }
                last = NULL;
                candidate_stack = popfree_option_stack(candidate_stack);
            }

            else if (str[0] == '=') {   /* probability of last derivation */
                /* just for sorting, reset later */
                /** ivan:  07/07/04 fixed bug introduced in dglc3bj2 - write as log P, read as P */
                last->ranking_lprob = atof(str + 1);
            }

            else if (last != NULL && last->previous_output == PROJOUT) {
                position = last->period;
                lab = atoi(str);
                if (lab < 0 || lab >= numlabels) {
                    sprintf(warning_text,
                            "input label %d out of range %d:%-2d",
                            lab, num_sent + 1, position);
                    warning(warning_text);
                }
                /* projection label chosen, pop and push nonterm */
#if HEADOUTS
                if (head == 0) {
#endif
                    out = projlabel_to_output(lab);
#if HEADOUTS
                } else {
                    out = hdprlabel_to_output(lab);
                }
#endif
                if (targetflag != 0 && last == target_ol->prev->elem &&
                    out == target_ol->elem->previous_output) {
                    opt = target_ol->elem;
                    target_ol = target_ol->next;
                } else {
                    opt =
                        create_option(position, -1, 0.0, 0.0, 0, out, last,
                                      lab, last->previous_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind, NULL,
                                      last->trigparent_option, last, NULL,
                                      1, 1, 0, 0);
                }
                last = opt;
            }

            else if (str[0] == '>') {   /* prediction, loop exit */
                position = last->period;
                if (position != 0) {    /* otherwise done above */
                    if (targetflag != 0 && last == target_ol->prev->elem &&
                        PREDOUT == target_ol->elem->previous_output) {
                        opt = target_ol->elem;
                        target_ol = target_ol->next;
                    } else {
                        opt =
                            create_option(position, last->phase, 0.0, 0.0,
                                          last->preterm, PREDOUT, last,
                                          last->trigger_label,
                                          last->trigger_output,
                                          last->stack_subtop,
                                          last->stack_rest_rel,
                                          last->prev_pred_ind,
                                          last->context_option,
                                          last->trigparent_option,
                                          last->trigchild_option,
                                          last->headchild_option, 0, 0, 0,
                                          1);
                    }
                    last = opt;
                }

                position++;
                if (position >= sequence_space - 2)     /* -2 to be safe */
                    expand_periods();

                str = next_token();     /* save tag and word spec */
		if (inputflag) {
		    strcpy(sentence_words[position], next_token());
		    /* now input line should be finished, so can reuse strtok*/
		    if (numcand == 0) {     /* otherwise redundant */
		        /* set input_hist[] */
		        out1 = read_tagword_spec(str, position); 
		    }
		    else {
		        out1 = input_hist[position].input1;
		    }
		}
		else {
		    out1 = input_hist[position].input1;
		    if (atoi(strtok(str, ", \t\n")) != out1)
		      warning("incompatible target and derivation input.");
		}

                pred_out = predtag_to_output(out1);
#if ! NOPREDICT
                dec_cluster = out_clusters_arr[out1]->root;
#else
                dec_cluster = NULL;
#endif
                trig_parent = last;

            /** creating the internal prediction nodes */
                for (idx = 0; idx < input_hist[position].cl_levels_out;
                     idx++) {

                    if (targetflag != 0 && last == target_ol->prev->elem &&
                        pred_out == target_ol->elem->previous_output) {
                        opt = target_ol->elem;
                        target_ol = target_ol->next;
                    } else {
                        opt =
                            create_option(position - 1, last->phase, 0.0,
                                          0.0, last->preterm, pred_out,
                                          last, last->trigger_label,
                                          last->trigger_output,
                                          last->stack_subtop,
                                          last->stack_rest_rel,
                                          last->prev_pred_ind,
                                          last->context_option,
                                          last->trigparent_option,
                                          last->trigchild_option,
                                          last->headchild_option, 0, 0, 0,
                                          1);

                        opt->dec_cluster = dec_cluster;

                        opt->memoffset = out1_first_word[out1] + dec_cluster->childs[0]->cluster_id;
                    }
                    child_id = WORDID_TO_CHILDID(input_hist[position].output2[idx], out1, 
                            dec_cluster->childs[0]->cluster_id);
                    dec_cluster = dec_cluster->childs[child_id];
                    
                    last = opt;

                    pred_out =  CHILDID_TO_MEMOUT(child_id);
                }


                /* tag/word prediction, push preterm */
                if (last->preterm != 0) {
                    error("can no longer predict from preterms");
                }

                if (targetflag != 0 && last == target_ol->prev->elem &&
                    pred_out == target_ol->elem->previous_output) {
                    opt = target_ol->elem;
                    target_ol = target_ol->next;
                } else {

                    opt = create_option(position, -1, 0.0, 0.0,
                                        1, pred_out, last,
                                        predtag_to_output(out1) -
                                        predtag_to_output(0) + numlabels,
                                        trig_parent->previous_output,
                                        last->stack_subtop,
                                        last->stack_rest_rel,
                                        last->prev_pred_ind, NULL,
                                        trig_parent, NULL, NULL, 1, 1, 0,
                                        0);
                }

                last = opt;



                break;          /* exit inner loop */

            }

            else if (str[0] == '|') {   /* projection */
                position = last->period;
#if HEADOUTS
                if (str[1] == 'h')
                    head = 1;
                else
                    head = 0;
#endif
                /* for PROJOUT/HDPROUT, delay pop and pushing nonterm until
                   computing labels */
#if HEADOUTS
                if (head == 0) {
#endif
                    if (targetflag != 0 && last == target_ol->prev->elem &&
                        PROJOUT == target_ol->elem->previous_output) {
                        opt = target_ol->elem;
                        target_ol = target_ol->next;
                    } else {
                        opt =
                            create_option(position, last->phase, 0.0, 0.0,
                                          last->preterm, PROJOUT, last,
                                          last->trigger_label,
                                          last->trigger_output,
                                          last->stack_subtop,
                                          last->stack_rest_rel,
                                          last->prev_pred_ind,
                                          last->context_option,
                                          last->trigparent_option,
                                          last->trigchild_option,
                                          last->headchild_option, 0, 0, 0,
                                          0);
                    }
#if HEADOUTS
                } else {
                    if (targetflag != 0 && last == target_ol->prev->elem &&
                        HDPROUT == target_ol->elem->previous_output) {
                        opt = target_ol->elem;
                        target_ol = target_ol->next;
                    } else {
                        opt =
                            create_option(current, last->phase, 0.0, 0.0,
                                          last->preterm, HDPROUT, last,
                                          last->trigger_label,
                                          last->trigger_output,
                                          last->stack_subtop,
                                          last->stack_rest_rel,
                                          last->prev_pred_ind,
                                          last->context_option,
                                          last->trigparent_option,
                                          last->trigchild_option,
                                          last->headchild_option, 0, 0, 0,
                                          0);
                    }
                }
#endif
                last = opt;
            }

            else if (str[0] == '^') {   /* attachment */
                position = last->period;
#if HEADOUTS
                if (str[1] == 'h')
                    head = 1;
                else
                    head = 0;
#endif
                /* for ATTOUT/HDATOUT, pop */
                newcont = last->trigparent_option;
                newtop = newcont->phase;
                newrest = last->stack_rest_rel;
#if HEADOUTS
                if (head == 0) {
#endif
                    if (targetflag != 0 && last == target_ol->prev->elem &&
                        ATTOUT == target_ol->elem->previous_output) {
                        opt = target_ol->elem;
                        target_ol = target_ol->next;
                    } else {
                        opt =
                            create_option(position, newtop, 0.0, 0.0,
                                          newcont->preterm, ATTOUT, last,
                                          newcont->trigger_label,
                                          newcont->trigger_output,
                                          -2, newrest,
                                          last->prev_pred_ind, newcont,
                                          newcont->trigparent_option,
                                          newcont->trigchild_option,
                                          newcont->headchild_option, 0, 1,
                                          0, 0);
                    }
#if HEADOUTS
                } else {
                    if (targetflag != 0 && last == target_ol->prev->elem &&
                        HDATOUT == target_ol->elem->previous_output) {
                        opt = target_ol->elem;
                        target_ol = target_ol->next;
                    } else {
                        opt =
                            create_option(position, newtop, 0.0, 0.0,
                                          newcont->preterm, HDATOUT, last,
                                          newcont->trigger_label,
                                          newcont->trigger_output,
                                          -2, newrest,
                                          last->prev_pred_ind, newcont,
                                          newcont->trigparent_option,
                                          newcont->trigchild_option,
                                          last, 0, 1, 0, 0);
                    }
                }
#endif
                last = opt;
            }

            else if (str[0] == '+') {   /* modification */
                position = last->period;
                newcont = last->trigparent_option;
                newtop = newcont->phase;
                newrest = last->stack_rest_rel;
                if (targetflag != 0 && last == target_ol->prev->elem &&
                    MODOUT == target_ol->elem->previous_output) {
                    opt = target_ol->elem;
                    target_ol = target_ol->next;
                } else {
                    opt =
                        create_option(position, newtop, 0.0, 0.0,
                                      newcont->preterm, MODOUT, last,
                                      newcont->trigger_label,
                                      newcont->trigger_output,
                                      -2, newrest,
                                      last->prev_pred_ind, newcont,
                                      newcont->trigparent_option,
                                      newcont->trigchild_option,
                                      newcont->headchild_option, 0, 1, 0,
                                      0);
                }
                last = opt;
            }

            else if (str[0] == '.') {   /* end of sentence, loop exit */
                if (candidate_ends != NULL)
                    position = 1 + candidate_ends->elem->period;
                else
                    position = 1;
                if (position >= sequence_space - 2)     /* -2 to be safe */
                    expand_periods();
		if (inputflag) {
		    set_end_input(position);
		    sentence_length = position;
		}

                stopflag = 1;   /* stopflag == 1 finishes outer loop */
                break;          /* exit inner loop */

            }

            else {
                if (candidate_ends != NULL)
                    position = 1 + candidate_ends->elem->period;
                else
                    position = 1;
                sprintf(warning_text,
                        "bad symbol in line: %s  %d:%-2d",
                        buffer, num_sent + 1, position);
                warning(warning_text);
            }

            str = next_token();
        }
    }

    if (candidate_ends != NULL)
        position = 1 + candidate_ends->elem->period;
    /* else hope already set */


    
    i = 0;
    while (candidate_ends != NULL) {
        if (i < candidate_predbeam) {
            if (targetflag && candidate_ends->elem->in_target != 0) {
                i--;    /* don't count target */

            }
        } else {
            /** @todo: ivan: shouldn't the condition be like that? Why not to free options when targetflag == 0?: 
             * if ((targetflag != 0 && candidate_ends->elem->in_target == 0) || targetflag == 0) { */
            if (targetflag != 0 && candidate_ends->elem->in_target == 0) {
                free_option_branch(candidate_ends->elem);
            }
        }
        candidate_ends = popfree_option_stack(candidate_ends);
        i++;
    }

    
#if DEBUGCAND2
    printf("candidate derivation options:\n");
    display_option_tree(candidate_derivations);
    printf("\n\n");
    check_option_tree(candidate_derivations);
#endif
    if (stopflag == 0) {
        if (position == 0) {
            if (candidate_derivations != NULL)
                free_option_branch(candidate_derivations);
            candidate_derivations = NULL;
            return (-1);        /* no sentence read */
        } else {
            sprintf(warning_text,
                    "not sufficient input in candidates: %s  %d:%-2d",
                    buffer, num_sent, position);
            warning(warning_text);
        }
    }

    return (position);
}



/* read the short output format for one sentence and set target_derivation.
     If no more sentences then return NULL.  Also sets input_hist[].  Returns
     sentence length, or -1 if no more sentences.*/
/* prev_pred_ind always MAXPREDBEAM */
int read_target_derivation(FILE * data)
{
    char buffer[MAXDATALINE];
    char *str;
    char *str2;
    int pos;
    int out;
    int newsubtop;
    int newtop;
    int newrest;
    int newpreterm;
    struct option *newcont;
    int inp1, out1, inp, inp2, out2, inpwrd, lab, em_word_code;
    int bin;
    int position;
    int pretermflag;
    int i, j;
    int head;
    struct option *opt;
    struct option *opt2;
    struct option *last;
    struct option *trig_parent;
    struct option_stack *ol;
    double pred_lprob;
    double pred_ranklprob;
    int pred_out;
    int pred_out1, *pred_word;
    double pred_word_lprob;
    int memoffset;
    int length;
    int stopflag;
    int idx;
    int prev_output_mem;
    int child_id;

    struct cluster *dec_cluster;
    char *out2_str;

    stopflag = 0;
    position = 0;

    target_derivation_next = target_derivation;
    /* root is phase 0, and already allocated */
    target_derivation_next->elem =
        create_option(position, 0, 1.0, 1.0, 0, PREDOUT, NULL,
                      label_id(""), ROOTFLAG, -2, -1, MAXPREDBEAM, NULL,
                      NULL, NULL, NULL, 1, 1, 0, 1);
    target_derivation_next =
        incr_option_stack_next(target_derivation_next);

    last = target_derivation_next->prev->elem;
    /* init_option_phases(last);  * not needed */
    strcpy(sentence_words[position], "^^^");

    while (stopflag == 0 && fgets(buffer, MAXDATALINE, data) != NULL) {


        str = first_token(buffer);
        if (str == NULL)        /* ignore blank lines */
            continue;

        /* first period of sentence starts with '>', 
           otherwise can't have two '>'s in a row */
        if (position != 0 && str[0] == '>') {
            sprintf(warning_text,
                    "shouldn't predict from a preterminal. %d:%-2d",
                    num_sent + 1, position);
            warning(warning_text);
        }

        /* loop until next prediction / end of line */
        while (1) {             /* enter this loop with str set */

            if (str[0] == '>') {        /* prediction, loop exit */

                if (position != 0) {    /* otherwise done above */
                    target_derivation_next->elem =
                        create_option(position, last->phase, 0.0, 0.0,
                                      last->preterm, PREDOUT, last,
                                      last->trigger_label,
                                      last->trigger_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind,
                                      last->context_option,
                                      last->trigparent_option,
                                      last->trigchild_option,
                                      last->headchild_option, 0, 0, 0, 1);
                    target_derivation_next =
                        incr_option_stack_next(target_derivation_next);

                    last = target_derivation_next->prev->elem;
                    init_option_phases(last);   /* I don't think this does anything */
                }

                trig_parent = target_derivation_next->prev->elem;

                position++;
                if (position >= sequence_space - 2)     /* -2 to be safe */
                    expand_periods();

                str = next_token();     /* save tag and word spec */
                strcpy(sentence_words[position], next_token());
                /* now input line should be finished, so can reuse strtok */
		out1 = read_tagword_spec(str, position); /* sets input_hist[]*/

                prev_output_mem = predtag_to_output(out1);
#if ! NOPREDICT
                dec_cluster = out_clusters_arr[out1]->root;
#else
                dec_cluster = NULL;
#endif


            /** creating the internal prediction nodes */
                for (idx = 0;
                     idx < input_hist[position].cl_levels_out
                     && idx < max_cl_levels; idx++) {

                    target_derivation_next->elem =
                        create_option(position - 1, last->phase, 0.0, 0.0,
                                      last->preterm, prev_output_mem, last,
                                      last->trigger_label,
                                      last->trigger_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind,
                                      last->context_option,
                                      last->trigparent_option,
                                      last->trigchild_option,
                                      last->headchild_option, 0, 0, 0, 1);

                    target_derivation_next->elem->dec_cluster =
                        dec_cluster;
                    target_derivation_next->elem->memoffset =
                        out1_first_word[out1] +
                        dec_cluster->childs[0]->cluster_id;

                    child_id =
                        WORDID_TO_CHILDID(input_hist[position].
                                          output2[idx], out1,
                                          dec_cluster->childs[0]->
                                          cluster_id);
                    dec_cluster = dec_cluster->childs[child_id];

                    target_derivation_next =
                        incr_option_stack_next(target_derivation_next);
                    last = target_derivation_next->prev->elem;
                    init_option_phases(last);

                    prev_output_mem = CHILDID_TO_MEMOUT(child_id);
                }


                /* tag/word prediction, push preterm */
                if (last->preterm != 0) {
                    error("can no longer predict from preterms");
                }
                target_derivation_next->elem =
                    create_option(position, -1, 0.0, 0.0,
                                  1, prev_output_mem, last,
                                  predtag_to_output(out1) -
                                  predtag_to_output(0) + numlabels,
                                  trig_parent->previous_output,
                                  last->stack_subtop, last->stack_rest_rel,
                                  last->prev_pred_ind, NULL, trig_parent,
                                  NULL, NULL, 1, 1, 0, 0);
                target_derivation_next =
                    incr_option_stack_next(target_derivation_next);

                last = target_derivation_next->prev->elem;
                init_option_phases(last);

                break;          /* exit inner loop */

            } else if (str[0] == '|') { /* projection */
#if HEADOUTS
                if (str[1] == 'h')
                    head = 1;
                else
                    head = 0;
#endif
                str = next_token();
                lab = atoi(str);
                if (lab < 0 || lab >= numlabels) {
                    sprintf(warning_text,
                            "input label %d out of range %d:%-2d",
                            lab, num_sent + 1, position);
                    warning(warning_text);
                }

                /* for PROJOUT/HDPROUT, delay pop and pushing nonterm until
                   computing labels */
#if HEADOUTS/*{{{ */
                if (head == 0) {
#endif/*}}} */
                    target_derivation_next->elem =
                        create_option(position, last->phase, 0.0, 0.0,
                                      last->preterm, PROJOUT, last,
                                      last->trigger_label,
                                      last->trigger_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind,
                                      last->context_option,
                                      last->trigparent_option,
                                      last->trigchild_option,
                                      last->headchild_option, 0, 0, 0, 0);
#if HEADOUTS/*{{{ */
                } else {
                    target_derivation_next->elem =
                        create_option(current, last->phase, 0.0, 0.0,
                                      last->preterm, HDPROUT, last,
                                      last->trigger_label,
                                      last->trigger_output,
                                      last->stack_subtop,
                                      last->stack_rest_rel,
                                      last->prev_pred_ind,
                                      last->context_option,
                                      last->trigparent_option,
                                      last->trigchild_option,
                                      last->headchild_option, 0, 0, 0, 0);
                }
#endif/*}}} */
                target_derivation_next =
                    incr_option_stack_next(target_derivation_next);

                last = target_derivation_next->prev->elem;
                init_option_phases(last);

                /* projection label chosen, pop and push nonterm */
#if HEADOUTS/*{{{ */
                if (head == 0) {
#endif/*}}} */
                    out = projlabel_to_output(lab);
#if HEADOUTS/*{{{ */
                } else {
                    out = hdprlabel_to_output(lab);
                }
#endif/*}}} */
                target_derivation_next->elem =
                    create_option(position, -1, 0.0, 0.0, 0, out, last,
                                  lab, last->previous_output,
                                  last->stack_subtop, last->stack_rest_rel,
                                  last->prev_pred_ind, NULL,
                                  last->trigparent_option, last, NULL, 1,
                                  1, 0, 0);
                target_derivation_next =
                    incr_option_stack_next(target_derivation_next);

                last = target_derivation_next->prev->elem;
                init_option_phases(last);
            } else if (str[0] == '^') { /* attachment */
#if HEADOUTS
                if (str[1] == 'h')
                    head = 1;
                else
                    head = 0;
#endif
                /* for ATTOUT/HDATOUT, pop */
                newcont = last->trigparent_option;
                newtop = newcont->phase;
                newrest = last->stack_rest_rel;
#if HEADOUTS
                if (head == 0) {
#endif
                    target_derivation_next->elem =
                        create_option(position, newtop, 0.0, 0.0,
                                      newcont->preterm, ATTOUT, last,
                                      newcont->trigger_label,
                                      newcont->trigger_output,
                                      -2, newrest,
                                      last->prev_pred_ind, newcont,
                                      newcont->trigparent_option,
                                      newcont->trigchild_option,
                                      newcont->headchild_option, 0, 1, 0,
                                      0);
#if HEADOUTS
                } else {
                    target_derivation_next->elem =
                        create_option(position, newtop, 0.0, 0.0,
                                      newcont->preterm, HDATOUT, last,
                                      newcont->trigger_label,
                                      newcont->trigger_output,
                                      -2, newrest,
                                      last->prev_pred_ind, newcont,
                                      newcont->trigparent_option,
                                      newcont->trigchild_option,
                                      last, 0, 1, 0, 0);
                }
#endif
                target_derivation_next =
                    incr_option_stack_next(target_derivation_next);

                last = target_derivation_next->prev->elem;
                init_option_phases(last);
            } else if (str[0] == '+') { /* modification */
                newcont = last->trigparent_option;
                newtop = newcont->phase;
                newrest = last->stack_rest_rel;
                target_derivation_next->elem =
                    create_option(position, newtop, 0.0, 0.0,
                                  newcont->preterm, MODOUT, last,
                                  newcont->trigger_label,
                                  newcont->trigger_output,
                                  -2, newrest,
                                  last->prev_pred_ind, newcont,
                                  newcont->trigparent_option,
                                  newcont->trigchild_option,
                                  newcont->headchild_option, 0, 1, 0, 0);
                target_derivation_next =
                    incr_option_stack_next(target_derivation_next);

                last = target_derivation_next->prev->elem;
                init_option_phases(last);
            } else if (str[0] == '.') { /* end of sentence, loop exit */
                position++;
                if (position >= sequence_space - 2)     /* -2 to be safe */
                    expand_periods();
		set_end_input(position);
                stopflag = 1;   /* stopflag == 1 finishes outer loop */
                break;          /* exit inner loop */

            } else {
                sprintf(warning_text,
                        "bad symbol in line: %s  %d:%-2d",
                        buffer, num_sent + 1, position);
                warning(warning_text);
            }
            str = next_token();
        }
    }
    /* set in_target flag */
    for (ol = target_derivation; ol != target_derivation_next;
         ol = ol->next) {
        ol->elem->in_target = 1;
    }

    if (stopflag == 0) {
        if (position == 0) {
            return (-1);        /* no sentence read */
        } else {
            sprintf(warning_text,
                    "not sufficient input: %s  %d:%-2d",
                    buffer, num_sent, position);
            warning(warning_text);
        }
    }

    sentence_length = last->period + 1;
    return (sentence_length);
}


/* compute all the network activations and probabilities for options.
     Assumes options list is in derivation order.   Does not compute the state
     for the last option.  */
void comp_option_list_act(struct option_stack *options,
                          struct option_stack *options_end)
{
    struct option_stack *ol;
    struct option *prevopt, *opt;
    int out, last_out_mem, first_out_mem;
    int memoffset;
    int pred_out1;
    int bin;
    double just_predlprob;
    double pred_word_lprob;
    int pred_out2;
    int i;
    struct cluster *dec_cluster;
    int dec_cluster_id;
    int out1;

    i = 0;
    for (ol = options; ol != options_end; ol = ol->next) {

        opt = ol->elem;
        prevopt = opt->previous_option;

        i++;
        if (prevopt == NULL) {
            opt->lprob = log(1.0);
            opt->ranking_lprob = log(1.0);
            opt->new_state = 1;
            comp_option_states(opt);    /* copies network states from init pattern */
        } else {
            out = opt->previous_output;


//EM_CLUSTERING

            if (out >= PREPREDFIRST && out <= PREPREDLAST) {
            /*** following seems to help performance, but shouldn't be here *
            if (opt->phase == 0 && opt->period == sentence_length) {
              opt->lprob = prevopt->lprob;
              opt->ranking_lprob = prevopt->ranking_lprob;
              opt->new_state = 0;
            }
            else {    
            ***/
                comp_output_act_range(prevopt, PREPREDFIRST, PREPREDLAST);
                opt->lprob =
                    prevopt->lprob + log(prevopt->outputs->act[out]);
                opt->ranking_lprob =
                    prevopt->ranking_lprob +
                    log(prevopt->outputs->act[out]);
                //here was bug in condition opt->period ==  sentence_length (fixed in dglc3e) 
                if (opt->phase == 0 && opt->period == (sentence_length - 1)) {
                    opt->new_state = 0; /* last option in derivatoin */
                } else {
                    comp_option_states(opt);    /* compute network states where needed */
                }
            } else if (out >= predtag_to_output(0) &&
                       (numout1 == 0 ||
                        out <= predtag_to_output(numout1 - 1))) {

                pred_out1 = input_hist[prevopt->period + 1].input1;

#if NOPREDICT
            /** should completely remove this option **/
                opt->lprob = prevopt->lprob;
                opt->ranking_lprob = prevopt->ranking_lprob;
#else
                comp_output_act_range(prevopt, predtag_to_output(0),
                                      predtag_to_output(numout1 - 1));

                opt->lprob =
                    prevopt->lprob + log(prevopt->outputs->act[out]);
                opt->ranking_lprob = opt->lprob;
#endif
                comp_option_states(opt);

                /* if prevopt->is_pred_opt and not tag prediction (previous case)  */
            } else if (prevopt->is_pred_opt) {


#if NOPREDICT
            /** should completely remove this option **/
                opt->lprob = prevopt->lprob;
                opt->ranking_lprob = opt->lprob;
#else
                first_out_mem = CHILDID_TO_MEMOUT(0);
                last_out_mem =
                    CHILDID_TO_MEMOUT(prevopt->dec_cluster->child_qty - 1);

                comp_output_act_range(prevopt, first_out_mem,
                                      last_out_mem);

                pred_out2 =
                    input_hist[prevopt->period +
                               1].output2[prevopt->dec_cluster->layer_num];
                opt->lprob =
                    prevopt->lprob +
                    log(prevopt->outputs->
                        act[WORDID_TO_MEMOUT
                            (pred_out2, prevopt->memoffset)]);
                opt->ranking_lprob = opt->lprob;

#endif

                comp_option_states(opt);

            } else if (out >= projlabel_to_output(0) &&
                       out <= projlabel_to_output(numlabels - 1)) {
                comp_output_act_range(prevopt, projlabel_to_output(0),
                                      projlabel_to_output(numlabels - 1));
                opt->lprob =
                    prevopt->lprob + log(prevopt->outputs->act[out]);
                opt->ranking_lprob =
                    prevopt->ranking_lprob +
                    log(prevopt->outputs->act[out]);
                comp_option_states(opt);        /* compute network states where needed */
            }
#if HEADOUTS
            else if (out >= hdprlabel_to_output(0) &&
                     out <= hdprlabel_to_output(numlabels - 1)) {
                comp_output_act_range(prevopt, hdprlabel_to_output(0),
                                      hdprlabel_to_output(numlabels - 1),
                                      0);
                opt->lprob =
                    prevopt->lprob + log(prevopt->outputs->act[out]);
                opt->ranking_lprob =
                    prevopt->ranking_lprob +
                    log(prevopt->outputs->act[out]);
                comp_option_states(opt);        /* compute network states where needed */
            }
#endif
            else
                error("bad output value in target derivation.");
        }
    }
}




/* read the short output format for one sentence into a tree.  If no more
     sentences then return NULL.  If format_flag != 0 then reintroduce
     chomsky adjunction for modifications and conflated nodes.  Also sets
     input_hist[]. */
/** maybe this can be replaced by read_target_derivation() and
      perform_derivation() and freeing.  **/
struct tree *read_target(FILE * data, int format_flag)
{
    struct tree *res;
    char buffer[MAXDATALINE];
    char *str;
    char *str2;
    char *out2_str;
    int pos;
    int inp1, out1, inp, inp2, out2, inpwrd, lab;
    int bin;
    int position;
    struct tree_list *tree_stack;
    struct tree_list *tl;
    struct tree *new_preterm;
    struct tree *new_nonterm;
    struct tree *top;
    struct tree *subtop;
    int em_word_code;
    int i, j;
    int head;

    position = 0;
    next_site_id = 0;

    tree_stack = start_root(position, next_site_id, NULL);
    res = tree_stack->elem;
    next_site_id++;
    position++;

    while (tree_stack != NULL && fgets(buffer, MAXDATALINE, data) != NULL) {


        str = first_token(buffer);
        if (str == NULL)        /* ignore blank lines */
            continue;



        /* first period of sentence starts with '>', 
           otherwise can't have two '>'s in a row */
        if (position != 1 && str[0] == '>') {
            sprintf(warning_text,
                    "shouldn't predict from a preterminal. %d:%-2d",
                    num_sent + 1, position);
            warning(warning_text);
        }


        while (1) {             /* enter this loop with str set */
            if (str[0] == '>') {        /* prediction, loop exit */
                new_preterm = create_tree();
                new_preterm->start = position;
                new_preterm->site_id = next_site_id;
                next_site_id++;
                str = next_token();     /* save tag and word spec */
                strcpy(sentence_words[position], next_token());
                /* now input line should be finished, so can reuse strtok */
		out1 = read_tagword_spec(str, position); /* sets input_hist[]*/
                strcpy(new_preterm->label, id_tag(out1));
                /* push preterm */
                tree_stack = create_tree_list(new_preterm, tree_stack);
                position++;
                if (position >= sequence_space - 2)     /* -2 to be safe */
                    expand_periods();

                break;          /* exit inner loop */

            } else if (str[0] == '|') { /* projection */
#if HEADOUTS
                if (str[1] == 'h')
                    head = 1;
                else
                    head = 0;
#endif
                str = next_token();
                lab = atoi(str);
                if (lab < 0 || lab >= numlabels) {
                    sprintf(warning_text,
                            "input label %d out of range %d:%-2d",
                            lab, num_sent + 1, position);
                    warning(warning_text);
                }
                top = tree_stack->elem;

                /* create projected nonterm */
                new_nonterm = create_tree();
                new_nonterm->start = top->start;
                new_nonterm->site_id = next_site_id;
                next_site_id++;
                strcpy(new_nonterm->label, id_label(lab));
                new_nonterm->children =
                    create_tree_list(top, new_nonterm->children);
#if HEADOUTS
                if (head != 0) {
                    new_nonterm->head_child = new_nonterm->children;
                    top->head = head;
                }
#endif

                /* set new projection site features */
                top->parent = new_nonterm;
                top->end = position;

                /* update tree_stack */
                tree_stack->elem = new_nonterm; /* pop and push */
            } else if (str[0] == '^') { /* attachment */
#if HEADOUTS
                if (str[1] == 'h')
                    head = 1;
                else
                    head = 0;
#endif
                top = tree_stack->elem;
                subtop = tree_stack->next->elem;
                subtop->children = create_tree_list(top, subtop->children);
#if HEADOUTS
                if (head != 0) {
                    subtop->head_child = subtop->children;
                    top->head = head;
                }
#endif
                top->parent = subtop;
                top->end = position;
                /* pop attached nonterm */
                tree_stack = popfree_tree_list(tree_stack);
            } else if (str[0] == '+') { /* modification */
                /* a modifier can't be a head */
                top = tree_stack->elem;
                subtop = tree_stack->next->elem;
                if (format_flag != 0) {
                    /* create Chomsky adjunction nonterm */
                    new_nonterm = create_tree();
                    new_nonterm->start = subtop->start;
                    new_nonterm->site_id = subtop->site_id;
                    /* set and reset labels */
                    for (i = 0;
                         subtop->label[i] != '\0' && i < MAXLABELLENGTH;
                         i++)
                        new_nonterm->label[i] = subtop->label[i];
                    new_nonterm->label[i] = '\0';
                    while (subtop->label[i] != '-' && i > 0)
                        i--;
                    if (i > 0) {        /* relabel site with last conflated label */
                        for (j = 0;
                             subtop->label[i + j + 1] != '\0'
                             && j < MAXLABELLENGTH; j++)
                            subtop->label[j] = subtop->label[i + j + 1];
                        subtop->label[j] = '\0';
                    }
                    /* do Chomsky adjunction */
                    new_nonterm->children =
                        create_tree_list(subtop, new_nonterm->children);
                    subtop->parent = new_nonterm;
                    subtop->end = top->start;
                    new_nonterm->children =
                        create_tree_list(top, new_nonterm->children);
                    top->parent = new_nonterm;
                    top->end = position;
                    /* update tree_stack */
                    /* pop modifier */
                    tree_stack = popfree_tree_list(tree_stack);
                    /* pop modification site */
                    tree_stack = popfree_tree_list(tree_stack);
                    /* push modification */
                    tree_stack = create_tree_list(new_nonterm, tree_stack);
                } else {
                    subtop->children =
                        create_tree_list(top, subtop->children);
                    top->parent = subtop;
                    top->end = position;
                    top->modifier = 1;
                    /* pop modifier */
                    tree_stack = popfree_tree_list(tree_stack);
                }
            } else if (str[0] == '.') { /* end of sentence, loop exit */
	        set_end_input(position);
                tree_stack->elem->end = position;
                if (tree_stack->elem != res) {
                    sprintf(warning_text,
                            "tree_stack not empty at end of sentence %d:%-2d",
                            num_sent + 1, position);
                    warning(warning_text);
                }
                while (tree_stack != NULL) {
                    /* empty tree_stack */
                    tree_stack = popfree_tree_list(tree_stack);
                }

                break;          /* exit inner loop */

                /* empty tree_stack finishes outer loop */

            } else {
                sprintf(warning_text,
                        "bad symbol in line: %s  %d:%-2d",
                        buffer, num_sent + 1, position);
                warning(warning_text);
            }
            str = next_token();
        }
    }

    if (tree_stack != NULL) {
        if (position == 1) {
            return (NULL);      /* no sentence read */
        } else {
            sprintf(warning_text,
                    "not sufficient input: %s  %d:%-2d",
                    buffer, num_sent, position);
            warning(warning_text);
        }
    }

    if (format_flag) {          /* expand conflations */
        expand_conflations(res);
    }

    sentence_length = res->end;

    return (res);
}


/* create a start preterm and add a root tree to stack with start
     position and site_id id, and add root to the children of the
     previous tree in stack, if one exists */
struct tree_list *start_root(int position, int id, struct tree_list *stack)
{
    struct tree_list *res;
    struct tree *newtree;
    struct tree_list *newchildren;
    struct tree *preterm;

    if (id != 0)
        warning("Root id should be 0 in start_root().");

    newtree = create_tree();
    newtree->start = position;
    newtree->site_id = id;

    preterm = create_tree();
    preterm->start = position;
    strcpy(preterm->label, "^^^");
    preterm->end = position + 1;

    preterm->parent = newtree;
    newchildren = create_tree_list(preterm, NULL);
    newtree->children = newchildren;

    if (stack != NULL) {        /* add tree to parent's children (in reverese order) */
        newtree->parent = stack->elem;
        newchildren = create_tree_list(newtree, stack->elem->children);
        stack->elem->children = newchildren;
    }

    res = create_tree_list(newtree, stack);
    return (res);
}


/* replace nodes labeled with a conflation by multiple nodes, one per
     conflated label. */
void expand_conflations(struct tree *root)
{
    struct tree *new_nonterm;
    struct tree_list *tl;
    int i, j;

    for (i = 0; root->label[i] != '\0' && i < MAXLABELLENGTH; i++)
        if (root->label[i] == '-' && i != 0) {
            root->label[i] = '\0';
            new_nonterm = create_tree();
            new_nonterm->start = root->start;
            new_nonterm->end = root->end;
            new_nonterm->site_id = root->site_id;
            for (j = 0;
                 root->label[j + i + 1] != '\0' && j < MAXLABELLENGTH; j++)
                new_nonterm->label[j] = root->label[j + i + 1];
            new_nonterm->label[j] = '\0';
            new_nonterm->children = root->children;
            for (tl = new_nonterm->children; tl != NULL; tl = tl->next) {
                tl->elem->parent = new_nonterm;
            }
            root->children = create_tree_list(new_nonterm, NULL);
            new_nonterm->parent = root;
            break;
        }

    for (tl = root->children; tl != NULL; tl = tl->next) {
        expand_conflations(tl->elem);
    }
}


/* allocate an option.  */
struct option *create_option(int period, int phase, double lprob,
                             double ranking_lprob, int preterm,
                             int previous_output,
                             struct option *previous_option,
                             int trigger_label, int trigger_output,
                             int stack_subtop, int stack_rest_rel,
                             int prev_pred_ind,
                             struct option *context_option,
                             struct option *trigparent_option,
                             struct option *trigchild_option,
                             struct option *headchild_option,
                             int new_phase, int new_state, int new_subtop,
                             int is_pred_opt)
{
    struct option *res;

#if NOPREDICT
    if (previous_output > predtag_to_output(0) &&
        (numout1 == 0 ||
         previous_output <= predtag_to_output(numout1 - 1)))
        previous_output = predtag_to_output(0);
#endif

    {
        res = create_null_option();
    }
    set_option(res, period, phase, lprob, ranking_lprob, preterm,
               previous_output, previous_option, trigger_label,
               trigger_output, stack_subtop, stack_rest_rel, prev_pred_ind,
               context_option, trigparent_option, trigchild_option,
               headchild_option, new_phase, new_state, new_subtop,
               is_pred_opt);
#if CHECKANTIGARBAGE
    anti_garbage_options =
        create_option_stack(res, anti_garbage_options, NULL);
#endif

#if PRINTDIST
    if (context_option == NULL) {
        if (preterm != 0) {
        } else if (previous_output == PREDOUT)
            printf("sc %3d\n", period - res->left_edge);
        else if (previous_output < NUMSTRUC)
            printf("se %3d\n", period - res->left_edge);
    } else if (previous_output == PREDOUT)
        printf("c %3d\n", period - context_option->period);
    else if (previous_output < NUMSTRUC)
        printf("e %3d\n", period - context_option->period);
#endif

    return (res);
}

/* set values in an option.  */
void set_option(struct option *res, int period, int phase, double lprob,
                double ranking_lprob, int preterm, int previous_output,
                struct option *previous_option,
                int trigger_label, int trigger_output,
                int stack_subtop,
                int stack_rest_rel, int prev_pred_ind,
                struct option *context_option,
                struct option *trigparent_option,
                struct option *trigchild_option,
                struct option *headchild_option, int new_phase,
                int new_state, int new_subtop, int is_pred_opt)
{
    struct option *opt;
    int pos;

    res->period = period;
    res->phase = phase;
    res->lprob = lprob;
    res->ranking_lprob = ranking_lprob;
    res->preterm = preterm;
    res->previous_output = previous_output;
    res->previous_option = previous_option;
    res->trigger_label = trigger_label;
    res->trigger_output = trigger_output;
    res->stack_subtop = stack_subtop;
    res->stack_rest_rel = stack_rest_rel;
    res->prev_pred_ind = prev_pred_ind;
    res->context_option = context_option;
    res->trigparent_option = trigparent_option;
    res->trigchild_option = trigchild_option;
    res->headchild_option = headchild_option;
    res->new_phase = new_phase;
    res->new_state = new_state;
    res->new_subtop = new_subtop;
    res->is_pred_opt = is_pred_opt;

    if (previous_option != NULL)
        previous_option->next_options =
            create_option_stack(res, previous_option->next_options, NULL);
    if (context_option != NULL)
        res->trigger_period = context_option->trigger_period;
    else
        res->trigger_period = period;
    if (trigchild_option == NULL)
        res->left_edge = period;
    else
        res->left_edge = trigchild_option->left_edge;
    res->next_options = NULL;
    res->in_best = 0;
    res->in_target = 0;
    res->num_descendents = 0;
    if (previous_output == PROJOUT || previous_output == HDPROUT)
        res->proj_depth = 1 + previous_option->proj_depth;
    else if (previous_output < NUMSTRUC)
        res->proj_depth = 0;
    else
        res->proj_depth = previous_option->proj_depth;
    res->backward_prob = 0.0;
#if PLUGLEAKS
    res->backward_knownout = 0.0;
#if ! LOCALESTIMATE
    res->pred_loss = 0.0;
    res->pred_loss_weight = 0.0;
#endif
#endif


    res->firstout = 0;
    res->lastout = -1;
    res->memoffset = 0;

    res->dec_cluster = NULL;

}

/* allocate an option without setting its values. */
struct option *create_null_option()
{
    struct option *res;

    res = (struct option *) malloc(sizeof(struct option));
    if (res == NULL)
        error("Cannot allocate enough space for max number of options.\n");

    res->state = NULL;
    res->subtop_state = NULL;
    res->outputs = NULL;

    return (res);
}


/* free opt, all the descendents of opt, and all the ancesters of opt which
     no longer have descendents.  Also removes freed options from
     best_options and **SHIFTS LOWER BEST_OPTIONS UP**.  */
void free_option_branch(struct option *opt)
{
    struct option_stack *ol;
    struct option **best;
    int b;

    if (opt->previous_option != NULL) {
        if (opt->previous_option->is_pred_opt) {
            /* remove from best_options */
            best = best_options[opt->period];
            for (b = 0; b < predbeam && best[b] != NULL; b++) {
                if (best[b] == opt) {
                    pluck_best_option(b, best);
                    break;
                }
            }
        }
        /* remove opt from previous option's next options */
        ol = opt->previous_option->next_options;
        if (ol != NULL) {
            if (ol->elem == opt)
                opt->previous_option->next_options =
                    popfree_option_stack(ol);
            else {
                for (ol = ol->next; ol != NULL; ol = ol->next)
                    if (ol->elem == opt) {
                        pluckfree_option_stack(ol);
                        break;
                    }
            }
        }
        /* recursive call to ancester */
        if (opt->previous_option->next_options == NULL) {
            free_option_branch(opt->previous_option);
        }
    }

    opt->previous_option = NULL;
    /* recursive call to descendents */
    free_option_descendents(opt);

}


/* free the structure for opt, any states which are introduced by it, and if
     opt is the birth of a phase then reclaim it for future allocation. */
void free_option(struct option *opt)
{
    struct option_stack *ol, *ol2;
    struct option *opt2;
    int pos;



    /* reclaim phase number */
    if (opt->phase >= 0 && opt->new_phase != 0) {
        push_stack(free_phases, opt->phase);
    }
    if (opt->stack_subtop >= 0 && opt->new_subtop != 0) {
        push_stack(free_phases, opt->stack_subtop);
    }

    /* free next_options list */
    if (opt->next_options != NULL) {
#if SAFENEXTOPTIONS
        warning("next_options not all freed");
#endif
        ol = opt->next_options;
        while (ol != NULL) {
            ol = popfree_option_stack(ol);
        }
        opt->next_options = NULL;
    }


    /* remove opt from previous next_options */
    if (opt->previous_option != NULL) {
        ol = opt->previous_option->next_options;
        if (ol != NULL) {
            if (ol->elem == opt)
                opt->previous_option->next_options =
                    popfree_option_stack(ol);
            else {
                for (ol = ol->next; ol != NULL; ol = ol->next)
                    if (ol->elem == opt) {
                        pluckfree_option_stack(ol);
                        break;
                    }
            }
        }
    }

    /* free layer activation arrays */
    if (opt->new_state != 0) {
        if (opt->state != NULL) {
#if ! NOFREEING
            free(opt->state);
#endif
        }
        if (opt->outputs != NULL) {
#if ! NOFREEING
            free(opt->outputs);
#endif
        }
    } else if (opt->is_pred_opt) {
        if (opt->outputs != NULL) {
#if ! NOFREEING
            free(opt->outputs);
#endif
        }

    }
    if (opt->subtop_state != NULL && opt->new_subtop != 0) {
#if ! NOFREEING
        free(opt->subtop_state);
#endif
    }


#if ! NOFREEING
    free(opt);
#endif


}




/* free opt and recursively free all the next_option descendents of opt */
void free_option_descendents(struct option *opt)
{
    struct option_stack *ol, *ol2;

    /* call recursively and free elements */
    ol = opt->next_options;
    while (ol != NULL) {
        ol2 = ol->next;         /* next_options modified when descendent freed */
        free_option_descendents(ol->elem);
        ol = ol2;
    }

    free_option(opt);
}


/* merge two ordered option lists into one ordered one and return it.
     destructive of original lists.  Ordered by ranking_lprob.  */
struct option_stack *mergesort_options(struct option_stack *l1,
                                       struct option_stack *l2)
{
    struct option_stack *res, *cur;

    if (l1 == NULL)
        return (l2);
    if (l2 == NULL)
        return (l1);

    if (l1->elem->ranking_lprob >= l2->elem->ranking_lprob) {
        cur = l1;
        l1 = l1->next;
    } else {
        cur = l2;
        l2 = l2->next;
    }

    res = cur;

    while (1) {                 /* loop until one list NULL */
        if (l1 == NULL) {
            cur->next = l2;
            l2->prev = cur;
            break;
        }
        if (l2 == NULL) {
            cur->next = l1;
            l1->prev = cur;
            break;
        }
        if (l1->elem->ranking_lprob >= l2->elem->ranking_lprob) {
            cur->next = l1;
            l1->prev = cur;
            l1 = l1->next;
        } else {
            cur->next = l2;
            l2->prev = cur;
            l2 = l2->next;
        }
        cur = cur->next;
    }

    return (res);
}


void display_option(struct option *opt, int shortformat)
{
    int ss;
    int prvt, prvp;

    if (opt->period > 150)
        warning("possible bad value for option's period");
    if (opt->phase > phase_space)
        warning("bad value for option's phase");

    if (opt->previous_option != NULL) {
        prvt = opt->previous_option->period;
        prvp = opt->previous_option->phase;
    } else {
        prvt = -2;
        prvp = -2;
    }
    if (opt->trigparent_option != NULL)
        ss = opt->trigparent_option->phase;
    else
        ss = -2;

    if (shortformat)
        printf("%2d,%2d,%3d; ",
               opt->period, opt->phase, opt->previous_output);
    else
        printf
            ("t=%2d, p=%2d, ptq=%d, out=%3d, lprob=%.0e, sub=%d,%d, prv=%2d,%2d, ppi=%d. (ptr=%d,prvptr=%d)\n",
             opt->period, opt->phase, opt->preterm, opt->previous_output,
             opt->lprob, opt->stack_subtop, ss, prvt, prvp,
             opt->prev_pred_ind, opt, opt->previous_option);
}

void display_symbolic_option(struct option *opt)
{
    printf("t=%2d, p=%2d, ptq=%d:  ",
           opt->period, opt->phase, opt->preterm);
    if (opt->previous_output == PROJOUT)
        printf("|      ");
    else if (opt->previous_output == HDPROUT)
        printf("|h     ");
    else if (opt->previous_output == PREDOUT)
        printf(">      ");
    else if (opt->previous_output == ATTOUT)
        printf("^      ");
    else if (opt->previous_output == HDATOUT)
        printf("^h     ");
    else if (opt->previous_output == MODOUT)
        printf("+      ");
    else if (opt->previous_output >= projlabel_to_output(0) &&
             opt->previous_output <= projlabel_to_output(numlabels - 1))
        printf("|:%-5s",
               id_label(output_to_projlabel(opt->previous_output)));
#if HEADOUTS
    else if (opt->previous_output >= hdprlabel_to_output(0) &&
             opt->previous_output <= hdprlabel_to_output(numlabels - 1))
        printf("|h:%-5s",
               id_label(output_to_hdprlabel(opt->previous_output)));
#endif
    else if (opt->previous_output >= predtag_to_output(0) &&
             (numout1 == 0 ||
              opt->previous_output <= predtag_to_output(numout1 - 1)))
        printf(">:%-5s", id_tag(output_to_predtag(opt->previous_output)));
    else
        printf("???    ");
    printf("lprob= %e,  rpr= %e\n", opt->lprob, opt->ranking_lprob);
}


/* allocate an option stack cons */
struct option_stack *create_option_stack(struct option *elem,
                                         struct option_stack *next,
                                         struct option_stack *prev)
{
    struct option_stack *res;
    struct option_stack *ol;

    /******************/
    for (ol = next; ol != NULL; ol = ol->next) {
        if (ol->elem == elem)
            warning("adding option twice to same list");
        if (ol != next && ol->prev->next != ol)
            warning("prev->next not valid");
    }

    if (garbage_option_stack != NULL) {
        res = garbage_option_stack;
        garbage_option_stack = garbage_option_stack->next;
        /* printf("%d reclaiming option_stack\n", res); */
    } else {
        res = (struct option_stack *) malloc(sizeof(struct option_stack));
        if (res == NULL)
            error
                ("Cannot allocate enough space for max number of options.\n");
    }

    res->elem = elem;
    res->next = next;
    res->prev = prev;
    if (res->next != NULL)
        res->next->prev = res;
    if (res->prev != NULL)
        res->prev->next = res;

    return (res);
}

/* allocate an option stack cons with NULL pointers */
struct option_stack *create_option_cons(struct option *elem)
{
    struct option_stack *res;

    if (garbage_option_stack != NULL) {
        res = garbage_option_stack;
        garbage_option_stack = garbage_option_stack->next;
        /* printf("%d reclaiming option_stack\n", res); */
    } else {
        res = (struct option_stack *) malloc(sizeof(struct option_stack));
        if (res == NULL)
            error
                ("Cannot allocate enough space for max number of options.\n");
    }

    res->elem = elem;
    res->next = NULL;
    res->prev = NULL;

    return (res);
}


/* free the cons pointed to by end and return a pointer to end->next.  does
     not free end->elem */
struct option_stack *popfree_option_stack(struct option_stack *end)
{
    struct option_stack *res;
    struct option_stack *ol;

    if (end == NULL) {
        warning("trying to pop an empty option stack.");
        res = NULL;
    } else {
        res = end->next;
#if CHECKGARBAGE
        for (ol = garbage_option_stack; ol != NULL; ol = ol->next)
            if (ol == end) {
                warning
                    ("trying to free an option_stack which is already free.");
                return (res);
            }
#endif
        end->next = garbage_option_stack;
        garbage_option_stack = end;
        /* printf("%d collecting option_stack\n", garbage_option_stack); */
        /**
#if ! NOFREEING
        free(end);
#endif
#if DEBUG_ALLOC
        printf("\n%d option stack cons freed\n", end);
#endif
         **/
        if (res != NULL)
            res->prev = NULL;
    }

    return (res);
}

/* free the cons pointed to by ol, but not its option.  update rest of
     stack to remove this cons */
void pluckfree_option_stack(struct option_stack *ol)
{
    struct option_stack *ol2;

    if (ol->prev != NULL)
        ol->prev->next = ol->next;
    if (ol->next != NULL)
        ol->next->prev = ol->prev;
#if CHECKGARBAGE
    for (ol2 = garbage_option_stack; ol2 != NULL; ol2 = ol2->next)
        if (ol2 == ol) {
            warning
                ("trying to free an option_stack which is already free.");
            return;
        }
#endif
    ol->next = garbage_option_stack;
    garbage_option_stack = ol;
    /* printf("%d collecting option_stack\n", garbage_option_stack); */
    /**
#if ! NOFREEING
    free(ol);
#endif
#if DEBUG_ALLOC
    printf("\n%d option stack cons freed\n", ol);
#endif
    **/
}

struct option_stack *incr_option_stack_next(struct option_stack *nxt)
{
    if (nxt->next == NULL)
        nxt->next = create_option_stack(NULL, NULL, nxt);       /* alloc */

    return (nxt->next);
}


/* allocate an option stack cons.  does not set anything in next,prev */
struct option_stack_stack *create_option_stack_stack(struct option_stack
                                                     *elem, struct
                                                     option_stack_stack
                                                     *next, struct
                                                     option_stack_stack
                                                     *prev)
{
    struct option_stack_stack *res;

    res = (struct option_stack_stack *)
        malloc(sizeof(struct option_stack_stack));
    if (res == NULL)
        error
            ("Cannot allocate enough space for max number of option_stacks.\n");

    res->elem = elem;
    res->next = next;
    res->prev = prev;

    return (res);
}

/* allocate an option_stack stack cons with NULL pointers */
struct option_stack_stack *create_option_stack_cons(struct option_stack
                                                    *elem)
{
    struct option_stack_stack *res;

    res = (struct option_stack_stack *)
        malloc(sizeof(struct option_stack_stack));
    if (res == NULL)
        error
            ("Cannot allocate enough space for max number of option_stacks.\n");

    res->elem = elem;
    res->next = NULL;
    res->prev = NULL;

    return (res);
}


/* free the cons pointed to by end and return a pointer to end->next.  does
     not free end->elem */
struct option_stack_stack *popfree_option_stack_stack(struct
                                                      option_stack_stack
                                                      *end)
{
    struct option_stack_stack *res;

    if (end == NULL) {
        warning("trying to pop an empty option_stack stack.");
        res = NULL;
    } else {
        res = end->next;
#if ! NOFREEING
        free(end);
#endif
        if (res != NULL)
            res->prev = NULL;
    }

    return (res);
}





/* compute a summary of the output errors for opt.  */
void opt_comp_err_summary(struct option *opt)
{
    int i, im, t, p;
    double outerr;
    double posbias;
    double *output_act;
    double *target_act;
    int memoffset, im_last, im_true;
    int wrd_last_mem;
    int bin;
    int out1;
    int *out2;
    double target;
    int *pred_word;

    t = opt->period;
    out1 = input_hist[t + 1].input1;

    if (opt->previous_option != NULL && opt->in_target != 0 &&
        opt->previous_option->firstout <= opt->previous_option->lastout) {
/*
#if WORDIDS
      if (opt->previous_option->previous_output == PREDOUT) {
        pred_word = input_hist[opt->period].output2;
        for (bin = 0; bin < numbins; bin++) {
          if (pred_word[bin] >= 0) {
            outerr = 
              - log(opt->previous_option->outputs->act[predword_to_output(pred_word[bin] - opt->previous_option->memoffset)]);
          }
          else
            break;
#if DEBUGTEST4
          if (outerr != 0.0)
            printf("error for output %d: %f  [%d:%d,%d]\n", 
                   predword_to_output(pred_word[bin]), outerr, 
                   num_sent+1, opt->previous_option->period, 
                   opt->previous_option->phase);
#endif
          total_local_err += outerr;
          total_targets += 1.0;
          total_word_err += outerr;
          total_word_targets += 1.0;
        }
      }
#endif
*/
        outerr =
            -log(opt->previous_option->outputs->act[opt->previous_output]);

        total_local_err += outerr;
        num_items +=
            opt->previous_option->lastout -
            opt->previous_option->firstout + 1;
        total_targets += 1.0;
        if (opt->previous_output < NUMSTRUC) {
            total_struc_err += outerr;
            total_struc_targets += 1.0;
        } else if (opt->previous_output >= projlabel_to_output(0) &&
                   opt->previous_output <=
                   projlabel_to_output(numlabels - 1)) {
            total_label_err += outerr;
            total_label_targets += 1.0;
        }
#if ! NOPREDICT
        else if (opt->previous_output >= predtag_to_output(0) &&
                 opt->previous_output <= predtag_to_output(numout1 - 1)) {
            total_tag_err += outerr;
            total_tag_targets += 1.0;
        }

        else if (IS_CLUSTER_MEMOUT(opt->previous_output)) {
            total_word_err += outerr;
            total_word_targets += 1.0;
        } else
            warning("output not of any type, in error calculation");
#endif
    }
}





/* compare target_tree with output_tree and compute precision/recall
     figures.  */
void comp_structure_summary(int length)
{
    int const_correct, const_target, const_output;
    int const_struct_correct;

    if (bracketed_output != 0) {
        if (target_tree != NULL)
	    write_tree(target_tree, gldfile, NULL);
        write_tree(output_tree, resfile, NULL);
    }

    /* sets const_correct, const_target, const_output */
    test_constituents(&const_correct, &const_target, &const_output,
                      length);

    /* Compute recall/precision on exact structural matches */
    const_struct_correct = 0;




    /* update the totals */
    total_const_correct += const_correct;
    total_const_output += const_output;
    total_const_target += const_target;
    if (length <= 100) {
        total100_const_correct += const_correct;
        total100_const_output += const_output;
        total100_const_target += const_target;
    }
    if (length <= 40) {
        total40_const_correct += const_correct;
        total40_const_output += const_output;
        total40_const_target += const_target;
    }
    total_const_struct_correct += const_struct_correct;
    if (const_correct == const_target && const_correct == const_output)
        total_sent_correct++;

}

/* see comp_structure_summary */
void test_constituents(int *numcorrect_pntr, int *numtarget_pntr,
                       int *numoutput_pntr, int length)
{
    struct tree *tst, *ost;
    struct tree_list *tstl, *ostl;
    struct tree_list *tl;
    int const_correct, const_target, const_output;

    if (target_tree != NULL)
      tstl = flatten_tree(target_tree, NULL); /* flattens to an ordered list */
    else
      tstl = NULL;
    ostl = flatten_tree(output_tree, NULL);     /* flattens to an ordered list */


    const_correct = 0;
    const_target = 0;
    const_output = 0;
    while (tstl != NULL && ostl != NULL) {
        tst = tstl->elem;
        ost = ostl->elem;
        /* don't count matrix root */
        if (tst->start == 0 && tst->end == length) {
            tstl = popfree_tree_list(tstl);
        } else if (ost->start == 0 && ost->end == length) {
            ostl = popfree_tree_list(ostl);
        }
        /* use ordering to search for matches: start then -length then label */
        else if (tst->start == ost->start) {
            if (tst->end == ost->end) {
                if (strcmp(tst->label, ost->label) == 0) {
                    const_correct++;
                    const_target++;
                    const_output++;
                    tstl = popfree_tree_list(tstl);
                    ostl = popfree_tree_list(ostl);
                } else if (strcmp(tst->label, ost->label) < 0) {
                    const_target++;
                    tstl = popfree_tree_list(tstl);
                } else {
                    const_output++;
                    ostl = popfree_tree_list(ostl);
                }
            } else if (tst->end > ost->end) {   /* -length < */
                const_target++;
                tstl = popfree_tree_list(tstl);
            } else {
                const_output++;
                ostl = popfree_tree_list(ostl);
            }
        } else if (tst->start < ost->start) {
            const_target++;
            tstl = popfree_tree_list(tstl);
        } else {
            const_output++;
            ostl = popfree_tree_list(ostl);
        }
    }
    while (tstl != NULL) {
        const_target++;
        tstl = popfree_tree_list(tstl);
    }
    while (ostl != NULL) {
        const_output++;
        ostl = popfree_tree_list(ostl);
    }

    *numcorrect_pntr = const_correct;
    *numtarget_pntr = const_target;
    *numoutput_pntr = const_output;
}


/* return a list of the subtrees of tree ordered by start then -length
     (largest to smallest length) then label, with tail appended.  Initially
     tail should be NULL.  Does not include preterms. */
struct tree_list *flatten_tree(struct tree *tree, struct tree_list *tail)
{
    struct tree_list *res;
    struct tree_list *tl;

    res = tail;

    if (tree->children != NULL) {       /* otherwise a preterm */
        /* note children are from right to left */
        for (tl = tree->children; tl != NULL; tl = tl->next) {
            res = flatten_tree(tl->elem, res);
        }
        /* nonbranching nodes need to be ordered by label */
        if (res != NULL &&
            res->elem->start == tree->start &&
            res->elem->end == tree->end &&
            strcmp(res->elem->label, tree->label) < 0) {
            for (tl = res; tl->next != NULL; tl = tl->next) {
                if (tl->next->elem->start > tree->start ||
                    tl->next->elem->end < tree->end ||
                    strcmp(tl->next->elem->label, tree->label) >= 0) {
                    break;
                }
            }
            tl->next = create_tree_list(tree, tl->next);
        } else {
            res = create_tree_list(tree, res);
        }
    }

    return (res);
}



/* return a pointer to a new tree. */
struct tree *create_tree()
{
    struct tree *res;

    res = (struct tree *) malloc(sizeof(struct tree));
    if (res == NULL)
        error("Cannot allocate enough space for max number of trees.\n");
#if DEBUG_ALLOC
    printf("\n%d tree created\n", res);
#endif

    res->parent = NULL;
    res->children = NULL;
    res->label[0] = '\0';
    res->trigger = NULL;        /* flag for trigger not set */
    res->trigger_child = NULL;
    res->adjunction_child = NULL;
    res->modification_child = NULL;
    res->attachment = NULL;
    res->conflation = 0;
    res->toplexical = 0;
    res->modifier = 0;
    res->head = 0;
    res->site_id = -1;

    return (res);
}

/* free the contents of tree t, including children list but not children */
void free_tree(struct tree *t)
{
    if (t != NULL) {
        free_tree_list(t->children);
#if ! NOFREEING
        free(t);
#endif
#if DEBUG_ALLOC
        printf("\n%d tree freed\n", t);
#endif
    }
}

/* free the contents of tree t, including children */
void free_all_tree(struct tree *t)
{
    if (t != NULL) {
        free_all_tree_list(t->children);
#if DEBUG_READOUT
        printf("Freeing node %d - %d\n", t->start, t->end);
#endif
#if ! NOFREEING
        free(t);                /* children list conses already freed */
#endif
#if DEBUG_ALLOC
        printf("\n%d tree freed\n", t);
#endif
    }
}

/* return a pointer to a tree_list cons includeing elem and rest. */
struct tree_list *create_tree_list(struct tree *elem,
                                   struct tree_list *rest)
{
    struct tree_list *res;

    res = (struct tree_list *) malloc(sizeof(struct tree_list));
    if (res == NULL)
        error("Cannot allocate enough space for max number of trees.\n");
#if DEBUG_ALLOC
    printf("\n%d tree list created\n", res);
#endif

    res->elem = elem;
    res->next = rest;
    res->prev = NULL;
    if (rest != NULL)
        rest->prev = res;

    return (res);
}

/* free the space used by all the cons's in tl, but not its trees. */
void free_tree_list(struct tree_list *tl)
{
    struct tree_list *head;
    struct tree_list *rest;

    for (head = tl; head != NULL; head = rest) {
        rest = head->next;
#if ! NOFREEING
        free(head);
#endif
#if DEBUG_ALLOC
        printf("\n%d tree list freed\n", head);
#endif
    }
}

/* free the space used by all the cons's in tl, plus its trees and subtrees. */
void free_all_tree_list(struct tree_list *tl)
{
    struct tree_list *head;
    struct tree_list *rest;

    for (head = tl; head != NULL; head = rest) {
        rest = head->next;
        free_all_tree(head->elem);
#if ! NOFREEING
        free(head);
#endif
#if DEBUG_ALLOC
        printf("\n%d tree list freed\n", head);
#endif
    }
}

/* free the space used by the cons tl, but not its tree.  update rest of
     list to remove this cons */
void pluckfree_tree_list(struct tree_list *tl)
{
    if (tl->prev != NULL)
        tl->prev->next = tl->next;
    if (tl->next != NULL)
        tl->next->prev = tl->prev;
#if ! NOFREEING
    free(tl);
#endif
#if DEBUG_ALLOC
    printf("\n%d tree list freed\n", tl);
#endif
}

/* free the space used by the first cons in tl and return a pointer to
     the next cons.  Does not free the element in the cons. */
struct tree_list *popfree_tree_list(struct tree_list *tl)
{
    struct tree_list *res;

    if (tl == NULL)
        error("Error: can't pop an empty tlist.");

    res = tl->next;
    if (res != NULL)
        res->prev = NULL;
#if ! NOFREEING
    free(tl);
#endif
#if DEBUG_ALLOC
    printf("\n%d tree list freed\n", tl);
#endif

    return (res);
}


/* write the results computed by testing_epoch to brafile as for cycle. */
void write_testing(int best_flag)
{
    FILE *bfp;
    double score;

#if ERRVALIDATE
    score = exp(-1.0 * total_err / total_targets);
#else
    score = ((double) (total_const_correct * 200))
        / ((double) (total_const_target + total_const_output));
#endif

    if ((bfp = fopen(brafile, "a")) == NULL)
        error("Error: Cannot open bracket file to write data.");

    fprintf(bfp,
            "\n        Results of testing on data file %s at cycle %d\n          (prediction beam = %d, prediction factor = %f, branching factor = %d)\n\n",
            testingfile, weight_age, predbeam, expected_max_pred_factor,
	    phasebeam);

#if ! ERRVALIDATE

    fprintf(bfp,
            "%3d: Headlines: const F1 %6.3f%%, rec %6.3f%%, prec %6.3f%%",
            weight_age, score,
            ((double) (total_const_correct * 100)) /
            ((double) total_const_target),
            ((double) (total_const_correct * 100)) /
            ((double) total_const_output));
    if (best_flag != 0)
        fprintf(bfp, ", best.\n");
    else
        fprintf(bfp, ".\n");

    fprintf(bfp, "%3d: Sentences correct: %d, out of %d (%6.3f%%).\n",
            weight_age, total_sent_correct, num_sent,
            ((double) (total_sent_correct * 100)) / ((double) num_sent));

    fprintf(bfp,
            "%3d: All Constituents recall, precision, (F1): %6.3f%%, %6.3f%%, (%6.3f%%)\n",
            weight_age,
            ((double) (total_const_correct * 100)) /
            ((double) total_const_target),
            ((double) (total_const_correct * 100)) /
            ((double) total_const_output),
            ((double) (total_const_correct * 200))
            / ((double) (total_const_target + total_const_output)));
    fprintf(bfp, "  correct: %d,  target: %d,  output: %d.\n",
            total_const_correct, total_const_target, total_const_output);

    fprintf(bfp,
            "%3d: Length<=100 Constituents rec, prec, (F1): %6.3f%%, %6.3f%%, (%6.3f%%)\n",
            weight_age,
            ((double) (total100_const_correct * 100)) /
            ((double) total100_const_target),
            ((double) (total100_const_correct * 100)) /
            ((double) total100_const_output),
            ((double) (total100_const_correct * 200))
            / ((double) (total100_const_target + total100_const_output)));
    fprintf(bfp, "  correct: %d,  target: %d,  output: %d.\n",
            total100_const_correct, total100_const_target,
            total100_const_output);

    fprintf(bfp,
            "%3d: Length<=40 Constituents  rec, prec, (F1): %6.3f%%, %6.3f%%, (%6.3f%%)\n",
            weight_age,
            ((double) (total40_const_correct * 100)) /
            ((double) total40_const_target),
            ((double) (total40_const_correct * 100)) /
            ((double) total40_const_output),
            ((double) (total40_const_correct * 200))
            / ((double) (total40_const_target + total40_const_output)));
    fprintf(bfp, "  correct: %d,  target: %d,  output: %d.\n",
            total40_const_correct, total40_const_target,
            total40_const_output);


#else

    fprintf(bfp,
            "%3d: Headlines: P ave %.8f, (struc %.4f, lab %.4f, tag %.4f, word %.4f)",
            weight_age, score, exp(-total_struc_err / total_struc_targets),
            exp(-total_label_err / total_label_targets),
            exp(-total_tag_err / total_tag_targets),
            exp(-total_word_err / total_word_targets));
    if (best_flag != 0)
        fprintf(bfp, ", best.\n");
    else
        fprintf(bfp, ".\n");

    fprintf(bfp, "%3d: Total error in data: %f=%f+%f\n",
            weight_age, total_err, total_local_err,
            total_local_err_correction);
    fprintf(bfp, "%3d: P(target)= %.8f ^%.0f (%d including zeros)\n",
            weight_age, exp(-1.0 * total_err / total_targets),
            total_targets, num_items);
    fprintf(bfp, "%3d: P(structure)= %.8f ^%.0f\n", weight_age,
            exp(-total_struc_err / total_struc_targets),
            total_struc_targets);
    fprintf(bfp, "%3d: P(labeling)= %.8f ^%.0f\n", weight_age,
            exp(-total_label_err / total_label_targets),
            total_label_targets);
    fprintf(bfp, "%3d: P(tagging)= %.8f ^%.0f\n", weight_age,
            exp(-total_tag_err / total_tag_targets), total_tag_targets);
    fprintf(bfp, "%3d: P(word predicting)= %.8f ^%.0f\n",
            weight_age, exp(-total_word_err / total_word_targets),
            total_word_targets);
    printf("\n");

#endif

    if (best_flag != 0)
        fprintf(bfp, "%3d: Weights saved to %s (%f)\n\n",
                weight_age, valweightfile, score);

    fclose(bfp);
}

/* read the most recent weight file. */
void read_last_weights()
{
    FILE *fp;

    if ((fp = fopen(newweightfile, "r")) != NULL) {
        fclose(fp);
        read_weights(newweightfile);
    } else
        read_weights(weightfile);
}


/* read link weights from wgtfile, or generate random weights if wgtfile
     doesn't exists (and save them).  read weight and bias deltas from
     wgtfile.mom, or zero them if doesn't exist.  If WEIGHTHEADER=1 then sets
     weight_age, eta_red, reg_red, and weight_pretrnerr.  Return score found in
     header. 
     if is_initial_file is set then fields in the header are skipped (but it should present in the file
     unless WEIGHTHEADER == 0), deltas for probabilistic table are not required to be found
 */
double read_weights(char *wgtfile)
{
    char momfile[FILENAMESIZE];
    FILE *fp;
    int i, j, k, p;
    int inp1, out1;
    char *str;
    double val;
    int ival;
    double test_score;
    char buffer[MAXWGTLINE];
    test_score = 0.0;

    if ((fp = fopen(wgtfile, "r")) == NULL) {
        /* if no .wgt file, generate random weights. */
        /* uses seed to locate a point in the random sequence              */
        /* this sequence is always same, i.e. same seed gives same numbers */
        srand48(seed);
        for (i = 0; i < numinp1; i++)
            for (j = 0; j < numrho; j++) {
                link_r1_r[j][i] = RAND_WEIGHT;
                link_t1_r[j][i] = RAND_WEIGHT;
#if HEADWORDS
                link_h1_r[j][i] = RAND_WEIGHT;
#endif
            }
        for (i = 0; i < numinp2; i++)
            for (j = 0; j < numrho; j++) {
#if ZEROWORDINP
                link_r2_r[j][i] = 0.0;  /* zero initial values for word links */
#else
                link_r2_r[j][i] = RAND_WEIGHT;
#endif
#if ZEROWORDINP
                link_t2_r[j][i] = 0.0;  /* zero initial values for word links */
#else
                link_t2_r[j][i] = RAND_WEIGHT;
#endif
#if HEADWORDS
#if ZEROWORDINP
                link_h2_r[j][i] = 0.0;  /* zero initial values for word links */
#else
                link_h2_r[j][i] = RAND_WEIGHT;
#endif
#endif
            }
#if ! LABELCONTEXTINP
        for (i = 0; i < numrho; i++)
            for (j = 0; j < numrho; j++)
                link_r_r[j][i] = RAND_WEIGHT;
#endif
        for (i = 0; i < numrho; i++)
            for (j = 0; j < numout; j++)
                link_r_o[j][i] = RAND_WEIGHT_OUT;
        for (i = 0; i < numrho; i++)
            bias_r[i] = RAND_WEIGHT;
#if INITOUTPUTBIAS
        for (i = 0; i < numout; i++)
            RAND_WEIGHT_OUT;
#else
        for (i = 0; i < numout; i++)
            bias_o[i] = RAND_WEIGHT_OUT;
#endif
        for (i = 0; i < numrho; i++)
            hidrho_initstate.act[i] = RAND_WEIGHT;
        for (j = 0; j < numrho; j++)
            link_bir_r[j] = RAND_WEIGHT;
        for (i = 0; i < numlink_tro; i++)
            for (j = 0; j < numrho; j++)
                link_tro_r[j][i] = RAND_WEIGHT;
        for (i = 0; i < numlink_trp; i++)
            for (j = 0; j < numrho; j++) {
                link_trp_r[j][i] = RAND_WEIGHT;
            }
        for (i = 0; i < numlink_trc; i++)
            for (j = 0; j < numrho; j++) {
                link_trc_r[j][i] = RAND_WEIGHT;
            }
        for (i = 0; i < numlink_lab; i++)
            for (j = 0; j < numrho; j++)
                link_lab_r[j][i] = RAND_WEIGHT;
        for (i = 0; i < numrho; i++)
            link_pred_r[i] = RAND_WEIGHT;
#if CHILDCONTEXT
        for (i = 0; i < numlink_chd; i++)
            for (j = 0; j < numrho; j++) {
                link_chd_r[j][i] = RAND_WEIGHT;
            }
#endif
#if HEADHIDDEN
        for (i = 0; i < numlink_hdc; i++)
            for (j = 0; j < numrho; j++) {
                link_hdc_r[j][i] = RAND_WEIGHT;
            }
#endif
#if CHOPCONTEXT
        for (i = 0; i < numlink_cho; i++)
            for (j = 0; j < numrho; j++) {
                link_cho_r[j][i] = RAND_WEIGHT;
            }
#endif
#if HEADOUTS
        for (j = 0; j < numrho; j++) {
            link_nhd_r[j] = RAND_WEIGHT;
        }
#endif
    } else
        /* if .wgt file exists, load weights in     */
    {


        fgets(buffer, MAXWGTLINE, fp);

            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt header. [type]");
            if (strcmp(str, "Ar") != 0)
                error
                    ("Error: .wgt file is not for a reduced type A network.");
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error
                    ("Error: Insufficient input in .wgt header. [weight_age]");
            ival = atoi(str);
            if (ival < 0) {
                sprintf(warning_text,
                        "weight age out of range; using %f.\n",
                        weight_age);
                warning(warning_text);
            } else
                weight_age = ival;
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error
                    ("Error: Insufficient input in .wgt header. [eta_red]");
            val = atof(str);
            if ((eta_inf * val < 0) || (eta_inf * val > 1)) {
                sprintf(warning_text,
                        "learning rate reduction out of range; using %f.\n",
                        eta_red);
                warning(warning_text);
            } else
                eta_red = val;
            eta = eta_inf * eta_red;
            decay = 1.0 - reg * eta;
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error
                    ("Error: Insufficient input in .wgt header. [reg_red]");
            val = atof(str);
            if ((val < 0.0) || (reg_inf - val > 0.9))
                error
                    ("Error: Regularization reduction out of range (or not present).\n");
            else
                reg_red = val;
            if (reg_inf - reg_red > 0.0)
                reg = reg_inf - reg_red;
            else
                reg = 0.0;
            decay = 1.0 - reg * eta;
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error
                    ("Error: Insufficient input in .wgt header. [weight_pretrnerr]");
            val = atof(str);
            if (val < -1.0) {
                sprintf(warning_text,
                        "previous training error out of range; using %f.\n",
                        weight_pretrnerr);
                warning(warning_text);
            } else
                weight_pretrnerr = val;
            if ((str = strtok(NULL, ", \t\n")) != NULL)
                test_score = atof(str);



        for (i = 0; i < numinp1; i++) {
            //skip all the lines started with EM  
            do {
                fgets(buffer, MAXWGTLINE, fp);
                if ((str = strtok(buffer, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [I0-R]");
            } while (strcmp(str, "EM") == 0);
            link_r1_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL) {
                    printf("j = %d, numrho = %d\n", j, numrho);
                    error("Error: Insufficient input .wgt. [I-R]");
                }
                link_r1_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp2; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [I0-R]");
            link_r2_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .wgt. [I-R]");
                link_r2_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp1; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [T0-R]");
            link_t1_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .wgt. [T-R]");
                link_t1_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp2; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [T0-R]");
            link_t2_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .wgt. [T-R]");
                link_t2_r[j][i] = atof(str);
            }
        }
#if HEADWORDS
        for (i = 0; i < numinp1; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [H0-R]");
            link_h1_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .wgt. [H-R]");
                link_h1_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp2; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [H0-R]");
            link_h2_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .wgt. [H-R]");
                link_h2_r[j][i] = atof(str);
            }
        }
#endif
#if ! LABELCONTEXTINP
        for (i = 0; i < numrho; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[R0-R]");
            link_r_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [R-R]");
                link_r_r[j][i] = atof(str);
            }
        }
#endif
        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .wgt. [BIR0-R]");
        link_bir_r[0] = atof(str);
        for (j = 1; j < numrho; j++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [BIR-R]");
            link_bir_r[j] = atof(str);
        }
        for (i = 0; i < numlink_tro; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[TRO0-R]");
            link_tro_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [TRO-R]");
                link_tro_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numlink_trp; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[TRP-R]");
            link_trp_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [TRP-R]");
                link_trp_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numlink_trc; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[TRC-R]");
            link_trc_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [TRC-R]");
                link_trc_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numlink_lab; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[LAB-R]");
            link_lab_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [LAB-R]");
                link_lab_r[j][i] = atof(str);
            }
        }
        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .wgt. [PRED0-R]");
        link_pred_r[0] = atof(str);
        for (j = 1; j < numrho; j++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [PRED-R]");
            link_pred_r[j] = atof(str);
        }
#if CHILDCONTEXT
        for (i = 0; i < numlink_chd; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[CHD-R]");
            link_chd_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [CHD-R]");
                link_chd_r[j][i] = atof(str);
            }
        }
#endif
#if HEADHIDDEN
        for (i = 0; i < numlink_hdc; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[HDC-R]");
            link_hdc_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [HDC-R]");
                link_hdc_r[j][i] = atof(str);
            }
        }
#endif
#if CHOPCONTEXT
        for (i = 0; i < numlink_cho; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt.[CHO-R]");
            link_cho_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [CHO-R]");
                link_cho_r[j][i] = atof(str);
            }
        }
#endif
#if HEADOUTS
        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .wgt.[NHD-R]");
        link_nhd_r[0] = atof(str);
        for (j = 1; j < numrho; j++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [NHD-R]");
            link_nhd_r[j] = atof(str);
        }
#endif
        for (i = 0; i < numrho; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [R0-O]");
            link_r_o[0][i] = atof(str);
            for (j = 1; j < numout; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL) {
                    sprintf(warning_text,
                            "Insufficient input in .wgt; randomizing. [R-O]\n");
                    warning(warning_text);
                    link_r_o[j][i] = RAND_WEIGHT_OUT;
                } else {
                    link_r_o[j][i] = atof(str);

                }
            }
        }

        /* now load in bias units */
        fgets(buffer, MAXWGTLINE, fp);
        if (strcmp(buffer, "Biases:\n") == 0)
            fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .wgt. [B0-R]");
        bias_r[0] = atof(str);
        for (i = 1; i < numrho; i++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .wgt. [B-R]");
            bias_r[i] = atof(str);
        }



        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .wgt. [B0-O]");
        bias_o[0] = atof(str);
        for (i = 1; i < numout; i++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL) {
                {
                    sprintf(warning_text,
                            "Insufficient input in .wgt; randomizing. [B-O]\n");
                    warning(warning_text);
                }
                warning("");
                bias_o[i] = RAND_WEIGHT_OUT;
            } else
                bias_o[i] = atof(str);
        }

        /* test for and load initialization states */
        if ((fgets(buffer, MAXWGTLINE, fp) != NULL) &&
            ((str = strtok(buffer, ", \t\n")) != NULL)) {
            hidrho_initstate.act[0] = atof(str);
            for (i = 1; i < numrho; i++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .wgt. [RI]");
                hidrho_initstate.act[i] = atof(str);
            }
        } else
            for (i = 0; i < numrho; i++)
                hidrho_initstate.act[i] = 0.0;


        fclose(fp);
    }

    /* initialize deltas with momentum */
    strcpy(momfile, wgtfile);
    strcat(momfile, ".mom");

    fp = NULL;
    if ((fp = fopen(momfile, "r")) == NULL)
    {   
        if (fp != NULL) {
            fclose(fp);
        }
            
        /* if no .mom file, zero deltas. */
        for (i = 0; i < numrho; i++)
            hidrho_initstate.err[i] = 0.0;
        for (i = 0; i < numinp1; i++) {
            for (j = 0; j < numrho; j++) {
                linkdel_r1_r[j][i] = 0.0;
                linkdel_t1_r[j][i] = 0.0;
#if HEADWORDS
                linkdel_h1_r[j][i] = 0.0;
#endif
            }
        }
        for (i = 0; i < numinp2; i++) {
            for (j = 0; j < numrho; j++) {
                linkdel_r2_r[j][i] = 0.0;
                linkdel_t2_r[j][i] = 0.0;
#if HEADWORDS
                linkdel_h2_r[j][i] = 0.0;
#endif
            }
        }
        for (i = 0; i < numrho; i++) {
            biasdel_r[i] = 0.0;
#if ! LABELCONTEXTINP
            for (j = 0; j < numrho; j++)
                linkdel_r_r[j][i] = 0.0;
#endif
        }
        for (j = 0; j < numrho; j++)
            linkdel_bir_r[j] = 0.0;
        for (i = 0; i < numlink_tro; i++)
            for (j = 0; j < numrho; j++)
                linkdel_tro_r[j][i] = 0.0;
        for (i = 0; i < numlink_trp; i++)
            for (j = 0; j < numrho; j++) {
                linkdel_trp_r[j][i] = 0.0;
            }
        for (i = 0; i < numlink_trc; i++)
            for (j = 0; j < numrho; j++) {
                linkdel_trc_r[j][i] = 0.0;
            }
        for (i = 0; i < numlink_lab; i++)
            for (j = 0; j < numrho; j++)
                linkdel_lab_r[j][i] = 0.0;
        for (j = 0; j < numrho; j++)
            linkdel_pred_r[j] = 0.0;
#if CHILDCONTEXT
        for (i = 0; i < numlink_chd; i++)
            for (j = 0; j < numrho; j++) {
                linkdel_chd_r[j][i] = 0.0;
            }
#endif
#if HEADHIDDEN
        for (i = 0; i < numlink_hdc; i++)
            for (j = 0; j < numrho; j++) {
                linkdel_hdc_r[j][i] = 0.0;
            }
#endif
#if CHOPCONTEXT
        for (i = 0; i < numlink_cho; i++)
            for (j = 0; j < numrho; j++) {
                linkdel_cho_r[j][i] = 0.0;
            }
#endif
#if HEADOUTS
        for (j = 0; j < numrho; j++) {
            linkdel_nhd_r[j] = 0.0;
        }
#endif
        for (i = 0; i < numrho; i++)
            for (j = 0; j < numout; j++)
                linkdel_r_o[j][i] = 0.0;
        for (i = 0; i < numout; i++)
            biasdel_o[i] = 0.0;
    } else
        /* if .mom file exists, load in deltas     */
    {
        /*
           printf(" starting to read momentum file\n");
         */
        for (i = 0; i < numinp1; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [I0-R]");
            linkdel_r1_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .mom. [I-R]");
                linkdel_r1_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp2; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [I0-R]");
            linkdel_r2_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .mom. [I-R]");
                linkdel_r2_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp1; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [T0-R]");
            linkdel_t1_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .mom. [T-R]");
                linkdel_t1_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp2; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [T0-R]");
            linkdel_t2_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .mom. [T-R]");
                linkdel_t2_r[j][i] = atof(str);
            }
        }
#if HEADWORDS
        for (i = 0; i < numinp1; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [H0-R]");
            linkdel_h1_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .mom. [H-R]");
                linkdel_h1_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numinp2; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [H0-R]");
            linkdel_h2_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input .mom. [H-R]");
                linkdel_h2_r[j][i] = atof(str);
            }
        }
#endif
#if ! LABELCONTEXTINP
        for (i = 0; i < numrho; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom.[R0 -R]");
            linkdel_r_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [R-R]");
                linkdel_r_r[j][i] = atof(str);
            }
        }
#endif
        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .mom. [N0-R]");
        linkdel_bir_r[0] = atof(str);
        for (j = 1; j < numrho; j++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [N-R]");
            linkdel_bir_r[j] = atof(str);
        }
        for (i = 0; i < numlink_tro; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [TRO0-R]");
            linkdel_tro_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [TRO-R]");
                linkdel_tro_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numlink_trp; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [TRP0-R]");
            linkdel_trp_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [TRP-R]");
                linkdel_trp_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numlink_trc; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [TRC0-R]");
            linkdel_trc_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [TRC-R]");
                linkdel_trc_r[j][i] = atof(str);
            }
        }
        for (i = 0; i < numlink_lab; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [LAB0-R]");
            linkdel_lab_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [LAB-R]");
                linkdel_lab_r[j][i] = atof(str);
            }
        }
        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .mom. [P0-R]");
        linkdel_pred_r[0] = atof(str);
        for (j = 1; j < numrho; j++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [P-R]");
            linkdel_pred_r[j] = atof(str);
        }
#if CHILDCONTEXT
        for (i = 0; i < numlink_chd; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [CHD0-R]");
            linkdel_chd_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [CHD-R]");
                linkdel_chd_r[j][i] = atof(str);
            }
        }
#endif
#if HEADHIDDEN
        for (i = 0; i < numlink_hdc; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [HDC0-R]");
            linkdel_hdc_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [HDC-R]");
                linkdel_hdc_r[j][i] = atof(str);
            }
        }
#endif
#if CHOPCONTEXT
        for (i = 0; i < numlink_cho; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [CHO0-R]");
            linkdel_cho_r[0][i] = atof(str);
            for (j = 1; j < numrho; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [CHO-R]");
                linkdel_cho_r[j][i] = atof(str);
            }
        }
#endif
#if HEADOUTS
        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .mom. [NHD0-R]");
        linkdel_nhd_r[0] = atof(str);
        for (j = 1; j < numrho; j++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [NHD-R]");
            linkdel_nhd_r[j] = atof(str);
        }
#endif
        for (i = 0; i < numrho; i++) {
            fgets(buffer, MAXWGTLINE, fp);
            if ((str = strtok(buffer, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [R0-O]");
            linkdel_r_o[0][i] = atof(str);
            for (j = 1; j < numout; j++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL) {
                    {
                        sprintf(warning_text,
                                "Insufficient input in .mom. [R-O]\n");
                        warning(warning_text);
                    }
                    warning("");
                    linkdel_r_o[j][i] = 0.0;
                } else
                    linkdel_r_o[j][i] = atof(str);
            }
        }

        /* now load in bias units */
        fgets(buffer, MAXWGTLINE, fp);
        if (strcmp(buffer, "Biases:\n") == 0)
            fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .mom. [B0-R]");
        biasdel_r[0] = atof(str);
        for (i = 1; i < numrho; i++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL)
                error("Error: Insufficient input in .mom. [B-R]");
            biasdel_r[i] = atof(str);
        }



        fgets(buffer, MAXWGTLINE, fp);
        if ((str = strtok(buffer, ", \t\n")) == NULL)
            error("Error: Insufficient input in .mom. [B0-O]");
        biasdel_o[0] = atof(str);
        for (i = 1; i < numout; i++) {
            if ((str = strtok(NULL, ", \t\n")) == NULL) {
                {
                    sprintf(warning_text,
                            "Insufficient input in .mom. [B-O]\n");
                    warning(warning_text);
                }
                warning("");
                biasdel_o[i] = 0.0;
            } else
                biasdel_o[i] = atof(str);
        }

        /* test for and load initialization states */
        if ((fgets(buffer, MAXWGTLINE, fp) != NULL) &&
            ((str = strtok(buffer, ", \t\n")) != NULL)) {
            hidrho_initstate.err[0] = atof(str);
            for (i = 1; i < numrho; i++) {
                if ((str = strtok(NULL, ", \t\n")) == NULL)
                    error("Error: Insufficient input in .mom. [RI]");
                hidrho_initstate.err[i] = atof(str);
            }
        } else
            for (i = 0; i < numrho; i++)
                hidrho_initstate.err[i] = 0.0;


        fclose(fp);
    }

    return (test_score);
}


/* write link and bias weights to wgtfile, and link and bias deltas to
     wgtfile.mom */
void save_weights(char *wgtfile, double score)
{
    char momfile[FILENAMESIZE];
    FILE *fp;
    int i, j, k;
    int inp1, out1;

    if ((fp = fopen(wgtfile, "w")) == NULL)
        error("Error: Cannot open weight file for output.");

    fprintf(fp, "Ar");
    fprintf(fp, ", %d", weight_age);
    fprintf(fp, ", %e", eta_red);
    fprintf(fp, ", %.20e", reg_red);  /* should change to multiplicative */
    fprintf(fp, ", %e", weight_pretrnerr);
    fprintf(fp, ", %f", score);
    fprintf(fp, "\n");


    for (i = 0; i < numinp1; i++) {
        fprintf(fp, "%.10f", link_r1_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_r1_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp2; i++) {
        fprintf(fp, "%.10f", link_r2_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_r2_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp1; i++) {
        fprintf(fp, "%.10f", link_t1_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_t1_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp2; i++) {
        fprintf(fp, "%.10f", link_t2_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_t2_r[j][i]);
        fprintf(fp, "\n");
    }
#if HEADWORDS
    for (i = 0; i < numinp1; i++) {
        fprintf(fp, "%.10f", link_h1_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_h1_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp2; i++) {
        fprintf(fp, "%.10f", link_h2_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_h2_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if ! LABELCONTEXTINP
    for (i = 0; i < numrho; i++) {
        fprintf(fp, "%.10f", link_r_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_r_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
    fprintf(fp, "%.10f", link_bir_r[0]);
    for (j = 1; j < numrho; j++)
        fprintf(fp, ", %.10f", link_bir_r[j]);
    fprintf(fp, "\n");
    for (i = 0; i < numlink_tro; i++) {
        fprintf(fp, "%.10f", link_tro_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_tro_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numlink_trp; i++) {
        fprintf(fp, "%.10f", link_trp_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_trp_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numlink_trc; i++) {
        fprintf(fp, "%.10f", link_trc_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_trc_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numlink_lab; i++) {
        fprintf(fp, "%.10f", link_lab_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_lab_r[j][i]);
        fprintf(fp, "\n");
    }
    fprintf(fp, "%.10f", link_pred_r[0]);
    for (j = 1; j < numrho; j++)
        fprintf(fp, ", %.10f", link_pred_r[j]);
    fprintf(fp, "\n");
#if CHILDCONTEXT
    for (i = 0; i < numlink_chd; i++) {
        fprintf(fp, "%.10f", link_chd_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_chd_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if HEADHIDDEN
    for (i = 0; i < numlink_hdc; i++) {
        fprintf(fp, "%.10f", link_hdc_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_hdc_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if CHOPCONTEXT
    for (i = 0; i < numlink_cho; i++) {
        fprintf(fp, "%.10f", link_cho_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", link_cho_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if HEADOUTS
    fprintf(fp, "%.10f", link_nhd_r[0]);
    for (j = 1; j < numrho; j++)
        fprintf(fp, ", %.10f", link_nhd_r[j]);
    fprintf(fp, "\n");
#endif
    for (i = 0; i < numrho; i++) {
        fprintf(fp, "%.10f", link_r_o[0][i]);
        for (j = 1; j < numout; j++)
            fprintf(fp, ", %.10f", link_r_o[j][i]);
        fprintf(fp, "\n");
    }

    fprintf(fp, "Biases:\n");

    fprintf(fp, "%.10f", bias_r[0]);
    for (i = 1; i < numrho; i++)
        fprintf(fp, ", %.10f", bias_r[i]);
    fprintf(fp, "\n");



    fprintf(fp, "%.10f", bias_o[0]);
    for (i = 1; i < numout; i++)
        fprintf(fp, ", %.10f", bias_o[i]);
    fprintf(fp, "\n");

    fprintf(fp, "%.10f", hidrho_initstate.act[0]);
    for (i = 1; i < numrho; i++)
        fprintf(fp, ", %.10f", hidrho_initstate.act[i]);
    fprintf(fp, "\n");


    fclose(fp);


    /* save deltas for momentum */
    strcpy(momfile, wgtfile);
    strcat(momfile, ".mom");
    if ((fp = fopen(momfile, "w")) == NULL)
        error("Error: Cannot open momentum file for output.");

    for (i = 0; i < numinp1; i++) {
        fprintf(fp, "%.6e", linkdel_r1_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.6e", linkdel_r1_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp2; i++) {
        fprintf(fp, "%.6e", linkdel_r2_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.6e", linkdel_r2_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp1; i++) {
        fprintf(fp, "%.6e", linkdel_t1_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.6e", linkdel_t1_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp2; i++) {
        fprintf(fp, "%.6e", linkdel_t2_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.6e", linkdel_t2_r[j][i]);
        fprintf(fp, "\n");
    }
#if HEADWORDS
    for (i = 0; i < numinp1; i++) {
        fprintf(fp, "%.6e", linkdel_h1_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.6e", linkdel_h1_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numinp2; i++) {
        fprintf(fp, "%.6e", linkdel_h2_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.6e", linkdel_h2_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if ! LABELCONTEXTINP
    for (i = 0; i < numrho; i++) {
        fprintf(fp, "%.6e", linkdel_r_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.6e", linkdel_r_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
    fprintf(fp, "%.6e", linkdel_bir_r[0]);
    for (j = 1; j < numrho; j++)
        fprintf(fp, ", %.6e", linkdel_bir_r[j]);
    fprintf(fp, "\n");
    for (i = 0; i < numlink_tro; i++) {
        fprintf(fp, "%.10f", linkdel_tro_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", linkdel_tro_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numlink_trp; i++) {
        fprintf(fp, "%.10f", linkdel_trp_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", linkdel_trp_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numlink_trc; i++) {
        fprintf(fp, "%.10f", linkdel_trc_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", linkdel_trc_r[j][i]);
        fprintf(fp, "\n");
    }
    for (i = 0; i < numlink_lab; i++) {
        fprintf(fp, "%.10f", linkdel_lab_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", linkdel_lab_r[j][i]);
        fprintf(fp, "\n");
    }
    fprintf(fp, "%.6e", linkdel_pred_r[0]);
    for (j = 1; j < numrho; j++)
        fprintf(fp, ", %.6e", linkdel_pred_r[j]);
    fprintf(fp, "\n");
#if CHILDCONTEXT
    for (i = 0; i < numlink_chd; i++) {
        fprintf(fp, "%.10f", linkdel_chd_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", linkdel_chd_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if HEADHIDDEN
    for (i = 0; i < numlink_hdc; i++) {
        fprintf(fp, "%.10f", linkdel_hdc_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", linkdel_hdc_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if CHOPCONTEXT
    for (i = 0; i < numlink_cho; i++) {
        fprintf(fp, "%.10f", linkdel_cho_r[0][i]);
        for (j = 1; j < numrho; j++)
            fprintf(fp, ", %.10f", linkdel_cho_r[j][i]);
        fprintf(fp, "\n");
    }
#endif
#if HEADOUTS
    fprintf(fp, "%.10f", linkdel_nhd_r[0]);
    for (j = 1; j < numrho; j++)
        fprintf(fp, ", %.10f", linkdel_nhd_r[j]);
    fprintf(fp, "\n");
#endif


    for (i = 0; i < numrho; i++) {
        fprintf(fp, "%.6e", linkdel_r_o[0][i]);
        for (j = 1; j < numout; j++)
            fprintf(fp, ", %.6e", linkdel_r_o[j][i]);
        fprintf(fp, "\n");
    }

    fprintf(fp, "Biases:\n");

    fprintf(fp, "%.6e", biasdel_r[0]);
    for (i = 1; i < numrho; i++)
        fprintf(fp, ", %.6e", biasdel_r[i]);
    fprintf(fp, "\n");



    fprintf(fp, "%.6e", biasdel_o[0]);
    for (i = 1; i < numout; i++)
        fprintf(fp, ", %.6e", biasdel_o[i]);
    fprintf(fp, "\n");

    fprintf(fp, "%.6e", hidrho_initstate.err[0]);
    for (i = 1; i < numrho; i++)
        fprintf(fp, ", %.6e", hidrho_initstate.err[i]);
    fprintf(fp, "\n");



    fclose(fp);
}


/* add random percentages to weights. */
/* weight *= 1 + [-range to range] */
void scatter_weights(double range)
{
    int i, j, p;

    /* adds 1 to seed to start at a different point in the random sequence */
    /* from read_weights. */
    for (i = 0; i < 1 + seed; i++)
        drand48();

    for (i = 0; i < numinp1; i++)
        for (j = 0; j < numrho; j++) {
            link_r1_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
            link_t1_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#if HEADWORDS
            link_h1_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#endif
        }
    for (i = 0; i < numinp2; i++)
        for (j = 0; j < numrho; j++) {
            link_r2_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
            link_t2_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#if HEADWORDS
            link_h2_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#endif
        }
#if ! LABELCONTEXTINP
    for (i = 0; i < numrho; i++)
        for (j = 0; j < numrho; j++)
            link_r_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#endif
    for (j = 0; j < numrho; j++)
        link_bir_r[j] *= 1.0 + range * 2.0 * drand48() - range;
    for (i = 0; i < numlink_tro; i++)
        for (j = 0; j < numrho; j++)
            link_tro_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
    for (i = 0; i < numlink_trp; i++)
        for (j = 0; j < numrho; j++)
            link_trp_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
    for (i = 0; i < numlink_trc; i++)
        for (j = 0; j < numrho; j++)
            link_trc_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
    for (i = 0; i < numlink_lab; i++)
        for (j = 0; j < numrho; j++)
            link_lab_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
    for (j = 0; j < numrho; j++)
        link_pred_r[i] *= 1.0 + range * 2.0 * drand48() - range;
#if CHILDCONTEXT
    for (i = 0; i < numlink_chd; i++)
        for (j = 0; j < numrho; j++)
            link_chd_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#endif
#if HEADHIDDEN
    for (i = 0; i < numlink_hdc; i++)
        for (j = 0; j < numrho; j++)
            link_hdc_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#endif
#if CHOPCONTEXT
    for (i = 0; i < numlink_cho; i++)
        for (j = 0; j < numrho; j++)
            link_cho_r[j][i] *= 1.0 + range * 2.0 * drand48() - range;
#endif
#if HEADOUTS
    for (j = 0; j < numrho; j++)
        link_nhd_r[j] *= 1.0 + range * 2.0 * drand48() - range;
#endif
    for (i = 0; i < numrho; i++)
        for (j = 0; j < numout; j++)
            link_r_o[j][i] *= 1.0 + range * 2.0 * drand48() - range;
    for (i = 0; i < numrho; i++)
        bias_r[i] *= 1.0 + range * 2.0 * drand48() - range;
    for (i = 0; i < numout; i++)
        bias_o[i] *= 1.0 + range * 2.0 * drand48() - range;
    for (i = 0; i < numrho; i++)
        hidrho_initstate.act[i] *= 1.0 + range * 2.0 * drand48() - range;

    /* zero deltas */
    for (i = 0; i < numrho; i++)
        hidrho_initstate.err[i] = 0.0;
    for (i = 0; i < numinp1; i++) {
        for (j = 0; j < numrho; j++) {
            linkdel_r1_r[j][i] = 0.0;
            linkdel_t1_r[j][i] = 0.0;
#if HEADWORDS
            linkdel_h1_r[j][i] = 0.0;
#endif
        }
    }
    for (i = 0; i < numinp2; i++) {
        for (j = 0; j < numrho; j++) {
            linkdel_r2_r[j][i] = 0.0;
            linkdel_t2_r[j][i] = 0.0;
#if HEADWORDS
            linkdel_h2_r[j][i] = 0.0;
#endif
        }
    }
    for (i = 0; i < numrho; i++) {
        biasdel_r[i] = 0.0;
#if ! LABELCONTEXTINP
        for (j = 0; j < numrho; j++)
            linkdel_r_r[j][i] = 0.0;
#endif
        for (j = 0; j < numout; j++)
            linkdel_r_o[j][i] = 0.0;
    }
    for (j = 0; j < numrho; j++)
        linkdel_bir_r[j] = 0.0;
    for (i = 0; i < numlink_tro; i++)
        for (j = 0; j < numrho; j++)
            linkdel_tro_r[j][i] = 0.0;
    for (i = 0; i < numlink_trp; i++)
        for (j = 0; j < numrho; j++)
            linkdel_trp_r[j][i] = 0.0;
    for (i = 0; i < numlink_trc; i++)
        for (j = 0; j < numrho; j++)
            linkdel_trc_r[j][i] = 0.0;
    for (i = 0; i < numlink_lab; i++)
        for (j = 0; j < numrho; j++)
            linkdel_lab_r[j][i] = 0.0;
    for (j = 0; j < numrho; j++)
        linkdel_pred_r[i] = 0.0;
#if CHILDCONTEXT
    for (i = 0; i < numlink_chd; i++)
        for (j = 0; j < numrho; j++)
            linkdel_chd_r[j][i] = 0.0;
#endif
#if HEADHIDDEN
    for (i = 0; i < numlink_hdc; i++)
        for (j = 0; j < numrho; j++)
            linkdel_hdc_r[j][i] = 0.0;
#endif
#if CHOPCONTEXT
    for (i = 0; i < numlink_cho; i++)
        for (j = 0; j < numrho; j++)
            linkdel_cho_r[j][i] = 0.0;
#endif
#if HEADOUTS
    for (j = 0; j < numrho; j++)
        linkdel_nhd_r[j] = 0.0;
#endif
    for (i = 0; i < numout; i++)
        biasdel_o[i] = 0.0;

}



/* set up the network data structures from inffile */
void initialize()
{
    FILE *fp;
    char buffer[MAXINFLINE];
    char *str;
    int t, b;
    int i, j;
    int ival;
    double val;
    int inp1, out1;
    int wrd, numwrd;
    int lab;
    int bin;
    int nb;
    int offset;
    int prbm;
    int sum;
    int max_branch_fact, max_cluster_qty, branch_fact;
    struct clustering *swap_clustering;



    if (testing_predbeam >= training_predbeam)
        prbm = testing_predbeam;
    else
        prbm = training_predbeam;

    /* read inf file */
    if ((fp = fopen(inffile, "r")) == NULL)
        error("Error: Cannot open .inf file.");

    /* learning rate */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [eta]");
    /* skip over network type, if present */
    if (str[0] >= 'A' && str[0] <= 'C') {
printf("network type being ignored\n");
        if (strcmp(str, "Ar") != 0)
	    error("Error: .inf file is not for a reduced type A network.");
	fgets(buffer, MAXINFLINE, fp);
	if ((str = strtok(buffer, ", \t\n")) == NULL)
	    error("Error: Insufficient input in .inf. [eta]");
    }
    val = atof(str);
    if ((val < 0) || (val > 1)) {
        sprintf(warning_text,
                "learning rate out of range; using %f.\n",
                eta_inf * eta_red);
        warning(warning_text);
    } else
        eta_inf = val;
    eta = eta_inf * eta_red;
    decay = 1.0 - reg * eta;
    /* momentum */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [momentum]");
    val = atof(str);
    if ((val < 0) || (val > 1)) {
        sprintf(warning_text, "momentum out of range; using %f.\n", mom);
        warning(warning_text);
    } else
        mom = val;
    /* regularization */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [reg]");
    val = atof(str);
    /* skip over tolerance if present (assumes reg is last before blank) */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) != NULL) {
        val = atof(str);
	/* blank */
	fgets(buffer, MAXINFLINE, fp);
	if (strtok(buffer, ", \t\n") != NULL) {
	    sprintf(warning_text,
		    "blank line expected before hidden layer size(s).\n");
	    warning(warning_text);
	}
    } /* else blank */
    if ((val - reg_red < 0) || (val - reg_red > 0.9))
        error("Error: Regularization out of range (or not present).\n");
    else
        reg_inf = val;
    reg = reg_inf - reg_red;
    decay = 1.0 - reg * eta;

    /* blank handled in reg */

    /* number rho hidden */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [numrho]");
    ival = atoi(str);
    if ((ival < 1) || (ival > MAXNUMRHO)) {
        sprintf(warning_text,
                "numrho=%d out of range; using %d.\n", ival, numrho);
        warning(warning_text);
    } else
        numrho = ival;

    /* blank */
    fgets(buffer, MAXINFLINE, fp);
    if (strtok(buffer, ", \t\n") != NULL) {
        sprintf(warning_text,
                "blank line expected before output label specs.\n");
        warning(warning_text);
    }
    /* number nonterminal labels */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [numlabels]");
    ival = atoi(str);
    if ((ival < 1) || (ival > MAXNUMTARMEM - NUMSTRUC)) {
        sprintf(warning_text,
                "numlabels=%d out of range; using %d.\n", ival, numlabels);
        warning(warning_text);
    } else
        numlabels = ival;
    /* number of input tags */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [numinp1]");
    ival = atoi(str);
    if ((ival < 1) || (ival > MAXNUMINP1)) {
        sprintf(warning_text,
                "numinp1=%d out of range; using %d.\n", ival, numinp1);
        warning(warning_text);
    } else
        numinp1 = ival;
#if NOPREDICT
    numout1 = 0;
#else
    numout1 = numinp1;
#endif
#if HEADOUTS
    numtarmem = (2 * numlabels) + NUMSTRUC + numout1;
#else
    numtarmem = numlabels + NUMSTRUC + numout1;
#endif
    if (numtarmem > MAXNUMTARMEM)
        error
            ("Error: Too many output units specifying structure and prediction outputs.");
#if HEADOUTS
    hdlabeloutputs = labeloutputs + numlabels;
    tagoutputs = hdlabeloutputs + numlabels;
#else
    tagoutputs = labeloutputs + numlabels;
#endif
    wordoutputs = tagoutputs + numout1;


    /* blank */
    fgets(buffer, MAXINFLINE, fp);
    if (strtok(buffer, ", \t\n") != NULL) {
        sprintf(warning_text,
                "blank line expected before input word specs.\n");
        warning(warning_text);
    }
    /* input word position */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [inpwrdpos]");
    ival = atoi(str);
    if (ival < 0) {
        sprintf(warning_text,
                "inpwrdpos out of range; using %d.\n", inpwrdpos);
        warning(warning_text);
    } else
        inpwrdpos = ival;
    /* number tag input words */
    fgets(buffer, MAXINFLINE, fp);
    offset = 0;
    if (inpwrdpos == 0) {
        for (inp1 = 0; inp1 < numinp1; inp1++) {
            inp1_first_word[inp1] = offset;
        }
    } else {
        str = strtok(buffer, ", \t\n");
        for (inp1 = 0; inp1 < numinp1; inp1++) {
            inp1_first_word[inp1] = offset;
            if (str == NULL)
                error
                    ("Error: Insufficient input in .inf. [inp1_first_word[inp1]]");
            ival = atoi(str);
            if (ival <= 0) {
                ival = 1;
                sprintf(warning_text,
                        "number of tag input words out of range; using %d.\n",
                        ival);
                warning(warning_text);
            }
            offset += ival;
            str = strtok(NULL, ", \t\n");
        }
    }
    inp1_first_word[numinp1] = offset;  /* set inp1+1 to last+1 */
    numinp2 = offset;

    /* blank */
    fgets(buffer, MAXINFLINE, fp);
    if (strtok(buffer, ", \t\n") != NULL) {
        sprintf(warning_text,
                "blank line expected before output word specs.\n");
        warning(warning_text);
    }
    /* output word position OPTIONAL */
    numpredwords = 0;

    out_clusters_arr =
        (struct clustering **) malloc(numout1 *
                                      sizeof(struct clustering *));
    CHECK_PTR(out_clusters_arr);

    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) != NULL) {
        ival = atoi(str);
        if (ival < 0) {
            sprintf(warning_text,
                    "outwrdpos out of range; using %d.\n", outwrdpos);
            warning(warning_text);
        } else {
            outwrdpos = ival;
        }


        if ((str = strtok(NULL, ", \t\n")) != NULL) {

            if (strcmp(str, "clustered") == 0) {
                /* assume reading clustered configuration */
                for (out1 = 0; out1 < numout1; out1++) {
                    fgets(buffer, MAXINFLINE, fp);
                    if ((out_clusters_arr[out1] =
                         cl_read_clustering(buffer, ", \t\n")) == NULL) {
                        error("Error when reading clustering from .inf.");
                    }
                }
            } else {
                printf("%s\n", str);
                error("Unexpected token in output configuration in .inf.");
            }

        } else {

            fgets(buffer, MAXINFLINE, fp);
            str = strtok(buffer, ", \t\n");
            for (out1 = 0; out1 < numout1; out1++) {
                if (str == NULL) {
                    error("Error: Insufficient input in .inf.");
                }

                ival = atoi(str);
                if (ival <= 0) {
                    ival = 1;
                    sprintf(warning_text,
                            "number of tag input words out of range; using %d.\n",
                            ival);
                    warning(warning_text);
                }
                if ((out_clusters_arr[out1] =
                     cl_create_one_level(ival)) == NULL) {
                    error
                        ("Error when creating clustering from simple format for outputs in .inf.");
                }

                str = strtok(NULL, ", \t\n");

            }                   /* eo if in simple format */
        }
        /* if no explicit definition of output configuration */
    } else {
        outwrdpos = inpwrdpos;

        for (out1 = 0; out1 < numout1; out1++) {
            out_clusters_arr[out1] =
                cl_create_one_level(inp1_first_word[out1 + 1] -
                                    inp1_first_word[out1]);
            if (out_clusters_arr[out1] == NULL) {
                error
                    ("Error when creating clustering from simple format for outputs in .inf.");
            }
        }

    }



    numpredwords = 0;
    max_branch_fact = 0;
    max_cluster_qty = 0;

#if ! NOPREDICT
    for (out1 = 0; out1 < numout1; out1++) {
        out1_first_word[out1] = numpredwords;
        numpredwords += out_clusters_arr[out1]->cluster_qty;

        branch_fact = cl_get_max_branch_fact(out_clusters_arr[out1]);
        if (branch_fact > max_branch_fact) {
            max_branch_fact = branch_fact;
        }
        if (out_clusters_arr[out1]->cluster_qty > max_cluster_qty) {
            max_cluster_qty = out_clusters_arr[out1]->cluster_qty;
        }
    }
#endif
    numout = numtarmem + numpredwords;

/*#if !EM_RESERVED_WORD_SPACE */
    numoutmem = numtarmem + max_cluster_qty;
/*#else
    if (max_cluster_qty > EM_RESERVED_WORD_SPACE) {
        fprintf(stderr, "Requested %d cluster outputs\n");
        error("Cluster output quantity is larger than EM_RESERVED_WORD_SPACE, adjust EM_RESERVED_WORD_SPACE");
    }
    numoutmem = numoutmem + EM_RESERVED_WORD_SPACE;
#endif */

    if (numoutmem > MAXNUMOUTMEM) {
        printf("Error: numoutmem=%d out of range (>%d).",
               numoutmem, MAXNUMOUTMEM);
        error("");
    }
    if (numout * 14 > MAXWGTLINE) {
        error("MAXWGTLINE not large enough for output.");
    }
    if (numout > MAXNUMOUT) {
        fprintf(stderr, "requested numout = %d, MAXNUMOUT = %d\n", numout,
                MAXNUMOUT);
        error("MAXNUMOUT not large enough");
    }
#if INITOUTPUTBIAS
  /* blank */
  fgets(buffer, MAXINFLINE, fp);
  if (strtok(buffer, ", \t\n") != NULL)
    {
      sprintf(warning_text,
	      "blank line expected before output count specs.\n"); 
      warning(warning_text);
    }
  /* structure output counts */
  /* biases reset later if INITOUTPUTBIAS=0 */
  structypecount = 0;
  fgets(buffer, MAXINFLINE, fp);
  if ((str=strtok(buffer, ", \t\n")) == NULL)
    error("Error: Insufficient input in .inf. [PREDOUT count]");
  ival = atoi(str);
  if (ival <= 0)
    {
      if (ival < 0) {
	sprintf(warning_text,
		"PREDOUT count out of range; using 1.\n");
	warning(warning_text);
      }
      ival = 1;
    }
  structypecount += ival;
  relregtags = (double) ival;
  bias_o[PREDOUT] = log((double) ival);
  if ((str=strtok(NULL, ", \t\n")) == NULL)
    error("Error: Insufficient input in .inf. [PROJOUT count]");
  ival = atoi(str);
  if (ival <= 0)
    {
      if (ival < 0) {
	sprintf(warning_text,
		"PROJOUT count out of range; using 1.\n");
	warning(warning_text);
      }
      ival = 1;
    }
  structypecount += ival;
  relreglabels = (double) ival;
  bias_o[PROJOUT] = log((double) ival);
  if ((str=strtok(NULL, ", \t\n")) == NULL)
    error("Error: Insufficient input in .inf. [ATTOUT count]");
  ival = atoi(str);
  if (ival <= 0)
    {
      if (ival < 0) {
	sprintf(warning_text,
		"ATTOUT count out of range; using 1.\n");
	warning(warning_text);
      }
      ival = 1;
    }
  structypecount += ival;
  bias_o[ATTOUT] = log((double) ival);
  if ((str=strtok(NULL, ", \t\n")) == NULL)
    error("Error: Insufficient input in .inf. [MODOUT count]");
  ival = atoi(str);
  if (ival <= 0)
    {
      if (ival < 0) {
	sprintf(warning_text,
		"MODOUT count out of range; using 1.\n");
	warning(warning_text);
      }
      ival = 1;
    }
  structypecount += ival;
  bias_o[MODOUT] = log((double) ival);
  /* divide by total */
  relregtags /= (double) structypecount;
  relreglabels /= (double) structypecount;
  /* divide by average */
  val = log((double) NUMSTRUC) - log((double) structypecount); 
  bias_o[PREDOUT] += val;
  bias_o[PROJOUT] += val;
  bias_o[ATTOUT] += val;
  bias_o[MODOUT] += val;
  /* tag output counts */
  fgets(buffer, MAXINFLINE, fp);
  if ((str=strtok(buffer, ", \t\n")) == NULL)
    error("Error: Insufficient input in .inf. [tag counts]");
  sum = 0;
  for (out1 = 0; out1 < numout1; out1++)
    {
      ival=atoi(str);
      if (ival <= 0)
	{
	  if (ival < 0) {
	    sprintf(warning_text,
		    "tag count for %d out of range; using 1.\n",
		    out1);
	    warning(warning_text);
	  }
	  ival = 1;
	}
      sum += ival;
      relregwords[out1] = (double) ival / (double) structypecount;
      bias_o[predtag_to_output(out1)] = log((double) ival);
      if ((str=strtok(NULL, ", \t\n")) == NULL && out1 != numout1 - 1)
	{
	  warning("Insufficient input in .inf. [tag counts]");
	  for (out1++; out1 < numout1; out1++) {
	    ival = 1;
	    sum += ival;
	    relregwords[out1] = (double) ival / (double) structypecount;
	    bias_o[predtag_to_output(out1)] = log((double) ival);
	  }
	  break;
	}
    }
  val = log((double) numout1) - log((double) sum); 
  for (out1 = 0; out1 < numout1; out1++)
    bias_o[predtag_to_output(out1)] += val;
  /* label output counts */
  fgets(buffer, MAXINFLINE, fp);
  if ((str=strtok(buffer, ", \t\n")) == NULL)
    error("Error: Insufficient input in .inf. [label counts]");
  sum = 0;
  for (lab = 0; lab < numlabels; lab++)
    {
      ival=atoi(str);
      if (ival <= 0)
	{
	  if (ival < 0) {
	    sprintf(warning_text,
		    "label count for %d out of range; using 1.\n",
		    lab);
	    warning(warning_text);
	  }
	  ival = 1;
	}
      sum += ival;
      bias_o[projlabel_to_output(lab)] = log((double) ival);
      if ((str=strtok(NULL, ", \t\n")) == NULL && lab != numlabels - 1)
	{
	  warning("Insufficient input in .inf. [label counts]");
	  for (lab++; lab < numlabels; lab++) {
	    ival = 1;
	    sum += ival;
	    bias_o[projlabel_to_output(lab)] = log((double) ival);
	  }
	  break;
	}
    }
  val = log((double) numlabels) - log((double) sum); 
  for (lab = 0; lab < numlabels; lab++)
    bias_o[projlabel_to_output(lab)] += val;
  /* word output counts */
  for (out1 = 0; out1 < numout1; out1++)
    {
      fgets(buffer, MAXINFLINE, fp);
      numwrd = out1_first_word[out1 + 1] - out1_first_word[out1];
      sum = 0;
      if ((str=strtok(buffer, ", \t\n")) == NULL) {
	warning("Insufficient input in .inf. [word counts]");
	for (wrd = 0; wrd < numwrd; wrd++) {
	  ival = 1;
	  sum += ival;
	  bias_o[predword_to_output(out1_first_word[out1] + wrd)] = 
	    log((double) ival);
	}
      }
      else {
	for (wrd = 0; wrd < numwrd; wrd++)
	  {
	    ival=atoi(str);
	    if (ival <= 0)
	      {
		if (ival < 0) {
		  sprintf(warning_text,
			  "word count for %d out of range; using 1.\n",
			  wrd);
		  warning(warning_text);
		}
		ival = 1;
	      }
	    sum += ival;
	    bias_o[predword_to_output(out1_first_word[out1] + wrd)] = 
	      log((double) ival);
	    if ((str=strtok(NULL, ", \t\n")) == NULL && wrd != numwrd - 1)
	      {
		warning("Insufficient input in .inf. [word counts]");
		for (wrd++; wrd < numwrd; wrd++) {
		  ival = 1;
		  sum += ival;
		  bias_o[predword_to_output(out1_first_word[out1] + wrd)] = 
		    log((double) ival);
		}
		break;
	      }
	  }
	val = log((double) numwrd) - log((double) sum); 
	for (wrd = 0; wrd < numwrd; wrd++)
	  bias_o[predword_to_output(out1_first_word[out1] + wrd)] += val;
      }
    }
#endif
#if DISPMEM
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp,
                "Required array space:  numout= %d, numoutmem= %d, numtarmem= %d=%d.\n",
                numout, numoutmem, numtarmem, predword_to_output(0));
        fclose(warnfp);
    } else
#endif
        printf
            ("Required array space:  numout= %d, numoutmem= %d, numtarmem= %d=%d.\n",
             numout, numoutmem, numtarmem, predword_to_output(0));
#endif

#if DISPINF
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp,
                "numoutputs = %d\nnumlabels = %d\nnuminp1 = %d\nnumoutputsmemory = %d\n",
                numout, numlabels, numinp1, numoutmem);
        fclose(warnfp);
    } else
#endif
        printf
            ("numoutputs = %d\nnumlabels = %d\nnuminp1 = %d\nnumoutputsmemory = %d\n",
             numout, numlabels, numinp1, numoutmem);
#if ! NOPREDICT && DISPMEM
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp, "each tag's first tag-word input id:\n");
        for (inp1 = 0; inp1 <= numinp1; inp1++) {
            fprintf(warnfp, " %d", inp1_first_word[inp1]);
        }
        fprintf(warnfp, "\n");
        fprintf(warnfp, "each tag's first tag-word output:\n");
        for (out1 = 0; out1 <= numout1; out1++) {
            fprintf(warnfp, " %d", wordoutputs + out1_first_word[out1]);
        }
        fprintf(warnfp, "\n");
        fprintf(warnfp, "numpredwords = %d\n", numpredwords);
        fclose(warnfp);
    } else
#endif
    {
        printf("each tag's first tag-word input id:\n");
        for (inp1 = 0; inp1 <= numinp1; inp1++) {
            printf(" %d", inp1_first_word[inp1]);
        }
        printf("\n");
        printf("each tag's first tag-word output:\n");
        for (out1 = 0; out1 <= numout1; out1++) {
            printf(" %d", wordoutputs + out1_first_word[out1]);
        }
        printf("\n");
        printf("numpredwords = %d\n", numpredwords);
    }
#endif
#endif

    fclose(fp);

/* initialize link numbers */
#if LABELTRIGPARENTINP
    numlink_trp = numlabels;
#else
    numlink_trp = numrho;
#endif
#if LABELTRIGCHILDINP
    numlink_trc = numlabels + numinp1;
#else
    numlink_trc = numrho;
#endif
    numlink_lab = numlabels + numinp1;
#if CHILDCONTEXT
#if LABELCHILDINP
    numlink_chd = numlabels + numinp1;
#else
    numlink_chd = numrho;
#endif
#endif
#if HEADHIDDEN
#if LABELHEADCHILDINP
    numlink_hdc = numlabels + numinp1;
#else
    numlink_hdc = numrho;
#endif
#endif

    /* initialize nullstate (never changed) */
    for (i = 0; i < numrho; i++) {
        hidrho_nullstate.act[i] = 0.0;
        hidrho_nullstate.err[i] = 0.0;
    }

    /* allocate arrays for phase control structures */
    best_options = (struct option ***)
        malloc(sizeof(struct option **[ALLOCSIZE + 1]));
    if (best_options == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
    for (i = 0; i < ALLOCSIZE + 1; i++) {
        best_options[i] = (struct option **)
            malloc(sizeof(struct option *[MAXPREDBEAM]));
        if (best_options[i] == NULL)
            error("Error: Cannot allocate enough space to begin.\n");
    }
    sentence_words = (char **)
        malloc(sizeof(char *[ALLOCSIZE + 1]));
    if (sentence_words == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
    for (i = 0; i < ALLOCSIZE + 1; i++) {
        sentence_words[i] =
            (char *) malloc(sizeof(char[MAXWORDLENGTH + 1]));
        if (sentence_words[i] == NULL)
            error("Error: Cannot allocate enough space to begin.\n");
    }
    sentence_tagwordids = (char **)
        malloc(sizeof(char *[ALLOCSIZE + 1]));
    if (sentence_tagwordids == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
    for (i = 0; i < ALLOCSIZE + 1; i++) {
        sentence_tagwordids[i] =
            (char *) malloc(sizeof(char[MAXWORDLENGTH + 1]));
        if (sentence_tagwordids[i] == NULL)
            error("Error: Cannot allocate enough space to begin.\n");
    }
    free_phases = (struct stack *)
        malloc(sizeof(struct stack));
    if (free_phases == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
    free_phases->content = (int *)
        malloc(sizeof(int[ALLOCSIZE]));
    if (free_phases->content == NULL)
        error("Error: Cannot allocate enough space to begin.\n");

    /* allocate arrays for each layer     */
    input_hist = (struct input *)
        malloc(sizeof(struct input[ALLOCSIZE + 1]));
    if (input_hist == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
    hidrho_constant_net = (struct hidrho *)
        malloc(sizeof(struct hidrho[ALLOCSIZE + 1]));
    if (hidrho_constant_net == NULL)
        error("Error: Cannot allocate enough space to begin.\n");

    /* allocate arrays for output interpretation */
    target_derivation = create_option_stack(NULL, NULL, NULL);
    target_derivation_next = target_derivation;
    training_options = create_option_stack(NULL, NULL, NULL);
    training_options_next = training_options;
    scratch_options = create_option_stack(NULL, NULL, NULL);
    scratch_options_next = scratch_options;

    /* allocate arrays for calculating precision-recall */
#if ! LOCALESTIMATE
    ave_finish_probs = (double *)
        malloc(sizeof(double[ALLOCSIZE + 1]));
    if (ave_finish_probs == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
    total_finish_probs = (double *)
        malloc(sizeof(double[ALLOCSIZE + 1]));
    if (total_finish_probs == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
    total_finish_prob_weights = (double *)
        malloc(sizeof(double[ALLOCSIZE + 1]));
    if (total_finish_prob_weights == NULL)
        error("Error: Cannot allocate enough space to begin.\n");
#endif

    sequence_space = ALLOCSIZE;
    phase_space = ALLOCSIZE;
    state_space = ALLOCSIZE;
    output_space = ALLOCSIZE;


}


/* read in any new values for mom, or eta from inffile */
void read_inf_changes()
{
    FILE *fp;
    FILE *ofp;
    char buffer[MAXINFLINE];
    char *str;
    int t;
    int ival;
    double val;

    if ((fp = fopen(inffile, "r")) == NULL) {
        save_weights(newweightfile, 0.0);
        error("Error: Cannot open .inf file.");
    }

    /* learning rate */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [eta]");
    /* skip over network type, if present */
    if (str[0] >= 'A' && str[0] <= 'C') {
        if (strcmp(str, "Ar") != 0)
	    error("Error: .inf file is not for a reduced type A network.");
	fgets(buffer, MAXINFLINE, fp);
	if ((str = strtok(buffer, ", \t\n")) == NULL)
	    error("Error: Insufficient input in .inf. [eta]");
    }
    val = atof(str);
    if (val != eta_inf) {
        if ((val < 0) || (val > 1)) {
            sprintf(warning_text,
                    "new learning rate out of range; using %f.\n", eta);
            warning(warning_text);
        } else {
            eta_inf = val;
            eta = eta_inf * eta_red;
            decay = 1.0 - reg * eta;
            if ((ofp = fopen(outfile, "a")) != NULL) {
                fprintf(ofp,
                        "New value in .inf file at cycle %d:  Eta = %.2e.\n",
                        weight_age, eta);
                fclose(ofp);
            }
        }
    }
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [momentum]");
    val = atof(str);
    if (val != mom) {
        if ((val < 0) || (val > 1)) {
            sprintf(warning_text,
                    "new momentum out of range; using %f.\n", mom);
            warning(warning_text);
        } else {
            mom = val;
            if ((ofp = fopen(outfile, "a")) != NULL) {
                fprintf(ofp,
                        "New value in .inf file at cycle %d:  Mom = %.2f.\n",
                        weight_age, mom);
                fclose(ofp);
            }
        }
    }
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) == NULL)
        error("Error: Insufficient input in .inf. [reg]");
    val = atof(str);
    /* skip over tolerance if present (assumes reg is last before blank) */
    fgets(buffer, MAXINFLINE, fp);
    if ((str = strtok(buffer, ", \t\n")) != NULL) {
        val = atof(str);
    } 
    if (val != reg_inf) {
        if ((val - reg_red < 0) || (val - reg_red > 0.9)) {
            sprintf(warning_text,
                    "new regularization out of range; using %f.\n", reg);
            warning(warning_text);
        } else {
            reg_inf = val;
            reg = reg_inf - reg_red;
            if ((ofp = fopen(outfile, "a")) != NULL) {
                fprintf(ofp,
                        "New value in .inf file at cycle %d:  Reg = %.1e.\n",
                        weight_age, reg);
                fclose(ofp);
            }
        }
    }
    decay = 1.0 - reg * eta;

    fclose(fp);
}


/* allocate new space for longer sequences */
void expand_periods()
{
    int t, b;
    int i;
    int relc, reln;
    int ps;
    int prbm;

    if (testing_predbeam >= training_predbeam)
        prbm = testing_predbeam;
    else
        prbm = training_predbeam;

#if DISPMEM
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp, "Expanding arrays to length %d.  [sent:%d]\n",
                sequence_space + ALLOCSIZE, num_sent + 1);
        fclose(warnfp);
    } else
#endif
        printf("Expanding arrays to length %d.  [sent:%d]\n",
               sequence_space + ALLOCSIZE, num_sent + 1);
#endif

    /* expand array for phase control structures */
    best_options = (struct option ***)
        realloc(best_options,
                sizeof(struct option **[sequence_space + ALLOCSIZE + 1]));
    if (best_options == NULL)
        error
            ("Error: Cannot allocate enough space for max sequence length.\n");
    for (i = sequence_space + 1; i < sequence_space + ALLOCSIZE + 1; i++) {
        best_options[i] = (struct option **)
            malloc(sizeof(struct option *[MAXPREDBEAM]));
        if (best_options[i] == NULL)
            error
                ("Error: Cannot allocate enough space for max sequence length.\n");
    }
    sentence_words = (char **)
        realloc(sentence_words,
                sizeof(char *[sequence_space + ALLOCSIZE + 1]));
    if (sentence_words == NULL)
        error
            ("Error: Cannot allocate enough space for max sequence length.\n");
    for (i = sequence_space + 1; i < sequence_space + ALLOCSIZE + 1; i++) {
        sentence_words[i] =
            (char *) malloc(sizeof(char[MAXWORDLENGTH + 1]));
        if (sentence_words[i] == NULL)
            error
                ("Error: Cannot allocate enough space for max sequence length.\n");
    }
    sentence_tagwordids = (char **)
        realloc(sentence_tagwordids,
                sizeof(char *[sequence_space + ALLOCSIZE + 1]));
    if (sentence_tagwordids == NULL)
        error
            ("Error: Cannot allocate enough space for max sequence length.\n");
    for (i = sequence_space + 1; i < sequence_space + ALLOCSIZE + 1; i++) {
        sentence_tagwordids[i] =
            (char *) malloc(sizeof(char[MAXWORDLENGTH + 1]));
        if (sentence_tagwordids[i] == NULL)
            error
                ("Error: Cannot allocate enough space for max sequence length.\n");
    }


    /* expand arrays for each layer     */
    input_hist = (struct input *)
        realloc(input_hist,
                sizeof(struct input[sequence_space + ALLOCSIZE + 1]));
    if (input_hist == NULL)
        error
            ("Error: Cannot allocate enough space for max sequence length.\n");


    /* expand arrays for calculating precision-recall.  content is lost */
#if ! LOCALESTIMATE
#if ! NOFREEING
    free(ave_finish_probs);
#endif
    ave_finish_probs = (double *)
        malloc(sizeof(double[sequence_space + ALLOCSIZE + 1]));
    if (ave_finish_probs == NULL)
        error
            ("Error: Cannot allocate enough space for max sequence length.\n");
#if ! NOFREEING
    free(total_finish_probs);
#endif
    total_finish_probs = (double *)
        malloc(sizeof(double[sequence_space + ALLOCSIZE + 1]));
    if (total_finish_probs == NULL)
        error
            ("Error: Cannot allocate enough space for max sequence length.\n");
#if ! NOFREEING
    free(total_finish_prob_weights);
#endif
    total_finish_prob_weights = (double *)
        malloc(sizeof(double[sequence_space + ALLOCSIZE + 1]));
    if (total_finish_prob_weights == NULL)
        error
            ("Error: Cannot allocate enough space for max sequence length.\n");
#endif

    sequence_space += ALLOCSIZE;
}


/* allocate new space for more phases.  birth is period where new phases are 
     needed.  birth=-1 means period dependent and combin/output arrays don't
     need expanding.   Return flag for success. */
int expand_phases(int birth)
{
    int t, p, b;
    int i;
    int relc, reln;
    int temp;
    int psb;
    int psm;

    psm = phase_space;
        psb = psm;
    if (psb + ALLOCSIZE > MAXPHASES) {
        sprintf(warning_text,
                "blocking allocation of more phases beyond %d.", psb);
        warning(warning_text);
        return (0);
    }

#if DISPMEM
#if FILEDISP
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp,
                "Expanding arrays for time %d to %d phases.  [sent:%d]\n",
                birth, psb + ALLOCSIZE, num_sent + 1);
        fclose(warnfp);
    } else
#endif
        printf("Expanding arrays for time %d to %d phases.  [sent:%d]\n",
               birth, psb + ALLOCSIZE, num_sent + 1);
#endif

    /* expand array for phase control structures */

    /* expand arrays for each layer     */
    if (psb + ALLOCSIZE > psm) {
        temp = (int) hidrho_constant_net;
        hidrho_constant_net = (struct hidrho *)
            realloc(hidrho_constant_net,
                    sizeof(struct hidrho[psb + ALLOCSIZE + 1]));
        if (hidrho_constant_net == NULL) {
            hidrho_constant_net = (struct hidrho *) temp;
            warning("Cannot allocate the desired number of phases. (7)");
            return (0);
        }
    }

    /* only if everything else succeeds, expand free phases */
    if (psb + ALLOCSIZE > psm) {
        realloc_stack_content(free_phases, psb + ALLOCSIZE);
    }

    if (psb + ALLOCSIZE > phase_space)
        phase_space = psb + ALLOCSIZE;
    if (birth >= 0 && psb + ALLOCSIZE > output_space)
        output_space = psb + ALLOCSIZE;

    return (1);
}



/* allocate new space for more outputs (and combination layers).  Return
     flag for success. */
/* not used */
int expand_outputs()
{
    struct output *temp;

    printf("Outputs expanded %d\n", num_sent);
    temp = output_hist;
    output_hist = (struct output *)
        realloc(output_hist,
                sizeof(struct output[output_space + ALLOCSIZE]));
    if (output_hist == NULL) {
        output_hist = temp;
        warning("Cannot allocate the desired number of outputs.");
        return (0);
    }

    output_space += ALLOCSIZE;

    return (1);
}




/* write to stream fp, the options starting from root which are in the
     next_options. */
void write_derivations(struct option *root, FILE * fp)
{
    struct option *opt;
    struct option_stack_stack *opt_stacks;

    opt_stacks = NULL;
    opt = root;

    while (1) {
        while (1) {

            write_option_spec(opt, fp);
            if (opt->next_options == NULL) {
                fprintf(fp, "=%e ", opt->lprob);
            }

            if (opt->next_options != NULL
                && opt->next_options->next != NULL) {
                fprintf(fp, "{ ");
                opt_stacks =
                    create_option_stack_stack(opt->next_options->next,
                                              opt_stacks, NULL);
            }
            if (opt->next_options != NULL)
                opt = opt->next_options->elem;
            else
                break;
        }
        while (opt_stacks != NULL && opt_stacks->elem == NULL) {
            fprintf(fp, "} ");
            opt_stacks = popfree_option_stack_stack(opt_stacks);
        }
        if (opt_stacks != NULL) {
            fprintf(fp, ";\n");
            opt = opt_stacks->elem->elem;
            opt_stacks->elem = opt_stacks->elem->next;
        } else
            break;
    }
    fprintf(fp, ".\n\n");
}

/* write input format for opt */
void write_option_spec(struct option *opt, FILE * fp)
{
    int bin;
    int position;
    int inp1;
    int *inp2;

    if (opt->previous_output == PREDOUT) {
        position = opt->period + 1;
	fprintf(fp, "> %s %s\n", 
		sentence_tagwordids[position], sentence_words[position]);
    } else if (opt->previous_output == PROJOUT) {
        fprintf(fp, "| ");
    }
#if HEADOUTS
    else if (opt->previous_output == HDPROUT) {
        fprintf(fp, "|h ");
    }
#endif
    else if (opt->previous_output >= projlabel_to_output(0) &&
             opt->previous_output <= projlabel_to_output(numlabels - 1)) {
        fprintf(fp, "%d\t", output_to_projlabel(opt->previous_output));
    }
#if HEADOUTS
    else if (opt->previous_output >= hdprlabel_to_output(0) &&
             opt->previous_output <= hdprlabel_to_output(numlabels - 1)) {
        fprintf(fp, "%d\t", output_to_hdprlabel(opt->previous_output));
    }
#endif
    else if (opt->previous_output == ATTOUT) {
        fprintf(fp, "^\t");
    }
#if HEADOUTS
    else if (opt->previous_output == HDATOUT) {
        fprintf(fp, "^h\t");
    }
#endif
    else if (opt->previous_output == MODOUT) {
        fprintf(fp, "+\t");
    } else {
        /* word prediction options ignored */
    }
}


/* display all the branches in root */
void write_tree(struct tree *root, char *name, char *label)
{
    FILE *wfp;

    if (stdout_bracketed_output == 0) {
        wfp = fopen(name, "a");
    }
    else {
        wfp = stdout;
    }
    if (wfp != NULL) {
        if (label != NULL)
            fprintf(wfp, label);
        if (root != NULL)
	    write_subtree(root, root->start, wfp);
        fprintf(wfp, "\n");
	if (stdout_bracketed_output == 0) {
	    fclose(wfp);
	}
    } else
        warning("Couldn't append to bracketed string file.");
}

/* returns end position of subtree */
int write_subtree(struct tree *root, int position, FILE * wfp)
{
    struct tree_list *ch;
    int newposition;

    if (strcmp(root->label, "^^^") == 0) {

        newposition = position + 1;     /* don't print initializing terminal. */
    } else if (root->conflation == 0) {

        if (root->modifier != 0)
            fprintf(wfp, "+");  /* shouldn't happen? */
        if (strcmp(root->label, "") == 0)
            fprintf(wfp, "(TOP");       /* special symbol for matrix root */
        else
            fprintf(wfp, " (%s", root->label);

        if (root->children == NULL) {
            if (root->start != position)
                fprintf(wfp, " ****missing terminal :%d!=%d\n  ",
                        root->start, position);
            fprintf(wfp, " %s", sentence_words[root->start]);
            fprintf(wfp, ")");
            newposition = root->start + 1;
        } else {
            ch = root->children;
            while (ch->next != NULL)
                ch = ch->next;
            newposition = position;
            for (; ch != NULL; ch = ch->prev) {
                newposition = write_subtree(ch->elem, newposition, wfp);
            }

            fprintf(wfp, ")");
        }
    } else {
        newposition = write_subtree(root->children->elem, position, wfp);
    }

    return (newposition);
}




/* display all the branches in root */
void display_tree(struct tree *root)
{
    char *indent;
    char blanks[71] =
        "                                                                      ";
    indent = blanks + 70;

    display_subtree(root, indent, root->start);
    printf("\n");

}

/* returns end position of subtree */
int display_subtree(struct tree *root, char *indent, int position)
{
    struct tree_list *ch;
    char *newindent;
    int newposition;

    if (strlen(indent) >= 70)
        indent += strlen(indent);

    if (strcmp(root->label, "^^^") == 0) {

        newposition = position + 1;     /* don't print initializing terminal. */
    } else if (root->conflation == 0) {

        printf("\n%s", indent);
        if (root->modifier != 0)
            printf("+");
        if (strcmp(root->label, "") == 0)
            printf("(TOP");     /* special symbol for matrix root */
        else
            printf("(%s", root->label);

        if (root->children == NULL) {
            if (root->start != position)
                printf(" ****missing terminal :%d!=%d\n%s  ",
                       root->start, position, indent);
            printf(" %s", sentence_words[root->start]);
#if TREEPOS
            printf(" :%2d)", root->start);
#endif
            printf(")");

            newposition = root->start + 1;
        } else {
            newindent = indent - 2;
            ch = root->children;
            while (ch->next != NULL)
                ch = ch->next;
            newposition = position;
            for (; ch != NULL; ch = ch->prev) {
                newposition =
                    display_subtree(ch->elem, newindent, newposition);
            }

            printf(" )");
        }
    } else {
        newposition =
            display_subtree(root->children->elem, indent, position);
    }

    return (newposition);
}


/* display the network's link weights, plus the mean and mean squared
     (i.e. stddev assuming 0 mean) values for each link type.  Print a * after
     suspect weights. */
void display_net()
{
    int i, j;
    double val;
    double n, m, ms;

    double maxweight = 5.0;
    double minweight = -5.0;


    printf("\nNetwork:\n");

    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numinp1; i++) {
        printf("I1R%d to R(0)=%f", i, val = link_r1_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_r1_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("I1R to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numinp2; i++) {
        printf("I2R%d to R(0)=%f", i, val = link_r2_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_r2_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("I2R to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numinp1; i++) {
        printf("I1T%d to R(0)=%f", i, val = link_t1_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_t1_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("I1T to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numinp2; i++) {
        printf("I2T%d to R(0)=%f", i, val = link_t2_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_t2_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("I2T to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#if HEADWORDS
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numinp1; i++) {
        printf("I1H%d to R(0)=%f", i, val = link_h1_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_h1_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("I1H to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numinp2; i++) {
        printf("I2H%d to R(0)=%f", i, val = link_h2_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_h2_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("I2H to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#endif
#if ! LABELCONTEXTINP
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numrho; i++) {
        printf("R%d to R(0)=%f", i, val = link_r_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_r_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("R to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#endif
    n = 0;
    m = 0.0;
    ms = 0.0;
    printf("N to R(0)=%f", val = link_bir_r[0]);
    if (val > maxweight || val < minweight || val == 0.0)
        printf("*");
    if (val > maxweight || val < minweight)
        printf("*");
    if (val > 2 * maxweight || val < 2 * minweight)
        printf("*");
    n++;
    m += val;
    ms += val * val;
    for (j = 1; j < numrho; j++) {
        printf(" (%d)=%f", j, val = link_bir_r[j]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
    }
    printf("\n");
    printf("N to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numlink_tro; i++) {
        printf("TRO%d to R(0)=%f", i, val = link_tro_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_tro_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("TRO to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numlink_trp; i++) {
        printf("TRP%d to R(0)=%f", i, val = link_trp_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_trp_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("TRP to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numlink_trc; i++) {
        printf("TRC%d to R(0)=%f", i, val = link_trc_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_trc_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("TRC to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numlink_lab; i++) {
        printf("LAB%d to R(0)=%f", i, val = link_lab_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_lab_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("LAB to R: mean = %f, mean squared = %f\n", m / n, ms / n);
    n = 0;
    m = 0.0;
    ms = 0.0;
    printf("P to R(0)=%f", val = link_pred_r[0]);
    if (val > maxweight || val < minweight || val == 0.0)
        printf("*");
    if (val > maxweight || val < minweight)
        printf("*");
    if (val > 2 * maxweight || val < 2 * minweight)
        printf("*");
    n++;
    m += val;
    ms += val * val;
    for (j = 1; j < numrho; j++) {
        printf(" (%d)=%f", j, val = link_pred_r[j]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
    }
    printf("\n");
    printf("P to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#if CHILDCONTEXT
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numlink_chd; i++) {
        printf("CHD%d to R(0)=%f", i, val = link_chd_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_chd_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("CHD to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#endif
#if HEADHIDDEN
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numlink_hdc; i++) {
        printf("HDC%d to R(0)=%f", i, val = link_hdc_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_hdc_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("HDC to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#endif
#if CHOPCONTEXT
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numlink_cho; i++) {
        printf("CHO%d to R(0)=%f", i, val = link_cho_r[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = link_cho_r[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("CHO to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#endif
#if HEADOUTS
    n = 0;
    m = 0.0;
    ms = 0.0;
    printf("NHD to R(0)=%f", val = link_nhd_r[0]);
    if (val > maxweight || val < minweight || val == 0.0)
        printf("*");
    if (val > maxweight || val < minweight)
        printf("*");
    if (val > 2 * maxweight || val < 2 * minweight)
        printf("*");
    n++;
    m += val;
    ms += val * val;
    for (j = 1; j < numrho; j++) {
        printf(" (%d)=%f", j, val = link_nhd_r[j]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
    }
    printf("\n");
    printf("NHD to R: mean = %f, mean squared = %f\n", m / n, ms / n);
#endif
    n = 0;
    m = 0.0;
    ms = 0.0;
    for (i = 0; i < numrho; i++) {
        printf("R%d to O(0)=%f", i, val = link_r_o[0][i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
        for (j = 1; j < numout; j++) {
            printf(" (%d)=%f", j, val = link_r_o[j][i]);
            if (val > maxweight || val < minweight || val == 0.0)
                printf("*");
            if (val > maxweight || val < minweight)
                printf("*");
            if (val > 2 * maxweight || val < 2 * minweight)
                printf("*");
            n++;
            m += val;
            ms += val * val;
        }
        printf("\n");
    }
    printf("R to O: mean = %f, mean squared = %f\n", m / n, ms / n);

    n = 0;
    m = 0.0;
    ms = 0.0;
    printf("Bias R(0)=%f", val = bias_r[0]);
    if (val > maxweight || val < minweight || val == 0.0)
        printf("*");
    if (val > maxweight || val < minweight)
        printf("*");
    if (val > 2 * maxweight || val < 2 * minweight)
        printf("*");
    n++;
    m += val;
    ms += val * val;
    for (i = 1; i < numrho; i++) {
        printf(" (%d)=%f", i, val = bias_r[i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
    }
    printf("\n");
    printf("Bias R: mean = %f, mean squared = %f\n", m / n, ms / n);



    n = 0;
    m = 0.0;
    ms = 0.0;
    printf("Bias O(0)=%f", val = bias_o[0]);
    if (val > maxweight || val < minweight || val == 0.0)
        printf("*");
    if (val > maxweight || val < minweight)
        printf("*");
    if (val > 2 * maxweight || val < 2 * minweight)
        printf("*");
    n++;
    m += val;
    ms += val * val;
    for (i = 1; i < numout; i++) {
        printf(" (%d)=%f", i, val = bias_o[i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
    }
    printf("\n");
    printf("Bias O: mean = %f, mean squared = %f\n", m / n, ms / n);

    n = 0;
    m = 0.0;
    ms = 0.0;
    printf("Initstate R(0)=%f", val = hidrho_initstate.act[0]);
    if (val > maxweight || val < minweight || val == 0.0)
        printf("*");
    if (val > maxweight || val < minweight)
        printf("*");
    if (val > 2 * maxweight || val < 2 * minweight)
        printf("*");
    n++;
    m += val;
    ms += val * val;
    for (i = 1; i < numrho; i++) {
        printf(" (%d)=%f", i, val = hidrho_initstate.act[i]);
        if (val > maxweight || val < minweight || val == 0.0)
            printf("*");
        if (val > maxweight || val < minweight)
            printf("*");
        if (val > 2 * maxweight || val < 2 * minweight)
            printf("*");
        n++;
        m += val;
        ms += val * val;
    }
    printf("\n");
    printf("Initstate R: mean = %f, mean squared = %f\n", m / n, ms / n);


    printf("\n");
}


/* display the network's link weight deltas */
void display_link_deltas()
{
    int i, j;
    double val;

    double maxdelta = 5.0;
    double mindelta = -5.0;

    printf("\n@@@@@@@@@@ Link deltas:\n");

    for (i = 0; i < numinp1; i++) {
        printf("Delta I1R%d to R(0)=%f", i, val = linkdel_r1_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_r1_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    for (i = 0; i < numinp2; i++) {
        printf("Delta I2R%d to R(0)=%f", i, val = linkdel_r2_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_r2_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    for (i = 0; i < numinp1; i++) {
        printf("Delta I1T%d to R(0)=%f", i, val = linkdel_t1_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_t1_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    for (i = 0; i < numinp2; i++) {
        printf("Delta I2T%d to R(0)=%f", i, val = linkdel_t2_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_t2_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
#if HEADWORDS
    for (i = 0; i < numinp1; i++) {
        printf("Delta I1H%d to R(0)=%f", i, val = linkdel_h1_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_h1_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    for (i = 0; i < numinp2; i++) {
        printf("Delta I2H%d to R(0)=%f", i, val = linkdel_h2_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_h2_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
#endif
#if ! LABELCONTEXTINP
    for (i = 0; i < numrho; i++) {
        printf("Delta R%d to R(0)=%f", i, val = linkdel_r_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_r_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
#endif
    printf("Delta N to R(0)=%f", val = linkdel_bir_r[0]);
    if (val > maxdelta || val < mindelta || val == 0.0)
        printf("*");
    if (val > maxdelta || val < mindelta)
        printf("*");
    if (val > 2 * maxdelta || val < 2 * mindelta)
        printf("*");
    for (j = 1; j < numrho; j++) {
        printf(" (%d)=%f", j, val = linkdel_bir_r[j]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
    }
    printf("\n");
    for (i = 0; i < numlink_tro; i++) {
        printf("Delta TRO%d to R(0)=%f", i, val = linkdel_tro_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_tro_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    for (i = 0; i < numlink_trp; i++) {
        printf("Delta TRP%d to R(0)=%f", i, val = linkdel_trp_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_trp_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    for (i = 0; i < numlink_trc; i++) {
        printf("Delta TRC%d to R(0)=%f", i, val = linkdel_trc_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_trc_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    for (i = 0; i < numlink_lab; i++) {
        printf("Delta LAB%d to R(0)=%f", i, val = linkdel_lab_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_lab_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
    printf("Delta P to R(0)=%f", val = linkdel_pred_r[0]);
    if (val > maxdelta || val < mindelta || val == 0.0)
        printf("*");
    if (val > maxdelta || val < mindelta)
        printf("*");
    if (val > 2 * maxdelta || val < 2 * mindelta)
        printf("*");
    for (j = 1; j < numrho; j++) {
        printf(" (%d)=%f", j, val = linkdel_pred_r[j]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
    }
    printf("\n");
#if CHILDCONTEXT
    for (i = 0; i < numlink_chd; i++) {
        printf("Delta CHD%d to R(0)=%f", i, val = linkdel_chd_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_chd_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
#endif
#if HEADHIDDEN
    for (i = 0; i < numlink_hdc; i++) {
        printf("Delta HDC%d to R(0)=%f", i, val = linkdel_hdc_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_hdc_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
#endif
#if CHOPCONTEXT
    for (i = 0; i < numlink_cho; i++) {
        printf("Delta CHO%d to R(0)=%f", i, val = linkdel_cho_r[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numrho; j++) {
            printf(" (%d)=%f", j, val = linkdel_cho_r[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }
#endif
#if HEADOUTS
    printf("Delta NHD to R(0)=%f", val = linkdel_nhd_r[0]);
    if (val > maxdelta || val < mindelta || val == 0.0)
        printf("*");
    if (val > maxdelta || val < mindelta)
        printf("*");
    if (val > 2 * maxdelta || val < 2 * mindelta)
        printf("*");
    for (j = 1; j < numrho; j++) {
        printf(" (%d)=%f", j, val = linkdel_nhd_r[j]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
    }
    printf("\n");
#endif
    for (i = 0; i < numrho; i++) {
        printf("Delta R%d to O(0)=%f", i, val = linkdel_r_o[0][i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
        for (j = 1; j < numout; j++) {
            printf(" (%d)=%f", j, val = linkdel_r_o[j][i]);
            if (val > maxdelta || val < mindelta || val == 0.0)
                printf("*");
            if (val > maxdelta || val < mindelta)
                printf("*");
            if (val > 2 * maxdelta || val < 2 * mindelta)
                printf("*");
        }
        printf("\n");
    }

    printf("Bias delta R(0): %f", val = biasdel_r[0]);
    if (val > maxdelta || val < mindelta || val == 0.0)
        printf("*");
    if (val > maxdelta || val < mindelta)
        printf("*");
    if (val > 2 * maxdelta || val < 2 * mindelta)
        printf("*");
    for (i = 1; i < numrho; i++) {
        printf(" (%d)=%f", i, val = biasdel_r[i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
    }
    printf("\n");



    printf("Bias delta O(0): %f", val = biasdel_o[0]);
    if (val > maxdelta || val < mindelta || val == 0.0)
        printf("*");
    if (val > maxdelta || val < mindelta)
        printf("*");
    if (val > 2 * maxdelta || val < 2 * mindelta)
        printf("*");
    for (i = 1; i < numout; i++) {
        printf(" (%d)=%f", i, val = biasdel_o[i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
    }
    printf("\n\n");

    printf("Initstate delta R(0): %f", val =
           eta * hidrho_initstate.err[0]);
    if (val > maxdelta || val < mindelta || val == 0.0)
        printf("*");
    if (val > maxdelta || val < mindelta)
        printf("*");
    if (val > 2 * maxdelta || val < 2 * mindelta)
        printf("*");
    for (i = 1; i < numrho; i++) {
        printf(" (%d)=%f", i, val = eta * hidrho_initstate.err[i]);
        if (val > maxdelta || val < mindelta || val == 0.0)
            printf("*");
        if (val > maxdelta || val < mindelta)
            printf("*");
        if (val > 2 * maxdelta || val < 2 * mindelta)
            printf("*");
    }
    printf("\n\n");

}

/* See check_output_act_del_range().  Derived from
     normalize_output_act().  */
void check_output_act_del(int current, int trainingflag)
{
    int wd, cn;
    int ptf, uhf, hpf;

    wd = current;

    for (cn = 0; cn <= max_phase[wd]; cn++)
        if (phase_hist[cn].nextcomp != NULL &&
            *phase_hist[cn].nextcomp == wd) {
            if (phase_hist[cn].trigger_op == PREDOUT) {
                ptf = 1;
                uhf = 0;
            }
#if UNHEADEDCOND && HEADOUTS
            else if (phase_hist[cn].head_child < 0 ||
                     phase_hist[phase_hist[cn].head_child].birth > wd) {
                ptf = 0;
                uhf = 1;
            }
#endif
            else {
                ptf = 0;
                uhf = 0;
            }
#if HEADEDPARENTCOND && HEADOUTS
            if (phase_hist[cn].trigger_parent >= 0 &&
                phase_hist[phase_hist[cn].trigger_parent].head_child >= 0
                && phase_hist[phase_hist[phase_hist[cn].trigger_parent].
                              head_child].death >= 0
                && phase_hist[phase_hist[phase_hist[cn].trigger_parent].
                              head_child].death < wd)
                hpf = 1;
            else
#endif
                hpf = 0;
            check_output_act_del_range(wd, cn, &output_hist[cn], 0,
                                       numoutmem - 1, ptf, uhf, hpf,
                                       trainingflag);
        }
}

/* check the outputs and deltas for period current and phase for units
     between first_mem and last_mem, inclusive.  Derived from
     normalize_output_act_range().  */
void check_output_act_del_range(int current, int phase,
                                struct output *outputs, int first_mem,
                                int last_mem, int pretermflag,
                                int unheadedflag, int hddparentflag,
                                int trainingflag)
{
    int i, j;
    int forend;
    int t, p;
    int inp1, out1;
    int bin;
    int binstart, binend;
    int memoffset;
    double *output_act, *target_act;
    double *output_del;

    double totact;
    double totdel;
    double grandnorm;
    double labelnorm;

    double margin = 0.0000000001;

    output_act = outputs->act;
    output_del = outputs->del;
    target_act = (target_hist[current])[phase].act;

    error
        ("check_output_act_del_range() is not supported - if required, memory model should be updated");

#if NOPREDICT
    if (last_mem >= predtag_to_output(0))
        printf
            ("argument last_mem = %d out of range (>= %d) in check_output_act_del_range().\n",
             last_mem, predtag_to_output(0));
#endif

    if (in_range(PREPREDFIRST, PREPREDLAST, first_mem, last_mem)) {
        /* check {PREDOUT, PROJOUT, HDPROUT, ATTOUT, HDATOUT, MODOUT} outputs */
        if ((trainingflag == 0 || target_act[ATTOUT] >= 0.0)
#if UNHEADEDCOND && HEADOUTS
            && unheadedflag == 0
#endif
            ) {
            totact = output_act[ATTOUT] + output_act[MODOUT];
            if (trainingflag != 0)
                totdel = output_del[ATTOUT] + output_del[MODOUT];
            else
                totdel = 0.0;
            totact += output_act[PREDOUT];
            if (trainingflag != 0)
                totdel += output_del[PREDOUT];
            totact += output_act[PROJOUT];
            if (trainingflag != 0)
                totdel += output_del[PROJOUT];
#if HEADOUTS
#if HEADEDPARENTCOND && HEADOUTS
            if (hddparentflag == 0)
#endif
            {
                totact += output_act[HDATOUT];
                if (trainingflag != 0)
                    totdel += output_del[HDATOUT];
            }
            totact += output_act[HDPROUT];
            if (trainingflag != 0)
                totdel += output_del[HDPROUT];
#endif
            totact -= 1.0;
            if (totact > margin || totact < -margin)
                printf
                    ("pre-prediction total outputs - 1.0: %e (%d:%d,%d)\n",
                     totact, num_sent + 1, current, phase);
            if (totdel > margin || totdel < -margin)
                printf
                    ("pre-prediction total output deltas: %e (%d:%d,%d)\n",
                     totdel, num_sent + 1, current, phase);
        }
    }


    if (in_range
        (projlabel_to_output(0), projlabel_to_output(numlabels - 1),
         first_mem, last_mem)) {
        /* check projected labels */
        if (trainingflag == 0 || target_act[projlabel_to_output(0)] >= 0.0) {
            totact = 0.0;
            totdel = 0.0;
            forend = projlabel_to_output(numlabels - 1);
            for (i = projlabel_to_output(0); i <= forend; i++) {
                totact += output_act[i];
                if (trainingflag != 0)
                    totdel += output_del[i];
            }
            totact -= 1.0;
            if (totact > margin || totact < -margin)
                printf("projlabel total outputs -1.0: %e (%d:%d,%d)\n",
                       totact, num_sent + 1, current, phase);
            if (totdel > margin || totdel < -margin)
                printf("projlabel total output deltas: %e (%d:%d,%d)\n",
                       totdel, num_sent + 1, current, phase);
        }
    }
#if HEADOUTS
    if (in_range
        (hdprlabel_to_output(0), hdprlabel_to_output(numlabels - 1),
         first_mem, last_mem)) {
        /* check head projected labels */
        if (trainingflag == 0 || target_act[hdprlabel_to_output(0)] >= 0.0) {
            totact = 0.0;
            totdel = 0.0;
            forend = hdprlabel_to_output(numlabels - 1);
            for (i = hdprlabel_to_output(0); i <= forend; i++) {
                totact += output_act[i];
                if (trainingflag != 0)
                    totdel += output_del[i];
            }
            totact -= 1.0;
            if (totact > margin || totact < -margin)
                printf("hdprlabel total outputs -1.0: %e (%d:%d,%d)\n",
                       totact, num_sent + 1, current, phase);
            if (totdel > margin || totdel < -margin)
                printf("hdprlabel total output deltas: %e (%d:%d,%d)\n",
                       totdel, num_sent + 1, current, phase);
        }
    }
#endif


#if ! NOPREDICT
    if (in_range(predtag_to_output(0), predtag_to_output(numout1 - 1),
                 first_mem, last_mem)) {
        /* check predicted tags */
        if (trainingflag == 0 || target_act[predtag_to_output(0)] >= 0.0) {
            totact = 0.0;
            totdel = 0.0;
            forend = predtag_to_output(numout1 - 1);
            for (i = predtag_to_output(0); i <= forend; i++) {
                totact += output_act[i];
                if (trainingflag != 0)
                    totdel += output_del[i];
            }
            totact -= 1.0;
            if (totact > margin || totact < -margin)
                printf("predtag total outputs - 1.0: %e (%d:%d,%d)\n",
                       totact, num_sent + 1, current, phase);
            if (totdel > margin || totdel < -margin)
                printf("predtag total output deltas: %e (%d:%d,%d)\n",
                       totdel, num_sent + 1, current, phase);
        }
    }
#endif


#if ! NOPREDICT
    if (last_mem >= predword_to_output(0)
        /* otherwise tag maybe undefined */
        /* always predict word whenever predicting tag */
        && (trainingflag == 0 || target_act[predtag_to_output(0)] >= 0.0)) {
        out1 = input_hist[current + 1].input1;
        memoffset = out1_first_word[out1];

        for (bin = 0; bin < numbins; bin++) {
            if (input_hist[current + 1].output2[bin] >= 0) {
                binstart =
                    predword_to_output(out1_bin_first_word[out1][bin]
                                       - memoffset);
                binend =
                    predword_to_output(out1_bin_first_word[out1][bin + 1]
                                       - memoffset - 1);

                if (in_range(binstart, binend, first_mem, last_mem)) {
                    /* check predicted words in bin */
                    totact = 0.0;
                    totdel = 0.0;
                    for (i = binstart; i <= binend; i++) {
                        totact += output_act[i];
                        if (trainingflag != 0)
                            totdel += output_del[i];
                    }
                    totact -= 1.0;
                    if (totact > margin || totact < -margin)
                        printf
                            ("tag %d bin %d predword total outputs -1.0: %e (%d:%d,%d)\n",
                             out1, bin + 1, totact, num_sent + 1, current,
                             phase);
                    if (totdel > margin || totdel < -margin)
                        printf
                            ("tag %d bin %d predword total output deltas: %e (%d:%d,%d)\n",
                             out1, bin + 1, totdel, num_sent + 1, current,
                             phase);
                }
            } else
                break;
        }
    }
#endif

}



/* display the deltas at period */
/** Not updated for virtual target values */
void display_deltas(int period)
{
    int i, p;

    printf("\n******** Deltas at period %d : \n", period);
    for (p = 0; p <= max_phase[period]; p++)
        if (test_comp(period, p)) {
            printf("**** Phase number: %d \n", p);
            for (i = 0; i < numoutmem; i++) {
                if ((target_hist[period])[p].act[i] >= 0.0)
                    printf("     Output: %2d, delta = %+f\n",
                           i, output_hist[p].del[i]);
                else if (output_hist[p].del[i] != 0.0)
                    printf
                        ("     Output: %2d, delta = %+f ***should be zero!***\n",
                         i, output_hist[p].del[i]);
            }
            /* combination deltas not recorded */
            for (i = 0; i < numrho; i++)
                printf(" Rho hidden: %2d, delta = %+f\n", i,
                       (hidrho_hist[period])[p].err[i] *
                       (hidrho_hist[period])[p].act[i] *
                       (1.0 - (hidrho_hist[period])[p].act[i]));
        }
    printf("******** End Deltas\n");
}


/* display the total deltas at period */
void display_total_deltas(int period)
{
    int i, p;
    double ptot, ntot;
    double val;

    printf("\n  Total deltas at period %d : \n", period);
    for (p = 0; p <= max_phase[period]; p++)
        if (test_comp(period, p)) {
            printf("    Phase number: %d \n", p);
            ptot = 0.0;
            ntot = 0.0;
            for (i = 0; i < numoutmem; i++) {
                val = output_hist[p].del[i];
                if (val >= 0)
                    ptot += val;
                else
                    ntot += val;
            }
            printf("      Output: pos = %+f, neg = %+f\n", ptot, ntot);
            /* combination deltas not recorded */
            ptot = 0.0;
            ntot = 0.0;
            for (i = 0; i < numrho; i++) {
                val =
                    (hidrho_hist[period])[p].err[i] *
                    (hidrho_hist[period])[p].act[i] *
                    (1.0 - (hidrho_hist[period])[p].act[i]);
                if (val >= 0)
                    ptot += val;
                else
                    ntot += val;
            }
            printf("      Hidrho: pos = %+f, neg = %+f\n", ptot, ntot);
        }
}



/* display the units' activation values at period */
/** Not updated for virtual target values */
void display_state(int period)
{
    int i, p;

    printf("\n######## Activations of units at time %d : \n", period);
    for (p = 0; p <= max_phase[period]; p++)
        if (test_comp(period, p)) {
            printf("#### Phase number: %d \n", p);
            for (i = 0; i < numoutmem; i++)
                printf("     Target: %2d, act = %f\n", i,
                       (target_hist[period])[p].act[i]);
            for (i = 0; i < numoutmem; i++)
                printf("     Output: %2d, act = %f\n", i,
                       output_hist[p].act[i]);
            for (i = 0; i < numrho; i++)
                printf(" Rho hidden: %2d, act = %f\n", i,
                       (hidrho_hist[period])[p].act[i]);
        }
    printf("#### Whole period\n");
    printf("     Input 1, letter = %d\n", input_hist[period].input1);
    printf("     Input input 2, letter = %d\n",
           input_hist[period].input2[0]);
    printf("     Output input 2, letter = %d\n",
           input_hist[period].output2[0]);
    printf("######## End Activations\n");

}


/* display the parentval and grandval histories */
/*
void display_histories(int length)
{
    int w, c;
    double val;
printf("\n***function shouldn't be used.\n");

    printf("\nChild:                           Parent \n      ");
    for (c = 0; c < length; c++)
      printf(" %3d ", c);
    printf("\n");
    for (w = 0; w <= length; w++)
      {
        printf(" %3d:", w);
        for (c = 0; c <= w; c++)
          {
            val = output_hist[c].act[PAROUT];
            if (val >= 1)
              printf(" %.2f", val);
            else
              printf(" .%03.0f", 1000.0*val-0.5);
          }
        printf("\n");
      }

    printf("\nChild:                           Grandparent \n      ");
    for (c = 0; c < length; c++)
      printf(" %3d ", c);
    printf("\n");
    for (w = 0; w <= length; w++)
      {
        printf(" %3d:", w);
        for (c = 0; c < w; c++)
          {
            val = output_hist[c].act[GRANDOUT];
            if (val >= 1)
              printf(" %.2f", val);
            else
              printf(" .%03.0f", 1000.0*val-0.5);
          }
        printf(" .%03.0f", 0.0);
        for (c = w+1; c < length; c++)
          {
            val = output_hist[w].act[SIBOUT];
            if (val >= 1)
              printf(" %.2f", val);
            else
              printf(" .%03.0f", 1000.0*val-0.5);
          }
        printf("\n");
      }

    printf("\n                                 Label \n      ");
    for (c = 0; c < length; c++)
      printf(" %3d ", c);
    printf("\n      ");
    for (c = 0; c < length; c++)
      printf(" %3d ", label_hist[c]);
    printf("\n      ");
    for (c = 0; c < length; c++)
      {
        if (output_hist[c].act[label_hist[c] + NUMSTRUC] >= 1)
          printf(" %.2f", output_hist[c].act[label_hist[c] + NUMSTRUC]);
        else
          printf(" .%03.0f", 1000.0*output_hist[c].act[label_hist[c] +
                                                           NUMSTRUC]-0.5);
      }
    printf("\n");
}
*/


/* display whether maximum outputs are targets within targetted groups, and
     cumulative probability of target derivation.  prob is previous
     probability and returns subsequent probability. */
/* Start with: printf("  Target and Output symbols:\n"); */
double display_symbolic_target_output(int period, double prob)
{
    int i, p;
    int *pp;
    double maxv;
    int maxi;
    int tari;
    double *target_act;
    double *output_act;

    printf("  Period %3d :", period);
    printf("      Input 1 = %s", id_tag(input_hist[period].input1));
    printf(" Input input 2 = %s", id_word(input_hist[period].input2[0]));
    printf(" Output input 2 = %s", id_word(input_hist[period].output2[0]));
    printf("\n");
    /* printf("      Target  Output\n"); */
    if (deriv_order[period].top != NULL) {

        pp = deriv_order[period].content;

        /* subsequent phases in deriv_order are pre-prediction steps */
        while (pp <= deriv_order[period].top) {
            p = *pp;
            pp++;
            printf("    Phase %3d: \n", p);
            target_act = (target_hist[period])[p].act;
            output_act = output_hist[p].act;

            if (target_act[PREDOUT] >= 0.0) {
                if (period != 0) {
                    if (target_act[PREDOUT] > 0.5) {
                        tari = PREDOUT;
                        printf("      >");
                    }
                    if (target_act[PROJOUT] > 0.5) {
                        tari = PROJOUT;
                        printf("      |");
                    }
                    if (target_act[ATTOUT] > 0.5) {
                        tari = ATTOUT;
                        printf("      ^");
                    }
#if HEADOUTS
                    if (target_act[HDPROUT] > 0.5) {
                        tari = HDPROUT;
                        printf("      |h");
                    }
                    if (target_act[HDATOUT] > 0.5) {
                        tari = HDATOUT;
                        printf("      ^h");
                    }
#endif
                    if (target_act[MODOUT] > 0.5) {
                        tari = MODOUT;
                        printf("      +");
                    }
                    maxi = PREDOUT;
                    maxv = output_act[PREDOUT];
                    if (output_act[PROJOUT] > maxv) {
                        maxv = output_act[PROJOUT];
                        maxi = PROJOUT;
                    }
                    if (output_act[ATTOUT] > maxv) {
                        maxv = output_act[ATTOUT];
                        maxi = ATTOUT;
                    }
#if HEADOUTS
                    if (output_act[HDPROUT] > maxv) {
                        maxv = output_act[HDPROUT];
                        maxi = HDPROUT;
                    }
                    if (output_act[HDATOUT] > maxv) {
                        maxv = output_act[HDATOUT];
                        maxi = HDATOUT;
                    }
#endif
                    if (output_act[MODOUT] > maxv) {
                        maxv = output_act[MODOUT];
                        maxi = MODOUT;
                    }
                    if (maxi == PREDOUT)
                        printf("       >");
                    if (maxi == PROJOUT)
                        printf("       |");
                    if (maxi == ATTOUT)
                        printf("       ^");
#if HEADOUTS
                    if (maxi == HDPROUT)
                        printf("       |h");
                    if (maxi == HDATOUT)
                        printf("       ^h");
#endif
                    if (maxi == MODOUT)
                        printf("       +");
                    if (tari != maxi)
                        printf("       (%f)",
                               output_act[maxi] - output_act[tari]);
                    else
                        printf("       ()        ");
                    prob *= output_act[tari];
                } else {        /* initial period always a prediction */
                    printf("      >");
                    printf("       >");
                    printf("       (def)     ");
                }
                printf("    tpr= %e", prob);
                printf("\n");
            }

            if (target_act[predtag_to_output(0)] >= 0.0) {
                tari = -1;
                for (i = predtag_to_output(0);
                     i <= predtag_to_output(numout1 - 1); i++)
                    if (target_act[i] > 0.5) {
                        tari = i;
                        break;
                    }
                if (tari < 0)
                    printf("      >:???  ");
                else
                    printf("      >:%-5s",
                           id_tag(tari - predtag_to_output(0)));
                maxi = predtag_to_output(0);
                maxv = output_act[predtag_to_output(0)];
                for (i = predtag_to_output(1);
                     i <= predtag_to_output(numout1 - 1); i++)
                    if (output_act[i] > maxv) {
                        maxv = output_act[i];
                        maxi = i;
                    }
                printf(" >:%-5s", id_tag(maxi - predtag_to_output(0)));
                if (tari != maxi)
                    printf(" (%f)", output_act[maxi] - output_act[tari]);
                else
                    printf(" ()        ");
                prob *= output_act[tari];
                printf("    tpr= %e", prob);
                printf("\n");
            }

            if (target_act[projlabel_to_output(0)] >= 0.0) {
                tari = -1;
                for (i = projlabel_to_output(0);
                     i <= projlabel_to_output(numlabels - 1); i++)
                    if (target_act[i] > 0.5) {
                        tari = i;
                        break;
                    }
                if (tari < 0)
                    printf("      |:???  ");
                else
                    printf("      |:%-5s",
                           id_label(tari - projlabel_to_output(0)));
                maxi = projlabel_to_output(0);
                maxv = output_act[projlabel_to_output(0)];
                for (i = projlabel_to_output(1);
                     i <= projlabel_to_output(numlabels - 1); i++)
                    if (output_act[i] > maxv) {
                        maxv = output_act[i];
                        maxi = i;
                    }
                printf(" |:%-5s", id_label(maxi - projlabel_to_output(0)));
                if (tari != maxi)
                    printf(" (%f)", output_act[maxi] - output_act[tari]);
                else
                    printf(" ()        ");
                prob *= output_act[tari];
                printf("    tpr= %e", prob);
                printf("\n");
            }
#if HEADOUTS
            if (target_act[hdprlabel_to_output(0)] >= 0.0) {
                tari = -1;
                for (i = hdprlabel_to_output(0);
                     i <= hdprlabel_to_output(numlabels - 1); i++)
                    if (target_act[i] > 0.5) {
                        tari = i;
                        break;
                    }
                if (tari < 0)
                    printf("      |h:???  ");
                else
                    printf("      |h:%-5s",
                           id_label(tari - hdprlabel_to_output(0)));
                maxi = hdprlabel_to_output(0);
                maxv = output_act[hdprlabel_to_output(0)];
                for (i = hdprlabel_to_output(1);
                     i <= hdprlabel_to_output(numlabels - 1); i++)
                    if (output_act[i] > maxv) {
                        maxv = output_act[i];
                        maxi = i;
                    }
                printf(" |h:%-5s",
                       id_label(maxi - hdprlabel_to_output(0)));
                if (tari != maxi)
                    printf(" (%f)", output_act[maxi] - output_act[tari]);
                else
                    printf(" ()        ");
                prob *= output_act[tari];
                printf("    tpr= %e", prob);
                printf("\n");
            }
#endif
        }
    }

    return (prob);
}


/* display the input, target, and output values at period. */
/** Not updated for virtual target values */
void display_target_output(int period)
{
    int i, p;

    printf("\n######## Activations of I/O at time %d : \n", period);
    printf("    Input 1 = %s\n", id_tag(input_hist[period].input1));
    printf("    Input input 2 = %s\n",
           id_word(input_hist[period].input2[0]));
    printf("    Output input 2 = %s\n",
           id_word(input_hist[period].output2[0]));
    for (p = 0; p <= max_phase[period]; p++)
        if (test_comp(period, p)) {
            printf("#### Phase number: %d \n", p);
            for (i = 0; i < numoutmem; i++)
                if ((target_hist[period])[p].act[i] >= 0)
                    printf
                        ("    Unit %2d:  target= %f, output= %f,  difference= %+f\n",
                         i, (target_hist[period])[p].act[i],
                         output_hist[p].act[i],
                         (target_hist[period])[p].act[i] -
                         output_hist[p].act[i]);
        }
    /*  printf("######## End Activations\n"); */
}


/* display each phases states list. */
void display_states(int maxphase)
{
    int p;
    struct stack *s;

    printf("Periods in which each phase has states,\n");
    for (p = 0; p <= maxphase; p++) {
        s = &phase_hist[p].states;
        printf("  phase %2d: ", p);
        display_stack(s);
    }
}


void display_opt_stack(struct option *opt)
{
    struct option *opt2;

    if (opt != NULL) {
        for (opt2 = opt; opt2->trigparent_option != NULL;
             opt2 = opt2->trigparent_option) {
            printf("%d, ", opt2->phase);
        }
        printf("%d\n", opt2->phase);
    } else
        printf("\n");
}

void display_stack(struct stack *s)
{
    int i;

    if (s->top != NULL) {
        if (s->top < s->content || s->top >= s->content + phase_space) {
            printf("**Bad stack format (%d %d %d)  ", s->top, s->subtop,
                   s->content);
            for (i = 0; i < phase_space - 1; i++)
                printf("%d, ", s->content[i]);
            printf("%d", s->content[phase_space - 1]);
            warning("  ");
        } else {
            for (i = 0; &s->content[i] != s->top; i++)
                printf("%d, ", s->content[i]);
            printf("%d\n", *s->top);
        }
    } else
        printf("\n");
}





/* Tag and label codes */

int tag_id(char *str)
{
    if (!strcmp(str, "#"))
        return (0);
    else if (!strcmp(str, "$"))
        return (1);
    else if (!strcmp(str, "''"))
        return (2);
    else if (!strcmp(str, ","))
        return (3);
    else if (!strcmp(str, "-LRB"))
        return (4);
    else if (!strcmp(str, "-RRB"))
        return (5);
    else if (!strcmp(str, "."))
        return (6);
    else if (!strcmp(str, ":"))
        return (7);
    else if (!strcmp(str, "^^^"))
        return (8);
    else if (!strcmp(str, "``"))
        return (9);
    else if (!strcmp(str, "CC"))
        return (10);
    else if (!strcmp(str, "CD"))
        return (11);
    else if (!strcmp(str, "DT"))
        return (12);
    else if (!strcmp(str, "EX"))
        return (13);
    else if (!strcmp(str, "FW"))
        return (14);
    else if (!strcmp(str, "IN"))
        return (15);
    else if (!strcmp(str, "JJ"))
        return (16);
    else if (!strcmp(str, "JJR"))
        return (17);
    else if (!strcmp(str, "JJS"))
        return (18);
    else if (!strcmp(str, "LS"))
        return (19);
    else if (!strcmp(str, "MD"))
        return (20);
    else if (!strcmp(str, "NN"))
        return (21);
    else if (!strcmp(str, "NNP"))
        return (22);
    else if (!strcmp(str, "NNPS"))
        return (23);
    else if (!strcmp(str, "NNS"))
        return (24);
    else if (!strcmp(str, "PDT"))
        return (25);
    else if (!strcmp(str, "POS"))
        return (26);
    else if (!strcmp(str, "PRP"))
        return (27);
    else if (!strcmp(str, "PRP$"))
        return (28);
    else if (!strcmp(str, "RB"))
        return (29);
    else if (!strcmp(str, "RBR"))
        return (30);
    else if (!strcmp(str, "RBS"))
        return (31);
    else if (!strcmp(str, "RP"))
        return (32);
    else if (!strcmp(str, "SYM"))
        return (33);
    else if (!strcmp(str, "TO"))
        return (34);
    else if (!strcmp(str, "UH"))
        return (35);
    else if (!strcmp(str, "VB"))
        return (36);
    else if (!strcmp(str, "VBD"))
        return (37);
    else if (!strcmp(str, "VBG"))
        return (38);
    else if (!strcmp(str, "VBN"))
        return (39);
    else if (!strcmp(str, "VBP"))
        return (40);
    else if (!strcmp(str, "VBZ"))
        return (41);
    else if (!strcmp(str, "WDT"))
        return (42);
    else if (!strcmp(str, "WP"))
        return (43);
    else if (!strcmp(str, "WP$"))
        return (44);
    else if (!strcmp(str, "WRB"))
        return (45);
    else
        return (-1);
}

int label_id(char *str)
{
    if (!strcmp(str, ""))
        return (0);
    else if (!strcmp(str, "ADJP"))
        return (1);
    else if (!strcmp(str, "ADVP"))
        return (2);
    else if (!strcmp(str, "CONJP"))
        return (3);
    else if (!strcmp(str, "FRAG"))
        return (4);
    else if (!strcmp(str, "INTJ"))
        return (5);
    else if (!strcmp(str, "LST"))
        return (6);
    else if (!strcmp(str, "NAC"))
        return (7);
    else if (!strcmp(str, "NP"))
        return (8);
    else if (!strcmp(str, "NX"))
        return (9);
    else if (!strcmp(str, "PP"))
        return (10);
    else if (!strcmp(str, "PRN"))
        return (11);
    else if (!strcmp(str, "PRT"))
        return (12);
    else if (!strcmp(str, "QP"))
        return (13);
    else if (!strcmp(str, "RRC"))
        return (14);
    else if (!strcmp(str, "S"))
        return (15);
    else if (!strcmp(str, "S-NP"))
        return (16);
    else if (!strcmp(str, "S-VP"))
        return (17);
    else if (!strcmp(str, "SBAR"))
        return (18);
    else if (!strcmp(str, "SBAR-S"))
        return (19);
    else if (!strcmp(str, "SBAR-S-VP"))
        return (20);
    else if (!strcmp(str, "SBARQ"))
        return (21);
    else if (!strcmp(str, "SINV"))
        return (22);
    else if (!strcmp(str, "SQ"))
        return (23);
    else if (!strcmp(str, "SQ-VP"))
        return (24);
    else if (!strcmp(str, "UCP"))
        return (25);
    else if (!strcmp(str, "VP"))
        return (26);
    else if (!strcmp(str, "WHADJP"))
        return (27);
    else if (!strcmp(str, "WHADVP"))
        return (28);
    else if (!strcmp(str, "WHNP"))
        return (29);
    else if (!strcmp(str, "WHPP"))
        return (30);
    else if (!strcmp(str, "X"))
        return (31);
    else if (!strcmp(str, "SBAR-S-NP"))         /** not actually in data! **/
        return (32);
    else
        return (-1);
}

char *id_tag(int id)
{
    if (id == 0)
        return ("#");
    else if (id == 1)
        return ("$");
    else if (id == 2)
        return ("''");
    else if (id == 3)
        return (",");
    else if (id == 4)
        return ("-LRB");
    else if (id == 5)
        return ("-RRB");
    else if (id == 6)
        return (".");
    else if (id == 7)
        return (":");
    else if (id == 8)
        return ("^^^");
    else if (id == 9)
        return ("``");
    else if (id == 10)
        return ("CC");
    else if (id == 11)
        return ("CD");
    else if (id == 12)
        return ("DT");
    else if (id == 13)
        return ("EX");
    else if (id == 14)
        return ("FW");
    else if (id == 15)
        return ("IN");
    else if (id == 16)
        return ("JJ");
    else if (id == 17)
        return ("JJR");
    else if (id == 18)
        return ("JJS");
    else if (id == 19)
        return ("LS");
    else if (id == 20)
        return ("MD");
    else if (id == 21)
        return ("NN");
    else if (id == 22)
        return ("NNP");
    else if (id == 23)
        return ("NNPS");
    else if (id == 24)
        return ("NNS");
    else if (id == 25)
        return ("PDT");
    else if (id == 26)
        return ("POS");
    else if (id == 27)
        return ("PRP");
    else if (id == 28)
        return ("PRP$");
    else if (id == 29)
        return ("RB");
    else if (id == 30)
        return ("RBR");
    else if (id == 31)
        return ("RBS");
    else if (id == 32)
        return ("RP");
    else if (id == 33)
        return ("SYM");
    else if (id == 34)
        return ("TO");
    else if (id == 35)
        return ("UH");
    else if (id == 36)
        return ("VB");
    else if (id == 37)
        return ("VBD");
    else if (id == 38)
        return ("VBG");
    else if (id == 39)
        return ("VBN");
    else if (id == 40)
        return ("VBP");
    else if (id == 41)
        return ("VBZ");
    else if (id == 42)
        return ("WDT");
    else if (id == 43)
        return ("WP");
    else if (id == 44)
        return ("WP$");
    else if (id == 45)
        return ("WRB");
    else
        return ("???");
}

char *id_label(int id)
{
    if (id == 0)
        return ("");
    else if (id == 1)
        return ("ADJP");
    else if (id == 2)
        return ("ADVP");
    else if (id == 3)
        return ("CONJP");
    else if (id == 4)
        return ("FRAG");
    else if (id == 5)
        return ("INTJ");
    else if (id == 6)
        return ("LST");
    else if (id == 7)
        return ("NAC");
    else if (id == 8)
        return ("NP");
    else if (id == 9)
        return ("NX");
    else if (id == 10)
        return ("PP");
    else if (id == 11)
        return ("PRN");
    else if (id == 12)
        return ("PRT");
    else if (id == 13)
        return ("QP");
    else if (id == 14)
        return ("RRC");
    else if (id == 15)
        return ("S");
    else if (id == 16)
        return ("S-NP");
    else if (id == 17)
        return ("S-VP");
    else if (id == 18)
        return ("SBAR");
    else if (id == 19)
        return ("SBAR-S");
    else if (id == 20)
        return ("SBAR-S-VP");
    else if (id == 21)
        return ("SBARQ");
    else if (id == 22)
        return ("SINV");
    else if (id == 23)
        return ("SQ");
    else if (id == 24)
        return ("SQ-VP");
    else if (id == 25)
        return ("UCP");
    else if (id == 26)
        return ("VP");
    else if (id == 27)
        return ("WHADJP");
    else if (id == 28)
        return ("WHADVP");
    else if (id == 29)
        return ("WHNP");
    else if (id == 30)
        return ("WHPP");
    else if (id == 31)
        return ("X");
    else if (id == 32)
        return ("SBAR-S-NP");
    else
        return ("???");
}


char *id_word(int wrd)
{
    char *res;
    int j, k, m;
    char str[100];

    return ("???");
}


void print_input(int t)
{
    printf("%3s", id_tag(input_hist[t].input1));
}



/* copy weight file and momentum file */
void copy_weights(char *fromfile, char *tofile)
{
    char momfromfile[FILENAMESIZE];
    char momtofile[FILENAMESIZE];
    FILE *ffp, *tfp;
    int c;

    if ((ffp = fopen(fromfile, "r")) == NULL)
        error("ERROR: couldn't open a file for copying from.");
    if ((tfp = fopen(tofile, "w")) == NULL)
        error("ERROR: couldn't open a file for copying to.");

    while ((c = getc(ffp)) != EOF)
        putc(c, tfp);

    fclose(ffp);
    fclose(tfp);

    /* copy momentum file */
    strcpy(momfromfile, fromfile);
    strcat(momfromfile, ".mom");
    strcpy(momtofile, tofile);
    strcat(momtofile, ".mom");
    if ((ffp = fopen(momfromfile, "r")) == NULL)
        error("ERROR: couldn't open a momentum file for copying from.");
    if ((tfp = fopen(momtofile, "w")) == NULL)
        error("ERROR: couldn't open a momentum file for copying to.");

    while ((c = getc(ffp)) != EOF)
        putc(c, tfp);

    fclose(ffp);
    fclose(tfp);
}


void copyfile(char *fromfile, char *tofile)
{
    FILE *ffp, *tfp;
    int c;

    if ((ffp = fopen(fromfile, "r")) == NULL)
        error("ERROR: couldn't open a file for copying from.");
    if ((tfp = fopen(tofile, "w")) == NULL)
        error("ERROR: couldn't open a file for copying to.");

    while ((c = getc(ffp)) != EOF)
        putc(c, tfp);

    fclose(ffp);
    fclose(tfp);
}


void error(char *str)
{
    printf("%s\n", str);
    exit(0);
}


void warning(char *str)
{
#if FILEWARN
    if ((warnfp = fopen(warnfile, "a")) != NULL) {
        fprintf(warnfp, "Warning: %s  [sent:%d]\n", str, num_sent + 1);
        fclose(warnfp);
    } else
#endif
        {
	    fprintf(stderr, "Warning: %s  [sent:%d]\n", str, num_sent + 1);
	    fflush(stderr);
	}
}


/** 
 * fills the input cl_levels_out and output2 fields basing on string representation out_str
 * @param inp input structure to fill
 * @param out1 output 1 (tag)
 * @param out_str the string which represents this clustering structure branch
 * param separaters separaters used in out_str
 */
void fill_input_clusterized(struct input *inp, int out1, char *out2_str,
                            char *separaters)
{


    char *child_id_str;
    int curr_cl_level;
    struct cluster *curr_cluster;
    int child_id;



    child_id_str = strtok(out2_str, "-");
    curr_cl_level = 0;
    curr_cluster = out_clusters_arr[out1]->root;

    while (child_id_str != NULL) {
        child_id = atoi(child_id_str);

        if (child_id < 0 || child_id >= curr_cluster->child_qty) {
            error
                ("Error: clustering in input file is wrong or doesn't correspond to description in .inf");
        }
        if (curr_cl_level >= MAXCLUSTERLEVELS) {
            error
                ("Too many levels in clustering or adjust MAXCLUSTERLEVELS and rebuild");
        }

        curr_cluster = curr_cluster->childs[child_id];

        inp->output2[curr_cl_level] =
            CLUSTERID_TO_WORDID(curr_cluster->cluster_id, out1);


        child_id_str = strtok(NULL, "-");
        curr_cl_level++;
    }

    inp->cl_levels_out = curr_cl_level;
}



/** returns the name of the weight file */
char *get_weight_file() {
    return weightfile;
}

/** set the prediction beam */
void set_pred_beam(int _predbeam) {
    predbeam = _predbeam;
}

/** sets the phase beam */
void set_phase_beam(int _phasebeam) {
    phasebeam = _phasebeam;
}

/** sets whethe joint error should be used  or not */
void set_joint_error_flag(int flag) {
    joint_error_flag = flag;
    if (joint_error_flag) {
        training_predbeam = 0;
    }
}

/** inits best_options */
void init_best_options(int length) {
    if (best_options == NULL) {
        error("Internal error: best_options must be allocated");
    }   

    int period, idx;
    for (period = 0; period <= length; period++) {
        for (idx = 0; idx < predbeam; idx++) {
            best_options[period][idx] = NULL;
        }
    }
}
